# This script works around an issue in Vivado 2015.2/2015.3 that occurs when all of
# the following are true:
#
# 1. A custom part is selected for DDR4 SDRAM Controller IP.
# 2. The Vivado project was closed before generating the output products of
#    the IP, and then reopened.
#
# At this point, when the output products of the IP are generated, an error
# occurs and implementation of the FPGA design will fail.
#
# To work around this, the following script should be source'd immediately before
# generating the output products of the IP. It removes the custom part properties
# and then restores them, which (provided that the Vivado project is not closed)
# allows the IP's output products to be generated correctly.
#

proc fix_custom_part {mig_ip_name} {
  set mig_ip [get_ips $mig_ip_name]
  
  set isCustom [get_property {CONFIG.C0.DDR4_isCustom} $mig_ip]
  if { !$isCustom } {
    puts "IP '$mig_ip_name' does not use a custom parts file. Nothing to do."
    return
  }
  
  # Save existing properties
  set CustomParts [get_property {CONFIG.C0.DDR4_CustomParts} $mig_ip]
  set MemoryType [get_property {CONFIG.C0.DDR4_MemoryType} $mig_ip]
  set MemoryPart [get_property {CONFIG.C0.DDR4_MemoryPart} $mig_ip]
  set MemoryVoltage [get_property {CONFIG.C0.DDR4_MemoryVoltage} $mig_ip]
  set Ecc [get_property {CONFIG.C0.DDR4_Ecc} $mig_ip]
  set DataMask [get_property {CONFIG.C0.DDR4_DataMask} $mig_ip]
  set DataWidth [get_property {CONFIG.C0.DDR4_DataWidth} $mig_ip]
  
  # Remove custom parts file
  set_property {CONFIG.C0.DDR4_isCustom} false $mig_ip
  
  # Apply custom parts file again
  set_property -dict [list \
    {CONFIG.C0.DDR4_isCustom} {true} \
    {CONFIG.C0.DDR4_CustomParts} $CustomParts \
    {CONFIG.C0.DDR4_MemoryType} $MemoryType \
    {CONFIG.C0.DDR4_MemoryPart} $MemoryPart \
    {CONFIG.C0.DDR4_MemoryVoltage} $MemoryVoltage \
    {CONFIG.C0.DDR4_Ecc} $Ecc \
    {CONFIG.C0.DDR4_DataMask} $DataMask \
    {CONFIG.C0.DDR4_DataWidth} $DataWidth \
  ] $mig_ip
  
  puts "Reapplied custom part properties to IP '$mig_ip_name'."
}


proc reset_output_products {mig_ip_name} {
  set mig_ip [get_ips $mig_ip_name]

  set isCustom [get_property {CONFIG.C0.DDR4_isCustom} $mig_ip]
  if { !$isCustom } {
    puts "IP '$mig_ip_name' does not use a custom parts file. Nothing to do."
    return
  }

  set ip_xci_file [get_property {IP_FILE} [get_ips $mig_ip]]
  reset_target all [get_files  $ip_xci_file]
  if {![string equal [get_filesets -quiet $mig_ip_name] ""]} {
    delete_ip_run [get_files -of_objects [get_fileset $mig_ip_name] $ip_xci_file]
  }
  update_compile_order -fileset sources_1

  puts "Reset output products for IP '$mig_ip_name'."
  puts "Now regenerate its output products without first closing the Vivado project."
}


fix_custom_part ddr4sdram
reset_output_products ddr4sdram
