/*
** dma_demo.cpp
**
** Test program for the DMA Demonstration FPGA Design (dma_demo)
**
** Notes
** =====
**
** 1. Uses the ADMXRC3UT library for opening the device. ADMXRC3UT is a set of
**    functions that contains commonly-used code in example / test programs,
**    but it is not part of the ADMXRC3 API.
**
**    ADMXRC3UT_OpenByIndex is similar to ADMXRC3_OpenEx from the ADMXRC3 API,
**    except that it also performs a check to see if the device is in an error
**    state and attempts to clear device errors if it is.
**
**    ADMXRC3UT_OpenBySerialNumber is similar to ADMXRC3_OpenEx from the
**    ADMXRC3 API, except that it searches for the device with a particular
**    serial number rather than opening by index. Like ADMXRC3UT_OpenByIndex,
**    it also checks to see if the device is in an error state and attempts to
**    clear device errors if it is.
**
** 2. This source file contains code for doing the actual test. In VxWorks,
**    the admxrc3DMADemoIndex & admxrc3DMADemoSN functions are entry points
**    that can be called from the VxWorks shell.
*/

#ifdef _WIN32

# include <windows.h>
# include <tchar.h>

typedef UINT8 uint8_t;
typedef UINT32 uint32_t;

#elif defined(__linux)

# include <stdint.h>

# define _T(x) x
# define _tmain main
# define _tcslen(x) strlen(x)
typedef char TCHAR;

#elif defined(__VXWORKS__) || defined(__vxworks)

# include <vxWorks.h>

# if _WRS_VXWORKS_MAJOR >= 6
#   include <stdint.h>
# endif

# define _T(x) x
# define _tmain main
# define _tcslen(x) strlen(x)
typedef char TCHAR;

#endif

#include <cassert>
#include <cstring>
#include <string>
#include <iostream>
#include <algorithm>

#include <exappevent.h>
#include <exappthread.h>
#include <exapptime.h>

#include <admxrc3.h>
#include <admxrc3ut.h>

#include "dma_demo.h"

using namespace AlphaData::AppFramework;

#ifndef FALSE
# define FALSE 0
#endif
#ifndef TRUE
# define TRUE 1
#endif

#ifdef _UNICODE
typedef std::wstring SType;
typedef std::wstringstream SStreamType;
#define SSEMPTY(ss) ((ss).str(std::basic_string<wchar_t>()))
#else
typedef std::string SType;
typedef std::stringstream SStreamType;
#define SSEMPTY(ss) ((ss).str(std::string()))
#endif

#define MAX_NUM_DMA_ENGINE (32)

CExAppMessageDisplay g_dmaDemoMessageDisplay;

#ifdef _WIN32
static const TCHAR* g_pPathSeparator = _T("\\");
#else
static const TCHAR* g_pPathSeparator = _T("/");
#endif

static const TCHAR* g_pDesignName = _T("dma_demo");

class CDMADemoThread : protected CExAppThread {
public:
  CDMADemoThread(
    CExAppMessageDisplay* pMessageDisplay,
    ADMXRC3_HANDLE hDevice,
    void* pFpgaBlockRam,
    size_t blockRamSize,
    size_t dmaTransferSize)
    : CExAppThread(), m_pMessageDisplay(pMessageDisplay)
  {
    m_hDevice = hDevice;
    m_pFpgaBlockRam = pFpgaBlockRam;
    m_blockRamSize = blockRamSize;
    m_dmaTransferSize = dmaTransferSize;
    m_pPatternBuffer = NULL;
    m_pCheckBuffer = NULL;
    m_pDmaBuffer = NULL;
    m_hDmaBuffer = 0;
  } 

  virtual ~CDMADemoThread() {
    if (IsRunning()) {
      Shutdown();
    }
    Cleanup();
  }

  bool Initialize() {
    m_pPatternBuffer = malloc(m_dmaTransferSize);
    if (NULL == m_pPatternBuffer) {
      SStreamType ss;
      ss << _T("Failed to allocate 0x") << std::hex << m_dmaTransferSize << _T(" bytes for test pattern buffer.");
      m_pMessageDisplay->AddError(ss.str());
      Cleanup();
      return false;
    }

    m_pCheckBuffer = malloc(m_dmaTransferSize);
    if (NULL == m_pCheckBuffer) {
      SStreamType ss;
      ss << _T("Failed to allocate 0x") << std::hex << m_dmaTransferSize << _T(" bytes for check buffer.");
      m_pMessageDisplay->AddError(ss.str());
      Cleanup();
      return false;
    }

    m_pDmaBuffer = malloc(m_dmaTransferSize);
    if (NULL == m_pDmaBuffer) {
      SStreamType ss;
      ss << _T("Failed to allocate 0x") << std::hex << m_dmaTransferSize << _T(" bytes for DMA buffer.");
      m_pMessageDisplay->AddError(ss.str());
      Cleanup();
      return false;
    }

    ADMXRC3_STATUS status = ADMXRC3_Lock(m_hDevice, m_pDmaBuffer, m_dmaTransferSize, &m_hDmaBuffer);
    if (ADMXRC3_SUCCESS != status) {
      SStreamType ss;
      ss << _T("Failed to lock DMA buffer of 0x") << std::hex << m_dmaTransferSize << _T(" bytes: ") << ADMXRC3_GetStatusString(status, TRUE);
      m_pMessageDisplay->AddError(ss.str());
      Cleanup();
      return false;
    }

    m_parameters.m_bTerminate = FALSE;

    Start();

    return true;
  }

  void SetParameters(
    unsigned int dmaEngineIndex,
    bool bReadFromFpga,
    unsigned int durationMs)
  {
    m_parameters.m_dmaEngineIndex = dmaEngineIndex;
    m_parameters.m_bReadFromFpga = bReadFromFpga;
    m_parameters.m_durationMs = durationMs;
    if (0 != m_parameters.m_durationMs) {
      m_parameters.m_duration = CExAppTime((double)durationMs / 1000.0, true);
      if (m_parameters.m_duration.IsZero()) {
        // Don't allow zero ticks for measurement period.
        m_parameters.m_duration.Increment();
      }
    }

    /*
    ** Generate a test pattern.
    **
    ** For DMA engine N, each 32-bit word is constructed as follows,
    ** where pattern is considered to be uint32_t*:
    **
    **   pattern[i][7:0]   = (N+1)^(i[7:0])
    **   pattern[i][15:8]  = (N+1)^(i[15:8])
    **   pattern[i][23:16] = (N+1)^(i[23:16])
    **   pattern[i][31:24] = (N+1)^(i[31:24])
    **
    ** where i is iterated over the range [0, m_dmaTransferSize/4-1].
    **
    ** For example, for DMA engine 1, the first few words of the test
    ** pattern are:
    **
    **   0x02020202, 0x03030303, 0x00000000, 0x01010101, 0x06060606, ...
    */
    uint32_t* p32 = (uint32_t*)m_pPatternBuffer;
    uint32_t index = (uint32_t)m_parameters.m_dmaEngineIndex + 1;
    size_t count32 = m_dmaTransferSize / sizeof(uint32_t);
    size_t remainder32 = m_dmaTransferSize % sizeof(uint32_t);
    uint32_t channelPattern = (index<< 0) | (index << 8) | (index << 16) | (index << 24);
    for (size_t i = 0; i < count32; i++) {
      *p32++ = channelPattern ^ (uint32_t)i;
    }
    uint8_t* p8 = (uint8_t*)p32;
    channelPattern ^= count32;
    for (size_t i = 0; i < remainder32; i++) {
      *p8++ = (uint8_t)(channelPattern >> (i * 8));
    }

    if (m_parameters.m_bReadFromFpga) {
      // Copy test pattern to FPGA BlockRAM, via Direct Slave channel.
      memcpy((uint8_t*)m_pFpgaBlockRam + m_blockRamSize * m_parameters.m_dmaEngineIndex, m_pPatternBuffer, m_dmaTransferSize);
    } else {
      // Copy test pattern to DMA buffer.
      memcpy(m_pDmaBuffer, m_pPatternBuffer, m_dmaTransferSize);
    }
  }

  void StartTest() {
    m_result.m_dmaStatus = ADMXRC3_SUCCESS;
    m_startEvent.Set();
  }

  void WaitTest() {
    m_finishedEvent.Wait();
  }

  void* Shutdown() {
    m_parameters.m_bTerminate = true;
    m_startEvent.Set();
    return Join();
  }

  unsigned int CheckData(unsigned int maxErrorReported) {
    uint32_t* pActual;

    if (m_parameters.m_bReadFromFpga) {
      // Data to verify against test pattern is in DMA buffer.
      pActual = (uint32_t*)m_pDmaBuffer;
    } else {
      // Copy data in FPGA BlockRAM to check buffer, via Direct Slave channel.
      memcpy(m_pCheckBuffer, (uint8_t*)m_pFpgaBlockRam + m_blockRamSize * m_parameters.m_dmaEngineIndex, m_dmaTransferSize);
      // Data to verify against test pattern is in check buffer.
      pActual = (uint32_t*)m_pCheckBuffer;
    }

    // Verify the data against the test pattern buffer.
    size_t n32 = (m_dmaTransferSize + sizeof(uint32_t) - 1) / sizeof(uint32_t);
    size_t remainder = m_dmaTransferSize % sizeof(uint32_t);
    uint32_t* pExpected = (uint32_t*)m_pPatternBuffer;
    unsigned numError = 0;
    for (size_t i = 0; i < n32; i++) {
      uint32_t expected = pExpected[i];
      uint32_t actual = pActual[i];
      if (i == n32 - 1 && 0 != remainder) {
        uint32_t mask = (0x1 << (8 * (int)remainder)) - 1;
        expected &= mask;
        actual &= mask;
      }
      if (expected != actual) {
        if (numError < maxErrorReported) {
          SStreamType ss;
          size_t offset = i * sizeof(uint32_t);
          ss << _T("Data verification error; offset=0x") << std::hex << std::setfill(_T('0')) << std::setw(8) << offset <<
            _T(" expected=0x") << std::hex << std::setfill(_T('0')) << std::setw(8) << expected <<
            _T(" actual=0x") << std::hex << std::setfill(_T('0')) << std::setw(8) << actual;
          m_pMessageDisplay->AddError(ss.str());
        } else if (numError == maxErrorReported) {
          m_pMessageDisplay->AddError(_T("[More data errors detected]"));
        }
        numError++;
      }
    }

    SStreamType ss;
    ss << numError << _T(" data error(s) detected for DMA engine ") << m_parameters.m_dmaEngineIndex;
    m_pMessageDisplay->AddInfo(ss.str());

    return numError;
  }

protected:
  void* ThreadFunction() {

    m_startEvent.Wait();

    while (!m_parameters.m_bTerminate) {
      m_result.m_transferred = 0;
      CExAppTime start, now = 0;
      start = CExAppTime::Now();
      do  {
        if (m_parameters.m_bReadFromFpga) {
          m_result.m_dmaStatus = ADMXRC3_ReadDMALockedEx(m_hDevice, m_parameters.m_dmaEngineIndex, 0, m_hDmaBuffer, 0, m_dmaTransferSize, 0);
        } else {
          m_result.m_dmaStatus = ADMXRC3_WriteDMALockedEx(m_hDevice, m_parameters.m_dmaEngineIndex, 0, m_hDmaBuffer, 0, m_dmaTransferSize, 0);
        }
        now = CExAppTime::Now() - start;
        if (ADMXRC3_SUCCESS != m_result.m_dmaStatus) {
          break;
        }
        m_result.m_transferred += m_dmaTransferSize;
      } while (0 == m_parameters.m_durationMs || now < m_parameters.m_duration);
      m_result.m_elapsedSec = (double)now;
      m_result.m_transferRate = (double)m_result.m_transferred / m_result.m_elapsedSec;

      m_finishedEvent.Set();

      m_startEvent.Wait();
    }

    return NULL;
  }

  void Cleanup() {
    if (0 != m_hDmaBuffer) {
      ADMXRC3_Unlock(m_hDevice, m_hDmaBuffer);
      m_hDmaBuffer = 0;
    }

    if (NULL != m_pDmaBuffer) {
      free(m_pDmaBuffer);
      m_pDmaBuffer = NULL;
    }

    if (NULL != m_pCheckBuffer) {
      free(m_pCheckBuffer);
      m_pCheckBuffer = NULL;
    }

    if (NULL != m_pPatternBuffer) {
      free(m_pPatternBuffer);
      m_pPatternBuffer = NULL;
    }
  }

public:
  struct {
    bool m_bTerminate;
    unsigned int m_dmaEngineIndex;
    bool m_bReadFromFpga;
    unsigned int m_durationMs;
    CExAppTime m_duration;
  } m_parameters;
  struct {
    ADMXRC3_STATUS m_dmaStatus;
    unsigned long long m_transferred;
    double m_elapsedSec;
    double m_transferRate;
  } m_result;

protected:
  CExAppMessageDisplay* m_pMessageDisplay;
  ADMXRC3_HANDLE m_hDevice;
  void* m_pFpgaBlockRam;
  size_t m_blockRamSize;
  size_t m_dmaTransferSize;
  void* m_pPatternBuffer;
  void* m_pCheckBuffer;
  void* m_pDmaBuffer;
  ADMXRC3_BUFFER_HANDLE m_hDmaBuffer;
  CExAppEvent m_startEvent;
  CExAppEvent m_finishedEvent;
};

static uint32_t /* Number of errors */
TestDirectSlaveRegion(
  CExAppMessageDisplay* pMessageDisplay,
  void* pFpgaBlockRam,
  unsigned int engineIndex,
  uint32_t blockRamSize,
  unsigned int maxDisplayedError)
{
  uint32_t n32 = blockRamSize / sizeof(uint32_t);
  uint32_t* p;
  uint32_t expected, actual;
  uint32_t numError = 0;

  pMessageDisplay->FormatInfo(
    _T("Testing BlockRAM %u using Direct Slave channel..."),
    engineIndex);

  // Write pattern
  p = (uint32_t*)pFpgaBlockRam;
  for (uint32_t i = 0; i < n32; i++) {
    *p++ = (engineIndex << 24) | i;
  }

  // Read back pattern and check
  p = (uint32_t*)pFpgaBlockRam;
  for (uint32_t i = 0; i < n32; i++) {
    expected = (engineIndex << 24) | i;
    actual = *p++;
    if (expected != actual) {
      SStreamType ss;
      if (0 == numError) {
        ss << _T("While verifying data written to BlockRAM ") << engineIndex << _T(" using Direct Slave channel:");
        pMessageDisplay->AddError(ss.str());
        ss.str(_T(""));
      }
      if (numError < maxDisplayedError) {
        ss << _T("Expected=0x") << std::hex << std::setfill(_T('0')) << std::setw(8) << expected << _T(" actual=0x") << std::hex << std::setw(8) << actual;
        pMessageDisplay->AddError(ss.str());
      } else if (numError == maxDisplayedError) {
        ss << _T("Further data verification errors detected, but not displayed (maximum ") << maxDisplayedError << _T(" such errors displayed)");
        pMessageDisplay->AddError(ss.str());
      }
      numError++;
    }
  }

  return numError;
}

static int
TestDirectSlave(
  CExAppMessageDisplay* pMessageDisplay,
  void* pFpgaBlockRam,
  uint32_t dmaEngineMask,
  uint32_t blockRamSize,
  unsigned int maxDisplayedError)
{
  int exitCode = ExitCodeSuccess;
  uint32_t numError, totalNumError = 0;

  for (unsigned int i = 0; i < 32; i++) {
    if ((0x1U << i) & dmaEngineMask) {
      numError = TestDirectSlaveRegion(pMessageDisplay, (uint8_t*)pFpgaBlockRam + i * blockRamSize, i, blockRamSize, maxDisplayedError);
      totalNumError += numError;
      maxDisplayedError -= (numError > maxDisplayedError) ? 0 : maxDisplayedError - numError;
    }
  }

  if (totalNumError) {
    SStreamType ss;
    ss << totalNumError << _T("(0x") << std::hex << totalNumError << _T(") error(s) were detected in initial test of data transfer to and from BlockRAMs using Direct Slave channel.");
    pMessageDisplay->AddWarning(ss.str());
    exitCode = ExitCodeDataErrors;
  } else {
    pMessageDisplay->AddInfo(_T("No errors were detected in initial test of data transfer to and from BlockRAMs using Direct Slave channel."));
  }

  return exitCode;
}

static int /* ExitCode */
admxrc3DMADemo(
  ADMXRC3_HANDLE hDevice,
  uint32_t dmaEngineMask,
  uint32_t dmaDirectionMask,
  uint32_t dmaTransferSize,
  unsigned int durationMs,
  unsigned int maxErrorDisplayed,
  int bVerifyData,
  const TCHAR* pBitPath,
  const TCHAR* pBitFile)
{
  CExAppMessageDisplay* pMessageDisplay = &g_dmaDemoMessageDisplay;
  int exitCode = ExitCodeSuccess;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFOEX cardInfo;
  void* pFpgaBlockRam = NULL;
  uint32_t dmaEnginesPresent;
  double totalMiB, totalRateMiB;
  unsigned int i, j, numDmaEngineUsed = 0;
  SStreamType resultSS;
  uint32_t blockRamSize = 0;
  size_t mapSize;
  CDMADemoThread* threads[MAX_NUM_DMA_ENGINE];
  SType sBitFile;

  // Ensure threads array is all NULL pointers initially, in case we have to cleanup early
  for (i = 0; i < sizeof(threads) / sizeof(threads[0]); i++) {
    threads[i] = NULL;
  }

  // Get card information; needed below.
  status = ADMXRC3_GetCardInfoEx(hDevice, &cardInfo);
  if (ADMXRC3_SUCCESS != status) {
    SStreamType ss;
    ss << _T("Failed to get card information: ") << ADMXRC3_GetStatusString(status, TRUE);
    pMessageDisplay->AddError(ss.str());
    exitCode = ExitCodeGetCardInfoFailed;
    goto done;
  }

  // Check that we are running on a supported model
  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRCKU1:
    blockRamSize = 512 * 1024;
    break;

  default:
    SStreamType ss;
    ss << _T("This program must be run on one of the following models: ADM-XRC-KU1. The device in use is: ") << ADMXRC3UT_GetModelName(cardInfo.Model);
    pMessageDisplay->AddError(ss.str());
    exitCode = ExitCodeUnsupportedModel;
    goto done;
  }

  if (0 != dmaTransferSize) {
    if (dmaTransferSize > blockRamSize) {
      SStreamType ss;
      ss << _T("Value ") << dmaTransferSize << _T(" for DMA transfer size must be no greater than size of BlockRAMs in FPGA (0x") << std::hex << blockRamSize << _T(")");
      pMessageDisplay->AddError(ss.str());
      exitCode = ExitCodeTransferSizeTooBig;
      goto done;
    }
  } else {
    dmaTransferSize = blockRamSize;
  }

  // Ensure that DMA engines that do not exist are not used.
  dmaEnginesPresent = (cardInfo.NumDmaChannel >= 32) ? 0xFFFFFFFFU : ((uint32_t)1 << cardInfo.NumDmaChannel) - 1;
  if (dmaEngineMask & ~dmaEnginesPresent) {
    SStreamType ss;
    ss << _T("Value 0x") << std::hex << dmaEngineMask << _T(" for DMA engine mask includes some DMA engines that do not exist. Corrected DMA engine mask is 0x");
    dmaEngineMask &= dmaEnginesPresent;
    ss << std::hex << dmaEngineMask;
    pMessageDisplay->AddWarning(ss.str());
  }

  // Count the number of DMA engines actually being used.
  numDmaEngineUsed = 0;
  if (0 == dmaEngineMask) {
    // If we get here, it is because we are in VxWorks and the dmaEngineMask parameter was omitted, so it is 0.
    // In this case, use the default.
    dmaEngineMask = 0x1;
  }
  for (i = 0; i < MAX_NUM_DMA_ENGINE; i++) {
    if (dmaEngineMask & (1 << i)) {
      numDmaEngineUsed++;
    }
  }
  if (0 == numDmaEngineUsed) {
    pMessageDisplay->AddError(_T("No DMA engines are selected. Aborting."));
    exitCode = ExitCodeNoDmaEnginesSelected;
    goto done;
  } else {
    SStreamType ss;
    ss << _T("Using ") << numDmaEngineUsed << _T(" DMA engine(s): ");
    for (i = 0, j = 0; i < MAX_NUM_DMA_ENGINE; i++) {
      if (dmaEngineMask & (1 << i)) {
        ss << i;
        if (j != numDmaEngineUsed - 1) {
          ss << _T(", ");
        }
        j++;
      }
    }
    pMessageDisplay->AddInfo(ss.str());
    SSEMPTY(ss);
    ss << _T("DMA transfer size is 0x") << std::hex << dmaTransferSize << _T("(") << std::dec << dmaTransferSize << _T(") byte(s)");
    pMessageDisplay->AddInfo(ss.str());
  }

  // Map the BlockRAMs, addressable by Direct Slave, into our virtual address space
  mapSize = blockRamSize * cardInfo.NumDmaChannel;
  status = ADMXRC3_MapWindow(hDevice, 0, 0, mapSize, &pFpgaBlockRam);
  if (ADMXRC3_SUCCESS != status) {
    SStreamType ss;
    ss << _T("Failed to map FPGA BlockRAMs (0x") << std::hex << mapSize << _T(" bytes ): ") << ADMXRC3_GetStatusString(status, TRUE);
    pMessageDisplay->AddError(ss.str());
    exitCode = ExitCodeMapWindowFailed;
    goto done;
  }

  // Configure the target FPGA with the .bit file for the "dma_demo" FPGA design.
  if (pBitFile == NULL) {
    SStreamType bitFile;
    //
    // Construct a .bit filename of the form:
    //   pBitPath + "/<design>-<device>[_<speed>][_<stepping>]/<design>.bit"
    //
    // Examples:
    //   pBitPath + "/dma_demo-ku115_2e/dma_demo.bit"
    //   pBitPath + "/dma_demo-ku060_2e_es2/dma_demo.bit"
    //
    if (pBitPath == NULL) {
      pMessageDisplay->AddError(_T("'pBitFile' and 'pBitPath' parameters cannot both be NULL."));
      exitCode = ExitCodeBadBitPath;
      goto done;
    }
    ADMXRC3_FPGA_INFO fpgaInfo;
    status = ADMXRC3_GetFpgaInfo(hDevice, 0, &fpgaInfo);
    if (ADMXRC3_SUCCESS != status) {
      SStreamType ss;
      ss << _T("Failed to get FPGA information: ") << ADMXRC3_GetStatusString(status, TRUE);
      pMessageDisplay->AddError(ss.str());
      exitCode = ExitCodeFpgaInfoFailed;
      goto done;
    }
    bitFile << pBitPath;
    bitFile << g_pPathSeparator;
    bitFile << g_pDesignName;
    bitFile << _T("-");
    switch (fpgaInfo.DeviceCode) {
    case ADMXRC3_FPGA_KU060:
      bitFile << _T("ku060");
      break;

    case ADMXRC3_FPGA_KU115:
      bitFile << _T("ku115");
      break;

    default:
      SStreamType ss;
      ss << _T("Unrecognized FPGA device code ") << fpgaInfo.DeviceCode << _T("(0x") << std::hex << fpgaInfo.DeviceCode << _T("). Cannot construct .bit filename automatically.");
      pMessageDisplay->AddError(ss.str());
      exitCode = ExitCodeUnrecognizedFpga;
      goto done;
    }
    if (_tcslen(fpgaInfo.SpeedGrade) > 0) {
      bitFile << _T("_");
      bitFile << fpgaInfo.SpeedGrade;
    }
    if (_tcslen(fpgaInfo.Stepping) > 0) {
      bitFile << _T("_");
      bitFile << fpgaInfo.Stepping;
    }
    bitFile << g_pPathSeparator;
    bitFile << g_pDesignName;
    bitFile << _T(".bit");
    sBitFile = bitFile.str();
    std::transform(sBitFile.begin(), sBitFile.end(), sBitFile.begin(), ::tolower);
    pBitFile = sBitFile.c_str();
  }
  status = ADMXRC3_ConfigureFromFile(hDevice, 0, 0, pBitFile);
  if (ADMXRC3_SUCCESS != status) {
    SStreamType ss;
    ss << _T("Failed to configure target FPGA 0 with '") << pBitFile << ("': ") << ADMXRC3_GetStatusString(status, TRUE);
    pMessageDisplay->AddError(ss.str());
    exitCode = ExitCodeConfigureFpgaFailed;
    goto done;
  }

  if (bVerifyData) {
    // Test the BlockRAMs via Direct Slave
    exitCode = TestDirectSlave(pMessageDisplay, pFpgaBlockRam, dmaEngineMask, blockRamSize, maxErrorDisplayed);
    if (ExitCodeSuccess != exitCode) {
      goto done;
    }
  }

  // Construct threads
  for (i = 0, j = 0; i < MAX_NUM_DMA_ENGINE; i++) {
    if (!(dmaEngineMask & (1 << i))) {
      continue;
    }
    CDMADemoThread* pThread = new CDMADemoThread(pMessageDisplay, hDevice, pFpgaBlockRam, blockRamSize, dmaTransferSize);
    if (NULL == pThread) {
      pMessageDisplay->FormatError(_T("Failed to create DMA thread %u."), j);
      exitCode = ExitCodeThreadCreateFailed;
      goto done;
    }
    if (!pThread->Initialize()) {
      delete pThread;
      pMessageDisplay->FormatError(_T("Failed to initialize DMA thread %u."), j);
      exitCode = ExitCodeThreadCreateFailed;
      goto done;
    }
    pThread->SetParameters(i, (dmaDirectionMask & (1 << i)) ? true : false, durationMs);
    threads[j++] = pThread;
  }

  // Do the test.
  pMessageDisplay->AddInfo(_T("Doing DMA performance test..."));
  for (j = 0; j < numDmaEngineUsed; j++) {
    threads[j]->StartTest();
  }
  for (j = 0; j < numDmaEngineUsed; j++) {
    threads[j]->WaitTest();
  }

  if (bVerifyData) {
    // Check the data of the last DMA transfer of each of DMA engine for correctness
    unsigned int numDataError = 0;
    for (j = 0; j < numDmaEngineUsed; j++) {
      numDataError += threads[j]->CheckData(maxErrorDisplayed);
    }
    if (numDataError) {
      exitCode = ExitCodeDataErrors;
    }
  }

  // Display results.
  totalMiB = 0.0;
  totalRateMiB = 0.0;
  for (j = 0; j < numDmaEngineUsed; j++) {
    CDMADemoThread* pThread = threads[j];
    if (ADMXRC3_SUCCESS != pThread->m_result.m_dmaStatus) {
      SSEMPTY(resultSS);
      resultSS << _T("DMA thread ") << j << _T(" failed to perform a DMA transfer: ") << ADMXRC3_GetStatusString(pThread->m_result.m_dmaStatus, TRUE);
      pMessageDisplay->AddError(resultSS.str());
      continue;
    }
    double elapsedSec = pThread->m_result.m_elapsedSec;
    double transferredMiB = pThread->m_result.m_transferred / 1048576.0;
    double transferRateMiB = pThread->m_result.m_transferRate / 1048576.0;
    totalMiB += transferredMiB;
    totalRateMiB += transferRateMiB;
    SSEMPTY(resultSS);
    resultSS << _T("DMA engine ") << pThread->m_parameters.m_dmaEngineIndex;
    if (pThread->m_parameters.m_bReadFromFpga) {
      resultSS << _T(" read ") << transferredMiB << _T(" MiB from the FPGA in ");
    } else {
      resultSS << _T(" wrote ") << transferredMiB << _T(" MiB to the FPGA in ");
    }
    resultSS << elapsedSec << _T(" s at ") << transferRateMiB << _T(" MiB/s");
    pMessageDisplay->AddInfo(resultSS.str());
  }
  SSEMPTY(resultSS);
  resultSS << numDmaEngineUsed << _T(" DMA engine(s) transferred ") << totalMiB << _T(" MiB to/from the FPGA at ") << totalRateMiB << _T(" MiB/s");
  pMessageDisplay->AddInfo(resultSS.str());

done:
  // Terminate and clean up threads
  for (j = 0; j < numDmaEngineUsed; j++) {
    CDMADemoThread* pThread = threads[j];
    if (NULL != pThread) {
      pThread->Shutdown();
      delete pThread;
      threads[j] = NULL;
    }
  }

  if (NULL != pFpgaBlockRam) {
    ADMXRC3_UnmapWindow(hDevice, pFpgaBlockRam);
    pFpgaBlockRam = NULL;
  }

  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
    hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return exitCode;
}

int /* ExitCode */
admxrc3DMADemoIndex(
  const TCHAR* pBitPath,
  const TCHAR* pBitFile,
  unsigned int index,
  uint32_t dmaEngineMask,
  uint32_t dmaDirectionMask,
  uint32_t dmaTransferSize,
  unsigned int durationMs,
  unsigned int maxErrorDisplayed,
  int bVerifyData)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;

  status = ADMXRC3UT_OpenByIndex(index, FALSE, TRUE, 0, &hDevice);
  if (ADMXRC3_SUCCESS != status) {
    g_dmaDemoMessageDisplay.FormatError(
      _T("Failed to open device with index %u: %s"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    return ExitCodeCannotOpenIndex;
  }

  return admxrc3DMADemo(
    hDevice,
    dmaEngineMask,
    dmaDirectionMask,
    dmaTransferSize,
    durationMs,
    maxErrorDisplayed,
    bVerifyData,
    pBitPath,
    pBitFile);
}

int /* ExitCode */
admxrc3DMADemoSN(
  const TCHAR* pBitPath,
  const TCHAR* pBitFile,
  uint32_t serialNumber,
  uint32_t dmaEngineMask,
  uint32_t dmaDirectionMask,
  uint32_t dmaTransferSize,
  unsigned int durationMs,
  unsigned int maxErrorDisplayed,
  int bVerifyData)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;

  status = ADMXRC3UT_OpenBySerialNumber(serialNumber, FALSE, TRUE, 0, &hDevice);
  if (ADMXRC3_SUCCESS != status) {
    g_dmaDemoMessageDisplay.FormatError(
      _T("Failed to open device with serial number %lu: %s"),
      (unsigned long)serialNumber, ADMXRC3_GetStatusString(status, TRUE));
    return ExitCodeCannotOpenSerial;
  }

  return admxrc3DMADemo(
    hDevice,
    dmaEngineMask,
    dmaDirectionMask,
    dmaTransferSize,
    durationMs,
    maxErrorDisplayed,
    bVerifyData,
    pBitPath,
    pBitFile);
}



