--==============================================================================
--
--     File Name: axi4_256_pkg.vhd
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package axi4_256
--
--   Description: Package defining AXI4 256 profile.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2015 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 23 Oct 2015: Initial design
--
--==============================================================================

--
-- Exported functions:
--
--
--
-- Exported procedures:
--
--

library ieee;
use ieee.std_logic_1164.all;

package axi4_256 is

  -- Define AXI4 256 widths
  constant AXI4_256_RESP_WIDTH      : integer := 2;
  constant AXI4_256_ADDR_WIDTH      : integer := 64;
  constant AXI4_256_LEN_WIDTH       : integer := 8;
  constant AXI4_256_ID_WIDTH        : integer := 12;
  constant AXI4_256_DATA_WIDTH      : integer := 256;
  constant AXI4_256_DATA_WIDTH_LOG2 : integer := 8;
  constant AXI4_256_STRB_WIDTH      : integer := 32;
  constant AXI4_256_STRB_WIDTH_LOG2 : integer := 5;

  constant AXI4_256_SIZE_WIDTH      : integer := 3;
  constant AXI4_256_BURST_WIDTH     : integer := 2;
  constant AXI4_256_CACHE_WIDTH     : integer := 4;
  constant AXI4_256_PROT_WIDTH      : integer := 3;
  constant AXI4_256_LOCK_WIDTH      : integer := 1;
  constant AXI4_256_QOS_WIDTH       : integer := 4;
  constant AXI4_256_REGION_WIDTH    : integer := 4;


  -- Define AXI4 256 subtypes
  subtype axi4_256_resp_s is std_logic_vector(AXI4_256_RESP_WIDTH-1 downto 0);
  constant AXI4_256_RESP_OKAY   : axi4_256_resp_s := "00";
  constant AXI4_256_RESP_EXOKAY : axi4_256_resp_s := "01";
  constant AXI4_256_RESP_SLVERR : axi4_256_resp_s := "10";
  constant AXI4_256_RESP_DECERR : axi4_256_resp_s := "11";

  subtype axi4_size_s   is std_logic_vector(AXI4_256_SIZE_WIDTH-1 downto 0);
  subtype axi4_burst_s  is std_logic_vector(AXI4_256_BURST_WIDTH-1 downto 0);
  subtype axi4_cache_s  is std_logic_vector(AXI4_256_CACHE_WIDTH-1 downto 0);
  subtype axi4_prot_s   is std_logic_vector(AXI4_256_PROT_WIDTH-1 downto 0);
  subtype axi4_lock_s   is std_logic_vector(AXI4_256_LOCK_WIDTH-1 downto 0);
  subtype axi4_qos_s    is std_logic_vector(AXI4_256_QOS_WIDTH-1 downto 0);
  subtype axi4_region_s is std_logic_vector(AXI4_256_REGION_WIDTH-1 downto 0);

  constant AXI4_256_SIZE   : axi4_size_s   := "101";  -- 256 bits wide
  constant AXI4_256_BURST  : axi4_burst_s  := "01";   -- INCRemental
  constant AXI4_256_CACHE  : axi4_cache_s  := "0011"; -- Xilinx recommended value
  constant AXI4_256_PROT   : axi4_prot_s   := "000";  -- Xilinx recommended value
  constant AXI4_256_LOCK   : axi4_lock_s   := "0";    -- Xilinx recommended value
  constant AXI4_256_QOS    : axi4_qos_s    := "0000"; -- Xilinx recommended value
  constant AXI4_256_REGION : axi4_region_s := "0000"; -- Xilinx recommended value

  subtype axi4_256_addr_s   is std_logic_vector(AXI4_256_ADDR_WIDTH-1 downto 0);
  subtype axi4_256_len_s    is std_logic_vector(AXI4_256_LEN_WIDTH-1 downto 0);
  subtype axi4_256_id_s     is std_logic_vector(AXI4_256_ID_WIDTH-1 downto 0);
  subtype axi4_256_data_s   is std_logic_vector(AXI4_256_DATA_WIDTH-1 downto 0);
  subtype axi4_256_strb_s   is std_logic_vector(AXI4_256_STRB_WIDTH-1 downto 0);

  subtype axi4_256_size_s   is std_logic_vector(AXI4_256_SIZE_WIDTH-1 downto 0);
  subtype axi4_256_burst_s  is std_logic_vector(AXI4_256_BURST_WIDTH-1 downto 0);
  subtype axi4_256_cache_s  is std_logic_vector(AXI4_256_CACHE_WIDTH-1 downto 0);
  subtype axi4_256_prot_s   is std_logic_vector(AXI4_256_PROT_WIDTH-1 downto 0);
  subtype axi4_256_lock_s   is std_logic_vector(AXI4_256_LOCK_WIDTH-1 downto 0);
  subtype axi4_256_qos_s    is std_logic_vector(AXI4_256_QOS_WIDTH-1 downto 0);
  subtype axi4_256_region_s is std_logic_vector(AXI4_256_REGION_WIDTH-1 downto 0);


  -- AXI4 256 interface types
  -- Data flow from Master to Slave
  type axi4_256_m2s_t is record
    -- Write address channel
    awaddr   : axi4_256_addr_s; -- Write Address (16-byte aligned)
    awvalid  : std_logic;       -- Write Address Valid
    awid     : axi4_256_id_s;   -- Write Address ID
    awlen    : axi4_256_len_s;  -- Write Address burst length
    awsize   : axi4_256_size_s;
    awburst  : axi4_256_burst_s;
    awcache  : axi4_256_cache_s;
    awprot   : axi4_256_prot_s;
    awlock   : axi4_256_lock_s;
    awqos    : axi4_256_qos_s;
    awregion : axi4_256_region_s;
    -- Read address channel
    araddr   : axi4_256_addr_s; -- Read Address (16-byte aligned)
    arvalid  : std_logic;       -- Read Address Valid
    arid     : axi4_256_id_s;   -- Read Address ID
    arlen    : axi4_256_len_s;  -- Read Address burst length
    arsize   : axi4_256_size_s;
    arburst  : axi4_256_burst_s;
    arcache  : axi4_256_cache_s;
    arlock   : axi4_256_lock_s;
    arprot   : axi4_256_prot_s;
    arqos    : axi4_256_qos_s;
    arregion : axi4_256_region_s;
    -- Write data channel
    wdata    : axi4_256_data_s; -- Write Data
    wstrb    : axi4_256_strb_s; -- Write Data Strobes
    wvalid   : std_logic;       -- Write Data Valid
    wid      : axi4_256_id_s;   -- Write Data ID
    wlast    : std_logic;       -- Write Data Last
    -- Write response channel
    bready   : std_logic;       -- Write Response Ready
    -- Read data channel
    rready   : std_logic;       -- Read Data Ready
  end record;

  -- Data flow from Slave to Master
  type axi4_256_s2m_t is record
    -- Write address channel
    awready : std_logic;       -- Write Address Ready
    -- Read address channel
    arready : std_logic;       -- Read Address Ready
    -- Write data channel
    wready  : std_logic;       -- Write Data Ready
    -- Write response channel
    bresp   : axi4_256_resp_s; -- Write Response
    bvalid  : std_logic;       -- Write Response Valid
    bid     : axi4_256_id_s;   -- Write Response ID
    -- Read data channel
    rdata   : axi4_256_data_s; -- Read Data
    rresp   : axi4_256_resp_s; -- Read Data Response
    rvalid  : std_logic;       -- Read Data Valid
    rid     : axi4_256_id_s;   -- Read Data ID
    rlast   : std_logic;       -- Read Data Last
  end record;

  type axi4_256_m2s_t_array is array (natural range <>) of axi4_256_m2s_t;
  type axi4_256_s2m_t_array is array (natural range <>) of axi4_256_s2m_t;


  -- Define AXI4 256 interface type defaults
  constant AXI4_256_DEFAULT_M2S : axi4_256_m2s_t := (
    awaddr   => (others => '-'),
    awvalid  => '0',
    awid     => (others => '-'),
    awlen    => (others => '-'),
    awsize   => (others => '-'),
    awburst  => (others => '-'),
    awcache  => (others => '-'),
    awprot   => (others => '-'),
    awlock   => (others => '-'),
    awqos    => (others => '-'),
    awregion => (others => '-'),
    --
    araddr   => (others => '-'),
    arvalid  => '0',
    arid     => (others => '-'),
    arlen    => (others => '-'),
    arsize   => (others => '-'),
    arburst  => (others => '-'),
    arcache  => (others => '-'),
    arprot   => (others => '-'),
    arlock   => (others => '-'),
    arqos    => (others => '-'),
    arregion => (others => '-'),
    --
    wdata    => (others => '-'),
    wstrb    => (others => '-'),
    wvalid   => '0',
    wid      => (others => '-'),
    wlast    => '-',
    --
    bready   => '0',
    rready   => '0');

  constant AXI4_256_DEFAULT_S2M : axi4_256_s2m_t := (
    awready => '0',
    arready => '0',
    wready  => '0',
    --
    bresp   => (others => '-'),
    bvalid  => '0',
    bid     => (others => '-'),
    --
    rdata   => (others => '-'),
    rresp   => (others => '-'),
    rvalid  => '0',
    rid     => (others => '-'),
    rlast   => '-');

  constant AXI4_256_FLOAT_M2S : axi4_256_m2s_t := (
    awaddr   => (others => 'W'),
    awvalid  => 'L',
    awid     => (others => 'W'),
    awlen    => (others => 'W'),
    awsize   => (others => 'W'),
    awburst  => (others => 'W'),
    awcache  => (others => 'W'),
    awprot   => (others => 'W'),
    awlock   => (others => 'W'),
    awqos    => (others => 'W'),
    awregion => (others => 'W'),
    --
    araddr   => (others => 'W'),
    arvalid  => 'L',
    arid     => (others => 'W'),
    arlen    => (others => 'W'),
    arsize   => (others => 'W'),
    arburst  => (others => 'W'),
    arcache  => (others => 'W'),
    arprot   => (others => 'W'),
    arlock   => (others => 'W'),
    arqos    => (others => 'W'),
    arregion => (others => 'W'),
    --
    wdata    => (others => 'W'),
    wstrb    => (others => 'W'),
    wvalid   => 'L',
    wid      => (others => 'W'),
    wlast    => 'W',
    --
    bready   => 'L',
    rready   => 'L');

end package axi4_256;

package body axi4_256 is

end package body axi4_256;