library ieee;
use ieee.std_logic_1164.all;

library work;
use work.axi4_128.all;

entity mptl_tgt_ku_axi_sim is
port(
  -- Reset in
  async_sys_rst            : in    std_logic;
  -- AXI4
  aclk                     : out   std_logic;
  aresetn                  : out   std_logic;
  -- Direct slave
  ds_axi_awaddr            : out   std_logic_vector(63 downto 0);
  ds_axi_awlen             : out   std_logic_vector(7 downto 0);
  ds_axi_awsize            : out   std_logic_vector(2 downto 0);
  ds_axi_awburst           : out   std_logic_vector(1 downto 0);
  ds_axi_awcache           : out   std_logic_vector(3 downto 0);
  ds_axi_awprot            : out   std_logic_vector(2 downto 0);
  ds_axi_awvalid           : out   std_logic;
  ds_axi_wdata             : out   std_logic_vector(127 downto 0);
  ds_axi_wstrb             : out   std_logic_vector(15 downto 0);
  ds_axi_wlast             : out   std_logic;
  ds_axi_wvalid            : out   std_logic;
  ds_axi_bready            : out   std_logic;
  ds_axi_araddr            : out   std_logic_vector(63 downto 0);
  ds_axi_arlen             : out   std_logic_vector(7 downto 0);
  ds_axi_arsize            : out   std_logic_vector(2 downto 0);
  ds_axi_arburst           : out   std_logic_vector(1 downto 0);
  ds_axi_arcache           : out   std_logic_vector(3 downto 0);
  ds_axi_arprot            : out   std_logic_vector(2 downto 0);
  ds_axi_arvalid           : out   std_logic;
  ds_axi_rready            : out   std_logic;
  ds_axi_awready           : in    std_logic;
  ds_axi_wready            : in    std_logic;
  ds_axi_bresp             : in    std_logic_vector(1 downto 0);
  ds_axi_bvalid            : in    std_logic;
  ds_axi_arready           : in    std_logic;
  ds_axi_rdata             : in    std_logic_vector(127 downto 0);
  ds_axi_rresp             : in    std_logic_vector(1 downto 0);
  ds_axi_rlast             : in    std_logic;
  ds_axi_rvalid            : in    std_logic;
  ds_axi_awlock            : out   std_logic_vector(0 downto 0);
  ds_axi_awqos             : out   std_logic_vector(3 downto 0);
  ds_axi_arlock            : out   std_logic_vector(0 downto 0);
  ds_axi_arqos             : out   std_logic_vector(3 downto 0);
  ds_axi_awregion          : out   std_logic_vector(3 downto 0);
  ds_axi_arregion          : out   std_logic_vector(3 downto 0);
  -- MPTL
  mptl_tx_p                : out   std_logic_vector(3 downto 0);
  mptl_tx_n                : out   std_logic_vector(3 downto 0);
  mptl_rx_p                : in    std_logic_vector(3 downto 0);
  mptl_rx_n                : in    std_logic_vector(3 downto 0);
  mptl_clk_p               : in    std_logic;
  mptl_clk_n               : in    std_logic;
  mptl_clk_out             : out   std_logic;
  -- MPTL initialisation
  mptl_target_configured_n : out   std_logic;
  mptl_target_gtp_online_n : out   std_logic;
  mptl_bridge_gtp_online_n : in    std_logic;
   -- MPTL sideband
  gpio_t2b                 : in    std_logic_vector(15 downto 0);
  gpio_b2t                 : out   std_logic_vector(15 downto 0));
end;


architecture behav of mptl_tgt_ku_axi_sim is

  -- We generate AXI clock in this module.
  constant aclk_period : time := 5 ns;
  signal aclk_int : std_logic := '0';
  
  -- AXI constants/types package axi4_128
  signal direct_slave_m2s : axi4_128_m2s_t := AXI4_128_DEFAULT_M2S;
  signal direct_slave_s2m : axi4_128_s2m_t;
 
  signal mptl_clk_div2 : std_logic := '0';

  -- Configuration/online
  signal mptl_target_gtp_online_i_n : std_logic;
  signal mptl_target_configured_i_n : std_logic;
  signal mptl_online : std_logic;
  signal srst : std_logic;

  -- Host application
  component tb_simple_stim
  port(
    -- AXI
    axi_clk          : in    std_logic;
    direct_slave_m2s : out   axi4_128_m2s_t;
    direct_slave_s2m : in    axi4_128_s2m_t;
    -- Online
    mptl_online      : in    std_logic);
  end component;

begin
  
  aclk_int <= not aclk_int after aclk_period / 2;
  aclk <= aclk_int;
  
  mptl_target_configured_i_n <= '0';
  mptl_target_gtp_online_i_n <= mptl_bridge_gtp_online_n;
  
  -- Generate online signal
  mptl_online <= not(mptl_bridge_gtp_online_n or mptl_target_gtp_online_i_n or mptl_target_configured_i_n);

  srst_p : process(aclk_int)
  begin
    if rising_edge(aclk_int) then
      if (mptl_online = '0') then
        srst <= '1';
      else
        srst <= '0';
      end if;
    end if;
  end process;

  aresetn <= not srst;

  mptl_target_gtp_online_n <= mptl_target_gtp_online_i_n;
  mptl_target_configured_n <= mptl_target_configured_i_n;

  mptl_tx_p <= (others => '-');
  mptl_tx_n <= (others => '-');

  -- MPTL clock div 2
  mptl_clk_out_p : process(mptl_clk_p)
  begin
    if rising_edge(mptl_clk_p) then
      mptl_clk_div2 <= not(mptl_clk_div2);
    end if;
  end process;

  -- mptl_clk_out = half mptl_clk frequency
  mptl_clk_out <= mptl_clk_div2;
  
  gpio_b2t <= (others => '-');


  ds_axi_awaddr   <= direct_slave_m2s.awaddr;
  ds_axi_awlen    <= direct_slave_m2s.awlen;
  ds_axi_awvalid  <= direct_slave_m2s.awvalid;
  ds_axi_awsize   <= direct_slave_m2s.awsize;
  ds_axi_awburst  <= direct_slave_m2s.awburst;
  ds_axi_awcache  <= direct_slave_m2s.awcache;
  ds_axi_awprot   <= direct_slave_m2s.awprot;
  ds_axi_awlock   <= direct_slave_m2s.awlock;
  ds_axi_awqos    <= direct_slave_m2s.awqos;
  ds_axi_awregion <= direct_slave_m2s.awregion;
  ds_axi_wdata    <= direct_slave_m2s.wdata;
  ds_axi_wstrb    <= direct_slave_m2s.wstrb;
  ds_axi_wlast    <= direct_slave_m2s.wlast;
  ds_axi_wvalid   <= direct_slave_m2s.wvalid;
  ds_axi_bready   <= direct_slave_m2s.bready;
  ds_axi_araddr   <= direct_slave_m2s.araddr;
  ds_axi_arlen    <= direct_slave_m2s.arlen;
  ds_axi_arvalid  <= direct_slave_m2s.arvalid;
  ds_axi_arsize   <= direct_slave_m2s.arsize;
  ds_axi_arburst  <= direct_slave_m2s.arburst;
  ds_axi_arcache  <= direct_slave_m2s.arcache;
  ds_axi_arprot   <= direct_slave_m2s.arprot;
  ds_axi_arlock   <= direct_slave_m2s.arlock;
  ds_axi_arqos    <= direct_slave_m2s.arqos;
  ds_axi_arregion <= direct_slave_m2s.arregion;
  ds_axi_rready   <= direct_slave_m2s.rready;

  direct_slave_s2m.awready <= ds_axi_awready;
  direct_slave_s2m.wready  <= ds_axi_wready;
  direct_slave_s2m.bresp   <= ds_axi_bresp;
  direct_slave_s2m.bvalid  <= ds_axi_bvalid;
  direct_slave_s2m.arready <= ds_axi_arready;
  direct_slave_s2m.rdata   <= ds_axi_rdata;
  direct_slave_s2m.rresp   <= ds_axi_rresp;
  direct_slave_s2m.rlast   <= ds_axi_rlast;
  direct_slave_s2m.rvalid  <= ds_axi_rvalid;


  -- Simulation of host application
  tb_simple_stim_i : tb_simple_stim
  port map(
    -- AXI
    axi_clk          => aclk_int,         -- out   std_logic
    direct_slave_m2s => direct_slave_m2s, -- out   axi4_128_m2s_t
    direct_slave_s2m => direct_slave_s2m, -- in    axi4_128_s2m_t
    -- Online
    mptl_online      => mptl_online);     -- in    std_logic
  
end;