# Check that we are using 2015.4 or later
# See Xilinx quality alert at:
# http://www.xilinx.com/support/documentation/customer_notices/xcn15040.pdf
set viv_ver [version -short]
set viv_spl [split $viv_ver .]
set viv_maj [lindex $viv_spl 0]
set viv_min [lindex $viv_spl 1]

if { ($viv_maj == "2015" && $viv_min >= "4") || $viv_maj > "2015" } {
} else {
  puts "Vivado ${viv_ver} is not supported for model ADM-XRC-KU1."
  puts "Please review http://www.xilinx.com/support/documentation/customer_notices/xcn15040.pdf"
  return -1
}

# Get the directory containing this script. Source file paths are relative to this.
set origin_dir [file dirname [file normalize [info script]]]

# Define versions of dependencies in one convenient place
source "$origin_dir/version.tcl"

# Create project
set dir {simple_ocp-ku115_2e}
set part {xcku115-flva1517-2-e}
set ip_part {ku115_2e}
set name {simple_ocp}
create_project -force -part $part $name "$origin_dir/$dir"

# Set the directory path for the new project
set proj_dir [get_property directory [current_project]]

# Set project properties
set project [get_projects $name]
set_property "default_lib" "xil_defaultlib" $project
set_property "part" $part $project
set_property "target_language" "VHDL" $project

set repo_paths [list \
  [file normalize "$origin_dir/../../../../fpga/repo/vivado-2015.4"] \
]
set_property "ip_repo_paths" $repo_paths $project
update_ip_catalog


# Create 'sources_1' fileset (if not found)
if {[string equal [get_filesets -quiet sources_1] ""]} {
  create_fileset -srcset sources_1
}
set source_set [get_filesets sources_1]

# Set source set properties
#set_property "top" "simple" $source_set

# Add HDL source files
set hdl_files [list \
  [file normalize "$origin_dir/../src/adb3_ocp_pkg.vhd"] \
  [file normalize "$origin_dir/../src/adb3_ocp_tb_pkg.vhd"] \
  [file normalize "$origin_dir/../src/tb_simple_ocp_stim.vhd"] \
  [file normalize "$origin_dir/../src/mptl_tgt_ku_ocp_sim.vhd"] \
  [file normalize "$origin_dir/../src/mptl_tgt_ku_ocp_wrap.vhd"] \
  [file normalize "$origin_dir/../src/adb3_ocp_simple_bus_if.vhd"] \
  [file normalize "$origin_dir/../src/simple_ocp.vhd"] \
]
if { [llength $hdl_files] > 0 } {
  add_files -norecurse -fileset $source_set $hdl_files
}


# Create IP
create_ip -vlnv "alphadata:user:mptl_tgt_ku_slv:${hsocp_ver}" -module_name mptl_tgt_ku_slv_ip
set mptl_tgt_ku_slv_ip [get_ips mptl_tgt_ku_slv_ip]

set_property -dict { \
  CONFIG.break_out_ocp {false} \
  CONFIG.enable_bridge_master {true} \
  CONFIG.enable_dma {false} \
  CONFIG.enable_target_master {false} \
  CONFIG.wrapper_reset_signals {true} \
} $mptl_tgt_ku_slv_ip


# Create 'constrs_1' fileset (if not found)
if {[string equal [get_filesets -quiet constrs_1] ""]} {
  create_fileset -constrset constrs_1
}
set constraint_set [get_filesets constrs_1]

# Add constraints files
set constraint_files [list \
  [file normalize "$origin_dir/../src/refclk300.xdc"] \
  [file normalize "$origin_dir/../src/mptl-admxrcku1.xdc"] \
  [file normalize "$origin_dir/../src/bitstream.xdc"] \
  [file normalize "$origin_dir/../src/simple_ocp.xdc"] \
]
add_files -norecurse -fileset $constraint_set $constraint_files
set_property "target_constrs_file" [lindex $constraint_files end] $constraint_set


# Create 'sim_ocp' fileset (if not found)
if {[string equal [get_filesets -quiet sim_ocp] ""]} {
  create_fileset -simset sim_ocp
}
set sim_set [get_filesets sim_ocp]

# Add simulation HDL source files
set sim_files [list \
  [file normalize "$origin_dir/../src/tb_simple_ocp.vhd"] \
]
add_files -norecurse -fileset $sim_set $sim_files

# Set fileset properties
set_property "source_set" "sources_1" $sim_set
set_property "top" "tb_simple" $sim_set
set_property "generic" "sim_ocp=true" $sim_set

set_property "target_simulator" "ModelSim" $project
set_property "modelsim.simulate.runtime" "25 us" $sim_set
set_property "modelsim.simulate.custom_udo" [file normalize "$origin_dir/vsim_wave.udo"] $sim_set

set_property "target_simulator" "XSim" $project
set_property "xsim.simulate.runtime" "25 us" $sim_set
set_property "xsim.tclbatch" [file normalize "$origin_dir/xsim_wave.tcl"] $sim_set

set_property "simulator_language" "Mixed" $project

update_compile_order -fileset $sim_set

# Make sim_ocp active simset
current_fileset -simset [get_filesets sim_ocp]

if {![string equal [get_filesets -quiet sim_1] ""]} {
  delete_fileset sim_1
}


# Create 'synth_1' run (if not found)
if {[string equal [get_runs -quiet synth_1] ""]} {
  create_run -name synth_1 -part $part -flow {Vivado Synthesis 2015} -strategy "Vivado Synthesis Defaults" -constrset constrs_1
} else {
  set_property strategy "Vivado Synthesis Defaults" [get_runs synth_1]
  set_property flow "Vivado Synthesis 2015" [get_runs synth_1]
}
set synth_run [get_runs synth_1]
set_property "needs_refresh" "1" $synth_run
set_property "part" $part $synth_run
current_run -synthesis $synth_run


# Create 'impl_1' run (if not found)
if {[string equal [get_runs -quiet impl_1] ""]} {
  create_run -name impl_1 -part $part -flow {Vivado Implementation 2015} -strategy "Vivado Implementation Defaults" -constrset constrs_1 -parent_run synth_1
} else {
  set_property strategy "Vivado Implementation Defaults" [get_runs impl_1]
  set_property flow "Vivado Implementation 2015" [get_runs impl_1]
}
set impl_run [get_runs impl_1]
set_property "needs_refresh" "1" $impl_run
set_property "part" $part $impl_run
set_property -name {steps.route_design.args.more options} -value -timing_summary -objects $impl_run
current_run -implementation $impl_run

close_project

puts "INFO: Created project '$dir/$name'"
