--==============================================================================
--
--     File Name: adb3_ocp_simple_bus_if.vhd
--         Model: Synthesisable
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 OCP simple bus interface
--
--   Description: ADB3 OCP simple bus interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2013 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 05 May 2010: Initial design
--                0.2, KDC, 12 Mar 2012:
--                     - Added reset to entity outputs on ocp_rst
--                     - Added initial values to signals that are reset
--                0.3, KDC, 18 Jun 2013:
--                     - Converted block to use state machine.
--                     - Simple interface read transactions are now more
--                       efficient.
--                0.4, KDC, 30 Dec 2014:
--                     - Added missing don't care initial values to rw_addr and
--                       write_data_sr signals as they are primary outputs
--                     - Added missing don't care initial values to read_data_sr
--                       and read_tag signals as they are primary outputs
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;
use ieee.math_real.all;       -- Used for CEIL and LOG2 functions only

library work;
use work.adb3_ocp.all;

entity adb3_ocp_simple_bus_if is
generic(
  addr_width   : natural range 1 to ADB3_OCP_ADDR_WIDTH;
  data_width   : natural range 8 to ADB3_OCP_DATA_WIDTH;
  read_latency : natural := 1);
port(
  -- OCP interface
  ocp_rst : in    std_logic;
  ocp_clk : in    std_logic;
  ocp_m2s : in    adb3_ocp_m2sT;
  ocp_s2m : out   adb3_ocp_s2mT;
  -- Simple bus interface
  d       : out   std_logic_vector(data_width-1 downto 0);
  q       : in    std_logic_vector(data_width-1 downto 0);
  a       : out   std_logic_vector(addr_width-1 downto 0);  -- Byte address
  w       : out   std_logic;
  r       : out   std_logic;
  we      : out   std_logic_vector((data_width/8)-1 downto 0));
end;

architecture syn of adb3_ocp_simple_bus_if is

  -- We must ensure that when WORD_WIDTH = 1, WORD_CNT_WIDTH = 1
  -- round(A/B) == trunc((A+B/2)/B)
  -- WORD_CNT_WIDTH = 1 + log2(round(WORD_WIDTH/2)) = 1 + log2((WORD_WIDTH+1)/2)
  constant WORD_WIDTH     : natural := ADB3_OCP_DATA_WIDTH/data_width;
  constant WORD_CNT_WIDTH : natural := 1 + NATURAL(CEIL(LOG2(REAL((WORD_WIDTH+1)/2))));

  subtype word_cnt_s is std_logic_vector(WORD_CNT_WIDTH-1 downto 0);
  constant WORD_CNT_MAX : word_cnt_s   := std_logic_vector(to_unsigned(WORD_WIDTH-1,WORD_CNT_WIDTH));

  subtype local_addr_s is std_logic_vector(addr_width-1 downto 0);
  subtype local_data_s is std_logic_vector(data_width-1 downto 0);
  subtype local_be_s   is std_logic_vector((data_width/8)-1 downto 0);
  subtype local_lat_s  is std_logic_vector(read_latency downto 0);

  constant LOCAL_DATA_ZERO : local_data_s := (others => '0');
  constant LOCAL_BE_ZERO   : local_be_s   := (others => '0');

  -- simple bus interface => ocp_s2m signals
  signal cmd_accept   : std_logic       := ADB3_OCP_DEFAULT_S2M.CmdAccept;  -- ocp_s2m.CmdAccept  (output/reset)
  signal data_accept  : std_logic       := ADB3_OCP_DEFAULT_S2M.DataAccept; -- ocp_s2m.DataAccept (output/reset)
  signal resp         : ocp_RespT       := ADB3_OCP_DEFAULT_S2M.Resp;       -- ocp_s2m.Resp       (output/reset)
  signal read_data_sr : adb3_ocp_data_s := ADB3_OCP_DEFAULT_S2M.Data;       -- ocp_s2m.Data       (output)
  signal read_tag     : adb3_ocp_tag_s  := ADB3_OCP_DEFAULT_S2M.Tag;        -- ocp_s2m.Tag        (output)

  signal n_cmd_accept   : std_logic;
  signal n_data_accept  : std_logic;
  signal n_resp         : ocp_RespT;
  signal n_read_data_sr : adb3_ocp_data_s;
  signal n_read_tag     : adb3_ocp_tag_s;

  -- ocp_m2s => simple bus interface signals
  signal rw_addr       : local_addr_s    := (others => '-'); -- a  (output)
  signal write_data_sr : adb3_ocp_data_s := (others => '-'); -- d  (output)
  signal write_we_sr   : adb3_ocp_be_s   := (others => '0'); -- we (output/reset)
  signal write_data_v  : std_logic := '0';                   -- w  (output/reset)
  signal read_cmd_v    : std_logic := '0';                   -- r  (output/reset)

  signal n_rw_addr       : local_addr_s;
  signal n_write_data_sr : adb3_ocp_data_s;
  signal n_write_we_sr   : adb3_ocp_be_s;
  signal n_write_data_v  : std_logic;
  signal n_read_cmd_v    : std_logic;

  -- OCP state machine signals
  type ocp_state_t is (
    ocp_s_waiting_cmd,
    ocp_s_waiting_data,
    ocp_s_writing,
    ocp_s_waiting_resp,
    ocp_s_reading);

  signal ocp_state   : ocp_state_t := ocp_s_waiting_cmd; -- (reset)
  signal n_ocp_state : ocp_state_t;

  signal read_latency_1hot, n_read_latency_1hot : local_lat_s;
  signal write_addr,        n_write_addr        : local_addr_s;

  signal burst_cnt, n_burst_cnt : adb3_ocp_burst_s;
  signal word_cnt,  n_word_cnt  : word_cnt_s;
  signal cmd_cnt,   n_cmd_cnt   : word_cnt_s;

begin

  assert not(read_latency < 1)
    report "(adb3_ocp_simple_bus_if): Mimium read_latency is 1 clock cycle"
    severity error;

  assert not((data_width /= 128)and(data_width /= 64)and(data_width /= 32)and(data_width /= 16)and(data_width /= 8))
    report "(adb3_ocp_simple_bus_if): Generic data_width must be 8/16/32/64/128 bits"
    severity error;

  gen_ocp_state_p : process(ocp_rst, ocp_clk)
  begin
    if (ocp_rst = '1') then
      -- Simple bus interface
      write_we_sr  <= (others => '0');  -- we
      write_data_v <= '0';              -- w
      read_cmd_v   <= '0';              -- r

      -- Slave port
      cmd_accept  <= ADB3_OCP_DEFAULT_S2M.CmdAccept;
      data_accept <= ADB3_OCP_DEFAULT_S2M.DataAccept;
      resp        <= ADB3_OCP_DEFAULT_S2M.Resp;

      ocp_state <= ocp_s_waiting_cmd;
    elsif rising_edge(ocp_clk) then
      -- Simple bus interface
      write_we_sr  <= n_write_we_sr;   -- we
      write_data_v <= n_write_data_v;  -- w
      read_cmd_v   <= n_read_cmd_v;    -- r

      -- Slave port
      cmd_accept  <= n_cmd_accept;
      data_accept <= n_data_accept;
      resp        <= n_resp;

      ocp_state <= n_ocp_state;
    end if;

    if rising_edge(ocp_clk) then
      -- Simple bus interface
      rw_addr       <= n_rw_addr;       -- a
      write_data_sr <= n_write_data_sr; -- d

      -- Slave port
      resp         <= n_resp;
      read_data_sr <= n_read_data_sr;
      read_tag     <= n_read_tag;

      -- Internal signals
      read_latency_1hot <= n_read_latency_1hot;
      write_addr        <= n_write_addr;
      burst_cnt         <= n_burst_cnt;
      word_cnt          <= n_word_cnt;
      cmd_cnt           <= n_cmd_cnt;
    end if;
  end process;

  gen_ocp_n_state_p : process(ocp_state, ocp_m2s, q,
                              rw_addr, write_data_sr, write_we_sr, write_data_v, read_cmd_v,
                              cmd_accept, data_accept, resp, read_data_sr, read_tag,
                              read_latency_1hot, write_addr, burst_cnt, word_cnt, cmd_cnt)
  begin
    -- Define signal defaults
    n_ocp_state <= ocp_state;

    n_rw_addr       <= (others => '-'); -- a
    n_write_data_sr <= (others => '-'); -- d
    n_write_we_sr   <= (others => '0'); -- we
    n_write_data_v  <= '0';             -- w
    n_read_cmd_v    <= '0';             -- r

    n_cmd_accept   <= ADB3_OCP_DEFAULT_S2M.CmdAccept;
    n_data_accept  <= ADB3_OCP_DEFAULT_S2M.DataAccept;
    n_resp         <= ADB3_OCP_DEFAULT_S2M.Resp;
    n_read_data_sr <= ADB3_OCP_DEFAULT_S2M.Data;
    n_read_tag     <= ADB3_OCP_DEFAULT_S2M.Tag;

    n_read_latency_1hot <= (others => '-');
    n_write_addr        <= (others => '-');
    n_burst_cnt         <= (others => '-');
    n_word_cnt          <= (others => '-');
    n_cmd_cnt           <= (others => '-');

    case ocp_state is
      -- Wait for valid command on slave i/f and accept
      when ocp_s_waiting_cmd =>
        if (OR_reduce(ocp_m2s.Cmd) = '1')and(cmd_accept = '1') then
          case (ocp_m2s.Cmd) is
            when OCP_CMD_WRITE =>
              n_burst_cnt  <= ocp_m2s.BurstLength;
              n_write_addr <= ocp_m2s.Addr(addr_width-1 downto 0);
              n_word_cnt   <= (others => '0');

              n_data_accept <= '1';  -- Ready to accept write data on slave i/f
              n_ocp_state <= ocp_s_waiting_data;
            when OCP_CMD_READ =>
              n_read_tag <= ocp_m2s.Tag;

              n_rw_addr    <= ocp_m2s.Addr(addr_width-1 downto 0); -- a
              n_read_cmd_v <= '1';                                 -- r

              n_burst_cnt  <= ocp_m2s.BurstLength;
              n_cmd_cnt    <= (others => '0');
              n_word_cnt   <= (others => '0');

              n_read_latency_1hot(read_latency downto 1) <= (others => '0');
              n_read_latency_1hot(0)                     <= '1';
              n_ocp_state <= ocp_s_waiting_resp;
            when others =>
              n_ocp_state <= ocp_s_waiting_cmd;
          end case;
        else
          n_cmd_accept <= '1';
        end if;

      -- Wait for valid data on slave i/f
      when ocp_s_waiting_data =>
        if (ocp_m2s.DataValid = '1') then
          n_rw_addr       <= write_addr;         -- a
          n_write_data_sr <= ocp_m2s.Data;       -- d
          n_write_we_sr   <= ocp_m2s.DataByteEn; -- we
          n_write_data_v  <= '1';                -- w

          if (WORD_WIDTH > 1) then
            n_burst_cnt  <= burst_cnt - 1;
            n_write_addr <= write_addr + data_width/8;
            n_word_cnt   <= word_cnt + 1;

            n_ocp_state <= ocp_s_writing;
          else
            if (OR_reduce(burst_cnt(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '0') then
              n_cmd_accept <= '1';
              n_ocp_state <= ocp_s_waiting_cmd;
            else
              n_burst_cnt  <= burst_cnt - 1;
              n_write_addr <= write_addr + data_width/8;
              n_word_cnt   <= word_cnt;

              n_data_accept <= '1';
            end if;
          end if;
        else
          n_burst_cnt  <= burst_cnt;
          n_write_addr <= write_addr;
          n_word_cnt   <= word_cnt;

          n_data_accept <= data_accept;
        end if;

      -- Wait for all data writes on simple i/f
      when ocp_s_writing =>
        n_rw_addr       <= write_addr;                                                                   -- a
        n_write_data_sr <= LOCAL_DATA_ZERO & write_data_sr(ADB3_OCP_DATA_WIDTH-1 downto data_width);     -- d
        n_write_we_sr   <= LOCAL_BE_ZERO & write_we_sr((ADB3_OCP_DATA_WIDTH/8)-1 downto (data_width/8)); -- we
        n_write_data_v  <= '1';                                                                          -- w

        -- Wait until count of WORD_WIDTH-1
        if (word_cnt = WORD_CNT_MAX) then
          if (OR_reduce(burst_cnt) = '0') then
            n_cmd_accept <= '1';
            n_ocp_state <= ocp_s_waiting_cmd;
          else
            n_burst_cnt  <= burst_cnt;
            n_write_addr <= write_addr + data_width/8;
            n_word_cnt   <= (others => '0');

            n_data_accept <= '1';
            n_ocp_state <= ocp_s_waiting_data;
          end if;
        else
          n_burst_cnt  <= burst_cnt;
          n_write_addr <= write_addr + data_width/8;
          n_word_cnt   <= word_cnt + 1;
        end if;

      -- Wait for all valid responses on simple i/f
      when ocp_s_waiting_resp =>
        n_read_tag  <= read_tag;
        n_burst_cnt <= burst_cnt;

        if (cmd_cnt = WORD_CNT_MAX) then
          n_cmd_cnt    <= cmd_cnt;
          n_rw_addr    <= rw_addr;
        else
          n_cmd_cnt    <= cmd_cnt + 1;
          n_rw_addr    <= rw_addr + data_width/8; -- a
          n_read_cmd_v <= '1';                    -- r
        end if;

        if (read_latency_1hot(read_latency) = '1') then
          n_read_latency_1hot <= read_latency_1hot;

          if (WORD_WIDTH > 1) then
            n_read_data_sr <= q & read_data_sr(ADB3_OCP_DATA_WIDTH-1 downto data_width);
            n_word_cnt     <= word_cnt + 1;
          else
            n_read_data_sr <= q;
            n_word_cnt     <= word_cnt;
          end if;

          -- Wait until count of WORD_WIDTH-1
          if (word_cnt = WORD_CNT_MAX) then
            n_resp <= OCP_RESP_VALID;
            n_ocp_state <= ocp_s_reading;
          end if;
        else
          n_read_data_sr      <= read_data_sr;
          n_word_cnt          <= word_cnt;
          n_read_latency_1hot <= read_latency_1hot(read_latency-1 downto 0) & '0';
        end if;


      -- Wait for response accept on slave i/f
      when ocp_s_reading =>
        if (ocp_m2s.RespAccept = '1') then
          if (OR_reduce(burst_cnt(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '0') then
            n_cmd_accept <= '1';
            n_ocp_state <= ocp_s_waiting_cmd;
          else
            n_read_tag <= read_tag;

            n_rw_addr    <= rw_addr + data_width/8; -- a
            n_read_cmd_v <= '1';                    -- r

            n_burst_cnt <= burst_cnt - 1;
            n_word_cnt  <= (others => '0');
            n_cmd_cnt   <= (others => '0');

            n_read_latency_1hot(read_latency downto 1) <= (others => '0');
            n_read_latency_1hot(0) <= '1';
            n_ocp_state <= ocp_s_waiting_resp;
          end if;
        else
          n_rw_addr   <= rw_addr;
          n_burst_cnt <= burst_cnt;

          n_resp         <= resp;
          n_read_data_sr <= read_data_sr;
          n_read_tag     <= read_tag;
        end if;

      when others =>
        n_cmd_accept <= '1';  -- Ready to accept command on slave i/f
        n_ocp_state  <= ocp_s_waiting_cmd;
    end case;
  end process;


  a  <= rw_addr;
  d  <= write_data_sr(data_width-1 downto 0);
  we <= write_we_sr((data_width/8)-1 downto 0);
  w  <= write_data_v;
  r  <= read_cmd_v;

  ocp_s2m.CmdAccept  <= cmd_accept;
  ocp_s2m.DataAccept <= data_accept;
  ocp_s2m.Resp       <= resp;
  ocp_s2m.Data       <= read_data_sr;
  ocp_s2m.Tag        <= read_tag;

end;