/*
** simple_ocp.cpp
**
** Test program for the Simple (OCP) FPGA Design (simple_ocp)
**
** Notes
** =====
**
** 1. Uses the ADMXRC3UT library for opening the device. ADMXRC3UT is a set of
**    functions that contains commonly-used code in example / test programs,
**    but it is not part of the ADMXRC3 API.
**
**    ADMXRC3UT_OpenByIndex is similar to ADMXRC3_OpenEx from the ADMXRC3 API,
**    except that it also performs a check to see if the device is in an error
**    state and attempts to clear device errors if it is.
**
**    ADMXRC3UT_OpenBySerialNumber is similar to ADMXRC3_OpenEx from the
**    ADMXRC3 API, except that it searches for the device with a particular
**    serial number rather than opening by index. Like ADMXRC3UT_OpenByIndex,
**    it also checks to see if the device is in an error state and attempts to
**    clear device errors if it is.
**
** 2. This source file contains code for doing the actual test. In VxWorks,
**    the admxrc3SimpleOcpIndex & admxrc3SimpleOcpSN functions are entry points
**    that can be called from the VxWorks shell.
*/

#ifdef _WIN32

# include <windows.h>
# include <tchar.h>

typedef UINT8 uint8_t;
typedef UINT32 uint32_t;

#elif defined(__linux)

# include <stdint.h>

# define _T(x) x
# define _tmain main
# define _tcslen(x) strlen(x)
typedef char TCHAR;
# define _gettc getc
# define _tscanf_s scanf

#elif defined(__VXWORKS__) || defined(__vxworks)

# include <vxWorks.h>

# if _WRS_VXWORKS_MAJOR >= 6
#   include <stdint.h>
# endif

# define _T(x) x
# define _tmain main
# define _tcslen(x) strlen(x)
typedef char TCHAR;
# define _gettc getc
# define _tscanf_s scanf

#endif

#include <cstdio>
#include <cassert>
#include <cstring>
#include <string>
#include <iostream>
#include <algorithm>

#include <exappendian.h>

#include <admxrc3.h>
#include <admxrc3ut.h>

#include "simple_ocp.h"

using namespace AlphaData::AppFramework;

#ifndef FALSE
# define FALSE 0
#endif
#ifndef TRUE
# define TRUE 1
#endif

#ifdef _UNICODE
typedef std::wstring SType;
typedef std::wstringstream SStreamType;
#define SSEMPTY(ss) ((ss).str(std::basic_string<wchar_t>()))
#else
typedef std::string SType;
typedef std::stringstream SStreamType;
#define SSEMPTY(ss) ((ss).str(std::string()))
#endif

#define MAX_NUM_DMA_ENGINE (32)

CExAppMessageDisplay g_simpleOcpMessageDisplay;

#ifdef _WIN32
static const TCHAR* g_pPathSeparator = _T("\\");
#else
static const TCHAR* g_pPathSeparator = _T("/");
#endif

static const TCHAR* g_pDesignName = _T("simple_ocp");

static int /* ExitCode */
admxrc3SimpleOcp(
  ADMXRC3_HANDLE hDevice,
  const TCHAR* pBitPath,
  const TCHAR* pBitFile)
{
  const size_t mapSize = sizeof(uint32_t); // The nibble-reversal register in the FPGA is 32 bits wide
  volatile uint32_t* pNibbleRevRegister = NULL; // Pointer to the nibble-reversal register, mapped into our virtual address space
  CExAppMessageDisplay* pMessageDisplay = &g_simpleOcpMessageDisplay;
  int exitCode = ExitCodeSuccess;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFOEX cardInfo;
  SStreamType resultSS;
  SType sBitFile;
  int nConverted;

  // Get card information; needed below.
  status = ADMXRC3_GetCardInfoEx(hDevice, &cardInfo);
  if (ADMXRC3_SUCCESS != status) {
    SStreamType ss;
    ss << _T("Failed to get card information: ") << ADMXRC3_GetStatusString(status, TRUE);
    pMessageDisplay->AddError(ss.str());
    exitCode = ExitCodeGetCardInfoFailed;
    goto done;
  }

  // Check that we are running on a supported model
  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRCKU1:
    break;

  default:
    SStreamType ss;
    ss << _T("This program must be run on one of the following models: ADM-XRC-KU1. The device in use is: ") << ADMXRC3UT_GetModelName(cardInfo.Model);
    pMessageDisplay->AddError(ss.str());
    exitCode = ExitCodeUnsupportedModel;
    goto done;
  }

  // Map the BlockRAMs, addressable by Direct Slave, into our virtual address space
  status = ADMXRC3_MapWindow(hDevice, 0, 0, mapSize, (void**)&pNibbleRevRegister);
  if (ADMXRC3_SUCCESS != status) {
    SStreamType ss;
    ss << _T("Failed to map FPGA BlockRAMs (0x") << std::hex << mapSize << _T(" bytes ): ") << ADMXRC3_GetStatusString(status, TRUE);
    pMessageDisplay->AddError(ss.str());
    exitCode = ExitCodeMapWindowFailed;
    goto done;
  }

  // Configure the target FPGA with the .bit file for the "simple_ocp" FPGA design.
  if (pBitFile == NULL) {
    SStreamType bitFile;
    //
    // Construct a .bit filename of the form:
    //   pBitPath + "/<design>-<device>[_<speed>][_<stepping>]/<design>.bit"
    //
    // Examples:
    //   pBitPath + "/simple_ocp-ku115_2e/simple_ocp.bit"
    //   pBitPath + "/simple_ocp-ku060_2e_es2/simple_ocp.bit"
    //
    if (pBitPath == NULL) {
      pMessageDisplay->AddError(_T("'pBitFile' and 'pBitPath' parameters cannot both be NULL."));
      exitCode = ExitCodeBadBitPath;
      goto done;
    }
    ADMXRC3_FPGA_INFO fpgaInfo;
    status = ADMXRC3_GetFpgaInfo(hDevice, 0, &fpgaInfo);
    if (ADMXRC3_SUCCESS != status) {
      SStreamType ss;
      ss << _T("Failed to get FPGA information: ") << ADMXRC3_GetStatusString(status, TRUE);
      pMessageDisplay->AddError(ss.str());
      exitCode = ExitCodeFpgaInfoFailed;
      goto done;
    }
    bitFile << pBitPath << g_pPathSeparator << g_pDesignName << _T("-");
    switch (fpgaInfo.DeviceCode) {
    case ADMXRC3_FPGA_KU060:
      bitFile << _T("ku060");
      break;

    case ADMXRC3_FPGA_KU115:
      bitFile << _T("ku115");
      break;

    default:
      SStreamType ss;
      ss << _T("Unrecognized FPGA device code ") << fpgaInfo.DeviceCode << _T("(0x") << std::hex << fpgaInfo.DeviceCode << _T("). Cannot construct .bit filename automatically.");
      pMessageDisplay->AddError(ss.str());
      exitCode = ExitCodeUnrecognizedFpga;
      goto done;
    }
    if (_tcslen(fpgaInfo.SpeedGrade) > 0) {
      bitFile << _T("_") << fpgaInfo.SpeedGrade;
    }
    if (_tcslen(fpgaInfo.Stepping) > 0) {
      bitFile << _T("_") << fpgaInfo.Stepping;
    }
    bitFile << g_pPathSeparator << g_pDesignName << _T(".bit");
    sBitFile = bitFile.str();
    std::transform(sBitFile.begin(), sBitFile.end(), sBitFile.begin(), ::tolower);
    pBitFile = sBitFile.c_str();
  }
  status = ADMXRC3_ConfigureFromFile(hDevice, 0, 0, pBitFile);
  if (ADMXRC3_SUCCESS != status) {
    SStreamType ss;
    ss << _T("Failed to configure target FPGA 0 with '") << pBitFile << ("': ") << ADMXRC3_GetStatusString(status, TRUE);
    pMessageDisplay->AddError(ss.str());
    exitCode = ExitCodeConfigureFpgaFailed;
    goto done;
  }

  pMessageDisplay->FormatInfo(_T("========================="));
  pMessageDisplay->FormatInfo(_T("Enter values for I/O"));
#ifdef _WIN32
  pMessageDisplay->FormatInfo(_T("(CTRL-Z or enter 0 to exit)"));
#else
  pMessageDisplay->FormatInfo(_T("(CTRL-D or enter 0 to exit)"));
#endif
  pMessageDisplay->FormatInfo(_T("========================="));
  
  do {
    unsigned long dataOut;
    uint32_t dataIn;
    int c;

    nConverted = _tscanf_s(_T("%lX"), &dataOut);
    if (nConverted >= 1) {
      if (dataOut == 0) {
        break;
      }
      *pNibbleRevRegister = cpu_to_le32((uint32_t)dataOut);
      dataIn = *pNibbleRevRegister;
      dataIn = le32_to_cpu(dataIn);
      pMessageDisplay->FormatInfo(_T("OUT = 0x%08lX, IN = 0x%08lX"), dataOut, (unsigned long)dataIn);
    } else {
      if (0 == nConverted) {
        /* Skip to next line */
        do {
          c = _gettc(stdin);
          if (EOF == c) {
            goto done;
          }
        } while (c != _T('\n'));
      }
    }
  } while (EOF != nConverted);

done:
  if (NULL != pNibbleRevRegister) {
    ADMXRC3_UnmapWindow(hDevice, (void*)pNibbleRevRegister);
    pNibbleRevRegister = NULL;
  }

  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
    hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return exitCode;
}

int /* ExitCode */
admxrc3SimpleOcpIndex(
  const TCHAR* pBitPath,
  const TCHAR* pBitFile,
  unsigned int index)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;

  status = ADMXRC3UT_OpenByIndex(index, FALSE, TRUE, 0, &hDevice);
  if (ADMXRC3_SUCCESS != status) {
    g_simpleOcpMessageDisplay.FormatError(
      _T("Failed to open device with index %u: %s"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    return ExitCodeCannotOpenIndex;
  }

  return admxrc3SimpleOcp(
    hDevice,
    pBitPath,
    pBitFile);
}

int /* ExitCode */
admxrc3SimpleOcpSN(
  const TCHAR* pBitPath,
  const TCHAR* pBitFile,
  uint32_t serialNumber)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;

  status = ADMXRC3UT_OpenBySerialNumber(serialNumber, FALSE, TRUE, 0, &hDevice);
  if (ADMXRC3_SUCCESS != status) {
    g_simpleOcpMessageDisplay.FormatError(
      _T("Failed to open device with serial number %lu: %s"),
      (unsigned long)serialNumber, ADMXRC3_GetStatusString(status, TRUE));
    return ExitCodeCannotOpenSerial;
  }

  return admxrc3SimpleOcp(
    hDevice,
    pBitPath,
    pBitFile);
}
