library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.ddr4_sdram_model.all;
use work.ddr4_sdram_model_int.all;

entity ddr4_sdram_die is
  generic(
    -- String prepended to messages to give some indication where the message comes from (for multi-die chips)
    model_name          : in    string;
    -- Selects the DDR4 SDRAM part to be modeled.
    part                : in    part_t;
    -- The value returned (in all DQ bits) of a memory location that has not previously been written or initialized.
    read_undef_val      : in    std_logic := 'U';
    -- Assumes that simulation begins from power on, so that power on timing used; other uses reset timing.
    init_from_power_on  : in    boolean := false;
    -- Demotes warnings about deviations from recommended initialization sequence to informational messages.
    relaxed_init_timing : in    boolean := true 
  );
  port(
    reset_n             : in    std_logic;
    ck_t                : in    std_logic;
    ck_c                : in    std_logic;
    ten                 : in    std_logic := '0';
    
    cke                 : in    std_logic;
    cs_n                : in    std_logic;
    odt                 : in    std_logic;
    act_n               : in    std_logic;
    -- A16 functions as RAS_n when ACT_n is high.
    -- A15 functions as CAS_n when ACT_n is high.
    -- A14 functions as WE_n when ACT_n is high.
    a                   : in    std_logic_vector(part.pinout.a_width - 1 downto 0);
    ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0);
    bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0);
    par                 : in    std_logic;
    
    -- x4: DM_DBI_n ignored.
    -- x8 & x16: DBI on same pins as DM and enabled via MR5.
    -- x8 only: DM_DBI_n[0] & TDQS_t are on same pin; TDQS function enabled via MR1 (takes precedence over DM & DBI functions).
    dm_dbi_n            : inout std_logic_vector(part.pinout.dm_dbi_width - 1 downto 0) := (others => 'U');
    -- x4 & x16: TDQS_c ignored.
    -- x8 only: TDQS_c function enabled via MR1; connect if supported by part and used in design.
    tdqs_c              : in    std_logic := 'U';
    dq                  : inout std_logic_vector(4*part.pinout.dq_nibbles - 1 downto 0);
    dqs_t               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
    dqs_c               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
    
    alert_n             : out   std_logic
  );
end entity;

architecture behav of ddr4_sdram_die is

  constant reset_qualified : boolean := true; -- RESET_n must always be valid
  signal reset_X01_n : std_logic;
  signal reset_valid : boolean;
  
  signal ck_qualified : boolean := false;
  signal ck_t_valid : boolean;
  signal ck_c_valid : boolean;
  signal ck_valid : boolean;

  signal cke_X01 : std_logic;
  signal cke_prev_X01 : std_logic;
  signal cke_qualified : boolean;
  
  signal cs_X01_n : std_logic;
  signal cs_qualified : boolean;
  
  signal act_X01_n : std_logic;
  signal act_qualified : boolean;
  
  signal a_X01 : std_logic_vector(part.pinout.a_width - 1 downto 0);
  signal a_01 : std_logic_vector(part.pinout.a_width - 1 downto 0);
  alias a_row_X01 : std_logic_vector(part.pinout.row_width - 1 downto 0) is a_X01(part.pinout.row_width - 1 downto 0);
  signal a_row_qualified : boolean; -- Row address must be valid
  signal a_col_X01 : std_logic_vector(part.pinout.col_width - 1 downto 0);
  signal a_col_01 : std_logic_vector(part.pinout.col_width - 1 downto 0);
  signal a_col_qualified : boolean; -- A[13,11,9:0] (column address) must be valid
  alias a_bc_X01 : std_logic is a_X01(12);
  signal a_bc_qualified : boolean; -- A[12] (burst chop) must be valid
  alias a_ap_X01 : std_logic is a_X01(10);
  signal a_ap_qualified : boolean; -- A[10] (autoprecharge) must be valid
  alias a_mrs_X01 : std_logic_vector(13 downto 0) is a_X01(13 downto 0);
  signal a_mrs_qualified : boolean; -- A[13:0] must be valid for MRS command
  
  alias ras_X01_n : std_logic is a(16);
  alias cas_X01_n : std_logic is a(15);
  alias we_X01_n : std_logic is a(14);
  signal cmd_qualified : boolean;
  
  signal bg_X01 : std_logic_vector(part.pinout.bg_width - 1 downto 0);
  signal bg_01 : std_logic_vector(part.pinout.bg_width - 1 downto 0);
  signal ba_X01 : std_logic_vector(part.pinout.ba_width - 1 downto 0);
  signal ba_01 : std_logic_vector(part.pinout.ba_width - 1 downto 0);
  signal bank_qualified : boolean;
  
  signal mr0, mr1, mr2, mr3, mr4, mr5, mr6 : mode_reg_t;
  signal init_done : boolean;
  signal cl : cl_t;
  signal al : al_t;
  signal cwl : cwl_t;
  signal bl : burst_length_t;
  signal bt : burst_type_t;
  signal read_preamble : read_preamble_t;
  signal rd_dbi_enabled : boolean;
  signal wr_dbi_enabled : boolean;
  
  signal power_state : power_state_t;
  
  -- Bank states
  constant num_bank_group : natural := 2**part.pinout.bg_width;
  constant num_bank_per_group : natural := 2**part.pinout.ba_width;  
  constant num_bank_total : natural := num_bank_per_group * num_bank_group;
  signal bank_states : bank_state_array_t(0 to num_bank_total - 1);
  -- Currently activated rows for banks
  signal activated_rows : bank_row_array_t(0 to num_bank_total - 1);
  
  
begin

  -- Sanity checks on pinout
  assert part.pinout.a_width >= 17 severity failure;
  assert part.pinout.row_width <= part.pinout.a_width severity failure;
  assert part.pinout.col_width <= 12 severity failure;
  assert part.pinout.col_width <= part.pinout.a_width severity failure;
  assert part.pinout.bg_width >= 1 severity failure;
  assert part.pinout.ba_width >= 2 severity failure;
  
  -- Sanity checks on timing parameters
  assert part.timing.operational.t_ih >= 1 ps severity failure;
  assert part.timing.operational.t_is >= 1 ps severity failure;
  
  
  --
  -- TO DO
  --
  
  dq <= (others => 'Z');
  dqs_t <= (others => 'Z');
  dqs_c <= (others => 'Z');
  alert_n <= '1';

  
  --
  -- Make "safe" versions of inputs
  --
  
  reset_X01_n <= make_X01(reset_n);
  reset_valid <= not is_X(reset_n);

  ck_t_valid <= not is_X(ck_t);
  ck_c_valid <= not is_X(ck_c);
  ck_valid <= ck_t_valid and ck_c_valid;

  cke_X01 <= make_X01(cke);

  cs_X01_n <= make_X01(cs_n);
  act_X01_n <= make_X01(act_n);

  a_X01 <= make_X01(a);
  -- ras_X01_n, cas_X01_n, we_X01_n are aliases of particular bits of a_X01
  a_01 <= make_01(a);

  bg_X01 <= make_X01(bg);
  bg_01 <= make_01(bg);

  ba_X01 <= make_X01(ba);
  ba_01 <= make_01(ba);

  
  --
  -- Register CKE
  --
  
  register_cke : process(reset_n, ck_t)
  begin
    if reset_n = '0' then
      cke_prev_X01 <= 'X';
    elsif rising_edge(ck_t) then
      cke_prev_X01 <= cke_X01;
    end if;
  end process;
  
  
  --
  -- Check RESET_n validity
  --
  
  check_reset_valid : process
  begin
    loop
      wait until reset_qualified and not reset_valid;
      report "ddr4_sdram_chip: RESET_n invalid" severity error;
      wait until not (reset_qualified and not reset_valid);
    end loop;
  end process;
  
  --
  -- Check CK_t, CK_c validity
  --
  
  ck_qualified <= (cke_X01 = '1');
  
  check_ck_t_valid : process
  begin
    loop
      wait until ck_qualified and not ck_t_valid;
      report "ddr4_sdram_chip: CK_t invalid while CKE high" severity error;
      wait until not (ck_qualified and not ck_t_valid);
    end loop;
  end process;
  
  check_ck_c_valid : process
  begin
    loop
      wait until ck_qualified and not ck_c_valid;
      report "ddr4_sdram_chip: CK_c invalid while CKE high" severity error;
      wait until not (ck_qualified and not ck_c_valid);
    end loop;
  end process;
  
  --
  -- Check CKE validity & timing w.r.t. CK_t
  --
  
  cke_qualified <= (reset_X01_n /= '0');
  
  check_cke_timing : ddr4_sdram_su_h_check_bit
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "CKE",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => cke_qualified,
      sig => cke_X01
    );
    
  --
  -- Check CS_n validity & timing w.r.t. CK_t
  --
  
  cs_qualified <= cke_X01 = '1';
  
  check_cs_timing : ddr4_sdram_su_h_check_bit
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "CS_n",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => cs_qualified,
      sig => cs_X01_n
    );
    
  --
  -- Check ACT_n validity & timing w.r.t. CK_t
  --
  
  act_qualified <= cs_X01_n = '0' and cke_X01 = '1';
  
  check_act_timing : ddr4_sdram_su_h_check_bit
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "ACT_n",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => act_qualified,
      sig => act_X01_n
    );
    
  --
  -- Check RAS_n, CAS_n & WE_n validity & timing w.r.t. CK_t
  --
  
  cmd_qualified <= act_X01_n = '1' and cs_X01_n = '0' and cke_X01 = '1'; -- Any command except ACT, SRE, SRX, PDE, PDX
  
  check_ras_timing : ddr4_sdram_su_h_check_bit
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "RAS_n",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => cmd_qualified,
      sig => ras_X01_n
    );
    
  check_cas_timing : ddr4_sdram_su_h_check_bit
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "RAS_n",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => cmd_qualified,
      sig => cas_X01_n
    );
    
  check_we_timing : ddr4_sdram_su_h_check_bit
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "WE_n",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => cmd_qualified,
      sig => we_X01_n
    );
    
  --
  -- Check BG & BA validity & timing w.r.t. CK_t
  --
  
  bank_qualified <= cs_X01_n = '0' and cke_X01 = '1' and
                    ((act_X01_n = '1' and ras_X01_n = '0' and cas_X01_n = '0' and we_X01_n = '0') or                     -- MRS
                     (act_X01_n = '1' and ras_X01_n = '0' and cas_X01_n = '1' and we_X01_n = '0' and a_X01(10) = '0') or -- PRE
                     (act_X01_n = '1' and ras_X01_n = '1' and cas_X01_n = '0') or                                        -- RD or WR
                     (act_X01_n = '0'));                                                                                 -- ACT
                   
  check_bg_timing : ddr4_sdram_su_h_check_vec
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "BG",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => bank_qualified,
      sig => bg_X01
    );
    
  check_ba_timing : ddr4_sdram_su_h_check_vec
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "BA",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => bank_qualified,
      sig => ba_X01
    );
    
  --
  -- Check A validity & timing w.r.t. CK_t
  --
  
  -- Row address bits qualification
  a_row_qualified <= cs_X01_n = '0' and cke_X01 = '1' and act_X01_n = '0'; -- ACT

  check_a_row_timing : ddr4_sdram_su_h_check_vec
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "A",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => a_row_qualified,
      sig => a_row_X01
    );
    
  -- Column address bits qualification
  a_col_qualified <= cs_X01_n = '0' and cke_X01 = '1' and act_X01_n = '1' and ras_X01_n = '1' and cas_X01_n = '0'; -- RD or WR

  gen_a_col_10_bits : if part.pinout.col_width = 10 generate
    a_col_X01 <= a_X01(9 downto 0);
    a_col_01 <= a_01(9 downto 0);
  end generate;

  gen_a_col_11_bits : if part.pinout.col_width = 11 generate
    a_col_X01 <= a_X01(11) & a_X01(9 downto 0);
    a_col_01 <= a_01(11) & a_01(9 downto 0);
  end generate;

  gen_a_col_12_bits : if part.pinout.col_width = 12 generate
    a_col_X01 <= a_X01(13) & a_X01(11) & a_X01(9 downto 0);
    a_col_01 <= a_01(13) & a_01(11) & a_01(9 downto 0);
  end generate;

  check_a_col_timing : ddr4_sdram_su_h_check_col_addr
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "A",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => a_col_qualified,
      sig => a_col_X01
    );
  
  -- A[10]/AP address bit qualification
  a_ap_qualified <= cs_X01_n = '0' and cke_X01 = '1' and act_X01_n = '1' and
                    ((ras_X01_n = '0' and cas_X01_n = '1' and we_X01_n = '0') or -- PRE or PREA
                     (ras_X01_n = '1' and cas_X01_n = '0') or                    -- RD or WR
                     (ras_X01_n = '1' and cas_X01_n = '1' and we_X01_n = '0'));  -- ZQCL or ZQCS
  
  check_a_ap_timing : ddr4_sdram_su_h_check_bit
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "A[10]",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => a_ap_qualified,
      sig => a_ap_X01
    );
    
  -- A[12]/BC address bit qualification (applies when MR0 set for on-the-fly burst length)
  a_bc_qualified <= a_col_qualified and bl = burst_length_otf; -- RD or WR, as per column address qualification
  
  check_bc_col_timing : ddr4_sdram_su_h_check_bit
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "A[12]",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => a_bc_qualified,
      sig => a_bc_X01
    );

  -- Address bits [13:0] qualification, for MRS command
  a_mrs_qualified <= cs_X01_n = '0' and cke_X01 = '1' and act_X01_n = '1' and ras_X01_n = '0' and cas_X01_n = '0' and we_X01_n = '0'; -- MRS

  check_a_mrs_timing : ddr4_sdram_su_h_check_vec
    generic map(
      model_name => model_name,
      clk_name => "CK_t",
      sig_name => "A",
      posedge => true,
      tsu => part.timing.operational.t_is,
      th => part.timing.operational.t_ih
    )
    port map(
      clk => ck_t,
      qualified => a_mrs_qualified,
      sig => a_mrs_X01
    );
    
  
  --
  -- Bank state machines
  -- Track whether or not a given bank is activated, and the row for which each bank is activated.
  --
  
  gen_bank_groups : for bank_group in 0 to num_bank_group - 1 generate
  
    gen_banks_in_group : for bank in 0 to num_bank_per_group - 1 generate
  
      constant bank_index : natural := num_bank_per_group * bank_group + bank;
      
    begin
    
      bank_state_inst : ddr4_sdram_bank_state
        generic map(
          model_name => model_name,
          bank_group => bank_group,
          bank => bank,
          part => part
        )
        port map(
          reset_n => reset_X01_n,
          ck_t => ck_t,
          cke => cke_X01,
          cs_n => cs_X01_n,
          act_n => act_X01_n,
          a => a_01,
          ba => ba_01,
          bg => bg_01,
          ras_n => ras_X01_n,
          cas_n => cas_X01_n,
          we_n => we_X01_n,
          bank_state => bank_states(bank_index),
          activated_row => activated_rows(bank_index)
        );
        
    end generate;
    
  end generate;
  
  
  data_inst : ddr4_sdram_data
    generic map(
      model_name => model_name,
      part => part,
      num_bank_total => num_bank_total,
      read_undef_val => read_undef_val
    )
    port map(
      reset_n => reset_X01_n,
      ck_t => ck_t,
      cke => cke_X01,
      cs_n => cs_X01_n,
      act_n => act_X01_n,
      a_col => a_col_01,
      a_ap => a_ap_X01,
      a_bc => a_bc_X01,
      ba => ba_01,
      bg => bg_01,
      ras_n => ras_X01_n,
      cas_n => cas_X01_n,
      we_n => we_X01_n,
      dq => dq,
      dm_dbi_n => dm_dbi_n,
      dqs_t => dqs_t,
      dqs_c => dqs_c,
      bank_states => bank_states,
      activated_rows => activated_rows,
      cl => cl,
      al => al,
      cwl => cwl,
      bl => bl,
      bt => bt,
      read_preamble => read_preamble,
      rd_dbi_enabled => rd_dbi_enabled,
      wr_dbi_enabled => wr_dbi_enabled
    );
  
  
  power_state_inst : ddr4_sdram_power_state
    generic map(
      model_name => model_name,
      part => part
    )
    port map(
      reset_n => reset_X01_n,
      ck_t => ck_t,
      cke => cke_X01,
      cke_prev => cke_prev_X01,
      cs_n => cs_X01_n,
      act_n => act_X01_n,
      ras_n => ras_X01_n,
      cas_n => cas_X01_n,
      we_n => we_X01_n,
      power_state => power_state
    );
    
    
  ref_mon_inst : ddr4_sdram_refresh_monitor
    generic map(
      model_name => model_name,
      part => part,
      num_bank_total => num_bank_total
    )
    port map(
      reset_n => reset_X01_n,
      ck_t => ck_t,
      cke => cke_X01,
      cs_n => cs_X01_n,
      act_n => act_X01_n,
      ras_n => ras_X01_n,
      cas_n => cas_X01_n,
      we_n => we_X01_n,
      bank_states => bank_states
    );
  
  
  init_monitor_inst : ddr4_sdram_init_monitor
    generic map(
      model_name => model_name,
      part => part,
      init_from_power_on => init_from_power_on,
      relaxed_init_timing => relaxed_init_timing
    )
    port map(
      reset_n => reset_X01_n,
      ck_t => ck_t,
      cke => cke_X01,
      cs_n => cs_X01_n,
      act_n => act_X01_n,
      ras_n => ras_X01_n,
      cas_n => cas_X01_n,
      we_n => we_X01_n,
      bg => bg_X01,
      ba => ba_X01,
      a => a_X01,
      init_done => init_done,
      mr0 => mr0,
      mr1 => mr1,
      mr2 => mr2,
      mr3 => mr3,
      mr4 => mr4,
      mr5 => mr5,
      mr6 => mr6,
      cl => cl,
      al => al,
      cwl => cwl,
      bl => bl,
      bt => bt,
      read_preamble => read_preamble,
      rd_dbi_enabled => rd_dbi_enabled,
      wr_dbi_enabled => wr_dbi_enabled
    );
  
  
end architecture;
