library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library std;
use std.textio.all;

library work;
use work.ddr4_sdram_model.all;
use work.ddr4_sdram_model_int.all;

entity ddr4_sdram_init_monitor is
  generic(
    model_name          : in    string;
    part                : in    part_t;
    init_from_power_on  : in    boolean;
    relaxed_init_timing : in    boolean
  );
  port(
    reset_n             : in    std_logic; -- Should be _X01 version of signal
    ck_t                : in    std_logic;
    cke                 : in    std_logic; -- Should be _X01 version of signal
    cs_n                : in    std_logic; -- Should be _X01 version of signal
    act_n               : in    std_logic; -- Should be _X01 version of signal
    ras_n               : in    std_logic; -- Should be _X01 version of signal
    cas_n               : in    std_logic; -- Should be _X01 version of signal
    we_n                : in    std_logic; -- Should be _X01 version of signal
    a                   : in    std_logic_vector(part.pinout.a_width - 1 downto 0);  -- Should be _X01 version of signal
    ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0); -- Should be _X01 version of signal
    bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0); -- Should be _X01 version of signal
    init_done           : out   boolean;
    mr0                 : out   mode_reg_t;
    mr1                 : out   mode_reg_t;
    mr2                 : out   mode_reg_t;
    mr3                 : out   mode_reg_t;
    mr4                 : out   mode_reg_t;
    mr5                 : out   mode_reg_t;
    mr6                 : out   mode_reg_t;
    cl                  : out   cl_t;
    al                  : out   al_t;
    cwl                 : out   cwl_t;
    bl                  : out   burst_length_t;
    bt                  : out   burst_type_t;
    read_preamble       : out   read_preamble_t;
    rd_dbi_enabled      : out   boolean;
    wr_dbi_enabled      : out   boolean
  );
end entity;

architecture behav of ddr4_sdram_init_monitor is

  signal mr_addr : mode_reg_addr_t;
  -- Array of MR0 ... MR6
  signal mr_array : mode_reg_array_t;
  alias mr0_int : mode_reg_t is mr_array(0);
  alias mr0_bl : std_logic_vector(1 downto 0) is mr0_int(1 downto 0);
  alias mr0_bt : std_logic is mr0_int(3);
  signal mr0_cl : std_logic_vector(4 downto 0);
  signal i_cl : cl_t;
  alias mr1_int : mode_reg_t is mr_array(1);
  alias mr1_al : std_logic_vector(1 downto 0) is mr1_int(4 downto 3);
  alias mr2_int : mode_reg_t is mr_array(2);
  alias mr2_cwl : std_logic_vector(2 downto 0) is mr2_int(5 downto 3);
  alias mr3_int : mode_reg_t is mr_array(3);
  alias mr4_int : mode_reg_t is mr_array(4);
  alias mr4_read_preamble : std_logic is mr4_int(11);
  alias mr5_int : mode_reg_t is mr_array(5);
  alias mr5_wr_dbi : std_logic is mr5_int(11);
  alias mr5_rd_dbi : std_logic is mr5_int(12);
  alias mr6_int : mode_reg_t is mr_array(6);
  
  -- State machine checking that mode register initialization follows recommended sequence.
  type mr_init_state_t is (
    mr_init_state_establish_mr3,
    mr_init_state_establish_mr6,
    mr_init_state_establish_mr5,
    mr_init_state_establish_mr4,
    mr_init_state_establish_mr2,
    mr_init_state_establish_mr1,
    mr_init_state_establish_mr0,
    mr_init_state_done
  );
  signal mr_init_state : mr_init_state_t := mr_init_state_establish_mr3;
  signal mr_init_deviation : boolean := false;
  signal mr_established : boolean_array_t(0 to 6) := (others => false);
  signal mr_established_all : boolean := false;
  
  -- State machine checking for ZQ calibration compliance
  type zq_init_state_t is (
    zq_init_state_zq_cal,
    zq_init_state_done
  );
  signal zq_init_state : zq_init_state_t := zq_init_state_zq_cal;
  signal zq_init_deviation_mr : boolean := false;
  signal zq_init_deviation_zqcs : boolean := false;
  
  
begin

  mr_addr <= bg(0) & ba(1 downto 0);
  
  gen_mr_init_state : process(reset_n, ck_t)
    variable mr_addr_int : natural;
    variable warning_issued : boolean;
    variable otf : boolean;
    variable mrs_allow_time : time := 0 ns;
    variable mrs_allow_ck : natural := 0;
  begin
    if reset_n = '0' then
      mr_init_state <= mr_init_state_establish_mr3;
      mr_init_deviation <= false;
      mr_array <= (others => (others => 'U'));
      mr_established <= (others => false);
      warning_issued := false;
      mrs_allow_time := 0 ns;
      mrs_allow_ck := 0;
    elsif rising_edge(ck_t) then
      if mrs_allow_ck > 0 then
        mrs_allow_ck := mrs_allow_ck - 1;
      end if;
      
      otf := mr0_bl = MR0_BL_OTF;
      if cke = '1' then
        if cs_n = '0' and act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '0' then
          if not is_X(mr_addr) then
            mr_addr_int := to_integer(unsigned(mr_addr));
            mr_established(mr_addr_int) <= true;
            mr_array(mr_addr_int) <= a(13 downto 0);
            if not warning_issued and and_reduce(mr_established) and mr_init_deviation then
              report
                model_name & ": All mode registers established, but with deviation from recommended initialization sequence."
                severity warning;
              warning_issued := true;
            end if;
          end if;
          
          -- Check that tMRD is satisfied
          if now < mrs_allow_time or mrs_allow_ck > 0 then
            report model_name & ": tMRD violated by MRS command" severity error;
          end if;
          mrs_allow_time := now + part.timing.operational.t_mrd;
          mrs_allow_ck := part.timing.operational.n_mrd;
        end if;
          
        case mr_init_state is
          when mr_init_state_establish_mr3 =>
            if cs_n = '0' then
              if act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '0' and mr_addr = "011" then
                mr_init_state <= mr_init_state_establish_mr6;
              else
                if not mr_init_deviation then
                  report
                    model_name & ": Deviation from recommended initialization sequence - expecting MRS command for MR3; this command is " &
                      classify_command(otf, '1', '1', cs_n, act_n, ras_n, cas_n, we_n, bg, ba, a)                  
                    severity warning;
                  mr_init_deviation <= true;
                end if;
              end if;
            end if;
          
          when mr_init_state_establish_mr6 =>
            if cs_n = '0' then
              if act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '0' and mr_addr = "110" then
                mr_init_state <= mr_init_state_establish_mr5;
              else
                if not mr_init_deviation then
                  report
                    model_name & ": Deviation from recommended initialization sequence - expecting MRS command for MR6; this command is " &
                      classify_command(otf, '1', '1', cs_n, act_n, ras_n, cas_n, we_n, bg, ba, a)                  
                    severity warning;
                  mr_init_deviation <= true;
                end if;
              end if;
            end if;
          
          when mr_init_state_establish_mr5 =>
            if cs_n = '0' then
              if act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '0' and mr_addr = "101" then
                mr_init_state <= mr_init_state_establish_mr4;
              else
                if not mr_init_deviation then
                  report
                    model_name & ": Deviation from recommended initialization sequence - expecting MRS command for MR5; this command is " &
                      classify_command(otf, '1', '1', cs_n, act_n, ras_n, cas_n, we_n, bg, ba, a)                  
                    severity warning;
                  mr_init_deviation <= true;
                end if;
              end if;
            end if;
          
          when mr_init_state_establish_mr4 =>
            if cs_n = '0' then
              if act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '0' and mr_addr = "100" then
                mr_init_state <= mr_init_state_establish_mr2;
              else
                if not mr_init_deviation then
                  report
                    model_name & ": Deviation from recommended initialization sequence - expecting MRS command for MR4; this command is " &
                      classify_command(otf, '1', '1', cs_n, act_n, ras_n, cas_n, we_n, bg, ba, a)                  
                    severity warning;
                  mr_init_deviation <= true;
                end if;
              end if;
            end if;
          
          when mr_init_state_establish_mr2 =>
            if cs_n = '0' then
              if act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '0' and mr_addr = "010" then
                mr_init_state <= mr_init_state_establish_mr1;
              else
                if not mr_init_deviation then
                  report
                    model_name & ": Deviation from recommended initialization sequence - expecting MRS command for MR2; this command is " &
                      classify_command(otf, '1', '1', cs_n, act_n, ras_n, cas_n, we_n, bg, ba, a)                  
                    severity warning;
                  mr_init_deviation <= true;
                end if;
              end if;
            end if;
          
          when mr_init_state_establish_mr1 =>
            if cs_n = '0' then
              if act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '0' and mr_addr = "001" then
                mr_init_state <= mr_init_state_establish_mr0;
              else
                if not mr_init_deviation then
                  report
                    model_name & ": Deviation from recommended initialization sequence - expecting MRS command for MR1; this command is " &
                      classify_command(otf, '1', '1', cs_n, act_n, ras_n, cas_n, we_n, bg, ba, a)                  
                    severity warning;
                  mr_init_deviation <= true;
                end if;
              end if;
            end if;
          
          when mr_init_state_establish_mr0 =>
            if cs_n = '0' then
              if act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '0' and mr_addr = "000" then
                report
                  model_name & ": All mode registers established, and in recommended order."
                  severity note;
                mr_init_state <= mr_init_state_done;
              else
                if not mr_init_deviation then
                  report
                    model_name & ": Deviation from recommended initialization sequence - expecting MRS command for MR0; this command is " &
                      classify_command(otf, '1', '1', cs_n, act_n, ras_n, cas_n, we_n, bg, ba, a)                  
                    severity warning;
                  mr_init_deviation <= true;
                end if;
              end if;
            end if;
        
          when mr_init_state_done =>
            null;
        end case;
      end if;
    end if;
  end process;

  mr_established_all <= and_reduce(mr_established);
  
  
  gen_zq_state : process(reset_n, ck_t)
    variable otf : boolean;
  begin
    if reset_n = '0' then
      zq_init_state <= zq_init_state_zq_cal;
    elsif rising_edge(ck_t) then
      case zq_init_state is
        when zq_init_state_zq_cal =>
          if cke = '1' and cs_n = '0' and act_n = '1' and ras_n = '1' and cas_n = '1' and we_n = '0' then
            if a(10) = '1' then
              zq_init_state <= zq_init_state_done;
              if not mr_established_all then
                if not zq_init_deviation_mr then
                  report
                    model_name & ": Deviation from recommended initialization sequence - ZQCL command issued before MR0 .. MR6 established"
                    severity warning;
                  zq_init_deviation_mr <= true;
                end if;
              end if;
            else
              if not zq_init_deviation_zqcs then
                otf := mr0_bl = MR0_BL_OTF;
                report
                  model_name & ": Deviation from recommended initialization sequence - expecting ZQCL command; this command is " &
                    classify_command(otf, '1', '1', cs_n, act_n, ras_n, cas_n, we_n, bg, ba, a)                  
                  severity warning;
                zq_init_deviation_zqcs <= true;
              end if;
            end if;
          end if;

        when zq_init_state_done =>
          null;         
      end case;
    end if;
  end process;
  
  
  init_done <= mr_established_all and zq_init_state = zq_init_state_done;
  
  mr0 <= mr0_int;
  mr1 <= mr1_int;
  mr2 <= mr2_int;
  mr3 <= mr3_int;
  mr4 <= mr4_int;
  mr5 <= mr5_int;
  mr6 <= mr6_int;
  
  -- Decode MR0[1:0] - burst length
  with mr0_bl select bl <=
    burst_length_otf when "01",
    burst_length_bc4 when "10",
    burst_length_bl8 when others; -- Treat others as "00"
                
  -- Decode MR0[3] - burst type
  with mr0_bt select bt <=
    burst_type_interleave when '1',
    burst_type_nibble_seq when others; -- Treat others as '0'
                
  -- Decode MR0[12,6:4,2] - CAS latency
  mr0_cl <= make_01(mr0_int(12) & mr0_int(6 downto 4) & mr0_int(2));
  with mr0_cl select i_cl <=
    10 when "00001",
    11 when "00010",
    12 when "00011",
    13 when "00100",
    14 when "00101",
    15 when "00110",
    16 when "00111",
    18 when "01000",
    20 when "01001",
    22 when "01010",
    24 when "01011",
    23 when "01100",
    17 when "01101",
    19 when "01110",
    21 when "01111",
    25 when "10000",
    26 when "10001",
    27 when "10010",
    28 when "10011",
    29 when "10100",
    30 when "10101",
    31 when "10110",
    32 when "10111",
     9 when others; -- Treat others as "00000"
  cl <= i_cl;
  
  -- Decode MR1[5:3] - additive latency
  with mr1_al select al <=
    i_cl - 1 when "01",
    i_cl - 2 when "10",
           0 when others; -- Treat others as "00"
                      
  -- Decode MR2[5:3] - CAS write latency
  with mr2_cwl select cwl <=
    10 when "001",
    11 when "010",
    12 when "011",
    14 when "100",
    16 when "101",
    18 when "110",
    20 when "111",
     9 when others; -- Treat others as "000"
  
  -- Decode MR4[11] - Read preamble
  with mr4_read_preamble select read_preamble <=
    read_preamble_2ck when '1',
    read_preamble_1ck when others; -- Treat others as '0'

  -- Decode MR5[11] - Write data bus inversion
  wr_dbi_enabled <= true when mr5_wr_dbi = '1' and part.pinout.dq_nibbles >= 2 else false;
  
  check_wr_dbi_support : process(
    mr5_wr_dbi)
  begin
    if part.pinout.dq_nibbles < 2 and mr5_wr_dbi = '1' then
      report
        model_name & ": Illegal attempt to enable write DBI for x4 part"
        severity error;      
    end if;
  end process;
  
  -- Decode MR5[12] - Read data bus inversion
  rd_dbi_enabled <= true when mr5_rd_dbi = '1' and part.pinout.dq_nibbles >= 2 else false;
  
  check_rd_dbi_support : process(
    mr5_rd_dbi)
  begin
    if part.pinout.dq_nibbles < 2 and mr5_rd_dbi = '1' then
      report
        model_name & ": Illegal attempt to enable read DBI for x4 part"
        severity error;      
    end if;
  end process;
  
  -- Mode register 0 messages
  mr0_message : process
    variable l : line;
    variable cl : std_logic_vector(4 downto 0);
    variable wr_rtp : std_logic_vector(3 downto 0);
  begin
    wait on mr0_int;
    l := new string(1 to 0);
    write(l, string'(model_name & ": MR0=" & to_string_binary(mr0_int) & ";"));
    write(l, string'(" BL="));
    case mr0_int(1 downto 0) is
      when "00" =>
        write(l, string'("BL8"));
      when "01" =>
        write(l, string'("OTF"));
      when "10" =>
        write(l, string'("BC4"));
      when "11" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" BurstType="));
    case mr0_int(3) is
      when '0' =>
        write(l, string'("NibbleSeq"));
      when '1' =>
        write(l, string'("Interleave"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" CL="));
    cl := mr0_int(12) & mr0_int(6 downto 4) & mr0_int(2);
    case cl is
      when "00000" =>
        write(l, string'("9ck"));
      when "00001" =>
        write(l, string'("10ck"));
      when "00010" =>
        write(l, string'("11ck"));
      when "00011" =>
        write(l, string'("12ck"));
      when "00100" =>
        write(l, string'("13ck"));
      when "00101" =>
        write(l, string'("14ck"));
      when "00110" =>
        write(l, string'("15ck"));
      when "00111" =>
        write(l, string'("16ck"));
      when "01000" =>
        write(l, string'("18ck"));
      when "01001" =>
        write(l, string'("20ck"));
      when "01010" =>
        write(l, string'("22ck"));
      when "01011" =>
        write(l, string'("24ck"));
      when "01100" =>
        write(l, string'("23ck"));
      when "01101" =>
        write(l, string'("17ck"));
      when "01110" =>
        write(l, string'("19ck"));
      when "01111" =>
        write(l, string'("21ck"));
      when "10000" =>
        write(l, string'("25ck"));
      when "10001" =>
        write(l, string'("26ck"));
      when "10010" =>
        write(l, string'("27ck"));
      when "10011" =>
        write(l, string'("28ck"));
      when "10100" =>
        write(l, string'("29ck"));
      when "10101" =>
        write(l, string'("30ck"));
      when "10110" =>
        write(l, string'("31ck"));
      when "10111" =>
        write(l, string'("32ck"));
      when "11000" | "11001" | "11010" | "11011" | "11100" | "11101" | "11110" | "11111" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" TestMode="));
    case mr0_int(7) is
      when '0' =>
        write(l, string'("Normal"));
      when '1' =>
        write(l, string'("ON/Illeg."));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" DLLReset="));
    case mr0_int(8) is
      when '0' =>
        write(l, string'("No"));
      when '1' =>
        write(l, string'("Yes"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" WR/RTP="));
    wr_rtp := mr0_int(13) & mr0_int(11 downto 9);
    case wr_rtp is
      when "0000" =>
        write(l, string'("10/5ck"));
      when "0001" =>
        write(l, string'("12/6ck"));
      when "0010" =>
        write(l, string'("14/7ck"));
      when "0011" =>
        write(l, string'("16/8ck"));
      when "0100" =>
        write(l, string'("18/9ck"));
      when "0101" =>
        write(l, string'("20/10ck"));
      when "0110" =>
        write(l, string'("24/12ck"));
      when "0111" =>
        write(l, string'("22/11ck"));
      when "1000" =>
        write(l, string'("26/13ck"));
      when "1001" | "1010" | "1011" | "1100" | "1101" | "1110" | "1111" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'("ck"));
    report l.all severity note;
  end process;
  
  -- Mode register 1 messages
  mr1_message : process
    variable l : line;
  begin
    wait on mr1_int;
    l := new string(1 to 0);
    write(l, string'(model_name & ": MR1=" & to_string_binary(mr1_int) & ";"));
    write(l, string'(" DLLEnable="));
    case mr1_int(0) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Normal"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" ODI="));
    case mr1_int(2 downto 1) is
      when "00" =>
        write(l, string'("RZQ/7"));
      when "01" =>
        write(l, string'("RZQ/5"));
      when "10" | "11" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" AL="));
    case mr1_int(4 downto 3) is
      when "00" =>
        write(l, string'("0"));
      when "01" =>
        write(l, string'("CL-1"));
      when "10" =>
        write(l, string'("CL-2"));
      when "11" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RFU[6:5]=" & to_string_binary(mr1_int(6 downto 5))));
    write(l, string'(" WrLvl="));
    case mr1_int(7) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" NomODT="));
    case mr1_int(10 downto 8) is
      when "000" =>
        write(l, string'("Disabled"));
      when "001" =>
        write(l, string'("RZQ/4"));
      when "010" =>
        write(l, string'("RZQ/2"));
      when "011" =>
        write(l, string'("RZQ/6"));
      when "100" =>
        write(l, string'("RZQ/1"));
      when "101" =>
        write(l, string'("RZQ/5"));
      when "110" =>
        write(l, string'("RZQ/3"));
      when "111" =>
        write(l, string'("RZQ/7"));
      when others =>
        write(l, string'("?"));
    end case;
    if part.pinout.tdqs_supported then
      write(l, string'(" TDQS="));
      case mr1_int(11) is
        when '0' =>
          write(l, string'("Disabled"));
        when '1' =>
          write(l, string'("Enabled"));
        when others =>
          write(l, string'("?"));
      end case;
    else
      write(l, string'(" RFU[11]=" & to_string_binary(mr1_int(11 downto 11))));
    end if;
    write(l, string'(" Qoff="));
    case mr1_int(12) is
      when '0' =>
        write(l, string'("Enabled"));
      when '1' =>
        write(l, string'("Disabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RFU[13]=" & to_string_binary(mr1_int(13 downto 13))));
    report l.all severity note;
  end process;
  
  -- Mode register 2 messages
  mr2_message : process
    variable l : line;
    variable trr_bg : std_logic_vector(1 downto 0);
  begin
    wait on mr2_int;
    l := new string(1 to 0);
    write(l, string'(model_name & ": MR2=" & to_string_binary(mr2_int) & ";"));
    write(l, string'(" TRR_BA="));
    case mr2_int(1 downto 0) is
      when "00" =>
        write(l, string'("BA0"));
      when "01" =>
        write(l, string'("BA1"));
      when "10" =>
        write(l, string'("BA2"));
      when "11" =>
        write(l, string'("BA3"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" TRR_BG="));
    trr_bg := mr2_int(8) & mr2_int(2);
    case trr_bg is
      when "00" =>
        write(l, string'("BG0"));
      when "01" =>
        write(l, string'("BG1"));
      when "10" =>
        write(l, string'("BG2"));
      when "11" =>
        write(l, string'("BG3"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" CWL(1ck,2ck)="));
    case mr2_int(5 downto 3) is
      when "000" =>
        write(l, string'("9,N/A"));
      when "001" =>
        write(l, string'("10,N/A"));
      when "010" =>
        write(l, string'("11,N/A"));
      when "011" =>
        write(l, string'("12,N/A"));
      when "100" =>
        write(l, string'("14,14"));
      when "101" =>
        write(l, string'("16,16"));
      when "110" =>
        write(l, string'("18,18"));
      when "111" =>
        write(l, string'("20,20"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" LPASR="));
    case mr2_int(7 downto 6) is
      when "00" =>
        write(l, string'("ManualNorm"));
      when "01" =>
        write(l, string'("ManualRed"));
      when "10" =>
        write(l, string'("ManualExt"));
      when "11" =>
        write(l, string'("Automatic"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" DynODT="));
    case mr2_int(11 downto 9) is
      when "000" =>
        write(l, string'("Disabled"));
      when "001" =>
        write(l, string'("RZQ/2"));
      when "010" =>
        write(l, string'("RZQ/1"));
      when "011" =>
        write(l, string'("Hi-Z"));
      when "100" =>
        write(l, string'("RZQ/3"));
      when "101" | "110" | "111" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" WrCRC="));
    case mr2_int(12) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" TRR="));
    case mr2_int(13) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    report l.all severity note;
  end process;
  
  -- Mode register 3 messages
  mr3_message : process
    variable l : line;
  begin
    wait on mr3_int;
    l := new string(1 to 0);
    write(l, string'(model_name & ": MR3=" & to_string_binary(mr3_int) & ";"));
    write(l, string'(" MPRPageSel="));
    case mr3_int(1 downto 0) is
      when "00" =>
        write(l, string'("0"));
      when "01" =>
        write(l, string'("1"));
      when "10" =>
        write(l, string'("2"));
      when "11" =>
        write(l, string'("3"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" MPRSel="));
    case mr3_int(2) is
      when '0' =>
        write(l, string'("Normal"));
      when '1' =>
        write(l, string'("FromMPR"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" Geardown="));
    case mr3_int(3) is
      when '0' =>
        write(l, string'("1:1"));
      when '1' =>
        write(l, string'("2:1"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" PerDRAMAdr="));
    case mr3_int(4) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" TempSens="));
    case mr3_int(5) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" FineRef="));
    case mr3_int(8 downto 6) is
      when "000" =>
        write(l, string'("Normal"));
      when "001" =>
        write(l, string'("Fixed2X"));
      when "010" =>
        write(l, string'("Fixed4X"));
      when "011" | "100" | "111" =>
        write(l, string'("(reserved)"));
      when "101" =>
        write(l, string'("OTF:1X/2X"));
      when "110" =>
        write(l, string'("OTF:1X/4X"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" WrCmdLatCRC="));
    case mr3_int(10 downto 9) is
      when "00" =>
        write(l, string'("4ck"));
      when "01" =>
        write(l, string'("5ck"));
      when "10" =>
        write(l, string'("6ck"));
      when "11" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" MPRFmt="));
    case mr3_int(12 downto 11) is
      when "00" =>
        write(l, string'("Serial"));
      when "01" =>
        write(l, string'("Parallel"));
      when "10" =>
        write(l, string'("Stagger"));
      when "11" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RFU[13]=" & to_string_binary(mr3_int(13 downto 13))));
    report l.all severity note;
  end process;
  
  -- Mode register 4 messages
  mr4_message : process
    variable l : line;
  begin
    wait on mr4_int;
    l := new string(1 to 0);
    write(l, string'(model_name & ": MR4=" & to_string_binary(mr4_int) & ";"));
    write(l, string'(" RFU[0]=" & to_string_binary(mr4_int(0 downto 0))));
    write(l, string'(" MaxPwrSave="));
    case mr4_int(1) is
      when '0' =>
        write(l, string'("Normal"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" TmpCtlRefRng="));
    case mr4_int(2) is
      when '0' =>
        write(l, string'("Norm"));
      when '1' =>
        write(l, string'("Ext"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" TmpCtlRefMode="));
    case mr4_int(3) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" IntVRefMon="));
    case mr4_int(4) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" sPRR="));
    case mr4_int(5) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" CmdAdrLat="));
    case mr4_int(8 downto 6) is
      when "000" =>
        write(l, string'("0ck"));
      when "001" =>
        write(l, string'("3ck"));
      when "010" =>
        write(l, string'("4ck"));
      when "011" =>
        write(l, string'("5ck"));
      when "100" =>
        write(l, string'("6ck"));
      when "101" =>
        write(l, string'("8ck"));
      when "110" | "111" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" SelfRefAbrt="));
    case mr4_int(9) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RdPreambleTrn="));
    case mr4_int(10) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RdPreamble="));
    case mr4_int(11) is
      when '0' =>
        write(l, string'("1ck"));
      when '1' =>
        write(l, string'("2ck"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" WrPreamble="));
    case mr4_int(12) is
      when '0' =>
        write(l, string'("1ck"));
      when '1' =>
        write(l, string'("2ck"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" PPR="));
    case mr4_int(13) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    report l.all severity note;
  end process;
  
  -- Mode register 5 messages
  mr5_message : process
    variable l : line;
  begin
    wait on mr5_int;
    l := new string(1 to 0);
    write(l, string'(model_name & ": MR5=" & to_string_binary(mr5_int) & ";"));
    write(l, string'(" RFU[0]=" & to_string_binary(mr5_int(0 downto 0))));
    write(l, string'(" CAParLatMode="));
    case mr5_int(2 downto 0) is
      when "000" =>
        write(l, string'("Disabled"));
      when "001" =>
        write(l, string'("4ck"));
      when "010" =>
        write(l, string'("5ck"));
      when "011" =>
        write(l, string'("6ck"));
      when "100" =>
        write(l, string'("8ck"));
      when "101" | "110" | "111" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" CRCErrStat="));
    case mr5_int(3) is
      when '0' =>
        write(l, string'("Clear"));
      when '1' =>
        write(l, string'("Error"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" CAParErrStat="));
    case mr5_int(4) is
      when '0' =>
        write(l, string'("Clear"));
      when '1' =>
        write(l, string'("Error"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" ODTIpPwrDn="));
    case mr5_int(4) is
      when '0' =>
        write(l, string'("Enabled"));
      when '1' =>
        write(l, string'("Disabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" ParkODT="));
    case mr5_int(8 downto 6) is
      when "000" =>
        write(l, string'("Disabled"));
      when "001" =>
        write(l, string'("RZQ/4"));
      when "010" =>
        write(l, string'("RZQ/2"));
      when "011" =>
        write(l, string'("RZQ/6"));
      when "100" =>
        write(l, string'("RZQ/1"));
      when "101" =>
        write(l, string'("RZQ/5"));
      when "110" =>
        write(l, string'("RZQ/3"));
      when "111" =>
        write(l, string'("RZQ/7"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" CAParPrsErrMode="));
    case mr5_int(9) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" DM="));
    case mr5_int(10) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" WrDBI="));
    case mr5_int(11) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RdDBI="));
    case mr5_int(12) is
      when '0' =>
        write(l, string'("Disabled"));
      when '1' =>
        write(l, string'("Enabled"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RFU[13]=" & to_string_binary(mr5_int(13 downto 13))));
    report l.all severity note;
  end process;
  
  -- Mode register 6 messages
  mr6_message : process
    variable l : line;
  begin
    wait on mr6_int;
    l := new string(1 to 0);
    write(l, string'(model_name & ": MR6=" & to_string_binary(mr6_int) & ";"));
    write(l, string'(" VRefCalValue=" & to_string_binary(mr6_int(5 downto 0))));
    write(l, string'(" VRefCalRange="));
    case mr6_int(6) is
      when '0' =>
        write(l, string'("Range1"));
      when '1' =>
        write(l, string'("Range2"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" VRefCalEn="));
    case mr6_int(7) is
      when '0' =>
        write(l, string'("Disable"));
      when '1' =>
        write(l, string'("Enable"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RFU[9:8]=" & to_string_binary(mr6_int(9 downto 8))));
    write(l, string'(" tCCD_L="));
    case mr6_int(12 downto 10) is
      when "000" =>
        write(l, string'("4ck"));
      when "001" =>
        write(l, string'("5ck"));
      when "010" =>
        write(l, string'("6ck"));
      when "011" =>
        write(l, string'("7ck"));
      when "100" =>
        write(l, string'("8ck"));
      when "101" | "110" | "111" =>
        write(l, string'("(reserved)"));
      when others =>
        write(l, string'("?"));
    end case;
    write(l, string'(" RFU[13]=" & to_string_binary(mr6_int(13 downto 13))));
    report l.all severity note;
  end process;
  

end architecture;
