library ieee;
use ieee.std_logic_1164.all;

entity ddr4_sdram_su_h_check_bit is
  generic(
    model_name          : in    string;
    clk_name            : in    string;
    sig_name            : in    string;
    posedge             : in    boolean;
    tsu                 : in    time;
    th                  : in    time
  );
  port(
    clk                 : in    std_logic;
    sig                 : in    std_logic; -- Should be _X01 version of signal
    qualified           : in    boolean
  );
end entity;

architecture behav of ddr4_sdram_su_h_check_bit is

  signal hold : boolean := false;
  signal sig_valid : boolean;
  
  
begin

  sig_valid <= not is_X(sig);
  
  check_setup : process(clk)
  begin
    if qualified and ((posedge and rising_edge(clk)) or (not posedge and falling_edge(clk))) then
      if not sig_valid then
        report model_name & ": " & sig_name & " invalid at " & clk_name & " edge" severity error;
      end if;
      
      if not sig'stable(tsu) then
        report model_name & ": " & sig_name & " setup violation at " & clk_name & " edge" severity error;
      end if;
      
      hold <= transport true, false after th;
    end if;
  end process;
  
  check_hold : process(sig)
  begin
    if hold then
      report model_name & ": " & sig_name & " hold violation after " & clk_name & " edge" severity error;
    end if;
  end process;

  
end architecture;
---------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity ddr4_sdram_su_h_check_vec is
  generic(
    model_name          : in    string;
    clk_name            : in    string;
    sig_name            : in    string;
    posedge             : in    boolean;
    tsu                 : in    time;
    th                  : in    time
  );
  port(
    clk                 : in    std_logic;
    sig                 : in    std_logic_vector; -- Should be _X01 version of signal
    qualified           : in    boolean
  );
end entity;

architecture behav of ddr4_sdram_su_h_check_vec is

  constant lo : natural := sig'low;
  constant hi : natural := sig'high;
  
  signal hold : boolean := false;
  signal sig_valid : boolean;
  
  
begin

  sig_valid <= not is_X(sig);
  
  check_setup : process(clk)
  begin
    if qualified and ((posedge and rising_edge(clk)) or (not posedge and falling_edge(clk))) then
      if not sig_valid then
        report model_name & ": " & sig_name & "[" & natural'image(hi) & ":" & natural'image(lo) & "] has at least one invalid bit at " & clk_name & " edge" severity error;
      end if;
      
      if not sig'stable(tsu) then
        report model_name & ": " & sig_name & "[" & natural'image(hi) & ":" & natural'image(lo) & "] setup violation at " & clk_name & " edge" severity error;
      end if;
      
      hold <= transport true, false after th;
    end if;
  end process;
  
  check_hold : process(sig)
  begin
    if hold then
      report model_name & ": " & sig_name & "[" & natural'image(hi) & ":" & natural'image(lo) & "] hold violation after " & clk_name & " edge" severity error;
    end if;
  end process;

  
end architecture;
---------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;

entity ddr4_sdram_su_h_check_col_addr is
  generic(
    model_name          : in    string;
    clk_name            : in    string;
    sig_name            : in    string;
    posedge             : in    boolean;
    tsu                 : in    time;
    th                  : in    time
  );
  port(
    clk                 : in    std_logic;
    sig                 : in    std_logic_vector; -- Should be _X01 version of signal
    qualified           : in    boolean
  );
end entity;

architecture behav of ddr4_sdram_su_h_check_col_addr is

  function get_sig_name(
    constant s : in string;
    constant n : in natural
  ) return string is
  begin
    if n <= 10 then
      return s & "[" & natural'image(n - 1) & ":0]";
    elsif n = 11 then
      return s & "[11,9:0]";
    elsif n = 12 then
      return s & "[13,11,9:0]";
    else
      return s & "[" & natural'image(n + 1) & ":13,11,9:0]";
    end if;
  end;

  signal hold : boolean := false;
  signal sig_valid : boolean;
  
  
begin

  -- Sanity checks
  assert sig'low = 0 severity failure;
  assert sig'length <= 12 severity failure;


  sig_valid <= not is_X(sig);
  
  check_setup : process(clk)
  begin
    if qualified and ((posedge and rising_edge(clk)) or (not posedge and falling_edge(clk))) then
      if not sig_valid then
        report model_name & ": " & get_sig_name(sig_name, sig'length) & " has at least one invalid bit at " & clk_name & " edge" severity error;
      end if;
      
      if not sig'stable(tsu) then
        report model_name & ": " & get_sig_name(sig_name, sig'length) & " setup violation at " & clk_name & " edge" severity error;
      end if;
      
      hold <= transport true, false after th;
    end if;
  end process;
  
  check_hold : process(sig)
  begin
    if hold then
      report model_name & ": " & get_sig_name(sig_name, sig'length) & " hold violation after " & clk_name & " edge" severity error;
    end if;
  end process;

  
end architecture;