--
-- Limitations:
-- 1. MIG must be configured so that there is one word of data transferred
--   on the MIG UI (app_wdf_data, app_rd_data) per command entered via
--   app_en. This implies that when app_rd_data_valid is asserted,
--   app_rd_data_end must also be asserted.
-- 2. In Kintex Ultrascale MIG 6.1 (Vivado 2014.4) with AXI4 interface and
--    ECC, write strobes do not appear to be honoured by MIG. The end result
--    is that all byte lanes appear to be written regardless. Therefore,
--    partial_writes should be false for that configuration; otherwise, the
--    memory test will fail.
--
-- Constraints on parameters:
-- 3. memory_data_width & m_axi_xdata_width must be the same.
-- 4. m_axi_axaddr_width should be equal to or greater than (memory_addr_width + log2(m_axi_xdata_width / 8)).
--    - If less, then a full memory test where ctl_length has the maximum
--      value will fail due to address wraparound.
--    - If more, then the entire MIG address space cannot be tested.
--
-- Notes:
-- 5. MIG physical data width is assumed to be m_axi_xdata_width / pack_factor.
--    pack_factor must be correct in order for phases that generate specific
--    patterns (e.g. maximum toggle rate on physical data pins) to generate
--    the required logical data.
-- 6. Currently only transaction ID 0 is used, so that MIG cannot return read
--    data out-of-order.
-- 7. This module has uses a preferred burst length of 2**m_axi_axlen_width,
--    and aligns burst start and end to a boundary of 2**m_axi_axlen_width * (m_axi_xdata_width / 8)
--    bytes when possible. It also honours the AXI4 rule that a burst cannot
--    cross a 4kB boundary.
-- 8. 'pack_rearranged' must be set as follows:
--    - 7 Series MIG DDR3 SDRAM interface   => false
--    - Ultrascale MIG DDR3 SDRAM interface => true
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

entity memory_exerciser_axi4 is
  generic(
    memory_data_width  : in    natural; -- Must equal m_axi_xdata_width
    memory_addr_width  : in    natural;
    m_axi_axaddr_width : in    natural;
    m_axi_xdata_width  : in    natural; -- Must equal memory_data_width
    m_axi_xid_width    : in    natural;
    m_axi_axlen_width  : in    natural;
    pack_factor        : in    natural;
    pack_rearranged    : in    boolean; -- See note 8 above.
    partial_writes     : in    boolean;
    error_count_width  : in    natural; -- log2 of error count
    error_index_width  : in    natural  -- log2 of number of errors that can be stored in error log
  );
  port(
    aresetn            : in    std_logic;
    aclk               : in    std_logic;
    -- AXI4 MM interface (master)
    m_axi_araddr       : out   std_logic_vector(m_axi_axaddr_width - 1 downto 0);
    m_axi_arburst      : out   std_logic_vector(1 downto 0);
    m_axi_arcache      : out   std_logic_vector(3 downto 0);
    m_axi_arid         : out   std_logic_vector(m_axi_xid_width - 1 downto 0);
    m_axi_arlen        : out   std_logic_vector(m_axi_axlen_width - 1 downto 0);
    m_axi_arlock       : out   std_logic_vector(0 downto 0);
    m_axi_arprot       : out   std_logic_vector(2 downto 0);
    m_axi_arqos        : out   std_logic_vector(3 downto 0);
    m_axi_arready      : in    std_logic;
    m_axi_arregion     : out   std_logic_vector(3 downto 0);
    m_axi_arsize       : out   std_logic_vector(2 downto 0);
    m_axi_arvalid      : out   std_logic;
    m_axi_awaddr       : out   std_logic_vector(m_axi_axaddr_width - 1 downto 0);
    m_axi_awburst      : out   std_logic_vector(1 downto 0);
    m_axi_awcache      : out   std_logic_vector(3 downto 0);
    m_axi_awid         : out   std_logic_vector(m_axi_xid_width - 1 downto 0);
    m_axi_awlen        : out   std_logic_vector(m_axi_axlen_width - 1 downto 0);
    m_axi_awlock       : out   std_logic_vector(0 downto 0);
    m_axi_awprot       : out   std_logic_vector(2 downto 0);
    m_axi_awqos        : out   std_logic_vector(3 downto 0);
    m_axi_awready      : in    std_logic;
    m_axi_awregion     : out   std_logic_vector(3 downto 0);
    m_axi_awsize       : out   std_logic_vector(2 downto 0);
    m_axi_awvalid      : out   std_logic;
    m_axi_wdata        : out   std_logic_vector(m_axi_xdata_width - 1 downto 0);
    m_axi_wlast        : out   std_logic;
    m_axi_wstrb        : out   std_logic_vector(m_axi_xdata_width / 8 - 1 downto 0);
    m_axi_wready       : in    std_logic;
    m_axi_wvalid       : out   std_logic;
    m_axi_bid          : in    std_logic_vector(m_axi_xid_width - 1 downto 0);
    m_axi_bready       : out   std_logic;
    m_axi_bresp        : in    std_logic_vector(1 downto 0);
    m_axi_bvalid       : in    std_logic;
    m_axi_rdata        : in    std_logic_vector(m_axi_xdata_width - 1 downto 0);
    m_axi_rid          : in    std_logic_vector(m_axi_xid_width - 1 downto 0);
    m_axi_rlast        : in    std_logic;
    m_axi_rready       : out   std_logic;
    m_axi_rresp        : in    std_logic_vector(1 downto 0);
    m_axi_rvalid       : in    std_logic;
    -- Control interface
    ctl_go             : in    std_logic;                                            -- Pulse high for at least one cycle to initiate test
    ctl_offset         : in    std_logic_vector(memory_addr_width - 1 downto 0);     -- Starting logical address of test
    ctl_length         : in    std_logic_vector(memory_addr_width - 1 downto 0);     -- Number of logical words to test, minus 1
    ctl_ph_en          : in    std_logic_vector(7 downto 0) := (others => '1');      -- Enables various phases of test
    ctl_done           : out   std_logic;                                            -- High indicates finished or idle
    -- When 'ctl_done' is asserted, assert 'ctl_erd' to read out an error (latency 2 clock cycles)
    ctl_erd            : in    std_logic;
    -- Index of error to read out
    ctl_eindex         : in    std_logic_vector(error_index_width - 1 downto 0);
    -- Qualified by 'ctl_done', high indicates at least one error occurred
    ctl_error          : out   std_logic;
    -- Qualified by 'ctl_done', actual number of errors
    ctl_ecount         : out   std_logic_vector(error_count_width - 1 downto 0);
    -- Qualified by 'ctl_done', asserted if error log is full
    ctl_efull          : out   std_logic;
    -- Qualified by 'ctl_done', asserted if errors were lost due to error log full
    ctl_elost          : out   std_logic;
    -- Address of selected error
    ctl_eaddr          : out   std_logic_vector(memory_addr_width - 1 downto 0);
    -- Phase in which selected error occurred
    ctl_ephase         : out   std_logic_vector(2 downto 0);
    -- Bits that were incorrect when selected error occurred
    ctl_ebits          : out   std_logic_vector(memory_data_width - 1 downto 0);
    -- Actual data when selected error occurred
    ctl_edata          : out   std_logic_vector(memory_data_width - 1 downto 0)
  );
end entity;

architecture rtl of memory_exerciser_axi4 is

  function min(
    constant a, b : in natural)
  return natural is begin
    if a < b then
      return a;
    else
      return b;
    end if;
  end;

  function choice(
    constant a, b : in std_logic_vector;
    constant selb : in boolean)
  return std_logic_vector is begin
    if selb then
      return b;
    else
      return a;
    end if;
  end;

  component memory_exerciser_data_gen
    generic(
      memory_data_width  : in    natural;
      memory_addr_width  : in    natural;
      pack_factor        : in    natural
    );
    port(
      rst                : in    std_logic;
      clk                : in    std_logic;
      -- Generated data/expected data
      d                  : out   std_logic_vector(memory_data_width - 1 downto 0);     -- Data when writing to memory
      dbe                : out   std_logic_vector(memory_data_width / 8 - 1 downto 0); -- Byte enables when writing to memory
      dvalid             : out   std_logic;                                            -- Qualifies 'd', 'dbe'
      exp                : out   std_logic_vector(memory_data_width - 1 downto 0);     -- Expected data when reading from memory
      expvalid           : out   std_logic;                                            -- Qualifies 'exp'
      enable             : in    std_logic;                                            -- When deasserted, 'valid' must deassert within a cycle or two
      -- Control interface
      ctl_init_rng       : in    std_logic;                                            -- Assert to initialise random number generators
      ctl_go             : in    std_logic;                                            -- Assert for >= 1 cycle while 'done' is asserted to initiate test
      ctl_reading        : in    std_logic;                                            -- 0 => writing to memory, 1 => reading from memory
      ctl_phase          : in    std_logic_vector(2 downto 0);                         -- Selects data generation method
      ctl_offset         : in    std_logic_vector(memory_addr_width - 1 downto 0);     -- Address at which to begin testing
      ctl_length         : in    std_logic_vector(memory_addr_width - 1 downto 0);     -- Number of logical words to test, minus 1
      ctl_idle           : out   std_logic                                             -- High indicates finished or idle
    );
  end component;

  component memory_exerciser_error_log
    generic(
      phase_width        : in    natural;
      memory_data_width  : in    natural;
      memory_addr_width  : in    natural;
      count_width        : in    natural;
      index_width        : in    natural
    );
    port(
      rst                : in    std_logic;
      clk                : in    std_logic;
      -- Clears the error log
      sr                 : in    std_logic;
      -- Set initial address & phase for current phase
      init               : in    std_logic;
      init_phase         : in    std_logic_vector(phase_width - 1 downto 0);
      init_offset        : in    std_logic_vector(memory_addr_width - 1 downto 0);
      -- Expected and actual data
      chk_ce             : in    std_logic;
      chk_act_data       : in    std_logic_vector(memory_data_width - 1 downto 0);
      chk_exp_data       : in    std_logic_vector(memory_data_width - 1 downto 0);
      chk_done           : out   std_logic;
      -- Error readout interface
      rd_ce              : in    std_logic;
      rd_index           : in    std_logic_vector(index_width - 1 downto 0);
      rd_not_empty       : out   std_logic;
      rd_full            : out   std_logic;
      rd_overflow        : out   std_logic;
      rd_count           : out   std_logic_vector(count_width - 1 downto 0);
      rd_rec_phase       : out   std_logic_vector(phase_width - 1 downto 0);
      rd_rec_addr        : out   std_logic_vector(memory_addr_width - 1 downto 0);
      rd_rec_bad_bits    : out   std_logic_vector(memory_data_width - 1 downto 0);
      rd_rec_act_data    : out   std_logic_vector(memory_data_width - 1 downto 0)
    );
  end component;

  component memory_exerciser_fifo
    generic(
      order       : in    natural range 2 to 6;
      width       : in    natural
    );
    port(
      rst         : in    std_logic;                             -- Asynchronous reset
      clk         : in    std_logic;                             -- Clock
      sr          : in    std_logic;                             -- Synchronous reset
      -- Writer interface
      wr          : in    std_logic;                             -- '1' => write value on 'd' into FIFO
      wd          : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
      wfull       : out   std_logic;                             -- '1' => full (from point of view of write side)
      -- Accounting interface
      aadv        : in    std_logic;                             -- Advance accounting pointer
      aempty      : out   std_logic;                             -- '1' => empty
      alevel      : out   std_logic_vector(ORDER downto 0);      -- Number of items in FIFO
      -- Reader interface
      radv        : in    std_logic;                             -- '1' => present next item on 'q' to next item
      rempty      : out   std_logic;                             -- '1' => empty
      rq          : out   std_logic_vector(WIDTH - 1 downto 0)   -- Current item, qualified by 'valid'
    );
  end component;

  component memory_exerciser_dearranger
    generic(
      memory_data_width  : in    natural;
      pack_factor        : in    natural;
      pack_rearranged    : in    boolean
    );
    port(
      data_in            : in    std_logic_vector(memory_data_width - 1 downto 0);
      data_out           : out   std_logic_vector(memory_data_width - 1 downto 0)
    );
  end component;
  
  component memory_exerciser_rearranger
    generic(
      memory_data_width  : in    natural;
      pack_factor        : in    natural;
      pack_rearranged    : in    boolean
    );
    port(
      data_in            : in    std_logic_vector(memory_data_width - 1 downto 0);
      mask_in            : in    std_logic_vector(memory_data_width / 8 - 1 downto 0);
      data_out           : out   std_logic_vector(memory_data_width - 1 downto 0);
      mask_out           : out   std_logic_vector(memory_data_width / 8 - 1 downto 0)
    );
  end component;
  
  constant num_axi_byte_lane : natural := m_axi_xdata_width / 8;
  constant log2_num_axi_byte_lane : natural := natural(floor(log2(real(num_axi_byte_lane) + 0.5)));

  constant phases_enable : std_logic_vector(7 downto 0) := choice("01111011", "11111111", partial_writes);

  signal rst : std_logic;

  -- Executive state machine
  type ex_state_t is (ex_s_idle, ex_s_init_phase1, ex_s_init_phase2, ex_s_run_phase, ex_s_finish);
  signal ex_state, n_ex_state : ex_state_t;
  signal ex_idle, n_ex_idle : std_logic;
  signal ex_init_rng, n_ex_init_rng : std_logic;
  signal ex_init_count, n_ex_init_count : std_logic_vector(2 downto 0);
  signal ex_phase, n_ex_phase : std_logic_vector(3 downto 0);
  signal ex_ph_go, n_ex_ph_go : std_logic;

  -- AXI4 command issue state machine
  -- Preferred burst length must not result in bursts that cross a 4k boundary.
  constant preferred_burst_order : natural := min(m_axi_axlen_width, 12 - log2_num_axi_byte_lane);
  constant preferred_burst_length : natural := 2**preferred_burst_order;
  type cmd_state_t is (cmd_s_idle, cmd_s_writing, cmd_s_reading);
  signal cmd_state, n_cmd_state : cmd_state_t;
  signal cmd_axaddr, n_cmd_axaddr : std_logic_vector(m_axi_axaddr_width - 1 downto log2_num_axi_byte_lane);
  signal cmd_arvalid, n_cmd_arvalid : std_logic;
  signal cmd_awvalid, n_cmd_awvalid : std_logic;
  signal cmd_axlen, n_cmd_axlen : std_logic_vector(preferred_burst_order - 1 downto 0);
  signal cmd_bcount, n_cmd_bcount : std_logic_vector(memory_addr_width - 1 downto preferred_burst_order);
  signal cmd_idle, n_cmd_idle : std_logic;
  signal cmd_last, n_cmd_last : std_logic;
  -- The following all have values that is one less than the actual value;
  -- e.g. if cmd_per_phase = 3, then there are 4 bursts per phase.
  -- This applies to cmd_per_phase & cmd_length_* .
  -- A *_eq_0 value is just the corresponding signal OR_reduce'd, so that it
  -- doesn't have to be OR_reduced in the combinatorial logic of the state
  -- machines that need these values.
  signal cmd_per_phase : std_logic_vector(memory_addr_width - 1 downto preferred_burst_order);
  signal cmd_per_phase_eq_0 : std_logic;
  signal cmd_length_first : std_logic_vector(preferred_burst_order - 1 downto 0);
  signal cmd_length_first_eq_0 : std_logic;
  constant cmd_length_middle : std_logic_vector(preferred_burst_order - 1 downto 0) := (others => '1');
  constant cmd_length_middle_eq_0 : std_logic := '0';
  signal cmd_length_last : std_logic_vector(preferred_burst_order - 1 downto 0);
  signal cmd_length_last_eq_0 : std_logic;

  -- AXI4 write data state machine
  type wd_state_t is (wd_s_idle, wd_s_writing);
  signal wd_state, n_wd_state : wd_state_t;
  signal wd_blen, n_wd_blen : std_logic_vector(preferred_burst_order - 1 downto 0);
  signal wd_bcount, n_wd_bcount : std_logic_vector(memory_addr_width - 1 downto preferred_burst_order);
  signal wd_en_wvalid, n_wd_en_wvalid : std_logic;
  signal wd_idle, n_wd_idle : std_logic;
  signal wd_wlast, n_wd_wlast : std_logic;

  -- AXI4 write response state machine
  type wr_state_t is (wr_s_idle, wr_s_writing);
  signal wr_state, n_wr_state : wr_state_t;
  signal wr_bready, n_wr_bready : std_logic;
  signal wr_bcount, n_wr_bcount : std_logic_vector(memory_addr_width - 1 downto preferred_burst_order);
  signal wr_idle, n_wr_idle : std_logic;

  -- AXI4 read response state machine
  type rr_state_t is (rr_s_idle, rr_s_reading, rr_s_drain);
  signal rr_state, n_rr_state : rr_state_t;
  signal rr_blen, n_rr_blen : std_logic_vector(preferred_burst_order - 1 downto 0);
  signal rr_bcount, n_rr_bcount : std_logic_vector(memory_addr_width - 1 downto preferred_burst_order);
  signal rr_drain_count, n_rr_drain_count : std_logic_vector(2 downto 0);
  signal rr_en_rready, n_rr_en_rready : std_logic;
  signal rr_idle, n_rr_idle : std_logic;
  signal rr_exp_rlast, n_rr_exp_rlast : std_logic;

  -- Latched versions of 'ctl_xxx' signals
  signal reg_offset : std_logic_vector(memory_addr_width - 1 downto 0) := (others => '0');
  signal reg_length : std_logic_vector(memory_addr_width - 1 downto 0) := (others => '0');
  signal reg_limit : std_logic_vector(memory_addr_width - 1 downto 0) := (others => '0');
  signal reg_ph_en : std_logic_vector(7 downto 0);
  
  constant memory_be_width : natural := memory_data_width / 8;

  -- Data generator
  signal gen_idle : std_logic;
  signal gen_d : std_logic_vector(memory_data_width - 1 downto 0);
  signal gen_dbe : std_logic_vector(memory_be_width - 1 downto 0);
  signal gen_dvalid : std_logic;
  signal gen_exp : std_logic_vector(memory_data_width - 1 downto 0);
  signal gen_expvalid : std_logic;
  signal gen_enable : std_logic;

  constant d_be_fifo_order : natural := 6;
  constant d_be_fifo_width : natural := memory_data_width + memory_be_width;
  signal d_be_fifo_wfull : std_logic;
  signal d_be_fifo_wd : std_logic_vector(d_be_fifo_width - 1 downto 0);
  signal d_be_fifo_radv : std_logic;
  signal d_be_fifo_rempty : std_logic;
  signal d_be_fifo_rq : std_logic_vector(d_be_fifo_width - 1 downto 0);

  constant exp_fifo_order : natural := 6;
  constant exp_fifo_width : natural := memory_data_width;
  signal exp_fifo_wfull : std_logic;
  signal exp_fifo_radv : std_logic;
  signal exp_fifo_rempty : std_logic;
  signal exp_fifo_rq : std_logic_vector(exp_fifo_width - 1 downto 0);

  signal clear_error_log : std_logic;
  signal check_data : std_logic;
  signal actual_data : std_logic_vector(memory_data_width - 1 downto 0);
  signal error_log_not_empty : std_logic;

  signal logic0, logic1 : std_logic;

begin

  assert m_axi_xdata_width = memory_data_width severity failure;

  assert m_axi_axaddr_width = memory_addr_width + log2_num_axi_byte_lane
    report "m_axi_axaddr_width and memory_addr_width imply different byte sizes. This is not necessarily a problem."
    severity warning;


  logic0 <= '0';
  logic1 <= '1';

  rst <= not aresetn;


  --
  -- Drive the AXI4 signals that nobody cares about with sensible values.
  --

  m_axi_arburst <= "01"; -- Incrementing
  m_axi_arcache <= "0011"; -- As per recommendation in Xilinx UG761
  m_axi_arlock <= (others => '0');
  m_axi_arprot <= (others => '0');
  m_axi_arqos <= (others => '0');
  m_axi_arregion <= (others => '0');
  m_axi_awburst <= "01"; -- Incrementing
  m_axi_awcache <= "0011"; -- As per recommendation in Xilinx UG761
  m_axi_awlock <= (others => '0');
  m_axi_awprot <= (others => '0');
  m_axi_awqos <= (others => '0');
  m_axi_awregion <= (others => '0');

  --
  -- Currently, we only use TID of 0 and never use narrow bursts.
  --

  m_axi_arid <= (others => '0');
  m_axi_arsize <= std_logic_vector(to_unsigned(log2_num_axi_byte_lane, m_axi_arsize'length));
  m_axi_awid <= (others => '0');
  m_axi_awsize <= std_logic_vector(to_unsigned(log2_num_axi_byte_lane, m_axi_awsize'length));


  --
  -- Register input parameters so that they are stable during the
  -- course of the test.
  --

  register_param : process(aclk)
    variable offset_low, length_low : std_logic_vector(preferred_burst_order - 1 downto 0);
  begin
    if rising_edge(aclk) then
      if ex_idle = '1' and ctl_go = '1' then
        reg_offset <= ctl_offset;
        reg_length <= ctl_length;
        reg_ph_en <= ctl_ph_en;
      end if;
      offset_low := reg_offset(preferred_burst_order - 1 downto 0);
      length_low := reg_length(preferred_burst_order - 1 downto 0);

      reg_limit <= reg_length + std_logic_vector(resize(unsigned(offset_low), memory_addr_width));

      -- Figure out number of AXI4 commands to issue per phase,
      -- and first & last AXI4 burst lengths of a phase.
      -- 1st burst ends at preferred address boundary, if possible.
      -- Align 2nd & later bursts to preferred address boundary, if possible.
      cmd_per_phase <= reg_limit(memory_addr_width - 1 downto preferred_burst_order);
      cmd_per_phase_eq_0 <= not OR_reduce(cmd_per_phase);
      if cmd_per_phase_eq_0 = '0' then
        -- Two or more AXI4 commands per phase
        cmd_length_first <= not offset_low;
        cmd_length_last <= reg_limit(preferred_burst_order - 1 downto 0);
        cmd_length_last_eq_0 <= not OR_reduce(cmd_length_last);
      else
        -- One AXI4 command per phase
        cmd_length_first <= length_low;
        cmd_length_last <= (others => '-');
        cmd_length_last_eq_0 <= '-';
      end if;
      cmd_length_first_eq_0 <= not OR_reduce(cmd_length_first);
    end if;
  end process;


  --
  -- Executive state machine
  --

  gen_ex_state : process(rst, aclk)
  begin
    if rst = '1' then
      ex_state <= ex_s_idle;
      ex_idle <= '1';
      ex_init_rng <= '0';
      ex_init_count <= (others => '-');
      ex_phase <= (others => '-');
      ex_ph_go <= '0';
    elsif rising_edge(aclk) then
      ex_state <= n_ex_state;
      ex_idle <= n_ex_idle;
      ex_init_rng <= n_ex_init_rng;
      ex_init_count <= n_ex_init_count;
      ex_phase <= n_ex_phase;
      ex_ph_go <= n_ex_ph_go;
      
      -- synthesis translate_off
      if n_ex_state = ex_s_init_phase2 and ex_phase(0) = '0' then
        report "memory_exerciser_axi4: Beginning test phase " & natural'image(to_integer(unsigned(ex_phase(ex_phase'high downto 1)))) & "." severity note;
      end if;
      
      if ex_state = ex_s_finish then
        if error_log_not_empty = '1' then
          report "memory_exerciser_axi4: Memory test completed with at least one error." severity error;
        else
          report "memory_exerciser_axi4: Memory test completed without error." severity note;
        end if;
      end if;
      -- synthesis translate_on
    end if;
  end process;

  ctl_done <= ex_idle;

  gen_n_ex_state : process(
    ex_state,
    ex_init_count,
    ex_phase,
    ctl_go,
    gen_idle,
    cmd_idle,
    wd_idle,
    wr_idle,
    rr_idle,
    reg_ph_en)
    variable tmp8 : std_logic_vector(7 downto 0);
  begin
    case ex_state is
      when ex_s_idle =>
        if ctl_go = '1' then
          n_ex_state <= ex_s_init_phase1;
          n_ex_idle <= '0';
          n_ex_init_count <= (others => '0');
          n_ex_init_rng <= '1';
          n_ex_phase <= (others => '0');
          n_ex_ph_go <= '0';
        else
          n_ex_state <= ex_s_idle;
          n_ex_idle <= '1';
          n_ex_init_count <= (others => '-');
          n_ex_init_rng <= '0';
          n_ex_phase <= (others => '-');
          n_ex_ph_go <= '0';
        end if;

      when ex_s_init_phase1 =>
        -- We pause for a few cycles in this state to allow calculations
        -- for cmd_per_phase & cmd_length_*, which are based on ctl_*
        -- parameters, to finish.
        n_ex_idle <= '0';
        n_ex_init_rng <= '0';
        n_ex_phase <= ex_phase;
        if AND_reduce(ex_init_count) = '1' then
          n_ex_state <= ex_s_init_phase2;
          n_ex_init_count <= (others => '-');
          tmp8 := phases_enable and reg_ph_en;
          n_ex_ph_go <= tmp8(to_integer(unsigned(ex_phase(3 downto 1))));
        else
          n_ex_state <= ex_s_init_phase1;
          n_ex_init_count <= ex_init_count + 1;
          n_ex_ph_go <= '0';
        end if;

      when ex_s_init_phase2 =>
        n_ex_state <= ex_s_run_phase;
        n_ex_idle <= '0';
        n_ex_init_count <= (others => '-');
        n_ex_init_rng <= '0';
        n_ex_phase <= ex_phase;
        n_ex_ph_go <= '0';

      when ex_s_run_phase =>
        n_ex_idle <= '0';
        n_ex_init_rng <= '0';
        if gen_idle = '1' and cmd_idle = '1' and wd_idle = '1' and wr_idle = '1' and rr_idle = '1' then
          if AND_reduce(ex_phase) = '1' then
            n_ex_state <= ex_s_finish;
            n_ex_init_count <= (others => '-');
            n_ex_phase <= (others => '-');
            n_ex_ph_go <= '0';
          else
            n_ex_state <= ex_s_init_phase1;
            n_ex_init_count <= (others => '0');
            n_ex_phase <= ex_phase + 1;
            n_ex_ph_go <= '0';
          end if;
        else
          n_ex_state <= ex_s_run_phase;
          n_ex_init_count <= (others => '-');
          n_ex_phase <= ex_phase;
          n_ex_ph_go <= '0';
        end if;

      when ex_s_finish =>
        n_ex_state <= ex_s_idle;
        n_ex_idle <= '1';
        n_ex_init_count <= (others => '-');
        n_ex_init_rng <= '0';
        n_ex_phase <= (others => '-');
        n_ex_ph_go <= '0';
    end case;
  end process;


  --
  -- AXI4 command issue state machine
  --

  gen_cmd_state : process(rst, aclk)
  begin
    if rst = '1' then
      cmd_state <= cmd_s_idle;
      cmd_axaddr <= (others => '-');
      cmd_arvalid <= '0';
      cmd_awvalid <= '0';
      cmd_axlen <= (others => '-');
      cmd_bcount <= (others => '-');
      cmd_idle <= '1';
      cmd_last <= '-';
    elsif rising_edge(aclk) then
      cmd_state <= n_cmd_state;
      cmd_axaddr <= n_cmd_axaddr;
      cmd_arvalid <= n_cmd_arvalid;
      cmd_awvalid <= n_cmd_awvalid;
      cmd_axlen <= n_cmd_axlen;
      cmd_bcount <= n_cmd_bcount;
      cmd_idle <= n_cmd_idle;
      cmd_last <= n_cmd_last;
    end if;
  end process;

  gen_n_cmd_state : process(
    cmd_state,
    cmd_axaddr,
    cmd_axlen,
    cmd_bcount,
    cmd_last,
    cmd_per_phase,
    cmd_per_phase_eq_0,
    cmd_length_first,
    cmd_length_last,
    ex_ph_go,
    ex_phase,
    reg_offset,
    m_axi_arready,
    m_axi_awready)
  begin
    case cmd_state is
      when cmd_s_idle =>
        if ex_ph_go = '1' then
          if ex_phase(0) = '1' then
            n_cmd_state <= cmd_s_reading;
            n_cmd_arvalid <= '1';
            n_cmd_awvalid <= '0';
          else
            n_cmd_state <= cmd_s_writing;
            n_cmd_arvalid <= '0';
            n_cmd_awvalid <= '1';
          end if;
          n_cmd_axaddr <= reg_offset;
          n_cmd_axlen <= cmd_length_first;
          n_cmd_bcount <= cmd_per_phase;
          n_cmd_last <= cmd_per_phase_eq_0;
          n_cmd_idle <= '0';
        else
          n_cmd_axaddr <= (others => '-');
          n_cmd_axlen <= (others => '-');
          n_cmd_state <= cmd_s_idle;
          n_cmd_arvalid <= '0';
          n_cmd_awvalid <= '0';
          n_cmd_bcount <= (others => '-');
          n_cmd_last <= '-';
          n_cmd_idle <= '1';
        end if;

      when cmd_s_writing =>
        if m_axi_awready = '1' then
          if cmd_last = '1' then
            n_cmd_state <= cmd_s_idle;
            n_cmd_arvalid <= '0';
            n_cmd_awvalid <= '0';
            n_cmd_axaddr <= (others => '-');
            n_cmd_axlen <= (others => '-');
            n_cmd_bcount <= (others => '-');
            n_cmd_last <= '-';
            n_cmd_idle <= '1';
          else
            n_cmd_state <= cmd_s_writing;
            n_cmd_arvalid <= '0';
            n_cmd_awvalid <= '1';
            n_cmd_axaddr <= cmd_axaddr + cmd_axlen + 1;
            n_cmd_bcount <= cmd_bcount - 1;
            n_cmd_idle <= '0';
            if OR_reduce(cmd_bcount(cmd_bcount'high downto cmd_bcount'low + 1)) = '1' then
              n_cmd_axlen <= cmd_length_middle;
              n_cmd_last <= '0';
            else
              n_cmd_axlen <= cmd_length_last;
              n_cmd_last <= '1';
            end if;
          end if;
        else
          n_cmd_state <= cmd_s_writing;
          n_cmd_arvalid <= '0';
          n_cmd_awvalid <= '1';
          n_cmd_axaddr <= cmd_axaddr;
          n_cmd_axlen <= cmd_axlen;
          n_cmd_bcount <= cmd_bcount;
          n_cmd_last <= cmd_last;
          n_cmd_idle <= '0';
        end if;

      when cmd_s_reading =>
        if m_axi_arready = '1' then
          if cmd_last = '1' then
            n_cmd_state <= cmd_s_idle;
            n_cmd_arvalid <= '0';
            n_cmd_awvalid <= '0';
            n_cmd_axaddr <= (others => '-');
            n_cmd_axlen <= (others => '-');
            n_cmd_bcount <= (others => '-');
            n_cmd_last <= '-';
            n_cmd_idle <= '1';
          else
            n_cmd_state <= cmd_s_reading;
            n_cmd_arvalid <= '1';
            n_cmd_awvalid <= '0';
            n_cmd_axaddr <= cmd_axaddr + cmd_axlen + 1;
            n_cmd_bcount <= cmd_bcount - 1;
            n_cmd_idle <= '0';
            if OR_reduce(cmd_bcount(cmd_bcount'high downto cmd_bcount'low + 1)) = '1' then
              n_cmd_axlen <= cmd_length_middle;
              n_cmd_last <= '0';
            else
              n_cmd_axlen <= cmd_length_last;
              n_cmd_last <= '1';
            end if;
          end if;
        else
          n_cmd_state <= cmd_s_reading;
          n_cmd_arvalid <= '1';
          n_cmd_awvalid <= '0';
          n_cmd_axaddr <= cmd_axaddr;
          n_cmd_axlen <= cmd_axlen;
          n_cmd_bcount <= cmd_bcount;
          n_cmd_last <= cmd_last;
          n_cmd_idle <= '0';
        end if;
    end case;
  end process;

  m_axi_arvalid <= cmd_arvalid;
  m_axi_araddr <= cmd_axaddr & std_logic_vector(to_unsigned(0, log2_num_axi_byte_lane));
  m_axi_arlen <= std_logic_vector(resize(unsigned(cmd_axlen), m_axi_axlen_width));
  m_axi_awvalid <= cmd_awvalid;
  m_axi_awaddr <= cmd_axaddr & std_logic_vector(to_unsigned(0, log2_num_axi_byte_lane));
  m_axi_awlen <= std_logic_vector(resize(unsigned(cmd_axlen), m_axi_axlen_width));


  --
  -- AXI4 write data state machine
  --

  gen_wd_state : process(rst, aclk)
  begin
    if rst = '1' then
      wd_state <= wd_s_idle;
      wd_blen <= (others => '-');
      wd_bcount <= (others => '-');
      wd_en_wvalid <= '0';
      wd_idle <= '1';
      wd_wlast <= '0';
    elsif rising_edge(aclk) then
      wd_state <= n_wd_state;
      wd_blen <= n_wd_blen;
      wd_bcount <= n_wd_bcount;
      wd_en_wvalid <= n_wd_en_wvalid;
      wd_idle <= n_wd_idle;
      wd_wlast <= n_wd_wlast;
    end if;
  end process;

  m_axi_wvalid <= wd_en_wvalid and not d_be_fifo_rempty;
  m_axi_wlast <= wd_wlast;

  gen_n_wd_state : process(
    wd_state,
    wd_bcount,
    wd_blen,
    wd_wlast,
    cmd_per_phase,
    cmd_length_first,
    cmd_length_first_eq_0,
    cmd_length_last,
    cmd_length_last_eq_0,
    d_be_fifo_rempty,
    ex_ph_go,
    ex_phase,
    m_axi_wready)
  begin
    case wd_state is
      when wd_s_idle =>
        if ex_ph_go = '1' and ex_phase(0) = '0' then
          n_wd_state <= wd_s_writing;
          n_wd_blen <= cmd_length_first;
          n_wd_bcount <= cmd_per_phase;
          n_wd_en_wvalid <= '1';
          n_wd_idle <= '0';
          n_wd_wlast <= cmd_length_first_eq_0;
        else
          n_wd_state <= wd_s_idle;
          n_wd_blen <= (others => '-');
          n_wd_bcount <= (others => '-');
          n_wd_en_wvalid <= '0';
          n_wd_idle <= '1';
          n_wd_wlast <= '0';
        end if;

      when wd_s_writing =>
        if m_axi_wready = '1' and d_be_fifo_rempty = '0' then
          if wd_wlast = '1' then
            if OR_reduce(wd_bcount) = '0' then
              n_wd_state <= wd_s_idle;
              n_wd_blen <= (others => '-');
              n_wd_bcount <= (others => '-');
              n_wd_en_wvalid <= '0';
              n_wd_idle <= '1';
              n_wd_wlast <= '0';
            else
              n_wd_state <= wd_s_writing;
              n_wd_bcount <= wd_bcount - 1;
              n_wd_en_wvalid <= '1';
              n_wd_idle <= '0';
              if OR_reduce(wd_bcount(wd_bcount'high downto wd_bcount'low + 1)) = '1' then
                n_wd_blen <= cmd_length_middle;
                n_wd_wlast <= cmd_length_middle_eq_0;
              else
                n_wd_blen <= cmd_length_last;
                n_wd_wlast <= cmd_length_last_eq_0;
              end if;
            end if;
          else
            n_wd_state <= wd_s_writing;
            n_wd_blen <= wd_blen - 1;
            n_wd_bcount <= wd_bcount;
            n_wd_en_wvalid <= '1';
            n_wd_idle <= '0';
            n_wd_wlast <= not OR_reduce(wd_blen(wd_blen'high downto 1));
          end if;
        else
          n_wd_state <= wd_s_writing;
          n_wd_blen <= wd_blen;
          n_wd_bcount <= wd_bcount;
          n_wd_en_wvalid <= '1';
          n_wd_idle <= '0';
          n_wd_wlast <= wd_wlast;
        end if;
    end case;
  end process;


  --
  -- AXI4 write response state machine
  --

  gen_wr_state : process(rst, aclk)
  begin
    if rst = '1' then
      wr_state <= wr_s_idle;
      wr_bready <= '0';
      wr_bcount <= (others => '-');
      wr_idle <= '1';
    elsif rising_edge(aclk) then
      wr_state <= n_wr_state;
      wr_bready <= n_wr_bready;
      wr_bcount <= n_wr_bcount;
      wr_idle <= n_wr_idle;
    end if;
  end process;

  m_axi_bready <= wr_bready;

  gen_n_wr_state : process(
    wr_state,
    wr_bcount,
    cmd_per_phase,
    ex_ph_go,
    ex_phase,
    m_axi_bvalid)
  begin
    case wr_state is
      when wr_s_idle =>
        if ex_ph_go = '1' and ex_phase(0) = '0' then
          n_wr_state <= wr_s_writing;
          n_wr_bready <= '1';
          n_wr_bcount <= cmd_per_phase;
          n_wr_idle <= '0';
        else
          n_wr_state <= wr_s_idle;
          n_wr_bready <= '0';
          n_wr_bcount <= (others => '-');
          n_wr_idle <= '1';
        end if;

      when wr_s_writing =>
        if m_axi_bvalid = '1' then
          if OR_reduce(wr_bcount) = '0' then
            n_wr_state <= wr_s_idle;
            n_wr_bready <= '0';
            n_wr_bcount <= (others => '-');
            n_wr_idle <= '1';
          else
            n_wr_state <= wr_s_writing;
            n_wr_bready <= '1';
            n_wr_bcount <= wr_bcount - 1;
            n_wr_idle <= '0';
          end if;
        else
          n_wr_state <= wr_s_writing;
          n_wr_bready <= '1';
          n_wr_bcount <= wr_bcount;
          n_wr_idle <= '0';
        end if;
    end case;
  end process;

  -- TO DO: Check m_axi_bid & m_axi_bresp.


  --
  -- AXI4 read response state machine
  --

  gen_rr_state : process(rst, aclk)
  begin
    if rst = '1' then
      rr_state <= rr_s_idle;
      rr_blen <= (others => '-');
      rr_bcount <= (others => '-');
      rr_drain_count <= (others => '-');
      rr_en_rready <= '0';
      rr_idle <= '1';
      rr_exp_rlast <= '0';
    elsif rising_edge(aclk) then
      rr_state <= n_rr_state;
      rr_blen <= n_rr_blen;
      rr_bcount <= n_rr_bcount;
      rr_drain_count <= n_rr_drain_count;
      rr_en_rready <= n_rr_en_rready;
      rr_idle <= n_rr_idle;
      rr_exp_rlast <= n_rr_exp_rlast;
    end if;
  end process;

  m_axi_rready <= rr_en_rready and not exp_fifo_rempty;

  gen_n_rr_state : process(
    rr_state,
    rr_bcount,
    rr_blen,
    rr_drain_count,
    rr_exp_rlast,
    cmd_per_phase,
    cmd_length_first,
    cmd_length_first_eq_0,
    cmd_length_last,
    cmd_length_last_eq_0,
    exp_fifo_rempty,
    ex_ph_go,
    ex_phase,
    m_axi_rvalid)
  begin
    case rr_state is
      when rr_s_idle =>
        n_rr_drain_count <= (others => '-');
        if ex_ph_go = '1' and ex_phase(0) = '1' then
          n_rr_state <= rr_s_reading;
          n_rr_blen <= cmd_length_first;
          n_rr_bcount <= cmd_per_phase;
          n_rr_en_rready <= '1';
          n_rr_idle <= '0';
          n_rr_exp_rlast <= cmd_length_first_eq_0;
        else
          n_rr_state <= rr_s_idle;
          n_rr_blen <= (others => '-');
          n_rr_bcount <= (others => '-');
          n_rr_en_rready <= '0';
          n_rr_idle <= '1';
          n_rr_exp_rlast <= '0';
        end if;

      when rr_s_reading =>
        if m_axi_rvalid = '1' and exp_fifo_rempty = '0' then
          if rr_exp_rlast = '1' then
            if OR_reduce(rr_bcount) = '0' then
              n_rr_state <= rr_s_drain;
              n_rr_blen <= (others => '-');
              n_rr_bcount <= (others => '-');
              n_rr_drain_count <= (others => '0');
              n_rr_en_rready <= '0';
              n_rr_idle <= '0';
              n_rr_exp_rlast <= '0';
            else
              n_rr_state <= rr_s_reading;
              n_rr_bcount <= rr_bcount - 1;
              n_rr_drain_count <= (others => '-');
              n_rr_en_rready <= '1';
              n_rr_idle <= '0';
              if OR_reduce(rr_bcount(rr_bcount'high downto rr_bcount'low + 1)) = '1' then
                n_rr_blen <= cmd_length_middle;
                n_rr_exp_rlast <= cmd_length_middle_eq_0;
              else
                n_rr_blen <= cmd_length_last;
                n_rr_exp_rlast <= cmd_length_last_eq_0;
              end if;
            end if;
          else
            n_rr_state <= rr_s_reading;
            n_rr_blen <= rr_blen - 1;
            n_rr_bcount <= rr_bcount;
            n_rr_drain_count <= (others => '-');
            n_rr_en_rready <= '1';
            n_rr_idle <= '0';
            n_rr_exp_rlast <= not OR_reduce(rr_blen(rr_blen'high downto 1));
          end if;
        else
          n_rr_state <= rr_s_reading;
          n_rr_blen <= rr_blen;
          n_rr_bcount <= rr_bcount;
          n_rr_drain_count <= (others => '-');
          n_rr_en_rready <= '1';
          n_rr_idle <= '0';
          n_rr_exp_rlast <= rr_exp_rlast;
        end if;

      when rr_s_drain =>
        -- Here we wait for the pipeline in the data checker to drain.
        if AND_reduce(rr_drain_count) = '1' then
          n_rr_state <= rr_s_idle;
          n_rr_blen <= (others => '-');
          n_rr_bcount <= (others => '-');
          n_rr_drain_count <= (others => '-');
          n_rr_en_rready <= '0';
          n_rr_idle <= '1';
          n_rr_exp_rlast <= '0';
        else
          n_rr_state <= rr_s_drain;
          n_rr_blen <= (others => '-');
          n_rr_bcount <= (others => '-');
          n_rr_drain_count <= rr_drain_count + 1;
          n_rr_en_rready <= '0';
          n_rr_idle <= '0';
          n_rr_exp_rlast <= '0';
        end if;
    end case;
  end process;

  -- TO DO: check m_axi_rlast against rd_exp_rlast.
  -- TO DO: check m_axi_rid.
  -- TO DO: check m_axi_rresp.


  --
  -- Data generator
  --
  -- Generates either:
  -- (1) write data & byte enables when ctl_reading deasserted, or
  -- (2) expected read data when ctl_reading asserted
  --

  -- Flow control, to avoid overflowing FIFOs
  gen_enable <= not (d_be_fifo_wfull or exp_fifo_wfull);

  data_gen_inst : memory_exerciser_data_gen
    generic map(
      memory_data_width => memory_data_width,
      memory_addr_width => memory_addr_width,
      pack_factor => pack_factor
    )
    port map(
      rst => rst,
      clk => aclk,
      d => gen_d,
      dbe => gen_dbe,
      dvalid => gen_dvalid,
      exp => gen_exp,
      expvalid => gen_expvalid,
      enable => gen_enable,
      ctl_init_rng => ex_init_rng,
      ctl_go => ex_ph_go,
      ctl_reading => ex_phase(0),
      ctl_phase => ex_phase(3 downto 1),
      ctl_offset => reg_offset,
      ctl_length => reg_length,
      ctl_idle => gen_idle
    );


  --
  -- FIFO for write data & byte enables
  --

  d_be_fifo_wd <= gen_dbe & gen_d;
  d_be_fifo_radv <= wd_en_wvalid and not d_be_fifo_rempty and m_axi_wready;

  d_be_fifo_inst : memory_exerciser_fifo
    generic map(
      order => d_be_fifo_order,
      width => d_be_fifo_width
    )
    port map(
      rst => rst,
      clk => aclk,
      sr => ex_idle,
      wr => gen_dvalid,
      wd => d_be_fifo_wd,
      wfull => d_be_fifo_wfull,
      aadv => d_be_fifo_radv,
      aempty => open,
      alevel => open,
      radv => d_be_fifo_radv,
      rempty => d_be_fifo_rempty,
      rq => d_be_fifo_rq
    );

  -- Check d_be_fifo_radv & d_be_fifo_rempty
  check_d_be_fifo : process(rst, aclk)
  begin
    if rst = '1' then
      null;
    elsif rising_edge(aclk) then
      if ex_idle = '0' then
        if d_be_fifo_radv = '1' then
          assert d_be_fifo_rempty = '0' severity failure;
        end if;
      end if;
    end if;
  end process;

  rearranger_inst : memory_exerciser_rearranger
    generic map(
      memory_data_width => memory_data_width,
      pack_factor => pack_factor,
      pack_rearranged => pack_rearranged
    )
    port map(
      data_in => d_be_fifo_rq(memory_data_width - 1 downto 0),
      mask_in => d_be_fifo_rq(memory_data_width + memory_be_width - 1 downto memory_data_width),
      data_out => m_axi_wdata,
      mask_out => m_axi_wstrb
    );

  --
  -- FIFO for expected read data
  --

  exp_fifo_radv <= rr_en_rready and m_axi_rvalid and not exp_fifo_rempty;

  exp_fifo_inst : memory_exerciser_fifo
    generic map(
      order => exp_fifo_order,
      width => exp_fifo_width
    )
    port map(
      rst => rst,
      clk => aclk,
      sr => ex_idle,
      wr => gen_expvalid,
      wd => gen_exp,
      wfull => exp_fifo_wfull,
      aadv => exp_fifo_radv,
      aempty => open,
      alevel => open,
      radv => exp_fifo_radv,
      rempty => exp_fifo_rempty,
      rq => exp_fifo_rq
    );

  -- Check exp_fifo_radv & exp_fifo_rempty
  check_exp_fifo : process(rst, aclk)
  begin
    if rst = '1' then
      null;
    elsif rising_edge(aclk) then
      if ex_idle = '0' then
        if exp_fifo_radv = '1' then
          assert exp_fifo_rempty = '0' severity failure;
        end if;
      end if;
    end if;
  end process;


  --
  -- Verification of data from AXI4 slave
  --
  -- Compares actual read data from AXI4 slave with expected data from
  -- FIFO, and generates error signals including address of error
  -- and byte lanes that are in error.
  --

  clear_error_log <= ex_idle and ctl_go;

  check_data <= rr_en_rready and not exp_fifo_rempty and m_axi_rvalid;

  dearranger_inst : memory_exerciser_dearranger
    generic map(
      memory_data_width => memory_data_width,
      pack_factor => pack_factor,
      pack_rearranged => pack_rearranged
    )
    port map(
      data_in => m_axi_rdata,
      data_out => actual_data
    );
    
  error_log_inst : memory_exerciser_error_log
    generic map(
      phase_width => ex_phase'length - 1,
      memory_data_width => memory_data_width,
      memory_addr_width => memory_addr_width,
      count_width => error_count_width,
      index_width => error_index_width
    )
    port map(
      rst => rst,
      clk => aclk,
      sr => clear_error_log,
      init => ex_ph_go,
      init_phase => ex_phase(ex_phase'high downto 1),
      init_offset => reg_offset,
      chk_ce => check_data,
      chk_act_data => actual_data,
      chk_exp_data => exp_fifo_rq,
      chk_done => open,
      rd_ce => ctl_erd,
      rd_index => ctl_eindex,
      rd_not_empty => error_log_not_empty,
      rd_full => ctl_efull,
      rd_overflow => ctl_elost,
      rd_count => ctl_ecount,
      rd_rec_phase => ctl_ephase,
      rd_rec_addr => ctl_eaddr,
      rd_rec_bad_bits => ctl_ebits,
      rd_rec_act_data => ctl_edata
    );

  ctl_error <= error_log_not_empty;
  
  
  -- TO DO: check m_axi_rresp, m_axi_rlast


end architecture;
