--
-- Phases:
-- 0     => Single '1' bit walking across all physical data bits
-- 1     => 55/AA, changing every 16 words or so
-- 2     => AA/55, changing every 16 words or so, with single '1' bit walking across byte enables
-- 3     => Own address
-- 4     => 55/AA toggling at max. rate
-- 5     => 00/FF toggling at max. rate
-- 6     => Pseudorandom
-- 7     => Pseudorandom with randomised byte enables
-- other => reserved
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

entity memory_exerciser_data_gen is
  generic(
    memory_data_width  : in    natural;
    memory_addr_width  : in    natural;
    pack_factor        : in    natural
  );
  port(
    rst                : in    std_logic;
    clk                : in    std_logic;
    -- Generated data/expected data
    d                  : out   std_logic_vector(memory_data_width - 1 downto 0);      -- Data when writing to memory
    dbe                : out   std_logic_vector(memory_data_width / 8 - 1 downto 0);  -- Byte enables when writing to memory
    dvalid             : out   std_logic;                                             -- Qualifies 'd', 'dbe'
    exp                : out   std_logic_vector(memory_data_width - 1 downto 0);      -- Expected data when reading from memory
    expvalid           : out   std_logic;                                             -- Qualifies 'exp'
    enable             : in    std_logic;                                             -- When deasserted, 'valid' must deassert within a cycle or two
    -- Control interface
    ctl_init_rng       : in    std_logic;                                             -- Assert to initialise random number generators
    ctl_go             : in    std_logic;                                             -- Assert for >= 1 cycle while 'done' is asserted to initiate test
    ctl_reading        : in    std_logic;                                             -- 0 => writing to memory, 1 => reading from memory
    ctl_phase          : in    std_logic_vector(2 downto 0);                          -- Selects data generation method
    ctl_offset         : in    std_logic_vector(memory_addr_width - 1 downto 0);      -- Address at which to begin testing
    ctl_length         : in    std_logic_vector(memory_addr_width - 1 downto 0);      -- Number of logical words to test, minus 1
    ctl_idle           : out   std_logic                                              -- High indicates finished or idle
  );
end entity;

architecture rtl of memory_exerciser_data_gen is

  component  memory_exerciser_rng
    generic(
      seed               : in    std_logic_vector(31 downto 0)
    );
    port(
      sr                 : in    std_logic;
      clk                : in    std_logic;
      ce                 : in    std_logic;
      ready              : out   std_logic;
      x                  : out   std_logic_vector(31 downto 0)
    );
  end component;

  constant pack_factor_log2 : natural := natural(floor(log2(real(pack_factor) + 0.5)));
  constant phys_addr_width : natural := memory_addr_width + pack_factor_log2;
  subtype phys_addr_t is std_logic_vector(phys_addr_width - 1 downto 0);

  constant memory_be_width : natural := memory_data_width / 8;
  subtype be_word_t is std_logic_vector(memory_be_width - 1 downto 0);
  subtype data_word_t is std_logic_vector(memory_data_width - 1 downto 0);

  constant phys_data_width : natural := memory_data_width / pack_factor;
  subtype phys_data_t is std_logic_vector(phys_data_width - 1 downto 0);
  constant phys_be_width : natural := phys_data_width / 8;
  subtype phys_be_t is std_logic_vector(phys_be_width - 1 downto 0);

  function gen_55_word return phys_data_t is
    variable tmp : phys_data_t;
  begin
    for i in 0 to phys_data_width - 1 loop
      if i mod 2 = 0 then
        tmp(i) := '1';
      else
        tmp(i) := '0';
      end if;
    end loop;
    return tmp;
  end;

  constant phase_width : natural := ctl_phase'length;
  constant num_phase : natural := 2**phase_width;
  signal reg_phase_1h : std_logic_vector(num_phase - 1 downto 0);
  signal reg_offset : std_logic_vector(memory_addr_width - 1 downto 0);
  signal reg_length : std_logic_vector(memory_addr_width - 1 downto 0);
  signal reg_reading : std_logic;

  -- Executive state machine
  type ex_state_t is (ex_s_idle, ex_s_write_init1, ex_s_write_init2, ex_s_writing, ex_s_read_init1, ex_s_read_init2, ex_s_reading, ex_s_finish);
  signal ex_state, n_ex_state : ex_state_t;
  signal ex_count, n_ex_count : std_logic_vector(memory_addr_width - 1 downto 0);
  signal ex_idle, n_ex_idle : std_logic;
  signal ex_generate_en, n_ex_generate_en : std_logic;
  signal ex_gen_ce : std_logic;
  signal ex_init, n_ex_init : std_logic;
  signal ex_init_q : std_logic;
  signal ex_pos, n_ex_pos : std_logic_vector(memory_addr_width - 1 downto 0);
  signal ex_writing, n_ex_writing : std_logic;

  -- Data generation
  signal ready : std_logic;
  -- Walking 1
  signal d_walk_1_dq : data_word_t;
  signal be_walk_1_dq : be_word_t;
  signal dvalid_walk_1_dq : std_logic;
  signal exp_walk_1_dq : data_word_t;
  signal expvalid_walk_1_dq : std_logic;
  -- 55 pattern, changing to AA pattern every 4096 logical words
  signal d_55_aa_slow : data_word_t;
  signal be_55_aa_slow : be_word_t;
  signal dvalid_55_aa_slow : std_logic;
  signal exp_55_aa_slow : data_word_t;
  signal expvalid_55_aa_slow : std_logic;
  -- AA pattern, changing to 55 pattern every 4096 logical words, with stippled byte enables
  signal d_aa_55_stipple : data_word_t;
  signal be_aa_55_stipple : be_word_t;
  signal dvalid_aa_55_stipple : std_logic;
  signal exp_aa_55_stipple : data_word_t;
  signal expvalid_aa_55_stipple : std_logic;
  -- Own address pattern
  signal d_own_addr : data_word_t := (others => '0');
  signal be_own_addr : be_word_t;
  signal dvalid_own_addr : std_logic;
  signal exp_own_addr : data_word_t;
  signal expvalid_own_addr : std_logic;
  -- 55/AA pattern toggling at maximum rate on physical data bus
  signal d_55_aa_tmax : data_word_t;
  signal be_55_aa_tmax : be_word_t;
  signal dvalid_55_aa_tmax : std_logic;
  signal exp_55_aa_tmax : data_word_t;
  signal expvalid_55_aa_tmax : std_logic;
  -- 00/FF pattern toggling at maximum rate on physical data bus
  signal d_00_ff_tmax : data_word_t;
  signal be_00_ff_tmax : be_word_t;
  signal dvalid_00_ff_tmax : std_logic;
  signal exp_00_ff_tmax : data_word_t;
  signal expvalid_00_ff_tmax : std_logic;
  -- Pseudorandom data
  signal d_random : data_word_t;
  signal be_random : be_word_t;
  signal dvalid_random : std_logic;
  signal exp_random : data_word_t;
  signal expvalid_random : std_logic;
  signal ready_rng_random : std_logic;
  -- Pseudorandom data, with stippled byte enables
  signal d_random_stipple : data_word_t;
  signal be_random_stipple : be_word_t;
  signal dvalid_random_stipple : std_logic;
  signal exp_random_stipple : data_word_t;
  signal expvalid_random_stipple : std_logic;
  signal ready_rng_random_stipple: std_logic;

  constant zeroes36 : std_logic_vector(35 downto 0) := (others => '0');


begin

  ctl_idle <= ex_idle;

  register_param : process(clk)
  begin
    if rising_edge(clk) then
      if ex_idle = '1' AND ctl_go = '1' then
        reg_reading <= ctl_reading;
        reg_offset <= ctl_offset;
        reg_length <= ctl_length;
        for i in 0 to num_phase - 1 loop
          if ctl_phase = i then
            reg_phase_1h(i) <= '1';
          else
            reg_phase_1h(i) <= '0';
          end if;
        end loop;
        
      end if;
      
      ready <= or_reduce(reg_phase_1h(5 downto 0)) or
               (reg_phase_1h(6) and ready_rng_random) or
               (reg_phase_1h(7) and ready_rng_random_stipple);
    end if;
  end process;

  gen_ex_state : process(rst, clk)
  begin
    if rst = '1' then
      ex_state <= ex_s_idle;
      ex_count <= (others => '-');
      ex_idle <= '0';
      ex_generate_en <= '0';
      ex_init <= '0';
      ex_init_q <= '0';
      ex_pos <= (others => '-');
      ex_writing <= '-';
    elsif rising_edge(clk) then
      ex_state <= n_ex_state;
      ex_count <= n_ex_count;
      ex_idle <= n_ex_idle;
      ex_generate_en <= n_ex_generate_en;
      ex_init <= n_ex_init;
      ex_init_q <= ex_init;
      ex_pos <= n_ex_pos;
      ex_writing <= n_ex_writing;
    end if;
  end process;

  gen_n_ex_state : process(
    ex_state,
    ctl_go,
    ctl_reading,
    reg_length,
    enable,
    ready,
    ex_count,
    ex_pos)
  begin
    case ex_state is
      when ex_s_idle =>
        if ctl_go = '1' then
          if ctl_reading = '1' then
            n_ex_state <= ex_s_read_init1;
          else
            n_ex_state <= ex_s_write_init1;
          end if;
          n_ex_count <= (others => '-');
          n_ex_idle <= '0';
          n_ex_generate_en <= '0';
          n_ex_init <= '1';
          n_ex_pos <= (others => '-');
          n_ex_writing <= '1';
        else
          n_ex_state <= ex_s_idle;
          n_ex_count <= (others => '-');
          n_ex_idle <= '1';
          n_ex_generate_en <= '0';
          n_ex_init <= '0';
          n_ex_pos <= (others => '-');
          n_ex_writing <= '-';
        end if;

      when ex_s_write_init1 =>
        n_ex_state <= ex_s_write_init2;
        n_ex_count <= (others => '-');
        n_ex_idle <= '0';
        n_ex_generate_en <= '0';
        n_ex_init <= '0';
        n_ex_pos <= (others => '0');
        n_ex_writing <= '1';

      when ex_s_write_init2 =>
        if ready = '1' then
          n_ex_state <= ex_s_writing;
          n_ex_count <= reg_length;
          n_ex_generate_en <= '1';
        else
          n_ex_state <= ex_s_write_init2;
          n_ex_count <= (others => '-');
          n_ex_generate_en <= '0';
        end if;
        n_ex_idle <= '0';
        n_ex_init <= '0';
        n_ex_pos <= (others => '0');
        n_ex_writing <= '1';

      when ex_s_writing =>
        n_ex_idle <= '0';
        if enable = '1' then
          if ex_count = 0 then
            n_ex_state <= ex_s_finish;
            n_ex_count <= (others => '-');
            n_ex_generate_en <= '0';
            n_ex_init <= '0';
            n_ex_pos <= (others => '-');
            n_ex_writing <= '0';
          else
            n_ex_state <= ex_s_writing;
            n_ex_count <= ex_count - 1;
            n_ex_generate_en <= '1';
            n_ex_init <= '0';
            n_ex_pos <= ex_pos + 1;
            n_ex_writing <= '1';
          end if;
        else
          n_ex_state <= ex_s_writing;
          n_ex_count <= ex_count;
          n_ex_generate_en <= '1';
          n_ex_init <= '0';
          n_ex_pos <= ex_pos;
          n_ex_writing <= '1';
        end if;

      when ex_s_read_init1 =>
        n_ex_state <= ex_s_read_init2;
        n_ex_count <= (others => '-');
        n_ex_idle <= '0';
        n_ex_generate_en <= '0';
        n_ex_init <= '0';
        n_ex_pos <= (others => '0');
        n_ex_writing <= '0';

      when ex_s_read_init2 =>
        if ready = '1' then
          n_ex_state <= ex_s_reading;
          n_ex_count <= reg_length;
          n_ex_generate_en <= '1';
        else
          n_ex_state <= ex_s_read_init2;
          n_ex_count <= (others => '-');
          n_ex_generate_en <= '0';
        end if;
        n_ex_idle <= '0';
        n_ex_init <= '0';
        n_ex_pos <= (others => '0');
        n_ex_writing <= '0';

      when ex_s_reading =>
        n_ex_idle <= '0';
        n_ex_init <= '0';
        n_ex_writing <= '0';
        if enable = '1' then
          if ex_count = 0 then
            n_ex_state <= ex_s_finish;
            n_ex_count <= (others => '-');
            n_ex_generate_en <= '0';
            n_ex_pos <= (others => '-');
          else
            n_ex_state <= ex_s_reading;
            n_ex_count <= ex_count - 1;
            n_ex_generate_en <= '1';
            n_ex_pos <= ex_pos + 1;
          end if;
        else
          n_ex_state <= ex_s_reading;
          n_ex_count <= ex_count;
          n_ex_generate_en <= '1';
          n_ex_pos <= ex_pos;
        end if;

      when ex_s_finish =>
        n_ex_state <= ex_s_idle;
        n_ex_count <= (others => '-');
        n_ex_idle <= '1';
        n_ex_generate_en <= '0';
        n_ex_init <= '0';
        n_ex_pos <= (others => '-');
        n_ex_writing <= '0';
    end case;
  end process;

  ex_gen_ce <= ex_generate_en AND enable;


  --
  -- Phase 0 data generation
  --

  gen_d_walk_1_dq : process(clk)
    variable tmp_data : phys_data_t;
  begin
    if rising_edge(clk) then
      if reg_phase_1h(0) = '1' then
        if ex_init = '1' then
          for i in 0 to pack_factor - 1 loop
            tmp_data := (others => '0');
            tmp_data(i) := '1';
            d_walk_1_dq(phys_data_width * (i + 1) - 1 downto phys_data_width * i) <= tmp_data;
          end loop;
        else
          if ex_gen_ce = '1' then
            for i in 0 to pack_factor - 1 loop
              d_walk_1_dq(phys_data_width * i) <= d_walk_1_dq(phys_data_width * (i + 1) - 1);
              d_walk_1_dq(phys_data_width * (i + 1) - 1 downto phys_data_width * i + 1) <= d_walk_1_dq(phys_data_width * (i + 1) - 2 downto phys_data_width * i);
            end loop;
          end if;
        end if;
      end if;
    end if;
  end process;

  be_walk_1_dq <= (others => '1');
  dvalid_walk_1_dq <= ex_gen_ce and reg_phase_1h(0) and not reg_reading;

  exp_walk_1_dq <= d_walk_1_dq;
  expvalid_walk_1_dq <= ex_gen_ce and reg_phase_1h(0) and reg_reading;

  --
  -- Phase 1 data generation
  --
  -- This has data where each physical data word is initially a 0x55 pattern. After
  -- 4k data words, it inverts.
  --

  gen_d_55_aa_slow : process(clk)
    constant data_55 : phys_data_t := gen_55_word;
  begin
    if rising_edge(clk) then
      if reg_phase_1h(1) = '1' or reg_phase_1h(2) = '1' then
        if ex_init = '1' then
          for i in 0 to pack_factor - 1 loop
            d_55_aa_slow(phys_data_width * (i + 1) - 1 downto phys_data_width * i) <= data_55;
          end loop;
        else
          if ex_gen_ce = '1' then
            if AND_reduce(ex_pos(11 downto 0)) = '1' then
              d_55_aa_slow <= not d_55_aa_slow;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  be_55_aa_slow <= (others => '1');
  dvalid_55_aa_slow <= ex_gen_ce and reg_phase_1h(1) and not reg_reading;

  exp_55_aa_slow <= d_55_aa_slow;
  expvalid_55_aa_slow <= ex_gen_ce and reg_phase_1h(1) and reg_reading;

  --
  -- Phase 2 data generation
  --
  -- This has opposite data to phase 1, but mostly zero byte enables. A
  -- single '1' bit is walked across the byte enables, rotating left by 1
  -- bit on each successive physical data word.
  --

  gen_d_aa_55_stipple : process(clk)
    constant data_aa : phys_data_t := not gen_55_word;
    variable tmp_be : phys_be_t;
  begin
    if rising_edge(clk) then
      if reg_phase_1h(2) = '1' then
        if ex_init = '1' then
          for i in 0 to pack_factor - 1 loop
            d_aa_55_stipple(phys_data_width * (i + 1) - 1 downto phys_data_width * i) <= data_aa;
            tmp_be := (others => '0');
            tmp_be(i mod phys_be_width) := '1';
            be_aa_55_stipple(phys_be_width * (i + 1) - 1 downto phys_be_width * i) <= tmp_be;
          end loop;
        else
          if ex_gen_ce = '1' then
            if AND_reduce(ex_pos(11 downto 0)) = '1' then
              d_aa_55_stipple <= NOT d_aa_55_stipple;
            end if;
            for i in 0 to pack_factor - 1 loop
              be_aa_55_stipple(phys_be_width * (i + 1) - 1 downto phys_be_width * i + 1) <= be_aa_55_stipple(phys_be_width * (i + 1) - 2 downto phys_be_width * i);
              be_aa_55_stipple(phys_be_width * i) <= be_aa_55_stipple(phys_be_width * (i + 1) - 1);
            end loop;
          end if;
        end if;
      end if;
    end if;
  end process;

  dvalid_aa_55_stipple <= ex_gen_ce and reg_phase_1h(2) and not reg_reading;

  gen_exp_aa_55_stipple : process(clk)
  begin
    if rising_edge(clk) then
      if reg_phase_1h(2) = '1' then
        if ex_gen_ce = '1' then
          for i in 0 to memory_be_width - 1 loop
            if be_aa_55_stipple(i) = '1' then
              exp_aa_55_stipple(8 * (i + 1) - 1 downto 8 * i) <= d_aa_55_stipple(8 * (i + 1) - 1 downto 8 * i);
            else
              exp_aa_55_stipple(8 * (i + 1) - 1 downto 8 * i) <= d_55_aa_slow(8 * (i + 1) - 1 downto 8 * i);
            end if;
          end loop;
        end if;
      end if;

      expvalid_aa_55_stipple <= reg_phase_1h(2) and ex_gen_ce and reg_reading;
    end if;
  end process;


  --
  -- Phase 3 data generation
  --
  -- This generates data where each word of physical memory is written with something
  -- that relates to its address.
  --

  -- TO DO: think of a way to handle the case where the physical width of a memory
  -- word is not enough to hold an entire physical addresss.
  assert phys_addr_width <= phys_data_width
    report "Physical data width insufficient to hold entire physical address"
    severity error;

  gen_d_own_addr : process(clk)
    variable tmp_data : std_logic_vector(phys_data_width - pack_factor_log2 - 1 downto 0);
  begin
    if rising_edge(clk) then
      if reg_phase_1h(3) = '1' then
        if ex_init = '1' then
          for i in 0 to pack_factor - 1 loop
            d_own_addr(phys_data_width * i + phys_addr_width - 1 downto phys_data_width * i + pack_factor_log2) <= reg_offset;
          end loop;
        else
          if ex_gen_ce = '1' then
            for i in 0 to pack_factor - 1 loop
              d_own_addr(phys_data_width * i + phys_addr_width - 1 downto phys_data_width * i + pack_factor_log2) <=
                d_own_addr(phys_data_width * i + phys_addr_width - 1 downto phys_data_width * i + pack_factor_log2) + 1;
            end loop;
          end if;
        end if;
      end if;
    end if;
  end process;

  be_own_addr <= (others => '1');
  dvalid_own_addr <= ex_gen_ce and reg_phase_1h(3) and not reg_reading;

  exp_own_addr <= d_own_addr;
  expvalid_own_addr <= ex_gen_ce and reg_phase_1h(3) and reg_reading;

  --
  -- Phase 4 data generation
  --
  -- This generates a 55/AA toggle pattern, at maximum rate at the physical device pins.
  --

  gen_d_55_aa_tmax : process(clk)
    constant data_55 : phys_data_t := gen_55_word;
  begin
    if rising_edge(clk) then
      if reg_phase_1h(4) = '1' then
        if ex_init = '1' then
          for i in 0 to pack_factor - 1 loop
            if i mod 2 = 0 then
              d_55_aa_tmax(phys_data_width * (i + 1) - 1 downto phys_data_width * i) <= data_55;
            else
              d_55_aa_tmax(phys_data_width * (i + 1) - 1 downto phys_data_width * i) <= NOT data_55;
            end if;
          end loop;
        else
          if ex_gen_ce = '1' AND pack_factor < 2 then
            d_55_aa_tmax <= NOT d_55_aa_tmax;
          end if;
        end if;
      end if;
    end if;
  end process;

  be_55_aa_tmax <= (others => '1');
  dvalid_55_aa_tmax <= ex_gen_ce and reg_phase_1h(4) and not reg_reading;

  exp_55_aa_tmax <= d_55_aa_tmax;
  expvalid_55_aa_tmax <= ex_gen_ce and reg_phase_1h(4) and reg_reading;

  --
  -- Phase 5 data generation
  --
  -- This generates a 00/FF toggle pattern, at maximum rate at the physical device pins.
  --

  gen_d_00_ff_tmax : process(clk)
  begin
    if rising_edge(clk) then
      if reg_phase_1h(5) = '1' then
        if ex_init = '1' then
          for i in 0 to pack_factor - 1 loop
            if i mod 2 = 0 then
              d_00_ff_tmax(phys_data_width * (i + 1) - 1 downto phys_data_width * i) <= (others => '0');
            else
              d_00_ff_tmax(phys_data_width * (i + 1) - 1 downto phys_data_width * i) <= (others => '1');
            end if;
          end loop;
        else
          if ex_gen_ce = '1' AND pack_factor < 2 then
            d_00_ff_tmax <= NOT d_00_ff_tmax;
          end if;
        end if;
      end if;
    end if;
  end process;

  be_00_ff_tmax <= (others => '1');
  dvalid_00_ff_tmax <= ex_gen_ce and reg_phase_1h(5) and not reg_reading;

  exp_00_ff_tmax <= d_00_ff_tmax;
  expvalid_00_ff_tmax <= ex_gen_ce and reg_phase_1h(5) and reg_reading;

  --
  -- Phase 6 data generation
  --
  -- This generates pseudorandom data.
  --

  gen_d_random : block is

    constant num_rng : natural := (memory_data_width + 31) / 32;
    constant rng_data_width : natural := num_rng * 32;
    signal data_rng : std_logic_vector(rng_data_width - 1 downto 0);
    signal ready_rng : std_logic_vector(num_rng - 1 downto 0);
    signal exp_rng : std_logic_vector(rng_data_width - 1 downto 0);

  begin

    gen_rng : for i in 0 to num_rng - 1 generate

      constant seed : std_logic_vector(31 downto 0) := X"AAAAA" & std_logic_vector(to_unsigned(i + 1, 12));

      signal init_rng : std_logic_vector(1 downto 0);
      signal ce_data_rng : std_logic;
      signal ce_exp_rng : std_logic;

      attribute dont_touch : string;
      attribute dont_touch of init_rng : signal is "true";
      attribute dont_touch of ce_data_rng : signal is "true";
      attribute dont_touch of ce_exp_rng : signal is "true";
      
      
    begin

      retiming : process(clk)
      begin
        if rising_edge(clk) then
          init_rng <= (others => ctl_init_rng);
          ce_data_rng <= reg_phase_1h(6) and ex_gen_ce and not reg_reading and ready_rng(i);
          ce_exp_rng <= reg_phase_1h(6) and ex_gen_ce and reg_reading and ready_rng(i);
        end if;
      end process;
      
      data_inst : memory_exerciser_rng
        generic map(
          seed => seed)
        port map(
          sr => init_rng(0),
          clk => clk,
          ce => ce_data_rng,
          ready => ready_rng(i),
          x => data_rng(32 * (i + 1) - 1 downto 32 * i));

      exp_inst : memory_exerciser_rng
        generic map(
          seed => seed)
        port map(
          sr => init_rng(1),
          clk => clk,
          ce => ce_exp_rng,
          ready => open,
          x => exp_rng(32 * (i + 1) - 1 downto 32 * i));

    end generate;

    -- Can ignore the ready flags from other RNGs because they should all
    -- become ready simultaneously.
    ready_rng_random <= ready_rng(0);

    d_random <= std_logic_vector(resize(unsigned(data_rng), memory_data_width));
    be_random <= (others => '1');
    
    gen_dvalid_random : process(clk)
    begin
      if rising_edge(clk) then
        dvalid_random <= reg_phase_1h(6) and ex_gen_ce and not reg_reading;
      end if;
    end process;

    exp_random <= std_logic_vector(resize(unsigned(exp_rng), memory_data_width));

    gen_expvalid_random : process(clk)
    begin
      if rising_edge(clk) then
        expvalid_random <= reg_phase_1h(6) and ex_gen_ce and reg_reading;
      end if;
    end process;

  end block;

  --
  -- Phase 7 data generation
  --
  -- This generates pseudorandom data, with randomised byte enables
  --

  gen_d_random_stipple : block is

    constant num_rng : natural := (memory_data_width + 31) / 32;
    constant rng_data_width : natural := num_rng * 32;
    constant num_rng_be : natural := (memory_be_width + 31) / 32;
    constant rng_be_width : natural := num_rng_be * 32;
    signal data_rng : std_logic_vector(rng_data_width - 1 downto 0);
    signal be_rng : std_logic_vector(rng_be_width - 1 downto 0);
    signal exp1_rng : std_logic_vector(rng_data_width - 1 downto 0);
    signal exp2_rng : std_logic_vector(rng_data_width - 1 downto 0);
    signal expbe_rng : std_logic_vector(rng_be_width - 1 downto 0);
    signal ready_rng : std_logic_vector(num_rng - 1 downto 0);
    signal expvalid_random_stipple_early : std_logic;
    

  begin

    gen_data_rng : for i in 0 to num_rng - 1 generate

      constant seed1 : std_logic_vector(31 downto 0) := X"AAAAA" & std_logic_vector(to_unsigned(i + 1, 12)); -- Must be same seed as previous phase
      constant seed2 : std_logic_vector(31 downto 0) := X"55555" & std_logic_vector(to_unsigned(i + 1, 12));

      signal init_rng : std_logic_vector(2 downto 0);
      signal ce_data_rng : std_logic;
      signal ce_exp_rng : std_logic;

      attribute dont_touch : string;
      attribute dont_touch of init_rng : signal is "true";
      attribute dont_touch of ce_data_rng : signal is "true";
      attribute dont_touch of ce_exp_rng : signal is "true";
      
      
    begin

      retime_init_rng : process(clk)
      begin
        if rising_edge(clk) then
          init_rng <= (others => ctl_init_rng);
          ce_data_rng <= reg_phase_1h(7) and ex_gen_ce and not reg_reading and ready_rng(i);
          ce_exp_rng <= reg_phase_1h(7) and ex_gen_ce and reg_reading and ready_rng(i);
        end if;
      end process;
      
      -- Used to generate write data
      data_inst : memory_exerciser_rng
        generic map(
          seed => seed2)
        port map(
          sr => init_rng(0),
          clk => clk,
          ce => ce_data_rng,
          ready => ready_rng(i),
          x => data_rng(32 * (i + 1) - 1 downto 32 * i));

      -- Used to generate expected data
      exp1_inst : memory_exerciser_rng
        generic map(
          seed => seed1)
        port map(
          sr => init_rng(1),
          clk => clk,
          ce => ce_exp_rng,
          ready => open,
          x => exp1_rng(32 * (i + 1) - 1 downto 32 * i));

      -- Used to generate expected data
      exp2_inst : memory_exerciser_rng
        generic map(
          seed => seed2)
        port map(
          sr => init_rng(2),
          clk => clk,
          ce => ce_exp_rng,
          ready => open,
          x => exp2_rng(32 * (i + 1) - 1 downto 32 * i));

    end generate;

    gen_rng_be : for i in 0 to num_rng_be - 1 generate

      constant seed : std_logic_vector(31 downto 0) := X"5A5A5" & std_logic_vector(to_unsigned(i + 1, 12));

      signal init_rng : std_logic_vector(1 downto 0);
      signal ce_data_rng : std_logic;
      signal ce_exp_rng : std_logic;

      attribute dont_touch : string;
      attribute dont_touch of init_rng : signal is "true";
      attribute dont_touch of ce_data_rng : signal is "true";
      attribute dont_touch of ce_exp_rng : signal is "true";
      
      
    begin

      retime_init_rng : process(clk)
      begin
        if rising_edge(clk) then
          init_rng <= (others => ctl_init_rng);
          ce_data_rng <= reg_phase_1h(7) and ex_gen_ce and not reg_reading;
          ce_exp_rng <= reg_phase_1h(7) and ex_gen_ce and reg_reading;
        end if;
      end process;
      

      -- Used to generate write byte enables
      be_inst : memory_exerciser_rng
        generic map(
          seed => seed)
        port map(
          sr => init_rng(0),
          clk => clk,
          ce => ce_data_rng,
          ready => open,
          x => be_rng(32 * (i + 1) - 1 downto 32 * i));

      -- Used to generate expected data
      expbe_inst : memory_exerciser_rng
        generic map(
          seed => seed)
        port map(
          sr => init_rng(1),
          clk => clk,
          ce => ce_exp_rng,
          ready => open,
          x => expbe_rng(32 * (i + 1) - 1 downto 32 * i));

    end generate;

    -- Can ignore the ready flags from other RNGs because they should all
    -- become ready simultaneously.
    ready_rng_random_stipple <= ready_rng(0);

    d_random_stipple <= std_logic_vector(resize(unsigned(data_rng), memory_data_width));
    be_random_stipple <= std_logic_vector(resize(unsigned(be_rng), memory_be_width));
    
    gen_dvalid_random_stipple : process(clk)
    begin
      if rising_edge(clk) then
        dvalid_random_stipple <= reg_phase_1h(7) and ex_gen_ce and not reg_reading;
      end if;
    end process;
    
    gen_exp_random_stipple : process(clk)
    begin
      if rising_edge(clk) then
        expvalid_random_stipple_early <= reg_phase_1h(7) and ex_gen_ce and reg_reading;
        
        if expvalid_random_stipple_early = '1' then
          for i in 0 to memory_be_width - 1 loop
            if expbe_rng(i) = '1' then
              exp_random_stipple(8 * (i + 1) - 1 downto 8 * i) <= exp2_rng(8 * (i + 1) - 1 downto 8 * i);
            else
              exp_random_stipple(8 * (i + 1) - 1 downto 8 * i) <= exp1_rng(8 * (i + 1) - 1 downto 8 * i);
            end if;
          end loop;
        end if;

        expvalid_random_stipple <= expvalid_random_stipple_early;
      end if;
    end process;

  end block;

  mux_data : process(clk)
    variable tmp_d : data_word_t;
    variable tmp_dbe : be_word_t;
    variable tmp_exp : data_word_t;
  begin
    if rising_edge(clk) then
      tmp_d := (others => '0');
      tmp_dbe := (others => '0');
      if dvalid_walk_1_dq = '1' then
        tmp_d := tmp_d or d_walk_1_dq;
        tmp_dbe := tmp_dbe or be_walk_1_dq;
      end if;
      if dvalid_55_aa_slow = '1' then
        tmp_d := tmp_d or d_55_aa_slow;
        tmp_dbe := tmp_dbe or be_55_aa_slow;
      end if;
      if dvalid_aa_55_stipple = '1' then
        tmp_d := tmp_d or d_aa_55_stipple;
        tmp_dbe := tmp_dbe or be_aa_55_stipple;
      end if;
      if dvalid_own_addr = '1' then
        tmp_d := tmp_d or d_own_addr;
        tmp_dbe := tmp_dbe or be_own_addr;
      end if;
      if dvalid_55_aa_tmax = '1' then
        tmp_d := tmp_d or d_55_aa_tmax;
        tmp_dbe := tmp_dbe or be_55_aa_tmax;
      end if;
      if dvalid_00_ff_tmax = '1' then
        tmp_d := tmp_d or d_00_ff_tmax;
        tmp_dbe := tmp_dbe or be_00_ff_tmax;
      end if;
      if dvalid_random = '1' then
        tmp_d := tmp_d or d_random;
        tmp_dbe := tmp_dbe or be_random;
      end if;
      if dvalid_random_stipple = '1' then
        tmp_d := tmp_d or d_random_stipple;
        tmp_dbe := tmp_dbe or be_random_stipple;
      end if;
      d <= tmp_d;
      dbe <= tmp_dbe;
      dvalid <= dvalid_walk_1_dq or dvalid_55_aa_slow or dvalid_aa_55_stipple or dvalid_own_addr or
                dvalid_55_aa_tmax or dvalid_00_ff_tmax or dvalid_random or dvalid_random_stipple;

      tmp_exp := (others => '0');
      if expvalid_walk_1_dq = '1' then
        tmp_exp := tmp_exp or exp_walk_1_dq;
      end if;
      if expvalid_55_aa_slow = '1' then
        tmp_exp := tmp_exp or exp_55_aa_slow;
      end if;
      if expvalid_aa_55_stipple = '1' then
        tmp_exp := tmp_exp or exp_aa_55_stipple;
      end if;
      if expvalid_own_addr = '1' then
        tmp_exp := tmp_exp or exp_own_addr;
      end if;
      if expvalid_55_aa_tmax = '1' then
        tmp_exp := tmp_exp or exp_55_aa_tmax;
      end if;
      if expvalid_00_ff_tmax = '1' then
        tmp_exp := tmp_exp or exp_00_ff_tmax;
      end if;
      if expvalid_random = '1' then
        tmp_exp := tmp_exp or exp_random;
      end if;
      if expvalid_random_stipple = '1' then
        tmp_exp := tmp_exp or exp_random_stipple;
      end if;
      exp <= tmp_exp;
      expvalid <= expvalid_walk_1_dq or expvalid_55_aa_slow or expvalid_aa_55_stipple or expvalid_own_addr or
                  expvalid_55_aa_tmax or expvalid_00_ff_tmax or expvalid_random or expvalid_random_stipple;
    end if;
  end process;

end architecture;
