--
-- Limitations:
-- 1. MIG must be configured so that there is one word of data transferred
--   on the MIG UI (app_wdf_data, app_rd_data) per command entered via
--   app_en. This implies that when app_rd_data_valid is asserted,
--   app_rd_data_end must also be asserted.
--
-- Constraints on parameters:
-- 2. memory_data_width & mig_data_width must be the same.
-- 3. mig_addr_width should be equal to or greater than (memory_addr_width + log2(pack_factor)).
--    - If less, then a full memory test where ctl_length has the maximum
--      value will fail due to address wraparound.
--    - If more, then the entire MIG address space cannot be tested.
--
-- Notes:
-- 4. MIG physical data width is assumed to be mig_data_width / pack_factor.
--    pack_factor must be correct in order for phases that generate specific
--    patterns (e.g. maximum toggle rate on physical data pins) to generate
--    the required logical data.
-- 5. 'pack_rearranged' must be set as follows:
--    - 7 Series MIG DDR3 SDRAM interface   => false
--    - Ultrascale MIG DDR3 SDRAM interface => true
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

entity memory_exerciser_mig_native is
  generic(
    memory_addr_width  : in    natural; -- Logical address width
    memory_data_width  : in    natural; -- Logical data width; must equal mig_data_width
    mig_addr_width     : in    natural; -- MIG UI address width; see above
    mig_data_width     : in    natural; -- MIG UI data width; must equal memory_data_width
    mig_ctlr_has_ecc   : in    boolean; -- MIG DDR3 SDRAM controller has ECC logic
    pack_factor        : in    natural; -- Number of physical data words per logical data word
    pack_rearranged    : in    boolean; -- See note 5 above.
    partial_writes     : in    boolean; -- true => perform phases that exercise byte enables
    error_count_width  : in    natural; -- log2 of error count
    error_index_width  : in    natural  -- log2 of number of errors that can be stored in error log
  );
  port(
    rst                : in    std_logic;
    clk                : in    std_logic;
    -- MIG native interface
    app_en             : out   std_logic;
    app_hi_pri         : out   std_logic;
    app_autoprecharge  : out   std_logic;
    app_wdf_end        : out   std_logic;
    app_wdf_wren       : out   std_logic;
    app_rd_data_end    : in    std_logic;
    app_rd_data_valid  : in    std_logic;
    app_rdy            : in    std_logic;
    app_wdf_rdy        : in    std_logic;
    app_addr           : out   std_logic_vector(mig_addr_width - 1 downto 0);
    app_cmd            : out   std_logic_vector(2 downto 0);
    app_wdf_data       : out   std_logic_vector(mig_data_width - 1 downto 0);
    app_wdf_mask       : out   std_logic_vector(mig_data_width / 8 - 1 downto 0);
    app_rd_data        : in    std_logic_vector(mig_data_width - 1 downto 0);
    -- Control interface
    ctl_go             : in    std_logic;                                            -- Pulse high for at least one cycle to initiate test
    ctl_offset         : in    std_logic_vector(memory_addr_width - 1 downto 0);     -- Starting logical address of test
    ctl_length         : in    std_logic_vector(memory_addr_width - 1 downto 0);     -- Number of logical words to test, minus 1
    ctl_ph_en          : in    std_logic_vector(7 downto 0) := (others => '1');      -- Enables various phases of test
    ctl_done           : out   std_logic;                                            -- High indicates finished or idle
    -- When 'ctl_done' is asserted, assert 'ctl_erd' to read out an error (latency 2 clock cycles)
    ctl_erd            : in    std_logic;
    -- Index of error to read out
    ctl_eindex         : in    std_logic_vector(error_index_width - 1 downto 0);
    -- Qualified by 'ctl_done', high indicates at least one error occurred
    ctl_error          : out   std_logic;
    -- Qualified by 'ctl_done', actual number of errors
    ctl_ecount         : out   std_logic_vector(error_count_width - 1 downto 0);
    -- Qualified by 'ctl_done', asserted if error log is full
    ctl_efull          : out   std_logic;
    -- Qualified by 'ctl_done', asserted if errors were lost due to error log full
    ctl_elost          : out   std_logic;
    -- Address of selected error
    ctl_eaddr          : out   std_logic_vector(memory_addr_width - 1 downto 0);
    -- Phase in which selected error occurred
    ctl_ephase         : out   std_logic_vector(2 downto 0);
    -- Bits that were incorrect when selected error occurred
    ctl_ebits          : out   std_logic_vector(memory_data_width - 1 downto 0);
    -- Actual data when selected error occurred
    ctl_edata          : out   std_logic_vector(memory_data_width - 1 downto 0)
  );
end entity;

architecture rtl of memory_exerciser_mig_native is

  function choice(
    constant a, b : in std_logic_vector;
    constant selb : in boolean)
  return std_logic_vector is begin
    if selb then
      return b;
    else
      return a;
    end if;
  end;

  component memory_exerciser_data_gen
    generic(
      memory_data_width  : in    natural;
      memory_addr_width  : in    natural;
      pack_factor        : in    natural
    );
    port(
      rst                : in    std_logic;
      clk                : in    std_logic;
      -- Generated data/expected data
      d                  : out   std_logic_vector(memory_data_width - 1 downto 0);     -- Data when writing to memory
      dbe                : out   std_logic_vector(memory_data_width / 8 - 1 downto 0); -- Byte enables when writing to memory
      dvalid             : out   std_logic;                                            -- Qualifies 'd', 'dbe'
      exp                : out   std_logic_vector(memory_data_width - 1 downto 0);     -- Expected data when reading from memory
      expvalid           : out   std_logic;                                            -- Qualifies 'exp'
      enable             : in    std_logic;                                            -- When deasserted, 'valid' must deassert within a cycle or two
      -- Control interface
      ctl_init_rng       : in    std_logic;                                            -- Assert to initialise random number generators
      ctl_go             : in    std_logic;                                            -- Assert for >= 1 cycle while 'done' is asserted to initiate test
      ctl_reading        : in    std_logic;                                            -- 0 => writing to memory, 1 => reading from memory
      ctl_phase          : in    std_logic_vector(2 downto 0);                         -- Selects data generation method
      ctl_offset         : in    std_logic_vector(memory_addr_width - 1 downto 0);     -- Address at which to begin testing
      ctl_length         : in    std_logic_vector(memory_addr_width - 1 downto 0);     -- Number of logical words to test, minus 1
      ctl_idle           : out   std_logic                                             -- High indicates finished or idle
    );
  end component;

  component memory_exerciser_error_log
    generic(
      phase_width        : in    natural;
      memory_data_width  : in    natural;
      memory_addr_width  : in    natural;
      count_width        : in    natural;
      index_width        : in    natural
    );
    port(
      rst                : in    std_logic;
      clk                : in    std_logic;
      -- Clears the error log
      sr                 : in    std_logic;
      -- Set initial address & phase for current phase
      init               : in    std_logic;
      init_phase         : in    std_logic_vector(phase_width - 1 downto 0);
      init_offset        : in    std_logic_vector(memory_addr_width - 1 downto 0);
      -- Expected and actual data
      chk_ce             : in    std_logic;
      chk_act_data       : in    std_logic_vector(memory_data_width - 1 downto 0);
      chk_exp_data       : in    std_logic_vector(memory_data_width - 1 downto 0);
      chk_done           : out   std_logic;
      -- Error readout interface
      rd_ce              : in    std_logic;
      rd_index           : in    std_logic_vector(index_width - 1 downto 0);
      rd_not_empty       : out   std_logic;
      rd_full            : out   std_logic;
      rd_overflow        : out   std_logic;
      rd_count           : out   std_logic_vector(count_width - 1 downto 0);
      rd_rec_phase       : out   std_logic_vector(phase_width - 1 downto 0);
      rd_rec_addr        : out   std_logic_vector(memory_addr_width - 1 downto 0);
      rd_rec_bad_bits    : out   std_logic_vector(memory_data_width - 1 downto 0);
      rd_rec_act_data    : out   std_logic_vector(memory_data_width - 1 downto 0)
    );
  end component;

  component memory_exerciser_fifo
    generic(
      order       : in    natural range 2 to 6;
      width       : in    natural
    );
    port(
      rst         : in    std_logic;                             -- Asynchronous reset
      clk         : in    std_logic;                             -- Clock
      sr          : in    std_logic;                             -- Synchronous reset
      -- Writer interface
      wr          : in    std_logic;                             -- '1' => write value on 'd' into FIFO
      wd          : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
      wfull       : out   std_logic;                             -- '1' => full (from point of view of write side)
      -- Accounting interface
      aadv        : in    std_logic;                             -- Advance accounting pointer
      aempty      : out   std_logic;                             -- '1' => empty
      alevel      : out   std_logic_vector(ORDER downto 0);      -- Number of items in FIFO
      -- Reader interface
      radv        : in    std_logic;                             -- '1' => present next item on 'q' to next item
      rempty      : out   std_logic;                             -- '1' => empty
      rq          : out   std_logic_vector(WIDTH - 1 downto 0)   -- Current item, qualified by 'valid'
    );
  end component;

  component memory_exerciser_dearranger
    generic(
      memory_data_width  : in    natural;
      pack_factor        : in    natural;
      pack_rearranged    : in    boolean
    );
    port(
      data_in            : in    std_logic_vector(memory_data_width - 1 downto 0);
      data_out           : out   std_logic_vector(memory_data_width - 1 downto 0)
    );
  end component;
  
  component memory_exerciser_rearranger
    generic(
      memory_data_width  : in    natural;
      pack_factor        : in    natural;
      pack_rearranged    : in    boolean
    );
    port(
      data_in            : in    std_logic_vector(memory_data_width - 1 downto 0);
      mask_in            : in    std_logic_vector(memory_data_width / 8 - 1 downto 0);
      data_out           : out   std_logic_vector(memory_data_width - 1 downto 0);
      mask_out           : out   std_logic_vector(memory_data_width / 8 - 1 downto 0)
    );
  end component;
  
  constant pack_factor_log2 : natural := natural(floor(log2(real(pack_factor) + 0.5)));

  constant phases_enable : std_logic_vector(7 downto 0) := choice("01111011", "11111111", partial_writes);

  -- Executive state machine
  type ex_state_t is (ex_s_idle, ex_s_init_phase1, ex_s_init_phase2, ex_s_run_phase, ex_s_finish);
  signal ex_state, n_ex_state : ex_state_t;
  signal ex_idle, n_ex_idle : std_logic;
  signal ex_init_rng, n_ex_init_rng : std_logic;
  signal ex_phase, n_ex_phase : std_logic_vector(3 downto 0);
  signal ex_ph_go, n_ex_ph_go : std_logic;

  -- MIG command issue state machine
  type mig_state_t is (mig_s_idle, mig_s_writing, mig_s_wait_fifo, mig_s_reading, mig_s_reading_finish, mig_s_done);
  signal mig_state, n_mig_state : mig_state_t;
  signal mig_addr, n_mig_addr : std_logic_vector(memory_addr_width - 1 downto 0);
  signal mig_cmd_count, n_mig_cmd_count : std_logic_vector(memory_addr_width - 1 downto 0);
  signal mig_en_read, n_mig_en_read : std_logic;
  signal mig_en_write, n_mig_en_write : std_logic;
  signal mig_idle, n_mig_idle : std_logic;
  signal mig_rd_count, n_mig_rd_count : std_logic_vector(memory_addr_width - 1 downto 0);
  signal mig_exp_fifo_ready : std_logic;
  signal mig_rd_ready : std_logic;
  signal mig_wr_ready : std_logic;

  -- Latched versions of 'ctl_xxx' signals
  signal reg_offset : std_logic_vector(memory_addr_width - 1 downto 0) := (others => '0');
  signal reg_length : std_logic_vector(memory_addr_width - 1 downto 0) := (others => '0');
  signal reg_ph_en : std_logic_vector(7 downto 0);
  
  constant memory_be_width : natural := memory_data_width / 8;

  -- Data generator
  signal gen_idle : std_logic;
  signal gen_d : std_logic_vector(memory_data_width - 1 downto 0);
  signal gen_dbe : std_logic_vector(memory_be_width - 1 downto 0);
  signal gen_dbe_partial : std_logic;
  signal gen_dvalid : std_logic;
  signal gen_exp : std_logic_vector(memory_data_width - 1 downto 0);
  signal gen_expvalid : std_logic;
  signal gen_enable : std_logic;

  constant d_be_fifo_order : natural := 6;
  constant d_be_fifo_width : natural := memory_data_width + memory_be_width + 1;
  signal d_be_fifo_wfull : std_logic;
  signal d_be_fifo_aempty : std_logic;
  signal d_be_fifo_radv : std_logic;
  signal d_be_fifo_rempty : std_logic;
  signal d_be_fifo_wd : std_logic_vector(d_be_fifo_width - 1 downto 0);
  signal d_be_fifo_rq : std_logic_vector(d_be_fifo_width - 1 downto 0);
  signal d_be_fifo_rq_data : std_logic_vector(memory_data_width - 1 downto 0);
  signal d_be_fifo_rq_mask : std_logic_vector(memory_be_width - 1 downto 0);
  signal d_be_fifo_rq_partial : std_logic;

  constant exp_fifo_order : natural := 6;
  constant exp_fifo_width : natural := memory_data_width;
  signal exp_fifo_wfull : std_logic;
  signal exp_fifo_aadv : std_logic;
  signal exp_fifo_aempty : std_logic;
  signal exp_fifo_alevel : std_logic_vector(exp_fifo_order downto 0);
  signal exp_fifo_radv : std_logic;
  signal exp_fifo_rempty : std_logic;
  signal exp_fifo_rq : std_logic_vector(exp_fifo_width - 1 downto 0);

  signal actual_data : std_logic_vector(memory_data_width - 1 downto 0);
  signal data_checked : std_logic;
  signal clear_error_log : std_logic;
  signal error_log_not_empty : std_logic;
  
  signal logic0, logic1 : std_logic;


begin

  assert mig_data_width = memory_data_width severity failure;

  assert mig_addr_width = memory_addr_width + pack_factor_log2
    report "mig_addr_width and memory_addr_width imply different byte sizes. This is not necessarily a problem."
    severity warning;


  logic0 <= '0';
  logic1 <= '1';


  --
  -- Register input parameters so that they are stable during the
  -- course of the test.
  --

  register_param : process(clk)
  begin
    if rising_edge(clk) then
      if ex_idle = '1' and ctl_go = '1' then
        reg_offset <= ctl_offset;
        reg_length <= ctl_length;
        reg_ph_en <= ctl_ph_en;
      end if;
    end if;
  end process;


  --
  -- Executive state machine
  --

  gen_ex_state : process(rst, clk)
  begin
    if rst = '1' then
      ex_state <= ex_s_idle;
      ex_idle <= '1';
      ex_init_rng <= '0';
      ex_phase <= (others => '-');
      ex_ph_go <= '0';
    elsif rising_edge(clk) then
      ex_state <= n_ex_state;
      ex_idle <= n_ex_idle;
      ex_init_rng <= n_ex_init_rng;
      ex_phase <= n_ex_phase;
      ex_ph_go <= n_ex_ph_go;
      
      -- synthesis translate_off
      if n_ex_state = ex_s_init_phase2 and ex_phase(0) = '0' then
        report "memory_exerciser_mig_native: Beginning test phase " & natural'image(to_integer(unsigned(ex_phase(ex_phase'high downto 1)))) & "." severity note;
      end if;
      
      if ex_state = ex_s_finish then
        if error_log_not_empty = '1' then
          report "memory_exerciser_mig_native: Memory test completed with at least one error." severity error;
        else
          report "memory_exerciser_mig_native: Memory test completed without error." severity note;
        end if;
      end if;
      -- synthesis translate_on
    end if;
  end process;

  ctl_done <= ex_idle;

  gen_n_ex_state : process(
    ex_state,
    ex_phase,
    ctl_go,
    gen_idle,
    mig_idle,
    reg_ph_en)
    variable tmp8 : std_logic_vector(7 downto 0);
  begin
    case ex_state is
      when ex_s_idle =>
        if ctl_go = '1' then
          n_ex_state <= ex_s_init_phase1;
          n_ex_idle <= '0';
          n_ex_init_rng <= '1';
          n_ex_phase <= (others => '0');
          n_ex_ph_go <= '0';
        else
          n_ex_state <= ex_s_idle;
          n_ex_idle <= '1';
          n_ex_init_rng <= '0';
          n_ex_phase <= (others => '-');
          n_ex_ph_go <= '0';
        end if;

      when ex_s_init_phase1 =>
        n_ex_state <= ex_s_init_phase2;
        n_ex_idle <= '0';
        n_ex_init_rng <= '0';
        n_ex_phase <= ex_phase;
        tmp8 := phases_enable and reg_ph_en;
        n_ex_ph_go <= tmp8(to_integer(unsigned(ex_phase(3 downto 1))));

      when ex_s_init_phase2 =>
        n_ex_state <= ex_s_run_phase;
        n_ex_idle <= '0';
        n_ex_init_rng <= '0';
        n_ex_phase <= ex_phase;
        n_ex_ph_go <= '0';

      when ex_s_run_phase =>
        n_ex_idle <= '0';
        n_ex_init_rng <= '0';
        if gen_idle = '1' and mig_idle = '1' then
          if AND_reduce(ex_phase) = '1' then
            n_ex_state <= ex_s_finish;
            n_ex_phase <= (others => '-');
            n_ex_ph_go <= '0';
          else
            n_ex_state <= ex_s_init_phase1;
            n_ex_phase <= ex_phase + 1;
            n_ex_ph_go <= '0';
          end if;
        else
          n_ex_state <= ex_s_run_phase;
          n_ex_phase <= ex_phase;
          n_ex_ph_go <= '0';
        end if;

      when ex_s_finish =>
        n_ex_state <= ex_s_idle;
        n_ex_idle <= '1';
        n_ex_init_rng <= '0';
        n_ex_phase <= (others => '-');
        n_ex_ph_go <= '0';
    end case;
  end process;


  --
  -- MIG command issue state machine
  --

  gen_mig_state : process(rst, clk)
  begin
    if rst = '1' then
      mig_state <= mig_s_idle;
      mig_addr <= (others => '-');
      mig_cmd_count <= (others => '-');
      mig_en_read <= '0';
      mig_en_write <= '0';
      mig_idle <= '1';
      mig_rd_count <= (others => '-');
    elsif rising_edge(clk) then
      mig_state <= n_mig_state;
      mig_addr <= n_mig_addr;
      mig_cmd_count <= n_mig_cmd_count;
      mig_en_read <= n_mig_en_read;
      mig_en_write <= n_mig_en_write;
      mig_idle <= n_mig_idle;
      mig_rd_count <= n_mig_rd_count;
    end if;
  end process;

  -- Can send a write command to MIG when (a) MIG is ready for a command and
  -- (b) MIG is ready for data and (c) we have data ready in the data & byte
  -- enable FIFO.
  -- Used to decrement 'mig_cmd_count' when writing.
  mig_wr_ready <= app_wdf_rdy and app_rdy and not d_be_fifo_rempty;

  -- Can send a read command to MIG whenever MIG is ready for a command.
  -- We assume that expected data FIFO is not empty.
  -- Used to decrement 'mig_cmd_count' when reading.
  mig_rd_ready <= app_rdy;

  -- "exp_fifo_alevel > reg_length" is used instead of "exp_fifo_alevel >= reg_length"
  -- because reg_length holds number of words minus 1. This allows us to
  -- avoid computing (reg_length + 1).
  mig_exp_fifo_ready <= '1' when exp_fifo_alevel > reg_length or exp_fifo_wfull = '1' else '0';

  -- Issue a read or write command to MIG.
  app_en <= (app_wdf_rdy and app_rdy and not d_be_fifo_rempty and mig_en_write) or -- Write command to MIG.
            (app_rdy and mig_en_read);                                             -- Read command to MIG.
  app_addr <= std_logic_vector(resize(unsigned(mig_addr & std_logic_vector(to_unsigned(0, pack_factor_log2))), mig_addr_width));
  app_cmd <= "001" when mig_en_read = '1' else
             "011" when mig_ctlr_has_ecc and mig_en_read = '0' and d_be_fifo_rq_partial = '1' else
             "000";
  app_hi_pri <= '0';
  app_autoprecharge <= '0';

  -- Write data into MIG; should always be accompanied by 'app_en' assertion.
  app_wdf_wren <= app_wdf_rdy and app_rdy and not d_be_fifo_rempty and mig_en_write;
  -- Do not support multi-word commands at the moment.
  app_wdf_end <= '1';

  -- Data & byte enable FIFO control.
  -- Advance read ptr. when data & byte enables written to MIG.
  d_be_fifo_radv <= app_wdf_rdy and app_rdy and not d_be_fifo_rempty and mig_en_write;

  -- Expected data FIFO control.
  exp_fifo_aadv <= app_rdy and mig_en_read; -- Advance accounting ptr. when read command issued to MIG.
  exp_fifo_radv <= app_rd_data_valid;       -- Advance read ptr. when data valid from MIG.

  gen_n_mig_state : process(
    mig_state,
    mig_addr,
    mig_cmd_count,
    mig_exp_fifo_ready,
    mig_rd_count,
    mig_rd_ready,
    mig_wr_ready,
    reg_length,
    reg_offset,
    ex_ph_go,
    ex_phase,
    data_checked)
  begin
    case mig_state is
      when mig_s_idle =>
        n_mig_rd_count <= (others => '-');
        n_mig_en_read <= '0';
        n_mig_en_write <= '0';
        if ex_ph_go = '1' then
          n_mig_idle <= '0';
          if ex_phase(0) = '1' then
            n_mig_state <= mig_s_wait_fifo;
            n_mig_addr <= (others => '-');
            n_mig_cmd_count <= (others => '-');
          else
            n_mig_state <= mig_s_writing;
            n_mig_addr <= reg_offset;
            n_mig_cmd_count <= reg_length;
          end if;
        else
          n_mig_state <= mig_s_idle;
          n_mig_addr <= (others => '-');
          n_mig_cmd_count <= (others => '-');
          n_mig_idle <= '1';
        end if;

      when mig_s_writing =>
        n_mig_en_read <= '0';
        n_mig_idle <= '0';
        n_mig_rd_count <= (others => '-');
        if mig_wr_ready = '1' then
          if mig_cmd_count = 0 then
            n_mig_state <= mig_s_done;
            n_mig_addr <= (others => '-');
            n_mig_cmd_count <= (others => '-');
            n_mig_en_write <= '0';
          else
            n_mig_state <= mig_s_writing;
            n_mig_addr <= mig_addr + 1;
            n_mig_cmd_count <= mig_cmd_count - 1;
            n_mig_en_write <= '1';
          end if;
        else
          n_mig_state <= mig_s_writing;
          n_mig_addr <= mig_addr;
          n_mig_cmd_count <= mig_cmd_count;
          n_mig_en_write <= '1';
        end if;

      -- Here we wait for expected data FIFO to become full or for it to have
      -- at least (reg_length + 1) words in it.
      -- We guarantee that, once we start getting- read data from MIG, this
      -- FIFO cannot underflow, because the data_generator's data rate is at
      -- least as large as the data rate that MIG can sustain for reads.
      when mig_s_wait_fifo =>
        n_mig_en_write <= '0';
        n_mig_idle <= '0';
        if mig_exp_fifo_ready = '1' then
          n_mig_state <= mig_s_reading;
          n_mig_addr <= reg_offset;
          n_mig_cmd_count <= reg_length;
          n_mig_en_read <= '1';
          n_mig_rd_count <= reg_length;
        else
          n_mig_state <= mig_s_wait_fifo;
          n_mig_addr <= (others => '-');
          n_mig_cmd_count <= (others => '-');
          n_mig_en_read <= '0';
          n_mig_rd_count <= (others => '-');
        end if;

      -- Here we issue read commands to MIG.
      when mig_s_reading =>
        n_mig_en_write <= '0';
        n_mig_idle <= '0';
        if data_checked = '1' then
          n_mig_rd_count <= mig_rd_count - 1;
        else
          n_mig_rd_count <= mig_rd_count;
        end if;
        if mig_rd_ready = '1' then
          if mig_cmd_count = 0 then
            n_mig_state <= mig_s_reading_finish;
            n_mig_addr <= (others => '-');
            n_mig_cmd_count <= (others => '-');
            n_mig_en_read <= '0';
          else
            n_mig_state <= mig_s_reading;
            n_mig_addr <= mig_addr + 1;
            n_mig_cmd_count <= mig_cmd_count - 1;
            n_mig_en_read <= '1';
          end if;
        else
          n_mig_state <= mig_s_reading;
          n_mig_addr <= mig_addr;
          n_mig_cmd_count <= mig_cmd_count;
          n_mig_en_read <= '1';
        end if;

      -- Here we wait for MIG to give us the last few words of read data.
      when mig_s_reading_finish =>
        n_mig_addr <= (others => '-');
        n_mig_cmd_count <= (others => '-');
        n_mig_en_read <= '0';
        n_mig_en_write <= '0';
        n_mig_idle <= '0';
        if data_checked = '1' then
          if mig_rd_count = 0 then
            n_mig_state <= mig_s_done;
            n_mig_rd_count <= (others => '-');
          else
            n_mig_state <= mig_s_reading_finish;
            n_mig_rd_count <= mig_rd_count - 1;
          end if;
        else
          n_mig_state <= mig_s_reading_finish;
          n_mig_rd_count <= mig_rd_count;
        end if;

      -- We are here for a clock cycle before returning to idle.
      when mig_s_done =>
        n_mig_state <= mig_s_idle;
        n_mig_addr <= (others => '-');
        n_mig_cmd_count <= (others => '-');
        n_mig_en_read <= '0';
        n_mig_en_write <= '0';
        n_mig_idle <= '0';
        n_mig_rd_count <= (others => '-');
    end case;
  end process;

  --
  -- Data generator
  --
  -- Generates either:
  -- (1) write data & byte enables when ctl_reading deasserted, or
  -- (2) expected read data when ctl_reading asserted
  --

  -- Flow control, to avoid overflowing FIFOs
  gen_enable <= not (d_be_fifo_wfull or exp_fifo_wfull);

  data_gen_inst : memory_exerciser_data_gen
    generic map(
      memory_data_width => memory_data_width,
      memory_addr_width => memory_addr_width,
      pack_factor => pack_factor
    )
    port map(
      rst => rst,
      clk => clk,
      d => gen_d,
      dbe => gen_dbe,
      dvalid => gen_dvalid,
      exp => gen_exp,
      expvalid => gen_expvalid,
      enable => gen_enable,
      ctl_init_rng => ex_init_rng,
      ctl_go => ex_ph_go,
      ctl_reading => ex_phase(0),
      ctl_phase => ex_phase(3 downto 1),
      ctl_offset => reg_offset,
      ctl_length => reg_length,
      ctl_idle => gen_idle
    );

  gen_dbe_partial <= not AND_reduce(gen_dbe);


  --
  -- FIFO for write data & byte enables
  --

  d_be_fifo_wd <= gen_dbe_partial & gen_dbe & gen_d;

  d_be_fifo_inst : memory_exerciser_fifo
    generic map(
      order => d_be_fifo_order,
      width => d_be_fifo_width
    )
    port map(
      rst => rst,
      clk => clk,
      sr => ex_idle,
      wr => gen_dvalid,
      wd => d_be_fifo_wd,
      wfull => d_be_fifo_wfull,
      aadv => d_be_fifo_radv,
      aempty => d_be_fifo_aempty,
      alevel => open,
      radv => d_be_fifo_radv,
      rempty => d_be_fifo_rempty,
      rq => d_be_fifo_rq
    );

  -- Check d_be_fifo_radv & d_be_fifo_aempty
  -- Check d_be_fifo_radv & d_be_fifo_rempty
  check_d_be_fifo : process(rst, clk)
  begin
    if rst = '1' then
      null;
    elsif rising_edge(clk) then
      if ex_idle = '0' then
        if d_be_fifo_radv = '1' then
          assert d_be_fifo_aempty = '0' severity failure;
          assert d_be_fifo_rempty = '0' severity failure;
        end if;
      end if;
    end if;
  end process;

  d_be_fifo_rq_data <= d_be_fifo_rq(memory_data_width - 1 downto 0);
  d_be_fifo_rq_mask <= not d_be_fifo_rq(memory_data_width + memory_be_width - 1 downto memory_data_width);
  d_be_fifo_rq_partial <= d_be_fifo_rq(memory_data_width + memory_be_width);
  
  rearranger_inst : memory_exerciser_rearranger
    generic map(
      memory_data_width => memory_data_width,
      pack_factor => pack_factor,
      pack_rearranged => pack_rearranged
    )
    port map(
      data_in => d_be_fifo_rq_data,
      mask_in => d_be_fifo_rq_mask,
      data_out => app_wdf_data,
      mask_out => app_wdf_mask
    );
    
  --
  -- FIFO for expected read data
  --

  exp_fifo_inst : memory_exerciser_fifo
    generic map(
      order => exp_fifo_order,
      width => exp_fifo_width
    )
    port map(
      rst => rst,
      clk => clk,
      sr => ex_idle,
      wr => gen_expvalid,
      wd => gen_exp,
      wfull => exp_fifo_wfull,
      aadv => exp_fifo_aadv,
      aempty => exp_fifo_aempty,
      alevel => exp_fifo_alevel,
      radv => exp_fifo_radv,
      rempty => exp_fifo_rempty,
      rq => exp_fifo_rq
    );

  -- Check exp_fifo_aadv & exp_fifo_aempty
  -- Check exp_fifo_radv & exp_fifo_rempty
  check_exp_fifo : process(rst, clk)
  begin
    if rst = '1' then
      null;
    elsif rising_edge(clk) then
      if ex_idle = '0' then
        if exp_fifo_aadv = '1' then
          assert exp_fifo_aempty = '0' severity failure;
        end if;
        if exp_fifo_radv = '1' then
          assert exp_fifo_rempty = '0' severity failure;
        end if;
      end if;
    end if;
  end process;


  --
  -- Verification of data from MIG
  --
  -- Compares actual read data from MIG with expected data from
  -- FIFO, and generates error signals including address of error
  -- and byte lanes that are in error.
  --

  dearranger_inst : memory_exerciser_dearranger
    generic map(
      memory_data_width => memory_data_width,
      pack_factor => pack_factor,
      pack_rearranged => pack_rearranged
    )
    port map(
      data_in => app_rd_data,
      data_out => actual_data
    );
    
  clear_error_log <= ex_idle and ctl_go;

  error_log_inst : memory_exerciser_error_log
    generic map(
      phase_width => ex_phase'length - 1,
      memory_data_width => memory_data_width,
      memory_addr_width => memory_addr_width,
      count_width => error_count_width,
      index_width => error_index_width
    )
    port map(
      rst => rst,
      clk => clk,
      sr => clear_error_log,
      init => ex_ph_go,
      init_phase => ex_phase(ex_phase'high downto 1),
      init_offset => reg_offset,
      chk_ce => app_rd_data_valid,
      chk_act_data => actual_data,
      chk_exp_data => exp_fifo_rq,
      chk_done => data_checked,
      rd_ce => ctl_erd,
      rd_index => ctl_eindex,
      rd_not_empty => error_log_not_empty,
      rd_full => ctl_efull,
      rd_overflow => ctl_elost,
      rd_count => ctl_ecount,
      rd_rec_phase => ctl_ephase,
      rd_rec_addr => ctl_eaddr,
      rd_rec_bad_bits => ctl_ebits,
      rd_rec_act_data => ctl_edata
    );

  ctl_error <= error_log_not_empty;
  

end architecture;
