library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

entity ram_interface_axi4 is
  generic(
    s_axi_axaddr_width : in    natural;
    s_axi_xdata_width  : in    natural;
    s_axi_xid_width    : in    natural;
    s_axi_axlen_width  : in    natural;
    arbitration_mode   : in    string := "RR";
    ram_read_latency   : in    natural range 1 to 8;
    ram_order          : in    natural
  );
  port(
    aresetn            : in    std_logic;
    aclk               : in    std_logic;
    -- AXI4 MM slave interface (MODE = Slave, VLNV = xilinx.com:interface:aximm_rtl:1.0)
    s_axi_araddr       : in    std_logic_vector(s_axi_axaddr_width - 1 downto 0);
    s_axi_arburst      : in    std_logic_vector(1 downto 0);
    s_axi_arcache      : in    std_logic_vector(3 downto 0);
    s_axi_arid         : in    std_logic_vector(s_axi_xid_width - 1 downto 0);
    s_axi_arlen        : in    std_logic_vector(s_axi_axlen_width - 1 downto 0);
    s_axi_arlock       : in    std_logic_vector(0 downto 0);
    s_axi_arprot       : in    std_logic_vector(2 downto 0);
    s_axi_arqos        : in    std_logic_vector(3 downto 0);
    s_axi_arready      : out   std_logic;
    s_axi_arregion     : in    std_logic_vector(3 downto 0);
    s_axi_arsize       : in    std_logic_vector(2 downto 0);
    s_axi_arvalid      : in    std_logic;
    s_axi_awaddr       : in    std_logic_vector(s_axi_axaddr_width - 1 downto 0);
    s_axi_awburst      : in    std_logic_vector(1 downto 0);
    s_axi_awcache      : in    std_logic_vector(3 downto 0);
    s_axi_awid         : in    std_logic_vector(s_axi_xid_width - 1 downto 0);
    s_axi_awlen        : in    std_logic_vector(s_axi_axlen_width - 1 downto 0);
    s_axi_awlock       : in    std_logic_vector(0 downto 0);
    s_axi_awprot       : in    std_logic_vector(2 downto 0);
    s_axi_awqos        : in    std_logic_vector(3 downto 0);
    s_axi_awready      : out    std_logic;
    s_axi_awregion     : in    std_logic_vector(3 downto 0);
    s_axi_awsize       : in    std_logic_vector(2 downto 0);
    s_axi_awvalid      : in    std_logic;
    s_axi_wdata        : in    std_logic_vector(s_axi_xdata_width - 1 downto 0);
    s_axi_wlast        : in    std_logic;
    s_axi_wstrb        : in    std_logic_vector(s_axi_xdata_width / 8 - 1 downto 0);
    s_axi_wready       : out   std_logic;
    s_axi_wvalid       : in    std_logic;
    s_axi_bid          : out   std_logic_vector(s_axi_xid_width - 1 downto 0);
    s_axi_bready       : in    std_logic;
    s_axi_bresp        : out   std_logic_vector(1 downto 0);
    s_axi_bvalid       : out   std_logic;
    s_axi_rdata        : out   std_logic_vector(s_axi_xdata_width - 1 downto 0);
    s_axi_rid          : out   std_logic_vector(s_axi_xid_width - 1 downto 0);
    s_axi_rlast        : out   std_logic;
    s_axi_rready       : in    std_logic;
    s_axi_rresp        : out   std_logic_vector(1 downto 0);
    s_axi_rvalid       : out   std_logic;
    -- RAM interface (MODE = Master, VLNV = xilinx.com:interface:bram_rtl:1.0)
    ram_clk_a          : out   std_logic;
    ram_rst_a          : out   std_logic;
    ram_en_a           : out   std_logic;
    ram_addr_a         : out   std_logic_vector(ram_order - 1 downto 0);
    ram_we_a           : out   std_logic_vector(s_axi_xdata_width / 8 - 1 downto 0);
    ram_din_a          : out   std_logic_vector(s_axi_xdata_width - 1 downto 0);
    ram_dout_a         : in    std_logic_vector(s_axi_xdata_width - 1 downto 0)
  );
end entity;

architecture rtl of ram_interface_axi4 is

  -- Returns a mask for address incrementing, where a 1 bit holds the corresponding address bit constant.
  function gen_increment_mask(
    axburst : std_logic_vector(1 downto 0);
    axaddr  : std_logic_vector;
    axlen   : std_logic_vector)
  return std_logic_vector is
    constant n : natural := axaddr'length;
    variable mask : std_logic_vector(n - 1 downto 0);
  begin
    case axburst is
      when "00" =>
        -- FIXED: No address bits allowed to increment
        mask := (others => '1');
        
      when "10" =>
        -- WRAP: To get the increment mask for wrap case, sufficient to resize AXLEN to width of AXADDR and flip all bits.
        mask := not std_logic_vector(resize(unsigned(axlen), n));
        
      when others =>
        -- INCR: All address bits allowed to increment
        mask := (others => '0');
    end case;
    
    return mask;
  end;
  
  component ram_interface_fifo
    generic(
      ORDER       : in    natural range 2 to 6;
      WIDTH       : in    natural;
      AFULL_FREE  : in    natural := 0
    );
    port(
      rst         : in    std_logic;
      clk         : in    std_logic;
      wr          : in    std_logic;
      d           : in    std_logic_vector(WIDTH - 1 downto 0);
      adv         : in    std_logic;
      full        : out   std_logic;
      not_full    : out   std_logic;
      afull       : out   std_logic;
      not_afull   : out   std_logic;
      valid       : out   std_logic;
      q           : out   std_logic_vector(WIDTH - 1 downto 0)
    );  
  end component;

  type arb_mode_t is (arb_mode_rr, arb_mode_rdpri, arb_mode_wrpri);
  
  function arb_mode_from_string(
    constant arbitration_mode : in string)
  return arb_mode_t is
  begin
    if arbitration_mode = "RR" then
      return arb_mode_rr;
    elsif arbitration_mode = "RDPRI" then
      return arb_mode_rdpri;
    else
      return arb_mode_wrpri;
    end if;
  end;
  
  constant arb_mode : arb_mode_t := arb_mode_from_string(arbitration_mode);
  
  constant s_axi_wstrb_width : natural := s_axi_xdata_width / 8;
  constant num_byte_lane : natural := s_axi_wstrb_width;
  constant log2_num_byte_lane : natural := natural(floor(log2(real(num_byte_lane) + 0.5)));
  constant addr_width_words : natural := s_axi_axaddr_width - log2_num_byte_lane;
  
  -- If true, eliminates the dead cycle at the end of a burst, at the cost of greater
  -- logic complexity and lower FMAX in CSM.
  constant eliminate_idle_dead_cycle : boolean := true;
  
  
  constant cmd_fifo_order : natural := 3;  -- Command FIFO orders
  constant data_fifo_order : natural := 4; -- Data FIFO orders
  
  
  signal rst : std_logic;
  signal clk : std_logic;
  
  
  -- Read ID FIFO (RIDF)
  constant ridf_width : natural := s_axi_xid_width;
  constant ridf_order : natural := cmd_fifo_order;
  signal ridf_wr : std_logic;
  signal ridf_d : std_logic_vector(ridf_width - 1 downto 0);
  signal ridf_adv : std_logic;
  signal ridf_q : std_logic_vector(ridf_width - 1 downto 0);
  signal ridf_q_id : std_logic_vector(s_axi_xid_width - 1 downto 0);
  signal ridf_valid : std_logic;
  signal ridf_not_full : std_logic;
  
  
  -- Read Command FIFO (RCF)
  constant rcf_width : natural := 2 + s_axi_axlen_width + addr_width_words;
  constant rcf_order : natural := cmd_fifo_order;
  signal rcf_wr : std_logic;
  signal rcf_d : std_logic_vector(rcf_width - 1 downto 0);
  signal rcf_adv : std_logic;
  signal rcf_q : std_logic_vector(rcf_width - 1 downto 0);
  signal rcf_q_addr : std_logic_vector(s_axi_axaddr_width - 1 downto log2_num_byte_lane);
  signal rcf_q_burst : std_logic_vector(1 downto 0);
  signal rcf_q_len : std_logic_vector(s_axi_axlen_width - 1 downto 0);
  signal rcf_valid : std_logic;
  signal rcf_not_full : std_logic;
  
  
  -- Read Response FIFO (RRF)
  constant rrf_width : natural := 1 + s_axi_xdata_width;
  constant rrf_order : natural := data_fifo_order;
  signal rrf_wr : std_logic;
  signal rrf_d : std_logic_vector(rrf_width - 1 downto 0);
  signal rrf_adv : std_logic;
  signal rrf_q : std_logic_vector(rrf_width - 1 downto 0);
  signal rrf_q_data : std_logic_vector(s_axi_xdata_width - 1 downto 0);
  signal rrf_q_last : std_logic;
  signal rrf_valid : std_logic;
  signal rrf_not_afull : std_logic;
  
  
  -- Write Command FIFO (WCF)
  constant wcf_width : natural := 2 + s_axi_xid_width + s_axi_axlen_width + addr_width_words;
  constant wcf_order : natural := cmd_fifo_order;
  signal wcf_wr : std_logic;
  signal wcf_d : std_logic_vector(wcf_width - 1 downto 0);
  signal wcf_adv : std_logic;
  signal wcf_q : std_logic_vector(wcf_width - 1 downto 0);
  signal wcf_q_addr : std_logic_vector(s_axi_axaddr_width - 1 downto log2_num_byte_lane);
  signal wcf_q_burst : std_logic_vector(1 downto 0);
  signal wcf_q_id : std_logic_vector(s_axi_xid_width - 1 downto 0);
  signal wcf_q_len : std_logic_vector(s_axi_axlen_width - 1 downto 0);
  signal wcf_valid : std_logic;
  signal wcf_not_full : std_logic;

  
  -- Write Data FIFO (WDF)
  constant wdf_width : natural := s_axi_xdata_width + s_axi_wstrb_width;
  constant wdf_order : natural := data_fifo_order;
  signal wdf_wr : std_logic;
  signal wdf_d : std_logic_vector(wdf_width - 1 downto 0);
  signal wdf_adv : std_logic;
  signal wdf_q : std_logic_vector(wdf_width - 1 downto 0);
  signal wdf_q_data : std_logic_vector(s_axi_xdata_width - 1 downto 0);
  signal wdf_q_strb : std_logic_vector(s_axi_wstrb_width - 1 downto 0);
  signal wdf_valid : std_logic;
  signal wdf_not_full : std_logic;
  
  
  -- Write Response FIFO (WRF)
  constant wrf_width : natural := s_axi_xid_width;
  constant wrf_order : natural := cmd_fifo_order;
  signal wrf_wr : std_logic;
  signal wrf_d : std_logic_vector(wrf_width - 1 downto 0);
  signal wrf_adv : std_logic;
  signal wrf_q : std_logic_vector(wrf_width - 1 downto 0);
  signal wrf_q_id : std_logic_vector(s_axi_xid_width - 1 downto 0);
  signal wrf_valid : std_logic;
  signal wrf_not_afull : std_logic;
  
  
  -- Read/write arbitration state machine (ASM)
  type arb_state_t is (arb_read_last, arb_write_last);
  constant init_arb_state : arb_state_t := arb_read_last; -- TO DO: make initial state a parameter
  signal arb_state : arb_state_t := init_arb_state;
  
  -- Controller state machine (CSM)
  signal next_avail : std_logic;
  signal next_arb_adv : std_logic;
  signal next_is_read : std_logic;  
  type csm_state_t is (csm_idle, csm_reading, csm_writing);
  signal csm_state : csm_state_t := csm_idle;
  signal n_csm_state : csm_state_t;
  signal csm_addr, n_csm_addr : std_logic_vector(s_axi_axaddr_width - 1 downto log2_num_byte_lane);
  signal csm_bid, n_csm_bid : std_logic_vector(s_axi_xid_width - 1 downto 0);
  signal csm_blen, n_csm_blen : std_logic_vector(s_axi_axlen_width - 1 downto 0);
  signal csm_blast, n_csm_blast : std_logic;
  signal csm_incr_mask, n_csm_incr_mask : std_logic_vector(s_axi_axaddr_width - 1 downto log2_num_byte_lane);
  signal arb_adv : std_logic;
  
  
  -- Read validity pipeline
  signal rd_pipe_valid : std_logic_vector(ram_read_latency downto 0);
  signal rd_pipe_blast : std_logic_vector(ram_read_latency downto 0);
  
  
  -- VHDL delta cycle fudge delay, for ensuring that ram_*_a outputs (other than ram_clk_a) change after ram_clk_a
  signal delta_fudge_delay : time := 1 ps;
  signal ram_rst_a_int : std_logic;
  signal ram_en_a_int : std_logic;
  signal ram_addr_a_int : std_logic_vector(ram_order - 1 downto 0);
  signal ram_we_a_int : std_logic_vector(s_axi_xdata_width / 8 - 1 downto 0);
  signal ram_din_a_int : std_logic_vector(s_axi_xdata_width - 1 downto 0);
  signal ram_dout_a_int : std_logic_vector(s_axi_xdata_width - 1 downto 0);
  
  
begin

  assert s_axi_axaddr_width >= log2_num_byte_lane
    report "ram_interface_axi4: invalid value '" & natural'image(s_axi_axaddr_width) & "' for s_axi_axaddr_width; must be greater than or equal to log2(number of byte lanes)"
    severity error;
    
--  assert axi_interface_type = "AXI4MM" or axi_interface_type = "AXI4LITE"
--    report "ram_interface_axi4: invalid value '" & axi_interface_type & "' for axi_interface_type"
--    severity error;
    
  assert arbitration_mode = "RR" or arbitration_mode = "RDPRI" or arbitration_mode = "WRPRI"
    report "ram_interface_axi4: invalid value '" & arbitration_mode & "' for arbitration_mode"
    severity error;
    
    
  rst <= not aresetn;
  clk <= aclk;
  
  
  -- Output AXI4 signals
  s_axi_arready <= ridf_not_full and rcf_not_full;
  
  s_axi_awready <= wcf_not_full;
  
  s_axi_wready <= wdf_not_full;
  
  s_axi_rdata <= rrf_q_data;
  s_axi_rid <= ridf_q_id;
  s_axi_rlast <= rrf_q_last;
  s_axi_rresp <= "00";
  s_axi_rvalid <= rrf_valid and ridf_valid;
  
  s_axi_bid <= wrf_q_id;
  s_axi_bresp <= "00";
  s_axi_bvalid <= wrf_valid;
  
  
  --
  -- Read ID FIFO (RIDF)
  --
  
  ridf_wr <= ridf_not_full and rcf_not_full and s_axi_arvalid;
  ridf_d <= s_axi_arid;
  ridf_adv <= ridf_valid and rrf_valid and rrf_q_last and s_axi_rready;
  
  ridf_inst : ram_interface_fifo
    generic map(
      ORDER => ridf_order,
      WIDTH => ridf_width
    )
    port map(
      rst => rst,
      clk => clk,
      wr => ridf_wr,
      d => ridf_d,
      adv => ridf_adv,
      full => open,
      not_full => ridf_not_full,
      afull => open,
      not_afull => open,
      valid => ridf_valid,
      q => ridf_q
    );
      
  ridf_q_id <= ridf_q(s_axi_xid_width - 1 downto 0);
  
  
  --
  -- Read Command FIFO (RCF)
  --
  
  rcf_wr <= ridf_not_full and rcf_not_full and s_axi_arvalid;
  rcf_d <= s_axi_arburst & s_axi_arlen & s_axi_araddr(csm_addr'range);
  -- rcf_adv is generated by CSM
  
  rcf_inst : ram_interface_fifo
    generic map(
      ORDER => rcf_order,
      WIDTH => rcf_width
    )
    port map(
      rst => rst,
      clk => clk,
      wr => rcf_wr,
      d => rcf_d,
      adv => rcf_adv,
      full => open,
      not_full => rcf_not_full,
      afull => open,
      not_afull => open,
      valid => rcf_valid,
      q => rcf_q
    );
      
  rcf_q_addr <= rcf_q(addr_width_words - 1 downto 0);
  rcf_q_len <= rcf_q(s_axi_axlen_width + addr_width_words - 1 downto addr_width_words);
  rcf_q_burst <= rcf_q(2 + s_axi_axlen_width + addr_width_words - 1 downto s_axi_axlen_width + addr_width_words);
  
  
  --
  -- Read Response FIFO (RRF)
  --
  
  rrf_wr <= rd_pipe_valid(ram_read_latency);
  rrf_d <= rd_pipe_blast(ram_read_latency) & ram_dout_a_int;
  rrf_adv <= rrf_valid and ridf_valid and s_axi_rready;
  
  rrf_inst : ram_interface_fifo
    generic map(
      ORDER => rrf_order,
      WIDTH => rrf_width,
      AFULL_FREE => ram_read_latency + 1
    )
    port map(
      rst => rst,
      clk => clk,
      wr => rrf_wr,
      d => rrf_d,
      adv => rrf_adv,
      full => open,
      not_full => open,
      afull => open,
      not_afull => rrf_not_afull,
      valid => rrf_valid,
      q => rrf_q
    );
      
  rrf_q_data <= rrf_q(s_axi_xdata_width - 1 downto 0);
  rrf_q_last <= rrf_q(s_axi_xdata_width);
  
  
  --
  -- Write Command FIFO (WCF)
  --
  
  wcf_wr <= wcf_not_full and s_axi_awvalid;
  wcf_d <= s_axi_awburst & s_axi_awid & s_axi_awlen & s_axi_awaddr(csm_addr'range);
  -- wcf_adv is generated by CSM
  
  wcf_inst : ram_interface_fifo
    generic map(
      ORDER => wcf_order,
      WIDTH => wcf_width
    )
    port map(
      rst => rst,
      clk => clk,
      wr => wcf_wr,
      d => wcf_d,
      adv => wcf_adv,
      full => open,
      not_full => wcf_not_full,
      afull => open,
      not_afull => open,
      valid => wcf_valid,
      q => wcf_q
    );
      
  wcf_q_addr <= wcf_q(addr_width_words - 1 downto 0);
  wcf_q_len <= wcf_q(s_axi_axlen_width + addr_width_words - 1 downto addr_width_words);
  wcf_q_id <= wcf_q(s_axi_xid_width + s_axi_axlen_width + addr_width_words - 1 downto s_axi_axlen_width + addr_width_words);
  wcf_q_burst <= wcf_q(2 + s_axi_xid_width + s_axi_axlen_width + addr_width_words - 1 downto s_axi_xid_width + s_axi_axlen_width + addr_width_words);
  
  
  --
  -- Write Data FIFO (WDF)
  --
  
  wdf_wr <= wdf_not_full and s_axi_wvalid;
  wdf_d <= s_axi_wstrb & s_axi_wdata;
  -- wdf_adv is generated by CSM
  
  wdf_inst : ram_interface_fifo
    generic map(
      ORDER => wdf_order,
      WIDTH => wdf_width
    )
    port map(
      rst => rst,
      clk => clk,
      wr => wdf_wr,
      d => wdf_d,
      adv => wdf_adv,
      full => open,
      not_full => wdf_not_full,
      afull => open,
      not_afull => open,
      valid => wdf_valid,
      q => wdf_q
    );
      
  wdf_q_data <= wdf_q(s_axi_xdata_width - 1 downto 0);
  wdf_q_strb <= wdf_q(s_axi_wstrb_width + s_axi_xdata_width - 1 downto s_axi_xdata_width);
  
  
  --
  -- Write Response FIFO (WRF)
  --
  
  -- wrf_wr is generated by CSM;
  wrf_d <= csm_bid;
  wrf_adv <= wrf_valid and s_axi_bready;
  
  wrf_inst : ram_interface_fifo
    generic map(
      ORDER => wrf_order,
      WIDTH => wrf_width,
      AFULL_FREE => 1
    )
    port map(
      rst => rst,
      clk => clk,
      wr => wrf_wr,
      d => wrf_d,
      adv => wrf_adv,
      full => open,
      not_full => open,
      afull => open,
      not_afull => wrf_not_afull,
      valid => wrf_valid,
      q => wrf_q
    );
      
  wrf_q_id <= wrf_q(s_axi_xid_width - 1 downto 0);
  
  
  --
  -- Arbitration state machine (ASM)
  --
  -- Keeps track of who was last to perform a burst (read channel vs. write channel).
  --
  
  gen_arb_state : process(rst, clk)
  begin
    if rst = '1' then
      arb_state <= init_arb_state;
    elsif rising_edge(clk) then
      case arb_state is
        when arb_read_last =>
          if arb_adv = '1' then
            arb_state <= arb_write_last;
          end if;
          
        when arb_write_last =>
          if arb_adv = '1' then
            arb_state <= arb_read_last;
          end if;
      end case;
    end if;
  end process;
  
    
  --
  -- Controller state machine (CSM)
  --
  
  gen_csm_state : process(rst, clk)
  begin
    if rst = '1' then
      csm_state <= csm_idle;
      csm_addr <= (others => '-');
      csm_bid <= (others => '-');
      csm_blast <= '-';
      csm_blen <= (others => '-');
      csm_incr_mask <= (others => '-');
    elsif rising_edge(clk) then
      csm_state <= n_csm_state;
      csm_addr <= n_csm_addr;
      csm_bid <= n_csm_bid;
      csm_blast <= n_csm_blast;
      csm_blen <= n_csm_blen;
      csm_incr_mask <= n_csm_incr_mask;
    end if;
  end process;
  
  --
  -- Determine whether to get next command from read channel or write channel.
  -- Depends upon arbitration mode and whether RCF & WCF have valid outputs.
  --
  determine_next_cmd : process(
    arb_state,
    wcf_valid,
    rcf_valid)
  begin
    case arb_mode is
      when arb_mode_rr =>
        case arb_state is
          when arb_read_last =>
            if wcf_valid = '1' then
              next_avail <= '1';
              next_arb_adv <= '1';
              next_is_read <= '0';
            elsif rcf_valid = '1' then
              next_avail <= '1';
              next_arb_adv <= '0';
              next_is_read <= '1';
            else
              next_avail <= '0';
              next_arb_adv <= '-';
              next_is_read <= '-';
            end if;
            
          when arb_write_last =>
            if rcf_valid = '1' then
              next_avail <= '1';
              next_arb_adv <= '1';
              next_is_read <= '1';
            elsif wcf_valid = '1' then
              next_avail <= '1';
              next_arb_adv <= '0';
              next_is_read <= '0';
            else
              next_avail <= '0';
              next_arb_adv <= '-';
              next_is_read <= '-';
            end if;
        end case;
        
      when arb_mode_rdpri =>
        if rcf_valid = '1' then
          next_avail <= '1';
          next_arb_adv <= '0';
          next_is_read <= '1';
        elsif wcf_valid = '1' then
          next_avail <= '1';
          next_arb_adv <= '0';
          next_is_read <= '0';
        else
          next_avail <= '0';
          next_arb_adv <= '-';
          next_is_read <= '-';
        end if;
    
      when arb_mode_wrpri =>
        if wcf_valid = '1' then
          next_avail <= '1';
          next_arb_adv <= '0';
          next_is_read <= '0';
        elsif rcf_valid = '1' then
          next_avail <= '1';
          next_arb_adv <= '0';
          next_is_read <= '1';
        else
          next_avail <= '0';
          next_arb_adv <= '-';
          next_is_read <= '-';
        end if;
      
      when others =>
        next_avail <= '0';
        next_arb_adv <= '-';
        next_is_read <= '-';
    end case;
  end process;
  
  gen_n_csm_state : process(
    csm_state,
    csm_addr,
    csm_bid,
    csm_blast,
    csm_blen,
    csm_incr_mask,
    next_avail,
    next_arb_adv,
    next_is_read,
    rcf_q_addr,
    rcf_q_burst,
    rcf_q_len,
    rrf_not_afull,
    wcf_q_addr,
    wcf_q_burst,
    wcf_q_id,
    wcf_q_len,
    wdf_valid,
    wrf_not_afull)
  begin
    -- Establish defaults
    rcf_adv <= '0';
    wcf_adv <= '0';
    wdf_adv <= '0';
    wrf_wr <= '0';
    arb_adv <= '0';
    n_csm_addr <= csm_addr;
    n_csm_bid <= csm_bid;
    n_csm_blast <= csm_blast;
    n_csm_blen <= csm_blen;
    n_csm_incr_mask <= csm_incr_mask;
    
    case csm_state is
      when csm_idle =>
        if next_avail = '1' then
          if next_is_read = '1' then
            n_csm_state <= csm_reading;
            n_csm_addr <= rcf_q_addr(csm_addr'range);
            n_csm_bid <= (others => '-');
            n_csm_blast <= not or_reduce(rcf_q_len);
            n_csm_blen <= rcf_q_len;
            n_csm_incr_mask <= gen_increment_mask(rcf_q_burst, rcf_q_addr, rcf_q_len);
            arb_adv <= next_arb_adv;
            rcf_adv <= '1';
          else
            if wrf_not_afull = '1' then
              n_csm_state <= csm_writing;
              n_csm_addr <= wcf_q_addr(csm_addr'range);
              n_csm_bid <= wcf_q_id;
              n_csm_blast <= not or_reduce(wcf_q_len);
              n_csm_blen <= wcf_q_len;
              n_csm_incr_mask <= gen_increment_mask(wcf_q_burst, wcf_q_addr, wcf_q_len);
              arb_adv <= next_arb_adv;
              wcf_adv <= '1';
            else
              n_csm_state <= csm_idle;
              n_csm_addr <= (others => '-');
              n_csm_bid <= (others => '-');
              n_csm_blast <= '-';
              n_csm_blen <= (others => '-');
              n_csm_incr_mask <= (others => '-');
            end if;
          end if;
        else
          n_csm_state <= csm_idle;
          n_csm_addr <= (others => '-');
          n_csm_bid <= (others => '-');
          n_csm_blast <= '-';
          n_csm_blen <= (others => '-');
          n_csm_incr_mask <= (others => '-');
        end if;
        
      when csm_reading =>
        if rrf_not_afull = '1' then
          if csm_blast = '1' then
            if eliminate_idle_dead_cycle and next_avail = '1' then
              if next_is_read = '1' then
                n_csm_state <= csm_reading;
                n_csm_addr <= rcf_q_addr(csm_addr'range);
                n_csm_bid <= (others => '-');
                n_csm_blast <= not or_reduce(rcf_q_len);
                n_csm_blen <= rcf_q_len;
                n_csm_incr_mask <= gen_increment_mask(rcf_q_burst, rcf_q_addr, rcf_q_len);
                arb_adv <= next_arb_adv;
                rcf_adv <= '1';
              else
                if wrf_not_afull = '1' then
                  n_csm_state <= csm_writing;
                  n_csm_addr <= wcf_q_addr(csm_addr'range);
                  n_csm_bid <= wcf_q_id;
                  n_csm_blast <= not or_reduce(wcf_q_len);
                  n_csm_blen <= wcf_q_len;
                  n_csm_incr_mask <= gen_increment_mask(wcf_q_burst, wcf_q_addr, wcf_q_len);
                  arb_adv <= next_arb_adv;
                  wcf_adv <= '1';
                else
                  n_csm_state <= csm_idle;
                  n_csm_addr <= (others => '-');
                  n_csm_bid <= (others => '-');
                  n_csm_blast <= '-';
                  n_csm_blen <= (others => '-');
                  n_csm_incr_mask <= (others => '-');
                end if;
              end if;
            else
              n_csm_state <= csm_idle;
              n_csm_addr <= (others => '-');
              n_csm_bid <= (others => '-');
              n_csm_blast <= '-';
              n_csm_blen <= (others => '-');
              n_csm_incr_mask <= (others => '-');
            end if;
          else
            n_csm_state <= csm_reading;
            n_csm_addr <= (csm_addr and csm_incr_mask) or ((csm_addr + 1) and not csm_incr_mask);
            n_csm_blast <= not or_reduce(csm_blen(csm_blen'high downto 1));
            n_csm_blen <= csm_blen - 1;
          end if;
        else
          n_csm_state <= csm_reading;
        end if;
        
      when csm_writing =>
        wdf_adv <= wdf_valid;
        if wdf_valid = '1' then
          if csm_blast = '1' then
            wrf_wr <= '1';
            if eliminate_idle_dead_cycle and next_avail = '1' then
              if next_is_read = '1' then
                n_csm_state <= csm_reading;
                n_csm_addr <= rcf_q_addr(csm_addr'range);
                n_csm_bid <= (others => '-');
                n_csm_blast <= not or_reduce(rcf_q_len);
                n_csm_blen <= rcf_q_len;
                n_csm_incr_mask <= gen_increment_mask(rcf_q_burst, rcf_q_addr, rcf_q_len);
                arb_adv <= next_arb_adv;
                rcf_adv <= '1';
              else
                if wrf_not_afull = '1' then
                  n_csm_state <= csm_writing;
                  n_csm_addr <= wcf_q_addr(csm_addr'range);
                  n_csm_bid <= wcf_q_id;
                  n_csm_blast <= not or_reduce(wcf_q_len);
                  n_csm_blen <= wcf_q_len;
                  n_csm_incr_mask <= gen_increment_mask(wcf_q_burst, wcf_q_addr, wcf_q_len);
                  arb_adv <= next_arb_adv;
                  wcf_adv <= '1';
                else
                  n_csm_state <= csm_idle;
                  n_csm_addr <= (others => '-');
                  n_csm_bid <= (others => '-');
                  n_csm_blast <= '-';
                  n_csm_blen <= (others => '-');
                  n_csm_incr_mask <= (others => '-');
                end if;
              end if;
            else
              n_csm_state <= csm_idle;
              n_csm_addr <= (others => '-');
              n_csm_bid <= (others => '-');
              n_csm_blast <= '-';
              n_csm_blen <= (others => '-');
              n_csm_incr_mask <= (others => '-');
            end if;
          else
            n_csm_state <= csm_writing;
            n_csm_addr <= (csm_addr and csm_incr_mask) or ((csm_addr + 1) and not csm_incr_mask);
            n_csm_blast <= not or_reduce(csm_blen(csm_blen'high downto 1));
            n_csm_blen <= csm_blen - 1;
          end if;
        else
          n_csm_state <= csm_writing;
        end if;
    end case;
  end process;
  
  
  --
  -- Implement the RAM interface
  --
  
  ram_clk_a <= clk;
  ram_rst_a <= ram_rst_a_int after delta_fudge_delay;
  ram_en_a <= ram_en_a_int after delta_fudge_delay;
  ram_addr_a <= ram_addr_a_int after delta_fudge_delay;
  ram_din_a <= ram_din_a_int after delta_fudge_delay;
  ram_we_a <= ram_we_a_int after delta_fudge_delay;
  ram_dout_a_int <= ram_dout_a after delta_fudge_delay;
  
  ram_interface : process(clk)
  begin
    if rising_edge(clk) then
      ram_rst_a_int <= rst;
      
      case csm_state is
        when csm_idle =>
          if ram_read_latency >= 2 then
            ram_en_a_int <= or_reduce(rd_pipe_valid(ram_read_latency - 2 downto 0));
          else
            ram_en_a_int <= '0';
          end if;
          ram_addr_a_int <= (others => '-');
          ram_din_a_int <= (others => '-');
          ram_we_a_int <= (others => '0');
          
        when csm_reading =>
          if ram_read_latency >= 2 then
            ram_en_a_int <= rrf_not_afull or or_reduce(rd_pipe_valid(ram_read_latency - 2 downto 0));
          else
            ram_en_a_int <= rrf_not_afull;
          end if;
          ram_addr_a_int <= std_logic_vector(resize(unsigned(csm_addr(csm_addr'high downto log2_num_byte_lane)), ram_order));
          ram_din_a_int <= (others => '-');
          ram_we_a_int <= (others => '0');
        
        when csm_writing =>
          if ram_read_latency >= 2 then
            ram_en_a_int <= wdf_valid or or_reduce(rd_pipe_valid(ram_read_latency - 2 downto 0));
          else
            ram_en_a_int <= wdf_valid;
          end if;
          ram_addr_a_int <= std_logic_vector(resize(unsigned(csm_addr(csm_addr'high downto log2_num_byte_lane)), ram_order));
          ram_din_a_int <= wdf_q_data;
          ram_we_a_int <= wdf_q_strb;
      end case;
    end if;
  end process;
  
  ram_read_validity_pipe : process(rst, clk)
  begin
    if rst = '1' then
      rd_pipe_valid <= (others => '0');
      rd_pipe_blast <= (others => '-');
    elsif rising_edge(clk) then
      if csm_state = csm_reading then
        rd_pipe_valid(0) <= rrf_not_afull;
        rd_pipe_blast(0) <= csm_blast;
      else
        rd_pipe_valid(0) <= '0';
        rd_pipe_blast(0) <= '-';
      end if;
      rd_pipe_valid(ram_read_latency downto 1) <= rd_pipe_valid(ram_read_latency - 1 downto 0);
      rd_pipe_blast(ram_read_latency downto 1) <= rd_pipe_blast(ram_read_latency - 1 downto 0);
    end if;
  end process;
  
  
end architecture;
