library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity ram_interface_fifo is
  generic(
    ORDER       : in    natural range 2 to 6;                  -- Depth is 2**ORDER
    WIDTH       : in    natural;                               -- Width of FIFO, in bits
    AFULL_FREE  : in    natural := 0                           -- When 'AFULL_FREE' or fewer free locations, 'afull' is asserted
  );
  port(
    rst         : in    std_logic;                             -- Asynchronous reset
    clk         : in    std_logic;                             -- Clock
    wr          : in    std_logic;                             -- '1' => write value on 'd' into FIFO
    d           : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
    adv         : in    std_logic;                             -- '1' => present next item on 'q' to next item
    full        : out   std_logic;                             -- '1' => full
    not_full    : out   std_logic;                             -- Inverse of 'full'
    afull       : out   std_logic;                             -- '1' => almost full
    not_afull   : out   std_logic;                             -- Inverse of 'afull'
    valid       : out   std_logic;                             -- '1' => not empty
    q           : out   std_logic_vector(WIDTH - 1 downto 0)   -- Current item, qualified by 'valid'
  );  
end;

architecture rtl of ram_interface_fifo is

  constant DEPTH : natural := 2**ORDER;
  constant FULL_LEVEL : natural := DEPTH - 1;
  constant AFULL_LEVEL : natural := FULL_LEVEL - AFULL_FREE;

  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;

  signal wptr : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  signal rptr : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  signal rn : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  signal not_empty : std_logic;
  signal i_full : std_logic := '0';
  signal i_not_full : std_logic := '1';
  signal i_afull : std_logic := '0';
  signal i_not_afull : std_logic := '1';

  signal qvalid : std_logic := '0';
  signal qce : std_logic;
  signal rd : std_logic;
  signal i_q : std_logic_vector(WIDTH - 1 downto 0);

  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";
  signal sel_ram : ram_data_t;

begin

  valid <= qvalid;
  full <= i_full;
  not_full <= i_not_full;
  afull <= i_afull;
  not_afull <= i_not_afull;
  q <= i_q;

  write_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER - 1 downto 0);
  begin
    if rst = '1' then
      wptr <= (others => '0');
      i_not_full <= '1';
      i_full <= '0';
      i_not_afull <= '1';
      i_afull <= '0';
    elsif rising_edge(clk) then
      if wr = '1' then
        wptr <= wptr + 1;
      end if;

      if wr = '1' then
        assert conv_integer(wptr - rptr) < FULL_LEVEL severity error;
        tmp := wptr - rptr + 1;
      else
        tmp := wptr - rptr;
      end if;
      if tmp >= FULL_LEVEL then
        i_full <= '1';
        i_not_full <= '0';
      else
        i_full <= '0';
        i_not_full <= '1';
      end if;
      if tmp >= AFULL_LEVEL then
        i_afull <= '1';
        i_not_afull <= '0';
      else
        i_afull <= '0';
        i_not_afull <= '1';
      end if;
    end if;

    if rising_edge(clk) then
      if wr = '1' then
        ram(conv_integer(wptr)) <= d;
      end if;
    end if;
  end process;

  not_empty <= OR_reduce(rn);
  qce <= adv OR NOT qvalid;
  rd <= (adv OR NOT qvalid) AND not_empty;

  read_logic : process(rst, clk)
  begin
    if rst = '1' then
      rptr <= (others => '0');
    elsif rising_edge(clk) then
      if rd = '1' then
        rptr <= rptr + 1;
      end if;
    end if;

    if rst = '1' then
      null;
    elsif rising_edge(clk) then
      if rd = '1' then
        assert conv_integer(wptr - rptr) > 0 severity failure;
      end if;
    end if;

    if rst = '1' then
      rn <= (others => '0');
    elsif rising_edge(clk) then
      if rd = '1' then
        rn <= wptr - rptr - 1;
      else
        rn <= wptr - rptr;
      end if;
    end if;
  end process;

  output_valid : process(rst, clk)
  begin
    if rst = '1' then
      qvalid <= '0';
    elsif rising_edge(clk) then
      if qce = '1' then
        qvalid <= not_empty;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr));

  output_data : process(clk)
  begin
    if rising_edge(clk) then
      if qce = '1' then
        -- Assign i_q to undefined on FIFO empty to assist in simulation debug only
        if not_empty = '1' then
          i_q <= sel_ram;
        else
          i_q <= (others => '-');
        end if;
      end if;
    end if;
  end process;

end;
