#!/bin/bash

function find_device() {
  local device_dirs=`find /sys/bus/pci/devices -maxdepth 1 -mindepth 1 -name '*:*:*.*' -printf '%p '`
  local instance=0
  for device_dir in $device_dirs ; do
    local dev_vid=`cat $device_dir/vendor | awk '{ print tolower($0) }' | sed -e 's/^0x//g'`
    local dev_did=`cat $device_dir/device | awk '{ print tolower($0) }' | sed -e 's/^0x//g'`
    local dev_svid=`cat $device_dir/subsystem_vendor | awk '{ print tolower($0) }' | sed -e 's/^0x//g'`
    local dev_sdid=`cat $device_dir/subsystem_device | awk '{ print tolower($0) }' | sed -e 's/^0x//g'`
    local dev_class=`cat $device_dir/class | awk '{ print tolower($0) }' | sed -e 's/^0x//g'`
    if [ "$1" != "" ] && [ "$1" != "$dev_vid" ] ; then
      continue
    fi
    if [ "$2" != "" ] && [ "$2" != "$dev_did" ] ; then
      continue
    fi
    if [ "$3" != "" ] && [ "$3" != "$dev_svid" ] ; then
      continue
    fi
    if [ "$4" != "" ] && [ "$4" != "$dev_sdid" ] ; then
      continue
    fi
    if [ "$5" != "" ] && [ "$5" != "$dev_class" ] ; then
      continue
    fi
    if [ "$6" != "" ] && [ "$6" -ne "$instance" ] ; then
      instance=$((instance+1))
      continue
    fi
    echo "$device_dir"
    break
  done
}

function showhelp() {
  echo 'Usage: '$0' [option/predicate ...]'
  echo
  echo "Makes the Linux kernel reconfigure the PCI configuration space of a specified"
  echo "device (if not configured), by removing it from the kernel and rescanning the"
  echo "PCI bus hierarchy."
  echo
  echo 'where '"'"option"'"' can be the following:'
  echo '  -f, --force             Override the two or more predicates requirement'
  echo '  -r, --remove-only       Remove device from Linux kernel without rescanning PCI'
  echo '  -h, -help, --help       Show this help.'
  echo ''
  echo 'and '"'"predicate"'"' can be the following:'
  echo '  -vid <vendor ID>        Specify vendor ID, e.g. 4144.'
  echo '  -did <device ID>        Specify device ID, e.g. adb3.'
  echo '  -svid <sub. vendor ID>  Specify subsystem vendor ID, e.g. 4144.'
  echo '  -sdid <sub. device ID>  Specify subsystem device ID, e.g. 0706.'
  echo '  -class <class code>     Specify class code, e.g. 068000.'
  echo '  -instance <n>           Search for instance '"'n'"' (zero-based) of matching device.'
  echo ''
}

req_vid=''
req_did=''
req_svid=''
req_sdid=''
req_class=''
req_instance=0
force=0
predicates=0
remove_only=0

while (($# > 0)) ; do
  if [ "$1" == "-help" ] || [ "$1" == "-h" ] || [ "$1" == "--help" ] ; then
    showhelp
    exit 0
  elif [ "$1" == "-vid" ] ; then
    req_vid=`echo $2 | awk '{ print tolower($0) }' | sed -e 's/^0x//g' -e 's/.*[^A-Fa-f0-9].*/ERROR/g'`
    if [ "$req_vid" == "" ] || [ "$req_vid" == "ERROR" ] ; then
      echo "*** '-req_vid' option must be followed by a hexadecimal number, e.g. 4144 or 0x4144"
      echo "    Use '$0 -h' for help."
      exit 1
    fi
    predicates=$((predicates+1))
    shift 2
  elif [ "$1" == "-did" ] ; then
    req_did=`echo $2 | awk '{ print tolower($0) }' | sed -e 's/^0x//g' -e 's/.*[^A-Fa-f0-9].*/ERROR/g'`
    if [ "$req_did" == "" ] || [ "$req_did" == "ERROR" ] ; then
      echo "*** '-req_did' option must be followed by a hexadecimal number, e.g. adb3 or 0xadb3"
      echo "    Use '$0 -h' for help."
      exit 1
    fi
    predicates=$((predicates+1))
    shift 2
  elif [ "$1" == "-svid" ] ; then
    req_svid=`echo $2 | awk '{ print tolower($0) }' | sed -e 's/^0x//g' -e 's/.*[^A-Fa-f0-9].*/ERROR/g'`
    if [ "$req_svid" == "" ] || [ "$req_svid" == "ERROR" ] ; then
      echo "*** '-req_svid' option must be followed by a hexadecimal number, e.g. 4144 or 0x4144"
      echo "    Use '$0 -h' for help."
      exit 1
    fi
    predicates=$((predicates+1))
    shift 2
  elif [ "$1" == "-sdid" ] ; then
    req_sdid=`echo $2 | awk '{ print tolower($0) }' | sed -e 's/^0x//g' -e 's/.*[^A-Fa-f0-9].*/ERROR/g'`
    if [ "$req_sdid" == "" ] || [ "$req_sdid" == "ERROR" ] ; then
      echo "*** '-req_sdid' option must be followed by a hexadecimal number, e.g. 700 or 0x0700"
      echo "    Use '$0 -h' for help."
      exit 1
    fi
    predicates=$((predicates+1))
    shift 2
  elif [ "$1" == "-class" ] ; then
    req_class=`echo $2 | awk '{ print tolower($0) }' | sed -e 's/^0x//g' -e 's/.*[^A-Fa-f0-9].*/ERROR/g'`
    if [ "$req_class" == "" ] || [ "$req_class" == "ERROR" ] ; then
      echo "*** '-req_class' option must be followed by a hexadecimal number, e.g. 68000 or 0x068000"
      echo "    Use '$0 -h' for help."
      exit 1
    fi
    predicates=$((predicates+1))
    shift 2
  elif [ "$1" == "-instance" ] ; then
    req_instance=`echo $2 | sed -e 's/^0x//g' -e 's/.*[^0-9].*/ERROR/g'`
    if [ "$req_instance" == "" ] || [ "$req_instance" == "ERROR" ] ; then
      echo "*** '-req_instance' option must be followed by a zero-based instance number, e.g. 0 or 1."
      echo "    Use '$0 -h' for help."
      exit 1
    fi
    shift 2
  elif [ "$1" == "--force" ] || [ "$1" == "-f" ] ; then
    force=1
    shift 1
  elif [ "$1" == "--remove-only" ] || [ "$1" == "-r" ] ; then
    remove_only=1
    shift 1
  else
    echo "*** Unknown option: $1"
    exit 1
  fi
done

if [ "$predicates" -lt "2" ] && [ "$force" != "1" ] ; then
  echo "*** Less than two predicates specified, which may be dangerous."
  echo "    To override this failsafe, use the '-f' option."
  exit 2
fi

found=`find_device "$req_vid" "$req_did" "$req_svid" "$req_sdid" "$req_class" "$req_instance"`
if [ "$found" == "" ] ; then
  echo "*** Device not found."
  exit 3
else
  echo "Found device: $found"
fi

echo "Removing knowledge of $found from Linux kernel..."
echo '1' > "$found/remove"

echo "Verifying that device is no longer present..."
tmp_found=`find_device "$req_vid" "$req_did" "$req_svid" "$req_sdid" "$req_class" "$req_instance"`
if [ "$tmp_found" != "" ] ; then
  echo "WARNING: device is still present!"
fi

if [ "$remove_only" != "1" ] ; then
  echo "Forcing rescan of PCI bus hierarchy..."
  echo '1' > /sys/bus/pci/rescan

  echo "Verifying that device is present again..."
  tmp_found=`find_device "$req_vid" "$req_did" "$req_svid" "$req_sdid" "$req_class" "$req_instance"`
  if [ "$tmp_found" == "" ] ; then
    echo "WARNING: device not found after PCI rescan."
  fi
fi
