/*
** avr2util.c - utility to manipulate AVR2 microcontroller on certain reconfigurable
**              computing cards, at a low level.
**
** (C) Copyright 2016 Alpha Data
**
** Supported models:
**
**   o ADM-XRC-KU1
**   o ADM-PCIE-8V3
**   o ADM-PCIE-8K5
**
** Exit codes:
**
**   0   Success
**   1   Invalid command passed as 1st positional argument
**   2   Not enough positional arguments following command
**   3   Index value out of range or not a valid number
**   4   Frequency value out of range or not a valid number
**   5   Failed to read firmware file
**   6   Firmware file is too large for uC
**   7   Failed to allocate buffer for firmware or board config data
**   8   Errors found when verifying updated firmware or board config data
**   9   Attempting to use this utility on an unsupported model
** 100   Failed to open device
** 101   Failed to get AVR2 uC status
** 102   Failed to send command to AVR2 uC
** 103   AVR2 uC returned nonzero status for operation
** 104   AVR2 uC's response was too short (< 2 bytes) to be valid
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tcscpy_s(x, n, y) do { strncpy(x, y, n); (x)[(n) - 1] = '\0'; } while (0)
# if defined(__VXWORKS__) || defined(__vxworks)
#   include <ctype.h>
#   define _tcsicmp(pStr1, pStr2) ourStrCaseCmp(pStr1, pStr2)
# else
#   define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# endif
# define _tcsstr(x, y) strstr(x, y)
# define _tmain main
# define _tprintf printf
# define _stscanf_s sscanf
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>
#include <adb3.h>
#include <adb3/avr2.h>

#include <sdk_common.h>
#include <platform_time.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define FIRMWARE_BASE_ADDR     (0x0U) /* Address in Application Flash where firmware code lives */
#define FIRMWARE_LENGTH_LIMIT  (0x1FC00U) /* Maximum possible size of firmware code */

#define BOARD_CONFIG_BASE_ADDR    (0x1FC00U) /* Address in Application Flash where board configuration block lives */
#define BOARD_CONFIG_LENGTH_LIMIT (512U)     /* Maximum possible size of board configuration block */

#define VPD_BASE_ADDR    (0x1FF00U) /* Address in Application Flash where VPD lives */
#define VPD_LENGTH_LIMIT (128U)     /* Maximum possible size of VPD block */

#ifdef _WIN32
# define PROGRAM_NAME _T("avr2util.exe")
#else
# define PROGRAM_NAME _T("avr2util")
#endif

#if defined(__VXWORKS__) || defined(__vxworks)
static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}
#endif

#if !defined(_MSC_VER)
/* Emulate _tfopen_s ("secure file open") from MSVC++ */
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = fopen(pFilename, pMode);
  if (NULL == pFile) {
    return -1;
  }
  *ppFile = pFile;
  return 0;
}
#endif

static ADB3_STATUS
sampleAdb3OpenCard(
  unsigned int index,
  bool_t bByIndex,
  unsigned int serial,
  bool_t bBySerial,
  bool_t bReadOnly,
  ADB3_HANDLE* phCard)
{
  ADB3_STATUS status;
  ADB3_HANDLE hCard;
  ADB3_UNIQUE_ID id;
  unsigned int i;

  if (bByIndex && bBySerial) {
    /* Attempt to open card with specified index and serial number */
    status = ADB3_Open(index, bReadOnly, 0, &hCard);
    if (status != ADB3_SUCCESS) {
      return status;
    }
    status = ADB3_GetUniqueId(hCard, &id);
    if (status != ADB3_SUCCESS) {
      return status;
    }
    if (id.SerialNumber == serial) {
      *phCard = hCard;
      return ADB3_SUCCESS;
    }
    return ADB3_DEVICE_NOT_FOUND;
  }

  if (bBySerial) {
    /* Attempt to open card with specified serial number */
    for (i = 0; i < 256; i++) {
      status = ADB3_Open(i, bReadOnly, 0, &hCard);
      if (status != ADB3_SUCCESS) {
        if (status == ADB3_DEVICE_NOT_FOUND) {
          return status;
        }
        continue;
      }
      status = ADB3_GetUniqueId(hCard, &id);
      if (status != ADB3_SUCCESS) {
        continue;
      }
      if (id.SerialNumber == serial) {
        *phCard = hCard;
        return ADB3_SUCCESS;
      }
      ADB3_Close(hCard);
    }
    return ADB3_DEVICE_NOT_FOUND;
  } else {
    /* Attempt to open card with specified index */
    return ADB3_Open(index, bReadOnly, 0, phCard);
  }
}

static void
dumpAvr2Command(
  const uint8_t* pBuffer,
  uint32_t length)
{
  uint32_t i;

  _tprintf(_T("DEBUG: AVR2 command[%lu] = { "), (unsigned long)length);
  for (i = 0; i < length; i++) {
    _tprintf(_T("0x%02X "), (unsigned int)pBuffer[i]);
  }
  _tprintf(_T("}\n"));
}

static void
dumpAvr2Response(
  const uint8_t* pBuffer,
  uint32_t expectedLength,
  uint32_t actualLength)
{
  uint32_t i, length;

  length = (actualLength < expectedLength) ? actualLength : expectedLength;
  _tprintf(_T("DEBUG: AVR2 response[%lu] = { "), (unsigned long)length);
  for (i = 0; i < length; i++) {
    _tprintf(_T("0x%02X "), (unsigned int)pBuffer[i]);
  }
  _tprintf(_T("}\n"));
  if (actualLength > expectedLength) {
    _tprintf(_T("DEBUG: %lu extra unexpected characters of actual AVR2 response were discarded.\n"),
      (unsigned long)(actualLength - expectedLength));
  }
}

static int
sendAvr2Command(
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  uint32_t timeoutUs,
  const uint8_t* pCommand,
  uint32_t commandLength,
  const uint8_t expResponse[2],
  uint8_t* pResponse,
  uint32_t responseLength,
  uint32_t* pActualLength)
{
  ADB3_STATUS status;

  if (bVerbose) {
    dumpAvr2Command(pCommand, commandLength);
  }
  status = ADB3_CommandAVR2(hCard, 0, timeoutUs, commandLength, pCommand, responseLength, pResponse, pActualLength);
  if (ADB3_SUCCESS != status) {
    _tprintf(_T("*** Failed to send command to AVR2 uC: %s\n"), ADB3_GetStatusString(status, TRUE));
    return 102;
  }
  if (bVerbose) {
    dumpAvr2Response(pResponse, responseLength, *pActualLength);
  }
  if (*pActualLength != responseLength) {
    _tprintf(_T("+++ Unepected AVR2 response length %lu bytes; expected %lu bytes.\n"),
      (unsigned long)*pActualLength, (unsigned long)responseLength);
  }
  if (*pActualLength < 2) {
    _tprintf(_T("+++ AVR2 response too short to be valid.\n"));
    return 104;
  } else {
    if (pResponse[0] != expResponse[0] || pResponse[1] != expResponse[1]) {
      _tprintf(_T("+++ Unexpected AVR2 response code { 0x%02X, 0x%02X }; expected { 0x%02X, 0x%02X }.\n"),
        (unsigned int)pResponse[0], (unsigned int)pResponse[1],
        (unsigned int)expResponse[0], (unsigned int)expResponse[1]);
    }
  }

  return 0;
}

static int
versionCommand(
  ADB3_HANDLE hCard,
  bool_t bVerbose)
{
  const uint32_t timeoutUs = 100000;
  uint8_t command[2];
  uint8_t response[10];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint16_t version[4];
  int ret = 0;

  command[0] = 0x03; /* User command group */
  command[1] = 0x83; /* Version query */
  expResponse[0] = 0x03; /* User command group */
  expResponse[1] = 0x03; /* Version response */

  memset(response, 0, sizeof(response));
  ret = sendAvr2Command(
    hCard,
    bVerbose,
    timeoutUs,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (0 != ret) {
    return ret;
  }

  version[0] = ((uint16_t)response[2] << 0) | ((uint16_t)response[3] << 8);
  version[1] = ((uint16_t)response[4] << 0) | ((uint16_t)response[5] << 8);
  version[2] = ((uint16_t)response[6] << 0) | ((uint16_t)response[7] << 8);
  version[3] = ((uint16_t)response[8] << 0) | ((uint16_t)response[9] << 8);
  _tprintf(_T("AVR2 firmware version is %u.%u.%u.%u\n"), version[0], version[1], version[2], version[3]);

  return 0;
}

static int
getClkNvCommand(
  Arguments* pArguments,
  ADB3_HANDLE hCard,
  bool_t bVerbose)
{
  const uint32_t timeoutUs = 500000;
  uint8_t command[4];
  uint8_t response[7];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = 0;
  int nConverted;
  unsigned int index;
  uint32_t frequency;
  bool_t bIsOverridden;

  if (pArguments->nParam < 2) {
    _tprintf(_T("*** 'getclknv' command must be followed by a clock generator index.\n"));
    return 2;
  }

  nConverted = _stscanf_s(pArguments->param[1].pStringVal, _T("%u"), &index);
  if (1 != nConverted) {
    _tprintf(_T("*** Clock generator index value '%s' not recognized as a number.\n"), pArguments->param[1].pStringVal);
    return 3;
  }
  if (index > 0xFFU) {
    _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
    return 3;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x84; /* GetFrequencyEx query */
  command[2] = (uint8_t)index; /* Clock generator index; cast safe due to above range check. */
  command[3] = 0x1; /* Flags: 0x1 => Get NV frequency value */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x04; /* GetFrequencyEx response */

  memset(response, 0, sizeof(response));
  ret = sendAvr2Command(
    hCard,
    bVerbose,
    timeoutUs,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (0 != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid clock generator index");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned error status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    return 103;
  }

  frequency = ((uint32_t)response[3] << 0) | ((uint32_t)response[4] << 8) | ((uint32_t)response[5] << 16) | ((uint32_t)response[6] << 24);
  bIsOverridden = (frequency != 0U && frequency != 0xFFFFFFFFU);
  if (bIsOverridden) {
    _tprintf(_T("NV override frequency for clock generator %u is %lu Hz\n"),
      index, (unsigned long)frequency);
  } else {
    _tprintf(_T("NV override frequency for clock generator %u is %lu(0x%lX), which will not result in programming at power-up.\n"),
      index, (unsigned long)frequency, (unsigned long)frequency);
  }

  return 0;
}

static int
setClkNvCommand(
  Arguments* pArguments,
  ADB3_HANDLE hCard,
  bool_t bVerbose)
{
  const uint32_t timeoutUs = 500000;
  uint8_t command[8];
  uint8_t response[7];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = 0;
  int nConverted;
  unsigned int index;
  unsigned long frequency;
  bool_t bIsOverridden;

  if (pArguments->nParam < 3) {
    _tprintf(_T("*** 'setclknv' command must be followed by (1) a clock generator index and (2) an override frequency (Hz).\n"));
    return 2;
  }

  nConverted = _stscanf_s(pArguments->param[1].pStringVal, _T("%u"), &index);
  if (1 != nConverted) {
    _tprintf(_T("*** Clock generator index value '%s' not recognized as a number.\n"), pArguments->param[1].pStringVal);
    return 3;
  }
  if (index > 0xFFU) {
    _tprintf(_T("*** Clock generator index must be in the range 0 to 255.\n"));
    return 3;
  }

  nConverted = _stscanf_s(pArguments->param[2].pStringVal, _T("%lu"), &frequency);
  if (1 != nConverted) {
    _tprintf(_T("*** Override frequency value '%s' not recognized as a number.\n"), pArguments->param[2].pStringVal);
    return 4;
  }
  if (frequency > 0xFFFFFFFFU) {
    _tprintf(_T("*** Override frequency must be in the range 0 to 4294967295.\n"));
    return 4;
  }

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x83; /* SetFrequencyEx query */
  command[2] = (uint8_t)index; /* Clock generator index; cast safe due to above range check. */
  command[3] = (uint8_t)(frequency >> 0) & 0xFFU; /* Frequency bits [7:0] */
  command[4] = (uint8_t)(frequency >> 8) & 0xFFU; /* Frequency bits [15:8] */
  command[5] = (uint8_t)(frequency >> 16) & 0xFFU; /* Frequency bits [23:16] */
  command[6] = (uint8_t)(frequency >> 24) & 0xFFU; /* Frequency bits [31:24] */
  command[7] = 0x3; /* Flags: 0x1 => Set NV frequency value, 0x2 => Don't program clock generator now */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x03; /* SetFrequencyEx response */

  memset(response, 0, sizeof(response));
  ret = sendAvr2Command(
    hCard,
    bVerbose,
    timeoutUs,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (0 != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("Invalid clock generator index");
      break;

    case 2:
      pReason = _T("Clock generator frequency cannot be overridden");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    return 103;
  }

  bIsOverridden = (frequency != 0U && frequency != 0xFFFFFFFFU);

  if (bIsOverridden) {
    _tprintf(_T("NV override frequency for clock generator %u set to %lu Hz\n"),
      index, (unsigned long)frequency);
  } else {
    _tprintf(_T("NV override frequency for clock generator %u set to %lu(0x%lX), which will not result in programming at power-up.\n"),
      index, (unsigned long)frequency,  (unsigned long)frequency);
  }

  return 0;
}

static int
enterExitServiceMode(
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  bool_t bEnterNotExit)
{
  const uint32_t timeoutUs = 500000;
  uint8_t command[2];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = 0;

  command[0] = 0x03; /* User command group */
  expResponse[0] = 0x03; /* User command group */
  if (bEnterNotExit) {
    command[1] = 0x80; /* ServiceMode command */
    expResponse[1] = 0x00; /* ServiceMode acknowledge response */
  } else {
    command[1] = 0x82; /* ApplicationMode command */
    expResponse[1] = 0x02; /* ApplicationMode acknowledge response */
  }

  memset(response, 0, sizeof(response));
  ret = sendAvr2Command(
    hCard,
    bVerbose,
    timeoutUs,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (0 != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for %s command\n"),
      (unsigned int)avr2Status, bEnterNotExit ? _T("Enter Service Mode") : _T("Exit Service Mode"));
    return 102;
  }

  return 0;
}

#define FLASH_PAGE_ORDER (9)
#define FLASH_PAGE_SIZE (1U << FLASH_PAGE_ORDER)
#define FLASH_PAGE_MASK (FLASH_PAGE_SIZE - 1U)

static int
readApplicationFlashPage(
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  uint32_t pageAddress,
  uint8_t* pPageBuffer)
{
  const uint32_t timeoutUs = 1000000;
  uint8_t command[5];
  uint8_t response[3 + FLASH_PAGE_SIZE];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint16_t pageIndex;
  uint8_t avr2Status;
  int ret = 0;

  _tprintf(_T("Reading Application Flash page @ 0x%lX...\n"), (unsigned long)pageAddress);

  pageIndex = (uint16_t)(pageAddress >> FLASH_PAGE_ORDER); /* Cast should be safe because range of pageAddress is limited by callers */

  command[0] = 0x06; /* Boot Flash command group */
  command[1] = 0x80; /* Flash Application Read command */
  command[2] = FLASH_PAGE_ORDER;
  command[3] = (uint8_t)((pageIndex >> 0) & 0xFFU);
  command[4] = (uint8_t)((pageIndex >> 8) & 0xFFU);
  expResponse[0] = 0x06; /* Boot Flash command group */
  expResponse[1] = 0x00; /* Flash Application Read response */

  memset(response, 0, sizeof(response));
  ret = sendAvr2Command(
    hCard,
    bVerbose,
    timeoutUs,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (0 != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for Flash Application Read command\n"), avr2Status);
    return 102;
  }

  memcpy(pPageBuffer, &response[3], FLASH_PAGE_SIZE);

  return 0;
}

static int
writeApplicationFlashPage(
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  uint32_t pageAddress,
  const uint8_t* pPageBuffer)
{
  const uint32_t timeoutUs = 1000000;
  uint8_t command[5 + FLASH_PAGE_SIZE];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint16_t pageIndex;
  uint8_t avr2Status;
  int ret = 0;

  _tprintf(_T("Writing Application Flash page @ 0x%lX...\n"), (unsigned long)pageAddress);

  pageIndex = (uint16_t)(pageAddress >> FLASH_PAGE_ORDER); /* Cast should be safe because range of pageAddress is limited by callers */

  command[0] = 0x06; /* Boot Flash command group */
  command[1] = 0x82; /* Flash Application Write command */
  command[2] = FLASH_PAGE_ORDER;
  command[3] = (uint8_t)((pageIndex >> 0) & 0xFFU);
  command[4] = (uint8_t)((pageIndex >> 8) & 0xFFU);
  memcpy(&command[5], pPageBuffer, FLASH_PAGE_SIZE);
  expResponse[0] = 0x06; /* Boot Flash command group */
  expResponse[1] = 0x02; /* Flash Application Write response */

  memset(response, 0, sizeof(response));
  ret = sendAvr2Command(
    hCard,
    bVerbose,
    timeoutUs,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (0 != ret) {
    return ret;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    _tprintf(_T("*** AVR2 returned status 0x%02X for Flash Application Write command\n"), avr2Status);
    return 102;
  }

  return 0;
}

static int
writeApplicationFlash(
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  uint32_t address,
  const uint8_t* pBuffer,
  uint32_t length)
{
  uint8_t pageBuffer[512];
  uint32_t pageAddress, pageOffset, chunk;
  int ret = 0;

  while (length) {
    pageOffset = address & FLASH_PAGE_MASK;
    pageAddress = address & ~FLASH_PAGE_MASK;
    chunk = FLASH_PAGE_SIZE - pageOffset;
    if (chunk > length) {
      chunk = length;
    }

    if (chunk != FLASH_PAGE_SIZE) {
      ret = readApplicationFlashPage(hCard, bVerbose, pageAddress, pageBuffer);
      if (0 != ret) {
        return ret;
      }
    }
    memcpy(pageBuffer + pageOffset, pBuffer, chunk);

    ret = writeApplicationFlashPage(hCard, bVerbose, pageAddress, pageBuffer);
    if (0 != ret) {
      return ret;
    }

    length -= chunk;
    address += chunk;
    pBuffer += chunk;
  }

  return 0;
}

static int
verifyApplicationFlash(
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  uint32_t address,
  const uint8_t* pBuffer,
  uint32_t length)
{
  const unsigned int maxErrorReported = 20;
  uint8_t pageBuffer[512];
  uint32_t i, pageAddress, pageOffset, chunk;
  int ret = 0;
  uint32_t errorCount = 0;

  while (length) {
    pageOffset = address & FLASH_PAGE_MASK;
    pageAddress = address & ~FLASH_PAGE_MASK;
    chunk = FLASH_PAGE_SIZE - pageOffset;
    if (chunk > length) {
      chunk = length;
    }

    ret = readApplicationFlashPage(hCard, bVerbose, pageAddress, pageBuffer);
    if (0 != ret) {
      return ret;
    }

    for (i = 0; i < chunk; i++) {
      uint8_t expected = *pBuffer++;
      uint8_t actual = pageBuffer[pageOffset + i];

      if (actual != expected) {
        if (errorCount < maxErrorReported) {
          if (expected != actual) {
            _tprintf(_T("*** Verification error at address 0x%lX; expected=0x%02X actual=0x%02X\n"),
              (unsigned long)address, (unsigned int)expected, (unsigned int)actual);
          }
        } else if (errorCount == maxErrorReported) {
          _tprintf(_T("*** More than %u verification errors detected.\n"), maxErrorReported);
        }
        errorCount++;
      }
      address++;
    }

    length -= chunk;
  }

  if (errorCount) {
    _tprintf(_T("*** Total of %u verification errors detected.\n"), errorCount);
    ret = 8;
  }

  return 0;
}

static int
updateFirmwareCommand(
  Arguments* pArguments,
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  bool_t bInServiceMode)
{
  const unsigned int firmwareLengthLimit = FIRMWARE_LENGTH_LIMIT;
  const TCHAR* pFilename;
  FILE* pFile = NULL;
  uint8_t* pBuffer = NULL;
  long length;
  size_t nRead;
  int ret = 0;

  if (pArguments->nParam < 2) {
    _tprintf(_T("*** 'update-firmware' command must be followed by the firmware filename (.bin).\n"));
    return 2;
  }

  pFilename = pArguments->param[1].pStringVal;
  if (0 != _tfopen_s(&pFile, pFilename, _T("rb"))) {
    _tprintf(_T("*** Failed to open firmware file '%s' for reading.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_END)) {
    _tprintf(_T("*** Failed to seek to end of firmware file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  length = ftell(pFile);
  if (length < 0) {
    _tprintf(_T("*** Failed to get length of firmware file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_SET)) {
    _tprintf(_T("*** Failed to seek to beginning of firmware file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if ((size_t)length >= firmwareLengthLimit) { /* Cast to size_t safe due to positive check above */
    _tprintf(_T("*** Firmware file cannot be more than 0x%X bytes; actual length is 0x%llX.\n"),
      firmwareLengthLimit, (unsigned long long)length);
    ret = 6;
    goto done;
  }
  _tprintf(_T("Firmware file '%s' length is %u(0x%X) byte(s).\n"),
    pFilename, (unsigned int)length, (unsigned int)length); /* Cast safe due to previous range checks */

  pBuffer = (uint8_t*)malloc(length);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate firmware file buffer.\n"));
    ret = 7;
    goto done;
  }

  nRead = fread(pBuffer, 1, length, pFile);
  if (nRead != (size_t)length) {
    _tprintf(_T("*** Failed to read expected number of bytes from firmware file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

  if (!bInServiceMode) {
    _tprintf(_T("Entering service mode: "));
    fflush(stdout);
    ret = enterExitServiceMode(hCard, bVerbose, TRUE);
    if (0 != ret) {
      return ret;
    }
    sleepMs(50);
    _tprintf(_T("OK\n"));
  }

  ret = writeApplicationFlash(hCard, bVerbose, FIRMWARE_BASE_ADDR, pBuffer, (size_t)length);
  if (0 == ret) {
    ret = verifyApplicationFlash(hCard, bVerbose, FIRMWARE_BASE_ADDR, pBuffer, (size_t)length);
  }

  if (!bInServiceMode) {
    _tprintf(_T("Exiting service mode: "));
    fflush(stdout);
    enterExitServiceMode(hCard, bVerbose, FALSE);
    sleepMs(50);
    _tprintf(_T("OK\n"));
  }

done:
  if (NULL != pBuffer) {
    free(pBuffer);
  }
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

static int
updateBoardConfigCommand(
  Arguments* pArguments,
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  bool_t bInServiceMode)
{
  const unsigned int boardConfigLengthLimit = BOARD_CONFIG_LENGTH_LIMIT;
  const TCHAR* pFilename;
  FILE* pFile = NULL;
  uint8_t* pBuffer = NULL;
  long length;
  size_t nRead;
  int ret = 0;

  if (pArguments->nParam < 2) {
    _tprintf(_T("*** 'update-brdcfg' command must be followed by the board config. filename (.bin).\n"));
    return 2;
  }

  pFilename = pArguments->param[1].pStringVal;
  if (0 != _tfopen_s(&pFile, pFilename, _T("rb"))) {
    _tprintf(_T("*** Failed to open board config. file '%s' for reading.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_END)) {
    _tprintf(_T("*** Failed to seek to end of board config. file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  length = ftell(pFile);
  if (length < 0) {
    _tprintf(_T("*** Failed to get length of board config. file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_SET)) {
    _tprintf(_T("*** Failed to seek to beginning of board config. file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if ((size_t)length >= boardConfigLengthLimit) { /* Cast to size_t safe due to positive check above */
    _tprintf(_T("*** Board config. file cannot be more than 0x%X bytes; actual length is 0x%llX.\n"),
      boardConfigLengthLimit, (unsigned long long)length);
    ret = 6;
    goto done;
  }
  _tprintf(_T("Board config file '%s' length is %u(0x%X) byte(s).\n"),
    pFilename, (unsigned int)length, (unsigned int)length); /* Cast safe due to previous range checks */

  pBuffer = (uint8_t*)malloc(length);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate board config. file buffer.\n"));
    ret = 7;
    goto done;
  }

  nRead = fread(pBuffer, 1, length, pFile);
  if (nRead != (size_t)length) {
    _tprintf(_T("*** Failed to read expected number of bytes from board config. file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

  if (!bInServiceMode) {
    _tprintf(_T("Entering service mode: "));
    fflush(stdout);
    ret = enterExitServiceMode(hCard, bVerbose, TRUE);
    if (0 != ret) {
      return ret;
    }
    sleepMs(50);
    _tprintf(_T("OK\n"));
  }

  ret = writeApplicationFlash(hCard, bVerbose, BOARD_CONFIG_BASE_ADDR, pBuffer, (size_t)length);
  if (0 == ret) {
    ret = verifyApplicationFlash(hCard, bVerbose, BOARD_CONFIG_BASE_ADDR, pBuffer, (size_t)length);
  }

  if (!bInServiceMode) {
    _tprintf(_T("Exiting service mode: "));
    fflush(stdout);
    enterExitServiceMode(hCard, bVerbose, FALSE);
    sleepMs(50);
    _tprintf(_T("OK\n"));
  }

done:
  if (NULL != pBuffer) {
    free(pBuffer);
  }
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

static int
updateVpdCommand(
  Arguments* pArguments,
  ADB3_HANDLE hCard,
  bool_t bVerbose,
  bool_t bInServiceMode)
{
  const unsigned int vpdLengthLimit = VPD_LENGTH_LIMIT;
  const TCHAR* pFilename;
  FILE* pFile = NULL;
  uint8_t buffer[VPD_LENGTH_LIMIT];
  long length;
  size_t nRead;
  int ret = 0;

  if (pArguments->nParam < 2) {
    _tprintf(_T("*** 'update-vpd' command must be followed by the board config. filename (.bin).\n"));
    return 2;
  }

  pFilename = pArguments->param[1].pStringVal;
  if (0 != _tfopen_s(&pFile, pFilename, _T("rb"))) {
    _tprintf(_T("*** Failed to VPD file '%s' for reading.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_END)) {
    _tprintf(_T("*** Failed to seek to end of VPD file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  length = ftell(pFile);
  if (length < 0) {
    _tprintf(_T("*** Failed to get length of VPD file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if (0 != fseek(pFile, 0, SEEK_SET)) {
    _tprintf(_T("*** Failed to seek to beginning of VPD file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  if ((size_t)length >= vpdLengthLimit) { /* Cast to size_t safe due to positive check above */
    _tprintf(_T("*** VPD file cannot be more than 0x%X bytes; actual length is 0x%llX.\n"),
      vpdLengthLimit, (unsigned long long)length);
    ret = 6;
    goto done;
  }
  _tprintf(_T("VPD file '%s' length is %u(0x%X) byte(s).\n"),
    pFilename, (unsigned int)length, (unsigned int)length); /* Cast safe due to previous range checks */

  memset(buffer, 0, sizeof(buffer));
  nRead = fread(buffer, 1, length, pFile);
  if (nRead != (size_t)length) {
    _tprintf(_T("*** Failed to read expected number of bytes from VPD file '%s'.\n"), pFilename);
    ret = 5;
    goto done;
  }
  fclose(pFile);
  pFile = NULL;

  if (!bInServiceMode) {
    _tprintf(_T("Entering service mode: "));
    fflush(stdout);
    ret = enterExitServiceMode(hCard, bVerbose, TRUE);
    if (0 != ret) {
      return ret;
    }
    sleepMs(50);
    _tprintf(_T("OK\n"));
  }

  ret = writeApplicationFlash(hCard, bVerbose, VPD_BASE_ADDR, buffer, (size_t)length);
  if (0 == ret) {
    ret = verifyApplicationFlash(hCard, bVerbose, VPD_BASE_ADDR, buffer, (size_t)length);
  }

  if (!bInServiceMode) {
    _tprintf(_T("Exiting service mode: "));
    fflush(stdout);
    enterExitServiceMode(hCard, bVerbose, FALSE);
    sleepMs(50);
    _tprintf(_T("OK\n"));
  }

done:
  if (NULL != pFile) {
    fclose(pFile);
  }

  return ret;
}

#if 0
static int
rawCommand(
  Arguments* pArguments,
  ADB3_HANDLE hCard)
{
  return 1;
}
#endif

static int
modelSanityCheck(
  ADB3_HANDLE hCard)
{
  ADB3_STATUS status;
  ADB3_UNIQUE_ID id;

  status = ADB3_GetUniqueId(hCard, &id);
  if (ADB3_SUCCESS != status) {
    _tprintf(_T("*** Failed to get type of card: %s\n"), ADB3_GetStatusString(status, TRUE));
  }

  switch (id.Model) {
  case ADMXRC3_MODEL_ADMXRCKU1:
  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    return 0;

  default:
    _tprintf(_T("*** This version of %s can be used only with the following models:\n"), PROGRAM_NAME);
    _tprintf(_T("\n"));
    _tprintf(_T("  o ADM-XRC-KU1\n"));
    _tprintf(_T("  o ADM-PCIE-8V3\n"));
    _tprintf(_T("  o ADM-PCIE-8K5\n"));
    return 9;
  }
}

int _tmain(
  int argc,
  TCHAR* argv[])
{
  int ret = 0;
  ADB3_HANDLE hCard = ADB3_HANDLE_INVALID_VALUE;
  ADB3_STATUS status;
  ADB3_AVR2_STATUS avr2Status;
  Arguments arguments = g_initArguments;
  Option options[3];
  OptionType paramType[1];
  TCHAR* paramInfo[1];
  TCHAR* paramMeaning[1];
  unsigned int serial = 0, index = 0;
  bool_t bVerbose, bInServiceMode;
  TCHAR* pCommand;
  
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("verbose");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = 0;
  options[2].meaning = _T("TRUE => verbose output");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.bNoStdOptions = TRUE;
   
  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("One of 'version', 'getclknv', 'setclknv', 'update-firmware', 'update-brdcfg', 'update-vpd'");
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = 1;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  ret = sampleParseCommandLine(PROGRAM_NAME, argc, argv, &arguments);
  if (ret < 0) {
    goto done;
  }

  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bVerbose = options[2].value.booleanVal;

  pCommand = arguments.param[0].pStringVal;
  
  status = sampleAdb3OpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADB3_SUCCESS) {
    _tprintf(_T("*** Failed to open card: %s\n"), ADB3_GetStatusString(status, TRUE));
    ret = 100;
    goto done;
  }

  ret = modelSanityCheck(hCard);
  if (0 != ret) {
    goto done;
  }

  status = ADB3_GetAVR2Status(hCard, &avr2Status);
  if (status != ADB3_SUCCESS) {
    _tprintf(_T("*** Failed to get AVR2 status: %s\n"), ADB3_GetStatusString(status, TRUE));
    ret = 101;
    goto done;
  }
  bInServiceMode = avr2Status.Flags.ServiceMode ? TRUE : FALSE;
  _tprintf(_T("%s Service Mode\n"), bInServiceMode ? _T("In") : _T("Not in"));

  if (_tcsicmp(pCommand, _T("version")) == 0) {
    ret = versionCommand(hCard, bVerbose);
  } else if (_tcsicmp(pCommand, _T("getclknv")) == 0) {
    ret = getClkNvCommand(&arguments, hCard, bVerbose);
  } else if (_tcsicmp(pCommand, _T("setclknv")) == 0) {
    ret = setClkNvCommand(&arguments, hCard, bVerbose);
  } else if (_tcsicmp(pCommand, _T("update-firmware")) == 0) {
    ret = updateFirmwareCommand(&arguments, hCard, bVerbose, bInServiceMode);
  } else if (_tcsicmp(pCommand, _T("update-brdcfg")) == 0) {
    ret = updateBoardConfigCommand(&arguments, hCard, bVerbose, bInServiceMode);
  } else if (_tcsicmp(pCommand, _T("update-vpd")) == 0) {
    ret = updateVpdCommand(&arguments, hCard, bVerbose, bInServiceMode);
#if 0
  } else if (_tcsicmp(pCommand, _T("raw")) == 0) {
    ret = rawCommand(&arguments, hCard, bVerbose);
#endif
  } else {
    _tprintf(_T("*** Unrecognised command: %s\n"), pCommand);
    _tprintf(_T("    Use '%s -h' for help.\n"), PROGRAM_NAME);
    ret = 1;
    goto done;
  }
    
done: 
  if (hCard != ADB3_HANDLE_INVALID_VALUE) {
    ADB3_Close(hCard);
    hCard = ADB3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
