/*
** dmadump.c - utility to display or write data using a DMA engine on an
**             Alpha Data Gen 3 reconfigurable computing device.
**
** (C) Copyright 2014 Alpha Data
**
** This utility functions in a similar way to the "dump" utility, but uses DMA
** transfers (bus mastering) instead of Direct Slave (CPU-initiated) transfers.
**
** Note that each DMA engine has its own OCP address space, separate to the other
** DMA engines and to the Direct Slave OCP channel.
**
** Therefore, unless an FPGA design explicitly routes DMA OCP channels and Direct
** Slave channels to the same OCP endpoint (for example, the BlockRAM in the UBER
** FPGA design), data written using dump/dmadump will not be visible when
** attempting to read it back using dmadump/dump.
**
** This utility does not make any attempt to aggregate write data items into one
** large chunk before writing, so a write command that writes three 32-bit values
** will typically result in three DMA write transfers of four bytes each.
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>
# include <limits.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stprintf_s snprintf
#   define _stscanf_s(buffer, ...) _stscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
#   define _tcscpy_s(x, y, z) tcscpy(x, z)
#   define _tcsnset_s(buffer, max, c, n) _tcsnset(buffer, c, n)
#endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _istprint(x) isprint(x)
# define _stprintf_s snprintf
# define _stscanf_s sscanf
# define _tscanf_s scanf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# if defined(__VXWORKS__) || defined(__vxworks)
#   define _tcsicmp(pStr1, pStr2) ourStrCaseCmp(pStr1, pStr2)
# else
#   define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# endif
# define _tcslen(x) strlen(x)
# define _tcsnset_s(buffer, max, c, n) strset_s(buffer, max, c, n)
# define _tmain main
# define _tprintf printf
# define _tscanf(buffer, ...) scanf(buffer, __VA_ARGS__)
# define _totlower(c) tolower(c)
# define _gettc getc

#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <num_token.h>
#include <platform_endian.h>

#if defined(_UNICODE)
# define EOF_VALUE WEOF
#else
# define EOF_VALUE EOF
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define MIN(a, b) ((a) < (b) ? (a) : (b))

#define FILL_BUFFER_SIZE (0x10000U)

typedef enum _CommandType {
  CommandTypeFill,
  CommandTypeRead,
  CommandTypeWrite
} CommandType;

#if !defined(_WIN32)
static void
strset_s(
  TCHAR* pBuffer,
  size_t max,
  TCHAR c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = _T('\0');
}
#endif

#if defined(__VXWORKS__) || defined(__vxworks)
static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}
#endif

static bool_t
fillMemoryUsingDma(
  unsigned int radix,
  bool_t bBigEndian,
  ADMXRC3_HANDLE hDevice,
  unsigned int dmaChannel,
  uint64_t localAddress,
  size_t n,
  uint64_t fillValue)
{
  ADMXRC3_STATUS status;
  union {
    uint8_t asUint8[FILL_BUFFER_SIZE / sizeof(uint8_t)];
    uint16_t asUint16[FILL_BUFFER_SIZE / sizeof(uint16_t)];
    uint32_t asUint32[FILL_BUFFER_SIZE / sizeof(uint32_t)];
    uint64_t asUint64[FILL_BUFFER_SIZE / sizeof(uint64_t)];
  } fillBuffer;
  size_t i, m = n / radix, chunk;

  chunk = n > FILL_BUFFER_SIZE ? FILL_BUFFER_SIZE : n;
  m = chunk / radix;
  switch (radix) {
  case 1:
    if (bBigEndian) {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint8[i] = cpu_to_be8((uint8_t)fillValue);
      }
    } else {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint8[i] = cpu_to_le8((uint8_t)fillValue);
      }
    }
    break;

  case 2:
    if (bBigEndian) {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint16[i] = cpu_to_be16((uint16_t)fillValue);
      }
    } else {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint16[i] = cpu_to_le16((uint16_t)fillValue);
      }
    }
    break;

  case 4:
    if (bBigEndian) {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint32[i] = cpu_to_be32((uint32_t)fillValue);
      }
    } else {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint32[i] = cpu_to_le32((uint32_t)fillValue);
      }
    }
    break;

  case 8:
    if (bBigEndian) {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint64[i] = cpu_to_be64(fillValue);
      }
    } else {
      for (i = 0; i < m; i++) {
        fillBuffer.asUint64[i] = cpu_to_le64(fillValue);
      }
    }
    break;

  default:
    assert(FALSE);
  }

  while (n) {
    chunk = n > FILL_BUFFER_SIZE ? FILL_BUFFER_SIZE : n;
    status = ADMXRC3_WriteDMAEx(hDevice, dmaChannel, 0, (void*)fillBuffer.asUint8, chunk, localAddress);
      if (ADMXRC3_SUCCESS != status) {
        _tprintf(_T("*** fillMemoryUsingDma: Failed to perform write on DMA channel %u at local address 0x%08lX_%08lX: %s\n"),
          dmaChannel,
          (unsigned long)(uint32_t)(localAddress >> 32U), (unsigned long)(uint32_t)localAddress,
          ADMXRC3_GetStatusString(status, TRUE));
      return FALSE; /* Failure */
    }
    localAddress += chunk;
    n -= chunk;
  }

  return TRUE; /* OK */
}

static bool_t
dumpMemoryUsingDma(
  unsigned int indent,
  unsigned int nColumn,
  unsigned int radix,
  bool_t bBigEndian,
  bool_t bAlign,
  ADMXRC3_HANDLE hDevice,
  unsigned int dmaChannel,
  uint64_t localAddress,
  size_t n)
{
  static const unsigned int addressSize = sizeof(uint64_t);
  ADMXRC3_STATUS status;
  unsigned int addressStrLen = 3 + 2 * addressSize + (addressSize - 1) / 4;
  uint64_t limit, addressCol, m;
  unsigned int addressLo, columnMask;
  uint8_t* p8;
  TCHAR line[257];
  TCHAR* pPos;
  uint8_t columnBuf[256];
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int i, j, k, chunk;
  unsigned int rowBytes;
  int nwritten;
  size_t remaining;

  if (indent > 128) {
    _tprintf(_T("*** dumpMemoryUsingDma: invalid value '%lu' for indent\n"), (unsigned long)indent);
    return FALSE; /* Failure */
  }

  switch (nColumn) {
  case 1:
  case 2:
  case 4:
  case 8:
  case 16:
  case 32:
    break;
  default:
    _tprintf(_T("*** dumpMemoryUsingDma: invalid value '%lu' for nColumn\n"), (unsigned long)nColumn);
    return FALSE; /* Failure */
  }
  if (radix != 1 && radix != 2 && radix != 4 && radix != 8) {
    _tprintf(_T("*** dumpMemoryUsingDma: invalid value '%lu' for radix\n"), (unsigned long)radix);
    return FALSE; /* Failure */
  }

  rowBytes = nColumn * radix;

  /* Make sure everything is aligned */
  columnMask = rowBytes - 1U;
  limit = localAddress + n;
  if (bAlign) {
    addressCol = localAddress & ~(uint64_t)columnMask;
    addressLo = ((unsigned int)localAddress) & columnMask;
  } else {
    addressCol = localAddress;
    addressLo = 0;
  }

  m = limit - localAddress;

  pPos = line;
  pPos[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent);
  pPos += MIN(indent, remaining);
  remaining -= MIN(indent, remaining);
  nwritten = _stprintf_s(pPos, remaining, _T("Dump of memory at 0x%08lX_%08lX + %lu(0x%lX) bytes:\n"),
    (unsigned long)(uint32_t)(localAddress >> 32U), (unsigned long)(uint32_t)localAddress,
    (unsigned long)n, (unsigned long)n);
  if (nwritten > 0) {
    pPos += nwritten;
    remaining -= nwritten;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
  _tprintf(_T("%s"), line);

  pPos = line;
  line[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent + 1 + addressStrLen);
  pPos += MIN(indent + 1 + addressStrLen, remaining);
  remaining -= MIN(indent + 1 + addressStrLen, remaining);
  for (i = 0; i < rowBytes; i += radix) {
    switch (radix) {
    case 0x1U:
      nwritten = _stprintf_s(pPos, remaining, _T("%02lX "), (unsigned long)i);
      break;

    case 0x2U:
      nwritten = _stprintf_s(pPos, remaining, _T("  %02lX "), (unsigned long)i);
      break;

    case 0x4U:
      nwritten = _stprintf_s(pPos, remaining, _T("      %02lX "), (unsigned long)i);
      break;

    case 0x8U:
      nwritten = _stprintf_s(pPos, remaining, _T("               %02lX "), (unsigned long)i);
      break;
    }
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }
  }
  if (remaining) {
    *pPos++ = _T('\n');
    remaining--;
  }
  if (remaining) {
    *pPos++ = _T('\0');
    remaining--;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
  _tprintf(_T("%s"), line);

  while (m) {
    chunk = rowBytes - addressLo;
    if (chunk > m) {
      chunk = (unsigned int)m;
    }

    memset(columnBuf, 0, rowBytes);
    status = ADMXRC3_ReadDMAEx(hDevice, dmaChannel, 0, columnBuf + addressLo, chunk, localAddress);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** dumpMemoryUsingDma: Failed to perform read on DMA channel %u at local address 0x%08lX_%08lX: %s\n"),
        dmaChannel,
        (unsigned long)(uint32_t)(localAddress >> 32U), (unsigned long)(uint32_t)localAddress,
        ADMXRC3_GetStatusString(status, TRUE));
      return FALSE; /* Failure */
    }

    p8 = (uint8_t*)columnBuf;

    pPos = line;
    line[0] = _T('\0');
    remaining = ARRAY_LENGTH(line);
    _tcsnset_s(pPos, remaining, _T(' '), indent);
    pPos += MIN(indent, remaining);
    remaining -= MIN(indent, remaining);
    nwritten = _stprintf_s(pPos, remaining, _T("0x%08lX_%08lX: "), (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }

    j = addressLo + chunk;
    for (i = 0; i < rowBytes; i += radix) {
      switch (radix) {
        case 0x1U:
          val8 = *p8++;
          if (i >= addressLo && i < j) {
            nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)val8);
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
          } else {
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
          }
          break;

        case 0x2U:
          if (bBigEndian) {
            val16 = (((uint16_t)p8[0]) << 8U) | (((uint16_t)p8[1]) << 0U);
          } else {
            val16 = (((uint16_t)p8[0]) << 0U) | (((uint16_t)p8[1]) << 8U);
          }
          p8 += 2;
          for (k = 2; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val16 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
            if (k == 3) {
              if (remaining) {
                *pPos++ = _T('_');
                remaining--;
              }
            }
          }
          break;

        case 0x4U:
          if (bBigEndian) {
            val32 = (((uint32_t)p8[0]) << 24U) | (((uint32_t)p8[1]) << 16U) | (((uint32_t)p8[2]) <<  8U) | (((uint32_t)p8[3]) <<  0U);
          } else {
            val32 = (((uint32_t)p8[0]) <<  0U) | (((uint32_t)p8[1]) <<  8U) | (((uint32_t)p8[2]) << 16U) | (((uint32_t)p8[3]) << 24U);
          }
          p8 += 4;
          for (k = 4; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val32 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
          }
          break;

        case 0x8U:
          if (bBigEndian) {
            val64  = (((uint64_t)p8[0]) << 56U) | (((uint64_t)p8[1]) << 48U) | (((uint64_t)p8[2]) << 40U) | (((uint64_t)p8[3]) << 32U);
            val64 |= (((uint64_t)p8[4]) << 24U) | (((uint64_t)p8[5]) << 16U) | (((uint64_t)p8[6]) <<  8U) | (((uint64_t)p8[7]) <<  0U);
          } else {
            val64  = (((uint64_t)p8[0]) <<  0U) | (((uint64_t)p8[1]) <<  8U) | (((uint64_t)p8[2]) << 16U) | (((uint64_t)p8[3]) << 24U);
            val64 |= (((uint64_t)p8[4]) << 32U) | (((uint64_t)p8[5]) << 40U) | (((uint64_t)p8[6]) << 48U) | (((uint64_t)p8[7]) << 56U);
          }
          p8 += 8;
          for (k = 8; k != 0;) {
            k--;
            if (i + k >= addressLo && i + k < j) {
              nwritten = _stprintf_s(pPos, remaining, _T("%02lX"), (unsigned long)(uint8_t)(val64 >> (k << 3)));
              if (nwritten > 0) {
                pPos += nwritten;
                remaining -= nwritten;
              }
            } else {
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
              if (remaining) {
                *pPos++ = _T('.');
                remaining--;
              }
            }
            if (k == 4) {
              if (remaining) {
                *pPos++ = _T('_');
                remaining--;
              }
            }
          }
          break;
      }
      if (remaining) {
        *pPos++ = _T(' ');
        remaining--;
      }
    }

    p8 = (uint8_t*)columnBuf;
    for (i = 0; i < rowBytes; i++) {
      char c = (char)*p8++;

      if (remaining) {
        if (_istprint(c) && c != _T('\t')) {
          *pPos++ = c;
        } else {
          *pPos++ = _T('.');
        }
        remaining--;
      }
    }
    if (remaining) {
      *pPos++ = _T('\n');
      remaining--;
    }
    if (remaining) {
      *pPos++ = _T('\0');
      remaining--;
    }
    line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
    _tprintf(_T("%s"), line);
    addressLo = 0;
    addressCol += rowBytes;
    m -= chunk;
    localAddress += chunk;
  }

  return TRUE; /* Success */
}

static bool_t
writeValueUsingDma(
  ADMXRC3_HANDLE hDevice,
  unsigned int dmaChannel,
  unsigned int width,
  bool_t bBigEndian,
  uint64_t localAddress,
  uint64_t val64)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;

  switch (width) {
  case 1U:
    if (bBigEndian) {
      val8 = (uint8_t)cpu_to_be8((uint8_t)val64);
    } else {
      val8 = (uint8_t)cpu_to_le8((uint8_t)val64);
    }
    status = ADMXRC3_WriteDMAEx(hDevice, dmaChannel, 0, &val8, sizeof(val8), localAddress);
    break;

  case 2U:
    if (bBigEndian) {
      val16 = (uint16_t)cpu_to_be16((uint16_t)val64);
    } else {
      val16 = (uint16_t)cpu_to_le16((uint16_t)val64);
    }
    status = ADMXRC3_WriteDMAEx(hDevice, dmaChannel, 0, &val16, sizeof(val16), localAddress);
    break;

  case 4U:
    if (bBigEndian) {
      val32 = (uint32_t)cpu_to_be32((uint32_t)val64);
    } else {
      val32 = (uint32_t)cpu_to_le32((uint32_t)val64);
    }
    status = ADMXRC3_WriteDMAEx(hDevice, dmaChannel, 0, &val32, sizeof(val32), localAddress);
    break;

  case 8U:
    if (bBigEndian) {
      val64 = (uint64_t)cpu_to_be64(val64);
    } else {
      val64 = (uint64_t)cpu_to_le64(val64);
    }
    status = ADMXRC3_WriteDMAEx(hDevice, dmaChannel, 0, &val64, sizeof(val64), localAddress);
    break;

  default:
    assert(FALSE);
    break;
  }

  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** writeValueUsingDma: Failed to perform write on DMA channel %u at local address 0x%08lX_%08lX: %s\n"),
      dmaChannel,
      (unsigned long)(uint32_t)(localAddress >> 32U), (unsigned long)(uint32_t)localAddress,
      ADMXRC3_GetStatusString(status, TRUE));
    return FALSE; /* Failure */
  }

  return TRUE; /* OK */
}

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  SampleTokenState state = g_sampleTokenStateInit;
  unsigned int tokenFlags = SAMPLE_TOKEN_FLAG_NEWLINE_EMPTY_TOKEN;
  Option options[4];
  Arguments arguments = g_initArguments;
  OptionType paramType[5];
  TCHAR* paramInfo[5];
  TCHAR* paramMeaning[5];
  bool_t bBigEndian, bInterpretHex;
  CommandType command;
  unsigned int n, width, index, serial, dmaChannel;
  uint64_t localAddress;
  TCHAR* pCommand;
  int ret = 0;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("be");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("TRUE => use big-endian memory format");
  options[3].key = _T("hex");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
  options[3].meaning = _T("TRUE => interpret values as being hexadecimal");
  arguments.nOption = 4;
  arguments.option = options;
  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("Command to perform: fb, fw, fd, fq, rb, rw, rd, rq, wb, ww, wd or wq");
  paramType[1] = OptionUInt;
  paramInfo[1] = _T("channel");
  paramMeaning[1] = _T("Index of DMA channel to use");
  paramType[2] = OptionUInt64;
  paramInfo[2] = _T("address");
  paramMeaning[2] = _T("Local / OCP address to fill, read or write");
  paramType[3] = OptionUInt;
  paramInfo[3] = _T("length");
  paramMeaning[3] = _T("Number of bytes to fill, read or write");
  paramType[4] = OptionString;
  paramInfo[4] = _T("item ...");
  paramMeaning[4] = _T("Fill value or data items to write");
  arguments.nParamType = 5;
  arguments.minNParam = 3;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;
  
  if (sampleParseCommandLine(_T("dump"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bBigEndian = options[2].value.booleanVal;
  bInterpretHex = options[3].value.booleanVal;
  if (bInterpretHex) {
    tokenFlags |= SAMPLE_TOKEN_FLAG_INTERPRET_HEX;
  }
  pCommand = arguments.param[0].pStringVal;
  dmaChannel = arguments.param[1].uintVal;
  localAddress = arguments.param[2].uint64Val;

  if (_tcsicmp(pCommand, _T("fb")) == 0) {
    width = 1;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fw")) == 0) {
    width = 2;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fd")) == 0) {
    width = 4;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("fq")) == 0) {
    width = 8;
    command = CommandTypeFill;
  } else if (_tcsicmp(pCommand, _T("rb")) == 0) {
    width = 1;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rw")) == 0) {
    width = 2;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rd")) == 0) {
    width = 4;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("rq")) == 0) {
    width = 8;
    command = CommandTypeRead;
  } else if (_tcsicmp(pCommand, _T("wb")) == 0) {
    width = 1;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("ww")) == 0) {
    width = 2;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("wd")) == 0) {
    width = 4;
    command = CommandTypeWrite;
  } else if (_tcsicmp(pCommand, _T("wq")) == 0) {
    width = 8;
    command = CommandTypeWrite;
  } else {
    _tprintf(_T("*** Invalid command: %s\n"), pCommand);
    ret = -1;
    goto done;
  }

  if (arguments.nParam < 4) {
    n = width;
  } else {
    n = arguments.param[3].uintVal;
  }
  if (n % width) {
    _tprintf(_T("+++ Warning: byte count %u(0x%x) is not an integer multiple of width (%u)\n"), n, n, width);
    n &= ~(width - 1U);
    _tprintf(_T("+++          byte count adjusted to %u(0x%x)\n"), n, n);
  }
  if (n == 0) {
    _tprintf(_T("+++ Warning: adjusted byte count is 0 (nothing to do).\n"));
    goto done;
  }

  status = sampleOpenCardNoClearError(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  if (command == CommandTypeFill) {
    TCHAR* pArg;
    TCHAR ch;
#if defined(_UNICODE)
    wint_t c;
#else
    int c;
#endif
    size_t j = 0;
    int i = 4;

    if (i < arguments.nParam) {
      /* Read token for fill value from remaining command line arguments */
      pArg = arguments.param[4].pStringVal;
      do {
        ch = pArg[j];
        j++;
        if (sampleTokenPumpState(&state, ch, tokenFlags)) {
          if (state.result.length == 0) {
            goto done;
          }
          _tprintf(_T("Fill value: 0x%llX\n"), (unsigned long long)state.result.value);
          if (!fillMemoryUsingDma(width, bBigEndian, hCard, dmaChannel, localAddress, n, (uint64_t)state.result.value)) {
            ret = -1;
            goto done;
          }
          goto done;
        }
      } while (_T('\0') != ch);
    }
    /* If we didn't get a token yet, read token for fill value from stdin */
    _tprintf(_T("Provide fill value: "));
    do {
      c = _gettc(stdin);
      ch = (c == EOF_VALUE) ? _T('\0') : (TCHAR)c;
      if (sampleTokenPumpState(&state, ch, tokenFlags)) {
        if (state.result.length == 0) {
          goto done;
        }
        if (!fillMemoryUsingDma(width, bBigEndian, hCard, dmaChannel, localAddress, n, (uint64_t)state.result.value)) {
          ret = -1;
          goto done;
        }
        goto done;
      }
    } while (c != EOF_VALUE);
  } else if (command == CommandTypeWrite) {
    TCHAR* pArg;
    TCHAR ch;
#if defined(_UNICODE)
    wint_t c;
#else
    int c;
#endif
    size_t j = 0;
    int i = 4;

    while (i < arguments.nParam && n > 0) {
      /* Read token for next data item to write from remaining command line arguments */
      pArg = arguments.param[i].pStringVal;
      do {
        ch = pArg[j];
        j++;
        if (sampleTokenPumpState(&state, ch, tokenFlags)) {
          if (state.result.length == 0) {
            goto done;
          }
          _tprintf(_T("0x%08lX_%08lX: 0x%llX\n"),
            (unsigned long)(uint32_t)(localAddress >> 32), (unsigned long)(uint32_t)localAddress,
            state.result.value);
          if (!writeValueUsingDma(hCard, dmaChannel, width, bBigEndian, localAddress, (uint64_t)state.result.value)) {
            ret = -1;
            break;
          }
          n -= width;
          localAddress += width;
        }
      } while (_T('\0') != ch);
      i++;
    }
    if (n == 0) {
      goto done;
    }

    /* Read tokens for data items from stdin to satisfy remaining byte count (if nonzero) */
    _tprintf(_T("Provide data for 0x%08lX_%08lX: "),
      (unsigned long)(uint32_t)(localAddress >> 32), (unsigned long)(uint32_t)localAddress);
    fflush(stdout);
    do {
      c = _gettc(stdin);
      ch = (c == EOF_VALUE) ? _T('\0') : (TCHAR)c;
      if (sampleTokenPumpState(&state, ch, tokenFlags)) {
        if (state.result.length == 0) {
          goto done;
        }
        if (!writeValueUsingDma(hCard, dmaChannel, width, bBigEndian, localAddress, (uint64_t)state.result.value)) {
          ret = -1;
          break;
        }
        n -= width;
        localAddress += width;
        if (n > 0) {
          _tprintf(_T("Provide data for 0x%08lX_%08lX: "),
            (unsigned long)(uint32_t)(localAddress >> 32), (unsigned long)(uint32_t)localAddress);
          fflush(stdout);
        }
      }
    } while (n > 0 && c != EOF_VALUE);
  } else if (command == CommandTypeRead) {
    if (!dumpMemoryUsingDma(0, 16 / width, width, bBigEndian, TRUE, hCard, dmaChannel, localAddress, n)) {
      ret = -1;
    }
  } else {
    /* Should never get here */
    assert(FALSE);
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
