/*
** flash.c - utility to program, erase and verify the Flash memory on
**           an ADM-XRC series card.
**
** Warning: Although this utility performs checks in order to ensure
**          that you do not attempt to program the Flash memory with a
**          bitstream for the wrong type of FPGA, ensure that you do
**          not program the flash with an invalid bitstream. Doing so
**          and powering up the FPGA for an extended period of time
**          can damage devices on the card.
**
** Utility for:
**
**   o ADPe-XRC-5T
**   o ADM-XRC-6TL
**   o ADM-XRC-6T1
**   o ADM-XRC-6T-DA1
**   o ADM-XRC-6TGE & ADM-XRC-6TGEL
**   o ADM-XRC-6T-ADV8
**   o ADPE-XRC-6T & ADPE-XRC-6T-L
**   o ADPE-XRC-6T-ADV (Controller)
**   o ADM-XRC-7K1
**   o ADM-XRC-7V1
**   o ADM-VPX3-7V2
**   o ADM-PCIE-7V3
**   o ADM-XRC-KU1
**   o ADM-PCIE-KU3
**   o ADM-PCIE-8V3
**   o ADM-PCIE-8K5
**
** (C) Copyright Alpha Data 2009-2016
**
** NOTE: This utility applies only to the regions marked (*) in the comments
**       below, and cannot affect the other regions.
**
**  1. The Flash address map used on the ADM-XRC-6TL, ADM-XRC-6T1,
**     ADM-XRC-6T-DA1, ADM-XRC-6TGE, ADPE-XRC-6T(-L), ADM-XRC-7K1 &
**     ADM-XRC-7V1 is:
**
**     Flash bank 0:
**     0x0000000 - 0x07FFFFF  Alternate ADB3 PCI-E to OCP Bridge bitstream
**     0x0800000 - 0x0FFFFFF  Normal ADB3 PCI-E to OCP Bridge bitstream
**     0x1000000 - 0x10FFFFF  Alpha Data VPD region
**     0x1100000 - 0x11FFFFF  Customer VPD region
**     0x1200000 - 0x28FFFFF  Target FPGA bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA bitstream (*)
**
**  2. The Flash address map used on the ADM-XRC-6T-ADV8 is:
**
**     Flash bank 0:
**     0x0000000 - 0x1FFFFFF  Failsafe FPGA bitstream (*)
**     0x2000000 - 0x3FFFFFF  Normal FPGA bitstream (*)
**
**  3. The Flash address map used on the ADM-XRC-6T-ADV (Controller) is:
**
**     Flash bank 0 (U52):
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 0 bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 0 bitstream (*)
**
**     Flash bank 1 (U51):
**     0x0000000 - 0x0FFFFFF  Customer region
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 1 bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 1 bitstream (*)
**
**  4. The Flash address map used on the ADM-VPX3-7V2 is:
**
**     Flash bank 0 (U6) when size of bank is 64 MiB:
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 0 bitstream (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 0 bitstream (*)
**
**     Flash bank 0 (U6) when size of bank is 128 MiB:
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x48FFFFF  Target FPGA 0 bitstream (*)
**     0x4900000 - 0x7FFFFFF  Failsafe target FPGA 0 bitstream (*)
**
**     NOTE: The above address maps are logical, i.e. from the point of view of
**     somebody calling the Flash functions in the ADMXRC3 API. In order to
**     support both 64 MiB and 128 MiB devices without making regions
**     noncontiguous, the Alpha Data ADB3 Driver remaps the physical layout to
**     the above logical layout. The physical layout is as follows:
**
**     Flash bank 0 (U6):
**     0x0000000 - 0x07FFFFF  Alternate Controller FPGA bitstream
**     0x0800000 - 0x0FFFFFF  Normal Controller FPGA bitstream
**     0x1000000 - 0x10FFFFF  (reserved)
**     0x1100000 - 0x11FFFFF  Customer region
**     0x1200000 - 0x28FFFFF  Target FPGA 0 bitstream, lower 23 MiB (*)
**     0x2900000 - 0x3FFFFFF  Failsafe target FPGA 0 bitstream, lower 23 MiB (*)
**     (Next two regions present only in a 128 MiB bank)
**     0x4000000 - 0x5FFFFFF  Target FPGA 0 bitstream, upper 32 MiB (*)
**     0x6000000 - 0x7FFFFFF  Failsafe target FPGA 0 bitstream, upper 32 MiB (*)
**
**  5. The Flash address map used on the ADM-PCIE-7V3 & ADM-PCIE-KU3 is:
**
**     Flash bank 0:
**     0x0000000 - 0x1FFFFFF  Failsafe FPGA bitstream (*)
**     0x2000000 - 0x3FFFFFF  Normal FPGA bitstream (*)
**     0x4000000 - 0x5FFFFFF  User-definable (*)
**     0x6000000 - 0x7FFFFFF  User-definable (*)
**
**  6. The Flash address map used on the ADM-XRC-KU1 is:
**
**     Flash bank 0:
**     0x0000000 - 0x07FFFFF  Alternate ADB3 PCI-E to OCP/AXI4 Bridge bitstream
**     0x0800000 - 0x0FFFFFF  Normal ADB3 PCI-E to OCP/AXI4 Bridge bitstream
**     0x1000000 - 0x11FFFFF  (reserved)
**     0x1200000 - 0x4FFFFFF  Target FPGA bitstream (*)
**     0x5000000 - 0x7FFFFFF  Available for user data (*)
**
**  7. The Flash address map used on the ADM-PCIE-8V3 & ADM-PCIE-8K5 is:
**
**     Flash bank 0:
**     0x0000000 - 0x3FFFFFF  Failsafe FPGA bitstream (*)
**     0x4000000 - 0x7FFFFFF  Normal FPGA bitstream (*)
**
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# if defined(__VXWORKS__) || defined(__vxworks)
#   include <ctype.h>
#   define _tcsicmp(pStr1, pStr2) ourStrCaseCmp(pStr1, pStr2)
# else
#   define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# endif
# define _tcschr(x, c) strchr(x, c)
# define _tcslen(x) strlen(x)
# define _totlower(c) tolower(c)
# define _tmain main
# define _tprintf printf
# define _stscanf_s sscanf
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define BLANK_CHECK_BUFFER_SIZE (0x10000)
#define READBACK_BUFFER_SIZE (0x10000)
#define VERIFY_BUFFER_SIZE (0x10000)

const uint64_t sizeTMax = (uint64_t)(size_t)-1;

typedef enum _BootFlagType {
  /* ADM-XRC-6T-ADV8 */
  BootFlagTypeNone,          /* No boot flag is programmed into Flash */

  /* ADM-XRC-II */
  BootFlagType8Bit,          /* [7:0] = 0 */

  /*
  ** ADPe-XRC-5T, ADM-XRC-6TL, ADM-XRC-6T1, ADM-XRC-6T-DA1, ADM-XRC-6TGE,
  ** ADM-XRC-6T-ADV8, ADPE-XRC-6T(-L), ADPE-XRC-6T-ADV, ADM-XRC-7K1,
  ** ADM-XRC-7V1
  */
  BootFlagType24BitAddress,  /* [31:8] = length in bytes - 1, [7:0] = 0 */

  /* ADM-VPX3-7V2, ADM-XRC-KU1 */
  BootFlagType30BitAddress,  /* [31] = 0, [30] = (reserved), [29:0] = length in bytes - 1 */
} BootFlagType;

typedef enum _SwapFlags {
  SwapFlagsBitSwap   = 0x1 << 0,    /* Do bit-swapping inside bytes */
  SwapFlagsByteSwap2 = 0x1 << 1,    /* Do 2-byte swapping; mutually exclusive with SwapFlagsByteSwap4 */
  SwapFlagsByteSwap4 = 0x2 << 1,    /* Do 4-byte swapping; mutually exclusive with SwapFlagsByteSwap2 */
} SwapFlags;

static const SwapFlags SwapFlagsByteSwapMask = (SwapFlags)(0x3 << 1);

#if !defined(_MSC_VER)
/* Emulate _tfopen_s ("secure file open") from MSVC++ */
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = fopen(pFilename, pMode);
  if (NULL == pFile) {
    return -1;
  }
  *ppFile = pFile;
  return 0;
}
#endif

#if defined(__VXWORKS__) || defined(__vxworks)
static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}
#endif

/*
** Returns TRUE if device identifier in bitstream matches device identifier
** in target FPGA information, otherwise FALSE.
**
** To handle the fact that identifiers from Ultrascale bitstreams, unlike
** previous FPGA families, have the speed and temperature grade as suffices,
** consider the identifiers to match iff the identifier from target FPGA
** information is a prefix of the identifier from the bitstream.
*/
static bool_t
checkDeviceMatch(
  const TCHAR* pBitstreamIdentifier,
  const TCHAR* pFPGAIdentifier)
{
  size_t n = _tcslen(pBitstreamIdentifier);
  size_t m = _tcslen(pFPGAIdentifier);
  size_t i;

  if (n < m) {
    /* If FPGA identifier is shorter than bitstream identifier, one or the other must be truncated or malformed, so consider not matched. */
    return FALSE;
  }

  for (i = 0; i < m; i++) {
    TCHAR c1 = _totlower(pBitstreamIdentifier[i]);
    TCHAR c2 = _totlower(pFPGAIdentifier[i]);
    if (c1 != c2) {
      return FALSE;
    }
  }

  return TRUE;
}

/*
** Establish defaults, according to model, for:
**
**  o The default region used for erase/program/verify/chkblank commands.
**  o The failsafe region (for +failsafe option).
**  o The number of regions.
**  o The type of boot flag, if any.
**  o The length of the boot flag, if any.
**  o Any bit- or byte-swapping that needs to be performed on the SelectMap data.
*/
static void
getModelDefaults(
  ADMXRC3_MODEL_TYPE model,
  unsigned int* pDefaultRegion,
  unsigned int* pFailsafeRegion,
  unsigned int* pNumRegion,
  BootFlagType* pBootFlagType,
  unsigned int* pBootFlagLength,
  SwapFlags* pSwapFlags)
{
  unsigned int defaultRegion = 0U;
  unsigned int failsafeRegion = 0U;
  unsigned int numRegion = 0U;
  BootFlagType bootFlagType = 0;
  unsigned int bootFlagLength = 0;
  SwapFlags swapFlags = 0;

  switch (model) {
  case ADMXRC3_MODEL_ADMXRC2:
    defaultRegion = 0;
    numRegion = 1;
    bootFlagType = BootFlagType8Bit;
    bootFlagLength = 1;
    swapFlags = 0;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
    defaultRegion = 0;
    failsafeRegion = 1;
    numRegion = 2;
    bootFlagType = BootFlagType24BitAddress;
    bootFlagLength = 4;
    swapFlags = 0;
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    defaultRegion = 1;
    failsafeRegion = 0;
    numRegion = 2;
    bootFlagType = BootFlagTypeNone;
    bootFlagLength = 0;
    swapFlags = SwapFlagsBitSwap | SwapFlagsByteSwap2;
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    defaultRegion = 0;
    failsafeRegion = 1;
    numRegion = 2;
    bootFlagType = BootFlagType24BitAddress;
    bootFlagLength = 4;
    swapFlags = 0;
    break;

  case ADMXRC3_MODEL_ADMVPX37V2:
    defaultRegion = 0;
    failsafeRegion = 1;
    numRegion = 2;
    bootFlagType = BootFlagType30BitAddress;
    bootFlagLength = 4;
    swapFlags = 0;
    break;

  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
    defaultRegion = 1;
    failsafeRegion = 0;
    numRegion = 4;
    bootFlagType = BootFlagTypeNone;
    bootFlagLength = 0;
    swapFlags = SwapFlagsBitSwap | SwapFlagsByteSwap2;
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
    defaultRegion = 0;
    numRegion = 1;
    bootFlagType = BootFlagType30BitAddress;
    bootFlagLength = 4;
    swapFlags = 0;
    break;

  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    defaultRegion = 1;
    failsafeRegion = 0;
    numRegion = 2;
    bootFlagType = BootFlagTypeNone;
    bootFlagLength = 0;
    swapFlags = SwapFlagsBitSwap | SwapFlagsByteSwap2;
    break;

  default:
    assert(FALSE);
    break;
  }

  *pDefaultRegion = defaultRegion;
  *pFailsafeRegion = failsafeRegion;
  *pNumRegion = numRegion;
  *pBootFlagType = bootFlagType;
  *pBootFlagLength = bootFlagLength;
  *pSwapFlags = swapFlags;
}

/*
** Figure out the range of Flash addresses for a particular region.
** The address range depends upon the region and the model.
*/
static void
getEraseAddressRange(
  ADMXRC3_MODEL_TYPE model,
  const ADMXRC3_FLASH_INFO* pFlashInfo,
  unsigned int region,
  uint64_t* pEraseAddress,
  uint64_t* pEraseLimit)
{
  uint64_t eraseAddress = 0U;
  uint64_t eraseLimit = 0U;

  switch (model) {
  case ADMXRC3_MODEL_ADMXRC2:
    switch (region) {
    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + pFlashInfo->UseableLength;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    switch (region) {
    case 1:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0xC00000U;
      eraseLimit = pFlashInfo->UseableStart + 0x1800000U;
      break;

    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + 0xC00000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
    switch (region) {
    case 1:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x1700000U;
      eraseLimit = pFlashInfo->UseableStart + 0x2E00000U;
      break;

    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + 0x1700000U;
      break;
    
    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    switch (region) {
    case 1:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart + 0x2000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x4000000U;
      break;

    case 0:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x0U;
      eraseLimit = pFlashInfo->UseableStart + 0x2000000U;
      break;
    
    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    switch (region) {
    case 1:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x1700000U;
      eraseLimit = pFlashInfo->UseableStart + 0x2E00000U;
      break;

    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + 0x1700000U;
      break;
    
    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMVPX37V2:
    /* The driver remaps the physical Flash layout so that the normal and
    ** failsafe regions are exactly equal in size and contiguous, regardless
    ** of the size of Flash device fitted. */
    switch (region) {
    case 1:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + (pFlashInfo->UseableLength / 2);
      eraseLimit = eraseAddress + (pFlashInfo->UseableLength / 2);
      break;

    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = eraseAddress + (pFlashInfo->UseableLength / 2);
      break;
    
    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
    switch (region) {
    case 3:
      /* User-definable */
      eraseAddress = pFlashInfo->UseableStart + 0x6000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x8000000U;
      break;

    case 2:
      /* User-definable */
      eraseAddress = pFlashInfo->UseableStart + 0x4000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x6000000U;
      break;

    case 1:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart + 0x2000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x4000000U;
      break;

    case 0:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x0U;
      eraseLimit = pFlashInfo->UseableStart + 0x2000000U;
      break;
    
    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    switch (region) {
    case 1:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart + 0x4000000U;
      eraseLimit = pFlashInfo->UseableStart + 0x8000000U;
      break;

    case 0:
      /* Failsafe */
      eraseAddress = pFlashInfo->UseableStart + 0x0U;
      eraseLimit = pFlashInfo->UseableStart + 0x4000000U;
      break;
    
    default:
      assert(FALSE);
      break;
    }
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
    switch (region) {
    case 0:
      /* Default */
      eraseAddress = pFlashInfo->UseableStart;
      eraseLimit = pFlashInfo->UseableStart + 0x3E00000U;
      break;

    default:
      assert(FALSE);
      break;
    }
    break;

  default:
    assert(FALSE);
    break;
  }

  *pEraseAddress = eraseAddress;
  *pEraseLimit = eraseLimit;
}

/*
** Figure out the address at which to program the SelectMap data, and the
** address of the boot flag (if there is one). This is calculated as an
** offset, which depends upon the model, from the starting erase address.
*/
static void
getProgramAddress(
  ADMXRC3_MODEL_TYPE model,
  uint64_t eraseAddress,
  uint64_t* pBootFlagAddress,
  uint64_t* pProgramAddress)
{
  uint64_t bootFlagAddress = 0U;
  uint64_t programAddress = 0U;

  switch (model) {
  case ADMXRC3_MODEL_ADMXRC2:
    bootFlagAddress = eraseAddress + 0x8000U;
    programAddress = eraseAddress + 0x8001U;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC7V1:
  case ADMXRC3_MODEL_ADMVPX37V2:
  case ADMXRC3_MODEL_ADMXRCKU1:
    bootFlagAddress = eraseAddress;
    programAddress = eraseAddress + 0x20000U;
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    programAddress = eraseAddress;
    break;

  default:
    assert(FALSE);
    break;
  }

  *pBootFlagAddress = bootFlagAddress;
  *pProgramAddress = programAddress;
}

static ADMXRC3_STATUS
eraseFlash(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  uint64_t eraseStart,
  uint64_t eraseLimit)
{
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint64_t address, remaining;

  address = eraseStart;
  remaining = eraseLimit - eraseStart;
  while (address < eraseLimit) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"), (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    /*
    ** Adjust blockInfo address & length so that it does not erase anything
    ** outside of the address range [eraseStart, eraseLimit).
    */
    if (blockInfo.Address < address) {
      uint64_t difference = address - blockInfo.Address;
      blockInfo.Address = address;
      blockInfo.Length -= difference;
    }
    if (blockInfo.Length > remaining) {
      blockInfo.Length = remaining;
    }

    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_EraseFlash(hCard, flashIndex, 0, blockInfo.Address, blockInfo.Length);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to erase Flash at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)blockInfo.Address, (unsigned long long)(blockInfo.Address + blockInfo.Length - 1U), ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    /* Advance to next block */
    address = blockInfo.Address + blockInfo.Length;
    remaining = remaining - blockInfo.Length;
    if (address <= eraseStart) {
      /* Catch 'address' wrapping around to 0 */
      break;
    }
  }

  return status;
}

static ADMXRC3_STATUS
programBitstream(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_BITSTREAM* pBitstream,
  unsigned int flashIndex,
  uint64_t programStart)
{
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint8_t* pData;
  uint64_t address, chunk;
  size_t remaining;
  
  address = programStart;
  remaining = pBitstream->Length;
  pData = &pBitstream->Data[0];
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_WriteFlash(hCard, flashIndex, 0, address, (size_t)chunk, pData); /* Cast is safe because we guarantee chunk <= maximum size_t value */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to write Flash at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      break;
    }

    address += chunk;
    remaining -= (size_t)chunk; /* Cast is safe because we guarantee chunk <= maximum size_t value */
    pData += (size_t)chunk; /* Cast is safe because we guarantee chunk <= maximum size_t value */
  }

  return status;
}

static ADMXRC3_STATUS
checkBlank(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  uint64_t checkStart,
  uint64_t checkLimit,
  uint64_t* pnError)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, remaining, nError, i, chunk;

  pBuffer = (uint8_t*)malloc(BLANK_CHECK_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate blank check buffer of size 0x%lx bytes\n"),
      (unsigned long)BLANK_CHECK_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  nError = 0;
  address = checkStart;
  remaining = checkLimit - checkStart;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"), (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > BLANK_CHECK_BUFFER_SIZE) {
      chunk = BLANK_CHECK_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, flashIndex, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    for (i = 0; i < chunk; i++) {
      if (pBuffer[i] != 0xffU) {
        if (nError < 10) {
          _tprintf(_T("*** Error at Flash address 0x%llX: expected=0x%02X, actual=0x%02X\n"),
            (unsigned long long)(address + i), 0xffU, pBuffer[i]);
        } else if (nError == 10) {
          _tprintf(_T("*** (more errors)\n"));
        }
        nError++;
      }
    }

    address += chunk;
    remaining -= chunk;
  }

  free(pBuffer);

  *pnError = nError;
  return status;
}

static ADMXRC3_STATUS
verify(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_BITSTREAM* pBitstream,
  unsigned int flashIndex,
  uint64_t verifyStart,
  uint64_t* pnError)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, chunk, nError;
  size_t i, offset, remaining;

  pBuffer = (uint8_t*)malloc(VERIFY_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate verify buffer of size 0x%lx bytes\n"),
      (unsigned long)VERIFY_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  nError = 0;
  offset = 0;
  address = verifyStart;
  remaining = pBitstream->Length;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > VERIFY_BUFFER_SIZE) {
      chunk = VERIFY_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, flashIndex, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    for (i = 0; i < chunk; i++) {
      if (pBuffer[i] != pBitstream->Data[offset + i]) {
        if (nError < 10) {
          _tprintf(_T("*** Error at Flash address 0x%llX: expected=0x%02X, actual=0x%02X\n"),
            (unsigned long long)(address + i), pBitstream->Data[offset + i], pBuffer[i]);
        } else if (nError == 10) {
          _tprintf(_T("*** (more errors)\n"));
        }
        nError++;
      }
    }

    address += chunk;
    offset += (size_t)chunk; /* Cast is safe because we guarantee chunk < size of buffer */
    remaining -= (size_t)chunk; /* Cast is safe because we guarantee chunk < size of buffer */
  }

  free(pBuffer);

  *pnError = nError;
  return status;
}

ADMXRC3_STATUS
readback(
  ADMXRC3_HANDLE hCard,
  unsigned int flashIndex,
  uint64_t readAddress,
  uint64_t readLength,
  FILE* pFile)
{
  uint8_t* pBuffer;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, remaining, chunk;
  size_t n;

  pBuffer = (uint8_t*)malloc(READBACK_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate verify buffer of size 0x%lX bytes\n"),
      (unsigned long)READBACK_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  address = readAddress;
  remaining = readLength;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, flashIndex, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llX: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > READBACK_BUFFER_SIZE) {
      chunk = READBACK_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llX\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, flashIndex, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llX-0x%llX: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    n = fwrite(pBuffer, 1, (size_t)chunk, pFile); /* Cast is safe because we guarantee chunk < size of buffer */
    if (n != chunk) {
      _tprintf(_T("*** Failed to write all data to file, %lu(0x%lX) actual vs. %lu(0x%lX) expected.\n"),
        (unsigned long)n, (unsigned long)n, (unsigned long)chunk, (unsigned long)chunk);
      free(pBuffer);
      return ADMXRC3_UNEXPECTED_ERROR;
    }

    address += chunk;
    remaining -= chunk;
  }

  free(pBuffer);

  return status;
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  int ret = 0;
  const double clockFreq = 33.0e6;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO	cardInfo;
  ADMXRC3_FLASH_INFO flashInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_BITSTREAM* pBitstream = NULL;
  TCHAR* pCommand;
  Arguments arguments = g_initArguments;
  Option options[6];
  OptionType paramType[5];
  TCHAR* paramInfo[5];
  TCHAR* paramMeaning[5];
  unsigned int index, serial, targetIndex, flashIndex = (unsigned int)-1;
  bool_t bShowInfo = FALSE, bDoCheckBlank = FALSE, bDoErase = FALSE, bDoProgram = FALSE, bDoReadback = FALSE, bDoVerify = FALSE;
  uint64_t bootFlagAddress = 0;
  uint64_t programAddress = 0, programLimit = 0;
  uint64_t eraseAddress = 0, eraseLimit  = 0;
  uint64_t nError = 0;
  bool_t bFpgaInfoValid = FALSE;
  BootFlagType bootFlagType = BootFlagTypeNone;
  unsigned int bootFlagLength = 0U;
  bool_t bBootFlagVerify = TRUE;
  bool_t bBootFlagOk = FALSE;
  bool_t bFailsafeSpecified, bFailsafeRegion;
  bool_t bForce;
  bool_t bRegionSpecified, bRangeSpecified;
  TCHAR* pOverrideRange;
  uint64_t overrideAddress = 0, overrideLimit = 0;
  unsigned int numRegion, defaultRegion, failsafeRegion, region;
  SwapFlags swapFlags;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("failsafe");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("program failsafe region");
  options[3].key = _T("force");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
  options[3].meaning = _T("ignore device mismatch");
  options[4].key = _T("region");
  options[4].type = OptionInt;
  options[4].def.intVal = -1;
  options[4].meaning = _T("Flash region to use for chkblank/erase/program/verify");
  options[5].key = _T("range");
  options[5].type = OptionString;
  options[5].def.pStringVal = _T("");
  options[5].meaning = _T("Override address range to use for chkblank/erase/program/verify\n                                     Expects argument in the form of <address>,<length>");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.bNoStdOptions = TRUE;

  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("One of erase, info, program, readback, verify or chkblank");
  paramType[1] = OptionUInt;
  paramInfo[1] = _T("target");
  paramMeaning[1] = _T("Index of target FPGA");
  paramType[2] = OptionPath;
  paramInfo[2] = _T("file");
  paramMeaning[2] = _T("Bitstream filename (program or verify commands only)");
  paramType[3] = OptionUInt64;
  paramInfo[3] = _T("offset");
  paramMeaning[3] = _T("Offset for readback (readback command only)");
  paramType[4] = OptionUInt64;
  paramInfo[4] = _T("length");
  paramMeaning[4] = _T("Length for readback (readback command only)");
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = 1;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  if (sampleParseCommandLine(_T("flash"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }

  index = (unsigned int)options[0].value.uintVal;
  serial = (unsigned int)options[1].value.uintVal;
  bFailsafeRegion = options[2].value.booleanVal;
  bFailsafeSpecified = options[2].specified;
  bForce = options[3].value.booleanVal;
  region = options[4].value.intVal;
  bRegionSpecified = options[4].specified;
  bRangeSpecified = options[5].specified;
  pOverrideRange = options[5].value.pStringVal;
  if (bRangeSpecified) {
    TCHAR* p;
    unsigned long long start, length;
    int n, m;
    
    p = _tcschr(pOverrideRange, _T(','));
    if (NULL == p) {
      _tprintf(_T("*** Malformed argument for '-range' option.\n"));
      ret = -1;
      goto done;
    }
    n = _stscanf_s(pOverrideRange, _T("%llx"), &start);
    m = _stscanf_s(p + 1, _T("%llx"), &length);
    if (n != 1 || m != 1) {
      _tprintf(_T("*** Malformed argument for '-range' option.\n"));
      ret = -1;
      goto done;
    }
    if (length == 0) {
      _tprintf(_T("*** Length part of argument for '-range' option cannot be 0.\n"));
      ret = -1;
      goto done;
    }
    overrideAddress = (uint64_t)start;
    overrideLimit = overrideAddress + (uint64_t)length;
    if (overrideLimit != 0 && overrideLimit < overrideAddress) {
      _tprintf(_T("*** Argument for '-range' option wraps to 0.\n"));
      ret = -1;
      goto done;
    }
  }
  pCommand = arguments.param[0].pStringVal;
  targetIndex = (unsigned int)arguments.param[1].intVal;

  if (bFailsafeRegion && bRegionSpecified) {
    _tprintf(_T("*** +failsafe option is mutually exclusive with -region option.\n"));
    ret = -1;
    goto done;
  }

  if (bFailsafeRegion && bRangeSpecified) {
    _tprintf(_T("*** +failsafe option is mutually exclusive with -range option.\n"));
    ret = -1;
    goto done;
  }

  if (bRegionSpecified && bRangeSpecified) {
    _tprintf(_T("*** -region option is mutually exclusive with -range option.\n"));
    ret = -1;
    goto done;
  }

  bShowInfo = _tcsicmp(pCommand, _T("info")) == 0;
  bDoProgram = _tcsicmp(pCommand, _T("program")) == 0;
  bDoVerify = _tcsicmp(pCommand, _T("verify")) == 0;
  bDoErase = _tcsicmp(pCommand, _T("erase")) == 0;
  bDoCheckBlank = _tcsicmp(pCommand, _T("chkblank")) == 0;
  bDoReadback = _tcsicmp(pCommand, _T("readback")) == 0;

  if (!bShowInfo && !bDoProgram && !bDoVerify && !bDoErase && !bDoCheckBlank && !bDoReadback) {
    _tprintf(_T("*** Please specify a valid command. Use flash /? to get help.\n"));
    ret = -1;
    goto done;
  }

  if (bDoProgram || bDoVerify) {
    if (arguments.nParam < 3) {
      _tprintf(_T("*** program or verify command requires (1) target FPGA index and (2) .BIT filename.\n"));
      ret = -1;
      goto done;
    }
  }

  if (bDoCheckBlank || bDoErase || bShowInfo) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** chkblank, erase or info command requires target FPGA index.\n"));
      ret = -1;
      goto done;
    }
  }

  if (bDoReadback) {
    if (arguments.nParam < 4) {
      _tprintf(_T("*** readback command requires (1) target FPGA index, (2) byte offset and (3) byte length.\n"));
      ret = -1;
      goto done;
    }
  }

  if (bFailsafeSpecified && (bShowInfo || bDoReadback)) {
    _tprintf(_T("+++ WARNING: +failsafe & -failsafe options have no effect on 'info' & 'readback' commands.\n"));
  }

  if (bRegionSpecified && (bShowInfo || bDoReadback)) {
    _tprintf(_T("+++ WARNING: -region option has no effect on 'info' & 'readback' commands.\n"));
  }

  status = sampleOpenCardNoClearError(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRC2:
    if (bFailsafeRegion) {
      _tprintf(_T("*** +failsafe option is not supported for this model.\n"));
      ret = -1;
      goto done;
    }
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = -1;
      goto done;
    }
    flashIndex = 0U;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
  case ADMXRC3_MODEL_ADMXRC6TADV8:
  case ADMXRC3_MODEL_ADPEXRC6T:
  case ADMXRC3_MODEL_ADPEXRC6TL:
  case ADMXRC3_MODEL_ADMXRC7K1:
  case ADMXRC3_MODEL_ADMXRC6TDA1:
  case ADMXRC3_MODEL_ADMXRC7V1:
  case ADMXRC3_MODEL_ADMVPX37V2:
  case ADMXRC3_MODEL_ADMPCIE7V3:
  case ADMXRC3_MODEL_ADMPCIEKU3:
  case ADMXRC3_MODEL_ADMPCIE8V3:
  case ADMXRC3_MODEL_ADMPCIE8K5:
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = -1;
      goto done;
    }
    flashIndex = 0U;
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
    if (bFailsafeRegion) {
      _tprintf(_T("*** +failsafe option is not supported for this model.\n"));
      ret = -1;
      goto done;
    }
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = -1;
      goto done;
    }
    flashIndex = 0U;
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    if (targetIndex >= 2) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = -1;
      goto done;
    }
    flashIndex = targetIndex;
    break;

  default:
    _tprintf(_T("*** This utility must be run on one of the following cards:\n\n"));
    _tprintf(_T("\tADM-XRC-II\n"));
    _tprintf(_T("\tADPe-XRC-5T\n"));
    _tprintf(_T("\tADM-XRC-6TL\n"));
    _tprintf(_T("\tADM-XRC-6T1\n"));
    _tprintf(_T("\tADM-XRC-6T-DA1\n"));
    _tprintf(_T("\tADM-XRC-6TGE / ADM-XRC-6TGEL\n"));
    _tprintf(_T("\tADM-XRC-6T-ADV8\n"));
    _tprintf(_T("\tADPE-XRC-6T / ADPE-XRC-6T-L\n"));
    _tprintf(_T("\tADPE-XRC-6T-ADV (Controller)\n"));
    _tprintf(_T("\tADM-XRC-7K1\n"));
    _tprintf(_T("\tADM-XRC-7V1\n"));
    _tprintf(_T("\tADM-VPX3-7V2\n"));
    _tprintf(_T("\tADM-PCIE-7V3\n"));
    _tprintf(_T("\tADM-XRC-KU1\n"));
    _tprintf(_T("\tADM-PCIE-KU3\n"));
    _tprintf(_T("\tADM-PCIE-8V3\n"));
    _tprintf(_T("\tADM-PCIE-8K5\n"));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetFpgaInfo(hCard, targetIndex, &fpgaInfo);
  if (status != ADMXRC3_SUCCESS) {
    if (!bForce) {
      _tprintf(_T("*** Failed to get FPGA %lu info: %s\n"), (unsigned long)targetIndex, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    } else {
      _tprintf(_T("+++ WARNING: Failed to get FPGA %lu info: %s\n"), (unsigned long)targetIndex, ADMXRC3_GetStatusString(status, TRUE));
          _tprintf(_T("         Overriding due to +force option specified on command line.\n"));
    }
  } else {
    _tprintf(_T("Target FPGA is device %s\n"), fpgaInfo.Identifier);
    bFpgaInfoValid = TRUE;
  }

  status = ADMXRC3_GetFlashInfo(hCard, flashIndex, &flashInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get Flash information: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  getModelDefaults(cardInfo.Model, &defaultRegion, &failsafeRegion, &numRegion, &bootFlagType, &bootFlagLength, &swapFlags);
  if (!bRegionSpecified) {
    region = defaultRegion;
  }
  if (bFailsafeRegion) {
    region = failsafeRegion;
  }
  if (region >= numRegion) {
    _tprintf(_T("*** %u is not a valid Flash region for this model.\n    Region must be between 0 and %u (inclusive).\n"),
      region, numRegion - 1);
    ret = -1;
    goto done;
  }
  getEraseAddressRange(cardInfo.Model, &flashInfo, region, &eraseAddress, &eraseLimit);
  if (bRangeSpecified) {
    eraseAddress = overrideAddress;
    eraseLimit = overrideLimit;
  }
  getProgramAddress(cardInfo.Model, eraseAddress, &bootFlagAddress, &programAddress);

  if (eraseLimit > flashInfo.UseableStart + flashInfo.UseableLength ||
      eraseAddress < flashInfo.UseableStart)
  {
    _tprintf(_T("*** Selected range 0x%08llX - 0x%08llX overruns Flash useable area.\n"),
      (unsigned long long)eraseAddress, (unsigned long long)eraseLimit);
    ret = -1;
    goto done;
  }

  _tprintf(_T("Flash type is %s, %llu(0x%llX) kiB\n"), flashInfo.Identifier,
    (unsigned long long)(flashInfo.Size / 1024), (unsigned long long)(flashInfo.Size / 1024));
  _tprintf(_T("Useable region is 0x%llX-0x%llX\n"),
    (unsigned long long)flashInfo.UseableStart, (unsigned long long)(flashInfo.UseableStart + flashInfo.UseableLength - 1U));

  if (bShowInfo) {
    goto done;
  }

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    status = ADMXRC3_SetClockFrequency(hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1fMHz: %s\n"), clockFreq / 1.0e6, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  if (bDoProgram || bDoVerify) {
    uint32_t i;
    uint8_t val8;
    uint8_t val8a, val8b, val8c, val8d;

    if (bootFlagType != BootFlagTypeNone) {
      if (bootFlagAddress != 0 && bootFlagAddress + bootFlagLength < bootFlagAddress) {
        _tprintf(_T("*** Boot flag @ 0x%llX - 0x%llX wraps to address 0.\n"),
          (unsigned long long)bootFlagAddress, (unsigned long long)(bootFlagAddress + bootFlagLength - 1));
        ret = -1;
        goto done;
      }
      if (bootFlagAddress + bootFlagLength > eraseLimit) {
        _tprintf(_T("*** Boot flag @ 0x%llX - 0x%llX overruns range 0x%llX - 0x%llX.\n"),
          (unsigned long long)bootFlagAddress, (unsigned long long)(bootFlagAddress + bootFlagLength - 1),
          (unsigned long long)eraseAddress, (unsigned long long)(eraseLimit - 1));
        ret = -1;
        goto done;
      }
    }

    /* Load the .BIT file specified on the command line */
    status = ADMXRC3_LoadBitstream(arguments.param[2].pPathVal, &pBitstream);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to load bitstream '%s': %s\n"),
        arguments.param[2].pPathVal, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    _tprintf(_T("Bitstream is for device %s, length %lu(0x%lX) bytes\n"),
      pBitstream->Identifier, (unsigned long)pBitstream->Length, (unsigned long)pBitstream->Length);

    programLimit = programAddress + pBitstream->Length;
    if (programLimit != 0 && programLimit < programAddress) {
      _tprintf(_T("*** Bitstream @ 0x%llX - 0x%llX wraps to address 0.\n"),
        (unsigned long long)programAddress, (unsigned long long)(programLimit - 1));
      ret = -1;
      goto done;
    }
    if (programLimit > eraseLimit) {
      _tprintf(_T("*** Bitstream @ 0x%llX - 0x%llX overruns range 0x%llX - 0x%llX.\n"),
          (unsigned long long)programAddress, (unsigned long long)(programLimit - 1),
          (unsigned long long)eraseAddress, (unsigned long long)(eraseLimit - 1));
      ret = -1;
      goto done;
    }

    if (bFpgaInfoValid) {
      if (!checkDeviceMatch(pBitstream->Identifier, fpgaInfo.Identifier)) {
        if (!bForce) {
          _tprintf(_T("*** Bitstream device does not match target FPGA device - aborting\n"));
          ret = -1;
          goto done;
        } else {
          _tprintf(_T("+++ WARNING: Device information in bitstream does not match target FPGA device.\n"));
          if (bDoProgram) {
            _tprintf(_T("             Overriding due to +force option specified on command line.\n"));
          }
        }
      }
    } else {
      assert(bForce);
      _tprintf(_T("+++ WARNING: Information for target FPGA was not found. Bitstream may or may not match actual target FPGA device.\n"));
      _tprintf(_T("             Overriding due to +force option specified on command line.\n"));
    }
    if (swapFlags & SwapFlagsBitSwap) {
      _tprintf(_T("Doing bit-swapping...\n"));
      for (i = 0; i < pBitstream->Length; i++) {
        val8 = pBitstream->Data[i];
        pBitstream->Data[i] = ((val8 & 0x80U) >> 7) | ((val8 & 0x40U) >> 5) | ((val8 & 0x20U) >> 3) | ((val8 & 0x10U) >> 1) |
                              ((val8 & 0x08U) << 1) | ((val8 & 0x04U) << 3) | ((val8 & 0x02U) << 5) | ((val8 & 0x01U) << 7);
      }
    }
    switch (swapFlags & SwapFlagsByteSwapMask) {
    case 2U:
      _tprintf(_T("Doing 16-bit endian-swapping...\n"));
      for (i = 0; i < pBitstream->Length; i += 2) {
        val8a = pBitstream->Data[i + 0];
        val8b = pBitstream->Data[i + 1];
        pBitstream->Data[i + 0] = val8b;
        pBitstream->Data[i + 1] = val8a;
      }
      break;

    case 4U:
      _tprintf(_T("Doing 32-bit endian-swapping...\n"));
      for (i = 0; i < pBitstream->Length; i += 2) {
        val8a = pBitstream->Data[i + 0];
        val8b = pBitstream->Data[i + 1];
        val8c = pBitstream->Data[i + 2];
        val8d = pBitstream->Data[i + 3];
        pBitstream->Data[i + 0] = val8d;
        pBitstream->Data[i + 1] = val8c;
        pBitstream->Data[i + 2] = val8b;
        pBitstream->Data[i + 3] = val8a;
      }
      break;
    }
  }

  if (bDoErase || bDoProgram) {
    /* For the 'program' or 'erase' commands, a region of the Flash is first erased  */
    _tprintf(_T("Erasing Flash from 0x%llX-0x%llX...\n"), (unsigned long long)eraseAddress, (unsigned long long)(eraseLimit - 1U));
    status = eraseFlash(hCard, flashIndex, eraseAddress, eraseLimit);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Erase failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  if (bDoProgram) {
    /* For the 'program' commands, the bitstream is written into a region of the Flash  */
    _tprintf(_T("Writing bitstream at 0x%llX-0x%llX...\n"),
      (unsigned long long)programAddress, (unsigned long long)(programAddress + pBitstream->Length - 1U));
    status = programBitstream(hCard, pBitstream, flashIndex, programAddress);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Programming failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  if (bDoErase || bDoCheckBlank) {
    /* For the 'erase' commands, a blank check is performed for a region of the Flash */
    _tprintf(_T("Blank checking from 0x%llX-0x%llX...\n"),
      (unsigned long long)eraseAddress, (unsigned long long)(eraseLimit - 1U));
    status = checkBlank(hCard, flashIndex, eraseAddress, eraseLimit, &nError);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Blank check failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    if (nError) {
      _tprintf(_T("*** Blank check failed with %llu(0x%llX) error(s)\n"),
        (unsigned long long)nError, (unsigned long long)nError);
      ret = -1;
      goto done;
    } else {
      _tprintf(_T("Blank check successful.\n"));
    }
  }

  if (bDoProgram || bDoVerify) {
    /* For the 'program' and 'verify' commands, the bitstream data is compared against the contents of a region of the Flash */
    _tprintf(_T("Verifying from 0x%llX-0x%llX...\n"),
      (unsigned long long)programAddress, (unsigned long long)(programAddress + pBitstream->Length - 1U));
    status = verify(hCard, pBitstream, flashIndex, programAddress, &nError);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Verification check failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    if (nError) {
      _tprintf(_T("*** Verification failed with %llu(0x%llX) error(s)\n"),
        (unsigned long long)nError, (unsigned long long)nError);
      ret = -1;
      goto done;
    } else {
      _tprintf(_T("Verification successful.\n"));
    }
  }

  if (bDoProgram) {
    uint32_t val32;
    uint8_t val8;

    /*
    ** For the 'program' command, the flag must be set in the Flash to cause the bitstream to be loaded
    ** at reset / power up.
    */
    switch (bootFlagType) {
    case BootFlagType24BitAddress:
    case BootFlagType30BitAddress:
      /*
      ** Set the flag that causes the FPGA to be loaded from flash,
      ** and bitstream length.
      */
      if (bootFlagType == BootFlagType24BitAddress) {
        val32 = (uint32_t)((pBitstream->Length - 1U) << 8);
      } else {
        val32 = (uint32_t)((pBitstream->Length - 1U) & 0x3FFFFFFFU);
      }
      _tprintf(_T("Setting boot flag at 0x%llX-0x%llX to 0x%08lX...\n"),
        (unsigned long long)bootFlagAddress, (unsigned long long)(bootFlagAddress + sizeof(val32) - 1U), (unsigned long)val32);
      val32 = cpu_to_le32(val32);
      status = ADMXRC3_WriteFlash(hCard, flashIndex, 0, bootFlagAddress, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("*** Failed to set boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        bBootFlagVerify = FALSE;
        ret = -1;
      }
      break;

    case BootFlagType8Bit:
      /* Set the flag that causes the FPGA to be loaded from flash */
      val8 = 0x0U;
      _tprintf(_T("Setting boot flag at 0x%llX to 0x%02X...\n"), (unsigned long long)bootFlagAddress, (unsigned int)val8);
      status = ADMXRC3_WriteFlash(hCard, flashIndex, 0, bootFlagAddress, sizeof(val8), &val8);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("*** Failed to set boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        bBootFlagVerify = FALSE;
        ret = -1;
      }
      break;

    case BootFlagTypeNone:
      break;

    default:
      assert(FALSE);
      break;
    }
  }

  if (bDoProgram || bDoVerify) {
    if (bBootFlagVerify) {
      uint32_t expected32, bootFlag32;
      uint8_t expected8, bootFlag8;

      /*
      ** For the 'program' and 'verify' commands, the flag in the Flash that causes the bitstream to be loaded
      ** at reset / power up must be verified.
      */
      switch (bootFlagType) {
      case BootFlagType24BitAddress:
      case BootFlagType30BitAddress:
        /*
        ** Set the flag that causes the FPGA to be loaded from flash,
        ** and bitstream length.
        */
        if (bootFlagType == BootFlagType24BitAddress) {
          expected32 = (uint32_t)((pBitstream->Length - 1U) << 8);
        } else {
          expected32 = (uint32_t)((pBitstream->Length - 1U) & 0x3FFFFFFFU);
        }
        status = ADMXRC3_ReadFlash(hCard, flashIndex, 0, bootFlagAddress, sizeof(bootFlag32), &bootFlag32);
        if (status != ADMXRC3_SUCCESS) {
          _tprintf(_T("*** Failed to read boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        } else {
          bootFlag32 = le32_to_cpu(bootFlag32);
          if (bootFlag32 != expected32) {
            _tprintf(_T("*** Boot flag verification failed; expecting 0x%08lX, found 0x%08lX\n"),
              (unsigned long)expected32, (unsigned long)bootFlag32);
          } else {
            bBootFlagOk = TRUE;
          }
        }
        break;

      case BootFlagType8Bit:
        expected8 = 0x0U;
        status = ADMXRC3_ReadFlash(hCard, flashIndex, 0, bootFlagAddress, sizeof(bootFlag8), &bootFlag8);
        if (status != ADMXRC3_SUCCESS) {
          _tprintf(_T("*** Failed to read back boot flag: %s\n"),
            ADMXRC3_GetStatusString(status, TRUE));
        } else {
          if (bootFlag8 != expected8) {
            _tprintf(_T("*** Boot flag verification failed; expecting 0x%02X, found 0x%02X\n"),
              (unsigned int)expected8, (unsigned int)bootFlag8);
          } else {
            bBootFlagOk = TRUE;
          }
        }
        break;

      case BootFlagTypeNone:
        bBootFlagOk = TRUE;
        break;

      default:
        assert(FALSE);
        break;
      }
    }

    if (bootFlagType != BootFlagTypeNone) {
      if (bBootFlagOk) {
        _tprintf(_T("Boot flag OK. The bitstream will be loaded on reset / power up.\n"));
      } else {
        _tprintf(_T("*** The bitstream may not be loaded on reset / power up.\n"));
        ret = -1;
        goto done;
      }
    }
  }

  if (bDoReadback) {
    uint64_t readAddress, readLength;
    TCHAR* pFilename;
    FILE* pFile;

    readAddress = arguments.param[3].uint64Val;
    readLength = arguments.param[4].uint64Val;
    pFilename = arguments.param[2].pPathVal;
    if (_tfopen_s(&pFile, pFilename, _T("wb"))) {
      pFile = NULL;
      ret = -1;
      _tprintf(_T("*** Failed to open file '%s' for writing.\n"), pFilename);
      goto done;
    }
    _tprintf(_T("Reading data from 0x%llX-0x%llX into '%s'...\n"),
      (unsigned long long)readAddress, (unsigned long long)(readAddress + readLength - 1U), pFilename);
    status = readback(hCard, flashIndex, readAddress, readLength, pFile);
    if (ADMXRC3_SUCCESS != status) {
      fclose(pFile);
      ret = -1;
      goto done;
    }
    fclose(pFile);
  }

done:
  if (NULL != pBitstream) {
    ADMXRC3_UnloadBitstream(pBitstream);
  }

  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    if ((unsigned int)-1 != flashIndex) {
      status = ADMXRC3_SyncFlash(hCard, flashIndex, 0);
      if (ADMXRC3_SUCCESS != status) {
        _tprintf(_T("+++ WARNING: Failed to synchronize Flash on exit: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      }
    }
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
