/*
** info.c - displays information about a Gen 3 reconfigurable computing device.
**
** (C) Copyright 2009-2016 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _stprintf_s snprintf
# endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tmain main
# define _tprintf printf
# define _totupper toupper
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const struct {
  ADMXRC3_MODEL_TYPE model;
  const TCHAR* pName;
} models[] = {
  { ADMXRC3_MODEL_GENERIC, _T("Generic reconfigurable computing card") },
  { ADMXRC3_MODEL_ADMXRC2, _T("ADM-XRC-II") },
  { ADMXRC3_MODEL_ADPEXRC5T, _T("ADPe-XRC-5T") },
  { ADMXRC3_MODEL_ADMXRC6TL, _T("ADM-XRC-6TL") },
  { ADMXRC3_MODEL_ADMXRC6T1, _T("ADM-XRC-6T1") },
  { ADMXRC3_MODEL_ADMXRC6TGE, _T("ADM-XRC-6TGE") },
  { ADMXRC3_MODEL_ADMXRC6TADV8, _T("ADM-XRC-6T-ADV8") },
  { ADMXRC3_MODEL_ADPEXRC6T, _T("ADPE-XRC-6T") },
  { ADMXRC3_MODEL_ADPEXRC6TL, _T("ADPE-XRC-6T-L") },
  { ADMXRC3_MODEL_ADPEXRC6TADV_C, _T("ADPE-XRC-6T-ADV (Controller)") },
  { ADMXRC3_MODEL_ADPEXRC6TADV_T, _T("ADPE-XRC-6T-ADV (Target)") },
  { ADMXRC3_MODEL_ADMXRC7K1, _T("ADM-XRC-7K1") },
  { ADMXRC3_MODEL_ADMXRC6TDA1, _T("ADM-XRC-6T-DA1") },
  { ADMXRC3_MODEL_ADMXRC7V1, _T("ADM-XRC-7V1") },
  { ADMXRC3_MODEL_ADMVPX37V2, _T("ADM-VPX3-7V2") },
  { ADMXRC3_MODEL_ADMXRC6TGEL, _T("ADM-XRC-6TGEL") },
  { ADMXRC3_MODEL_ADMPCIE7V3, _T("ADM-PCIE-7V3") },
  { ADMXRC3_MODEL_ADMPCIEKU3, _T("ADM-PCIE-KU3") },
  { ADMXRC3_MODEL_ADMXRC7Z1, _T("ADM-XRC-7Z1") },
  { ADMXRC3_MODEL_ADMXRC7Z2, _T("ADM-XRC-7Z2") },
  { ADMXRC3_MODEL_ADMXRCKU1, _T("ADM-XRC-KU1") },
  { ADMXRC3_MODEL_ADMXRCKU1_P5, _T("ADM-XRC-KU1 (P5 Endpoint)") },
  { ADMXRC3_MODEL_ADMXRCKU1_P6, _T("ADM-XRC-KU1 (P6 Endpoint)") },
  { ADMXRC3_MODEL_ADMPCIE8V3, _T("ADM-PCIE-8V3") },
  { ADMXRC3_MODEL_ADMPCIE8K5, _T("ADM-PCIE-8K5") }
};

static void
toUpperCase(
  TCHAR* pString,
  size_t bufferLength)
{
  size_t i = 0;

  while (i < bufferLength && pString[i] != _T('\0')) {
    pString[i] = _totupper(pString[i]);
    i++;
  }
}

static const TCHAR*
getModelName(
  ADMXRC3_MODEL_TYPE modelType)
{
  int i;

  for (i = 0; i < ARRAY_LENGTH(models); i++) {
    if (models[i].model == modelType) {
      return models[i].pName;
    }
  }

  return NULL;
}

static const TCHAR*
getSpaceName(
  unsigned int index,
  ADMXRC3_MODEL_TYPE model)
{
  switch (model) {
  case ADMXRC3_MODEL_GENERIC:
    if (index == 0) {
      return _T("Unknown prefetchable space");
    } else if (index == 1) {
      return _T("Unknown nonprefetchable space");
    } else if (index == 2) {
      return _T("Unknown model-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRC2:
    if (index == 0) {
      return _T("Target FPGA 0 space");
    } else if (index == 1) {
      return _T("ADM-XRC-II-specific registers");
    } else if (index == 2) {
      return _T("PCI9656 Registers");
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADPe-XRC-5T-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-XRC-6TL-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6T1:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-XRC-6T1-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TGE:
  case ADMXRC3_MODEL_ADMXRC6TGEL:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      if (ADMXRC3_MODEL_ADMXRC6TGEL == model) {
        return _T("ADM-XRC-6TGEL-specific registers");
      } else {
        return _T("ADM-XRC-6TGE-specific registers");
      }
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-XRC-6T-ADV8-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC6T:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADPE-XRC-6T-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC6TL:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADPE-XRC-6T-L-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_C:
    if (index == 0) {
      return _T("Controller FPGA prefetchable space");
    } else if (index == 1) {
      return _T("Controller FPGA nonprefetchable space");
    } else if (index == 2) {
      return _T("ADPE-XRC-6T-ADV-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC6TADV_T:
    if (index == 0) {
      return _T("Target FPGA prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA nonprefetchable space");
    } else if (index == 2) {
      return _T("ADPE-XRC-6T-ADV-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRC7K1:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-XRC-7K1-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRC6TDA1:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-XRC-6T-DA1-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRC7V1:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-XRC-7V1-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMVPX37V2:
    if (index == 0) {
      return _T("Target FPGA 0 prefetchable space");
    } else if (index == 1) {
      return _T("Target FPGA 0 nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-VPX3-7V2-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMPCIE7V3:
    if (index == 0) {
      return _T("User prefetchable space");
    } else if (index == 1) {
      return _T("User nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-PCIE-7V3-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMPCIEKU3:
    if (index == 0) {
      return _T("User prefetchable space");
    } else if (index == 1) {
      return _T("User nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-PCIE-KU3-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMXRCKU1:
  case ADMXRC3_MODEL_ADMXRCKU1_P5:
  case ADMXRC3_MODEL_ADMXRCKU1_P6:
    if (index == 0) {
      return _T("User prefetchable space");
    } else if (index == 1) {
      return _T("User nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-PCIE-KU1-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMPCIE8V3:
    if (index == 0) {
      return _T("User prefetchable space");
    } else if (index == 1) {
      return _T("User nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-PCIE-8V3-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  case ADMXRC3_MODEL_ADMPCIE8K5:
    if (index == 0) {
      return _T("User prefetchable space");
    } else if (index == 1) {
      return _T("User nonprefetchable space");
    } else if (index == 2) {
      return _T("ADM-PCIE-8K5-specific registers");
    } else if (index == 3) {
      return _T("ADB3 bridge registers");
    }
    break;

  default:
    break;
  }

  return _T("Unknown");
}

const TCHAR*
getUnit(
  ADMXRC3_UNIT_TYPE unit)
{
  switch (unit) {
  case ADMXRC3_UNIT_NONE:
    return _T("(unitless)");

  case ADMXRC3_UNIT_A:
    return _T("A");

  case ADMXRC3_UNIT_V:
    return _T("V");

  case ADMXRC3_UNIT_C:
    return _T("deg. C");

  case ADMXRC3_UNIT_HZ:
    return _T("Hz");

  case ADMXRC3_UNIT_RPM:
    return _T("RPM");

  default:
    return _T("?");
  }
}

const TCHAR*
getDataType(
  ADMXRC3_DATA_TYPE datatype)
{
  switch (datatype) {
  case ADMXRC3_DATA_BOOL:
    return _T("boolean");

  case ADMXRC3_DATA_DOUBLE:
    return _T("double");

  case ADMXRC3_DATA_INT32:
    return _T("int32");

  case ADMXRC3_DATA_UINT32:
    return _T("uint32");

  default:
    return _T("?");
  }
}

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  int ret = 0;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  ADMXRC3_BANK_INFO bankInfo;
  ADMXRC3_CARD_INFOEX cardInfoEx;
  ADMXRC3_FLASH_INFO flashInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_MODULE_INFO moduleInfo;
  ADMXRC3_SENSOR_INFO sensorInfo;
  ADMXRC3_VERSION_INFO versionInfo;
  ADMXRC3_WINDOW_INFO windowInfo;
  TCHAR modelName[80];
  Option options[5];
  Arguments arguments = g_initArguments;
  unsigned int i, index, serial;
  bool_t bShowFlashInfo, bShowIoModuleInfo, bShowSensorInfo;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("flash");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("show Flash bank information");
  options[3].key = _T("io");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
  options[3].meaning = _T("show I/O module information");
  options[4].key = _T("sensor");
  options[4].type = OptionBoolean;
  options[4].def.booleanVal = FALSE;
  options[4].meaning = _T("show sensor information");
  arguments.nOption = 5;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;
  arguments.bNoStdOptions = TRUE;
  
  if (sampleParseCommandLine(_T("info"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bShowFlashInfo = options[2].value.booleanVal;
  bShowIoModuleInfo = options[3].value.booleanVal;
  bShowSensorInfo = options[4].value.booleanVal;

  status = sampleOpenCardNoClearError(index, options[0].specified, serial, options[1].specified, TRUE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  _tprintf(_T("API information\n"));
  status = ADMXRC3_GetVersionInfo(hCard, &versionInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get version information: %s\n\n"),
      ADMXRC3_GetStatusString(status, TRUE));
  } else {
    _tprintf(_T("API library version           %lu.%lu.%lu\n"),
      (unsigned long)versionInfo.Api.Major,
      (unsigned long)versionInfo.Api.Minor,
      (unsigned long)versionInfo.Api.Bugfix);
    _tprintf(_T("Driver version                %lu.%lu.%lu\n"),
      (unsigned long)versionInfo.Driver.Major,
      (unsigned long)versionInfo.Driver.Minor,
      (unsigned long)versionInfo.Driver.Bugfix);
    _tprintf(_T("\n"));
  }

  status = ADMXRC3_GetCardInfoEx(hCard, &cardInfoEx);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card information: %s\n\n"),
      ADMXRC3_GetStatusString(status, TRUE));
  } else {
    _tprintf(_T("Card information\n"));
    if (NULL != getModelName(cardInfoEx.Model)) {
      _stprintf_s(modelName, ARRAY_LENGTH(modelName), _T("%s"), getModelName(cardInfoEx.Model));
    } else {
      _stprintf_s(modelName, ARRAY_LENGTH(modelName), _T("unknown (0x%lX)"), (unsigned long)cardInfoEx.Model);
    }
    _tprintf(_T("Model                         %s\n"), modelName);
    _tprintf(_T("Serial number                 %lu(0x%lX)\n"), (unsigned long)cardInfoEx.SerialNumber, (unsigned long)cardInfoEx.SerialNumber);
    _tprintf(_T("Number of programmable clocks %lu\n"), (unsigned long)cardInfoEx.NumClockGen);
    _tprintf(_T("Number of DMA channels        %lu\n"), (unsigned long)cardInfoEx.NumDmaChannel);
    _tprintf(_T("Number of target FPGAs        %lu\n"), (unsigned long)cardInfoEx.NumTargetFpga);
    _tprintf(_T("Number of local bus windows   %lu\n"), (unsigned long)cardInfoEx.NumWindow);
    _tprintf(_T("Number of sensors             %lu\n"), (unsigned long)cardInfoEx.NumSensor);
    _tprintf(_T("Number of I/O module sites    %lu\n"), (unsigned long)cardInfoEx.NumModuleSite);
    _tprintf(_T("Number of memory banks        %lu\n"), (unsigned long)cardInfoEx.NumMemoryBank);
    _tprintf(_T("Bank presence bitmap          0x%lX\n"), (unsigned long)cardInfoEx.MemoryBankPresent);
    if (bShowFlashInfo) {
      _tprintf(_T("Number of Flash memory banks  %lu\n"), (unsigned long)cardInfoEx.NumFlashBank);
    }
    _tprintf(_T("\n"));

    _tprintf(_T("Target FPGA information\n"));
    for (i = 0; i < cardInfoEx.NumTargetFpga; i++) {
      _tprintf(_T("FPGA %-2lu                       "), (unsigned long)i);
      status = ADMXRC3_GetFpgaInfo(hCard, i, &fpgaInfo);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("Failed to get info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      } else {
        toUpperCase(fpgaInfo.Identifier, ARRAY_LENGTH(fpgaInfo.Identifier));
        toUpperCase(fpgaInfo.SpeedGrade, ARRAY_LENGTH(fpgaInfo.SpeedGrade));
        toUpperCase(fpgaInfo.Stepping, ARRAY_LENGTH(fpgaInfo.Stepping));
        _tprintf(_T("%s"), fpgaInfo.Identifier);
        if (fpgaInfo.Flags & ADMXRC3_FPGA_SPEEDVALID) {
          _tprintf(_T("-%s"), fpgaInfo.SpeedGrade);
        }
        if (fpgaInfo.Flags & ADMXRC3_FPGA_STEPPINGVALID) {
          _tprintf(_T(", stepping %s"), fpgaInfo.Stepping);
        }
        if (fpgaInfo.Flags & ADMXRC3_FPGA_NOTCONFIGURABLE) {
          _tprintf(_T(", not configurable"));
        }
        _tprintf(_T("\n"));
      }
    }
    if (0 != cardInfoEx.NumTargetFpga) {
      _tprintf(_T("\n"));
    } else {
      _tprintf(_T("(none)\n\n"));
    }

    _tprintf(_T("Memory bank information\n"));
    for (i = 0; i < cardInfoEx.NumMemoryBank; i++) {
      _tprintf(_T("Bank %-2lu                       "), (unsigned long)i);
      status = ADMXRC3_GetBankInfo(hCard, i, &bankInfo);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("Failed to get info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      } else {
        if (bankInfo.Present) {
          TCHAR unknownTypeStr[40];
          TCHAR* pTypeStr;

          if (bankInfo.TypeMask == (ADMXRC3_BANK_ZBT_FT | ADMXRC3_BANK_ZBT_P)) {
            pTypeStr = _T("SSRAM, ZBT P+F");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_ZBT_FT) {
            pTypeStr = _T("SSRAM, ZBT F");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_ZBT_P) {
            pTypeStr = _T("SSRAM, ZBT P");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_SDR) {
            pTypeStr = _T("SDRAM, SDR");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_DDR) {
            pTypeStr = _T("SDRAM, DDR");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_DDR2) {
            pTypeStr = _T("SDRAM, DDR-II");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SSRAM_DDR2) {
            pTypeStr = _T("SSRAM, DDR-II");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SSRAM_QDR) {
            pTypeStr = _T("SSRAM, QDR");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_DDR3) {
            pTypeStr = _T("SDRAM, DDR3");
          } else if (bankInfo.TypeMask == ADMXRC3_BANK_SDRAM_DDR4) {
            pTypeStr = _T("SDRAM, DDR4");
          } else {
            _stprintf_s(unknownTypeStr, ARRAY_LENGTH(unknownTypeStr), _T("Unknown memory type (0x%lX)"), (unsigned long)bankInfo.TypeMask);
            pTypeStr = unknownTypeStr;
          }
          _tprintf(_T("%s, %llu(0x%llx) kiW x %lu + %lu bits\n"),
            pTypeStr, (unsigned long long)(bankInfo.PhysicalSize >> 10), (unsigned long long)(bankInfo.PhysicalSize >> 10),
            (unsigned long)bankInfo.PhysicalDataWidth, (unsigned long)bankInfo.PhysicalECCWidth);
          if (0 != bankInfo.MaximumFrequency) {
            _tprintf(_T("                              %.1lf MHz - %.1lf MHz\n"),
              (double)bankInfo.MinimumFrequency / 1000000.0, (double)bankInfo.MaximumFrequency / 1000000.0);
          }
          _tprintf(_T("                              Connectivity mask 0x%lX\n"), (unsigned long)bankInfo.ConnectivityMask);
        } else {
          _tprintf(_T("Not present\n"));
        }
      }
    }
    if (0 != cardInfoEx.NumMemoryBank) {
      _tprintf(_T("\n"));
    } else {
      _tprintf(_T("(none)\n\n"));
    }

    _tprintf(_T("Local bus window information\n"));
    for (i = 0; i < cardInfoEx.NumWindow; i++) {
      TCHAR name[128];

      if (_stprintf_s(name, ARRAY_LENGTH(name), _T("(%s)"), getSpaceName(i, cardInfoEx.Model)) > 0) {
        _tprintf(_T("Window %-2lu %-.18s  "), (unsigned long)i, name);
      } else {
        _tprintf(_T("Window %-2lu %-.18s  "), (unsigned long)i, _T("(\?\?\?)"));
      }
      status = ADMXRC3_GetWindowInfo(hCard, i, &windowInfo);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("Failed to get info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      } else {
        _tprintf(_T("Bus base     0x%llX size 0x%llX\n"),
          (unsigned long long)windowInfo.BusBase, (unsigned long long)windowInfo.BusSize);
        _tprintf(_T("                              Local base   0x%llX size 0x%llX\n"),
          (unsigned long long)windowInfo.LocalBase, (unsigned long long)windowInfo.LocalSize);
        _tprintf(_T("                              Virtual size 0x%llX\n"),
          (unsigned long long)windowInfo.VirtualSize);
      }
    }
    if (0 != cardInfoEx.NumWindow) {
      _tprintf(_T("\n"));
    } else {
      _tprintf(_T("(none)\n\n"));
    }

    if (bShowFlashInfo) {
      _tprintf(_T("Flash bank information\n"));
      for (i = 0; i < cardInfoEx.NumFlashBank; i++) {
        _tprintf(_T("Bank %-2lu                       "), (unsigned long)i);
        status = ADMXRC3_GetFlashInfo(hCard, i, &flashInfo);
        if (status != ADMXRC3_SUCCESS) {
          _tprintf(_T("Failed to get info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        } else {
          _tprintf(_T("%s, %llu(0x%llX) kiB\n"),
            flashInfo.Identifier, (unsigned long long)(flashInfo.Size >> 10), (unsigned long long)(flashInfo.Size >> 10));
          _tprintf(_T("                              Useable area 0x%llX-0x%llX\n"),
            (unsigned long long)flashInfo.UseableStart,
            (unsigned long long)(flashInfo.UseableStart + flashInfo.UseableLength - 1));
        }
      }
      if (0 != cardInfoEx.NumFlashBank) {
        _tprintf(_T("\n"));
      } else {
        _tprintf(_T("(none)\n\n"));
      }
    }

    if (bShowIoModuleInfo) {
      _tprintf(_T("I/O module information\n"));
      for (i = 0; i < cardInfoEx.NumModuleSite; i++) {
        _tprintf(_T("Module %-2lu                     "), (unsigned long)i);
        status = ADMXRC3_GetModuleInfo(hCard, i, &moduleInfo);
        if (status != ADMXRC3_SUCCESS) {
          _tprintf(_T("Failed to get info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        } else {
          if (moduleInfo.Present) {
            if (moduleInfo.Flags & ADMXRC3_MODULE_LEGACYXRM) {
              _tprintf(_T("Legacy XRM (no further information available)\n"));
            } else {
              _tprintf(_T("Product %s\n"), moduleInfo.Product);
              _tprintf(_T("                  Part number %s\n"), moduleInfo.PartNumber);
              _tprintf(_T("                 Manufacturer %s\n"), moduleInfo.Manufacturer);
              _tprintf(_T("                Serial number %s\n"), moduleInfo.SerialNumber);
              _tprintf(_T("             Manufacture time %lu minutes since 00:00 1/1/1996\n"), (unsigned long) moduleInfo.ManufactureDate);
              _tprintf(_T("                  I/O voltage %.2f\n"), moduleInfo.IoVoltage);
            }
            _tprintf(_T("                        Flags 0x%lx%s%s%s%s%s\n"),
              (unsigned long)moduleInfo.Flags,
              moduleInfo.Flags ? _T(" => ") : _T(""),
              moduleInfo.Flags & ADMXRC3_MODULE_LEGACYXRM ? _T("LEGACYXRM ") : _T(""),
              moduleInfo.Flags & ADMXRC3_MODULE_FORCE2V5 ? _T("FORCE2V5 ") : _T(""),
              moduleInfo.Flags & ADMXRC3_MODULE_ROMERROR ? _T("ROMERROR ") : _T(""),
              moduleInfo.Flags & ADMXRC3_MODULE_CSUMERROR ? _T("CSUMERROR ") : _T(""));
          } else {
            _tprintf(_T("Not present\n"));
          }
        }
      }
      if (0 != cardInfoEx.NumModuleSite) {
        _tprintf(_T("\n"));
      } else {
        _tprintf(_T("(none)\n\n"));
      }
    }

    if (bShowSensorInfo) {
      _tprintf(_T("Sensor information\n"));
      for (i = 0; i < cardInfoEx.NumSensor; i++) {
        _tprintf(_T("Sensor %-2lu                     "), (unsigned long)i);
        status = ADMXRC3_GetSensorInfo(hCard, i, &sensorInfo);
        if (status != ADMXRC3_SUCCESS) {

          _tprintf(_T("Failed to get info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        } else {
          _tprintf(_T("%s\n"), sensorInfo.Description);
          _tprintf(_T("                              %s, %s, exponent %d, error %.3f\n"),
            getUnit(sensorInfo.Unit), getDataType(sensorInfo.DataType), sensorInfo.Exponent, sensorInfo.Error);
          if (sensorInfo.Capabilities) {
            _tprintf(_T("                              %s\n"), (sensorInfo.Capabilities & ADMXRC3_SENSOR_SWVALUE) ? "Software sensor" : "");
          }
        }
      }
      if (0 != cardInfoEx.NumSensor) {
        _tprintf(_T("\n"));
      } else {
        _tprintf(_T("(none)\n\n"));
      }
    }
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
